package com.taleo.integration.client.customstep;

import java.io.File;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import org.apache.commons.lang.StringUtils;

import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.ws.integration.client.GlobalConfig;

/**
 * Abstract custom step.
 * <p>
 * This step takes three parameters:
 * <ul>
 * <li><b>active</b>: Indicates if the step is active (<code>true</code> or
 * <code>false</code>, default: <code>true</code>).</li>
 * <br>
 * <br>
 * <li><b>NOTE</b>: Simple placeholder to add a descriptive note to the step.</li>
 * <br>
 * <br>
 * <li><b>encoding</b>: Optional - The encoding used for reading and writing
 * files (Default: UTF-8).</li>
 * </ul>
 * 
 * @author Romain Guay, Taleo Corporation
 */
public abstract class BaseCustomStep extends
		com.taleo.integration.client.step.BaseCustomStep {

	/**
	 * Constructor without arguments.
	 */
	public BaseCustomStep() {
		// Register valid parameters.
		registerParameterName("active");
		registerParameterName("NOTE");
		registerParameterName("encoding");
	}

	/**
	 * The global config.
	 */
	private GlobalConfig globalConfig;

	/**
	 * The list of valid parameter names.
	 */
	protected List parameterNames = new ArrayList();

	/**
	 * The active flag.
	 */
	private boolean active = true;

	/**
	 * The encoding used for reading ad writing files (Default: UTF-8).
	 * 
	 * @see java.nio.charset.Charset
	 */
	private String encoding = Constants.UTF8;

	/**
	 * Get the active flag.
	 * 
	 * @return
	 */
	public boolean isActive() {
		return active;
	}

	/**
	 * Set the active flag.
	 * 
	 * @param active
	 */
	public void setActive(boolean active) {
		this.active = active;
	}

	/**
	 * Get the encoding.
	 * 
	 * @return
	 */
	public String getEncoding() {
		return encoding;
	}

	/**
	 * Set the encoding.
	 * 
	 * @param encoding
	 */
	public void setEncoding(String encoding) {
		// Validate encoding by loading it into a Charset.
		Charset.forName(encoding);
		this.encoding = encoding;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);
		this.globalConfig = config;

		// Validate the parameter names.
		validateParameterNames();

		// Get the active parameter.
		active = Boolean.valueOf(
				StringUtils.defaultIfEmpty(parameters.getProperty("active"),
						"true")).booleanValue();

		encoding = parameters.getProperty("encoding", encoding);
	}

	/**
	 * Register a valid parameter name.
	 * 
	 * @param name
	 *            The name of a valid parameter.
	 */
	protected void registerParameterName(String name) {
		parameterNames.add(name);
	}

	/**
	 * Validate the parameters names.
	 */
	protected void validateParameterNames() {
		Enumeration keys = parameters.keys();
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			if (!parameterNames.contains(key)) {
				throw new IllegalArgumentException("Unknown parameter " + key
						+ ".");
			}
		}
	}

	/**
	 * Create a temporary file with the default "IntegrationClient" prefix.
	 * 
	 * @return
	 * @throws IOException
	 */
	protected File createTempFile() throws IOException {
		return createTempFile("IntegrationClient");
	}

	/**
	 * Create a temporary file with the specified prefix.
	 * 
	 * @param prefix
	 * @return
	 * @throws IOException
	 */
	protected File createTempFile(String prefix) throws IOException {
		boolean deleteOnExit = (globalConfig == null || globalConfig
				.getGeneral().getDeleteTemporaryFiles());
		File file = FileUtils.createTempFile(tempFolder, prefix, deleteOnExit);
		return file;
	}

	/**
	 * Get the temporary folder, defaulting to User temp folder.
	 * 
	 * @return
	 */
	protected File getTempFolder() {
		if (tempFolder == null) {
			return new File(System.getProperty("java.io.tmpdir"));
		} else {
			return new File(tempFolder);
		}
	}

}
