package com.taleo.integration.client.customstep.compress;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Post-processing step to zip the current workflow file or other files
 * corresponding to a pattern.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>filePattern</b>: Optional - The pattern for selecting the folder or
 * files to zip. Can be specified as an absolute or relative path and may use
 * wildcards as defined in
 * {@link org.apache.commons.io.filefilter.WildcardFileFilter} in the file names
 * only. Pattern is case sensitive. If absent, the current workflow document is
 * zipped.</li>
 * <br>
 * <li><b>outputFile</b>: Optional - The filename for the zipped file. If absent
 * the zipped file becomes the current document in the execution pipeline. This
 * value may be supplied as a relative or absolute path.</li>
 * <br>
 * <li><b>method</b>: Optional - The compression method - Deflated or Stored.
 * Default: Deflated.</li>
 * <br>
 * <li><b>deleteOriginal</b>: Optional - Delete the original file(s) after
 * compression (<code>true</code> or <code>false</code>, default:
 * <code>true</code>)</li>. This only applies if the filePattern parameter is
 * provided. In the case where the file is coming from the pipeline, it is not
 * deleted by this process, but by the normal TCC workflow cleanup process.
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here are two examples of how to define this custom step in the TCC
 * configuration file:
 * <p>
 * <img src="doc-files/ZipPostStep-1.png" /><br>
 * <img src="doc-files/ZipPostStep-2.png" />
 * 
 * @author Romain Guay, Taleo Corporation
 * @see java.util.zip
 */
public class ZipPostStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(ZipPostStep.class);

	/**
	 * The compression methods.
	 */
	public final int METHOD_DEFLATED = ZipOutputStream.DEFLATED;
	public final int METHOD_STORED = ZipOutputStream.STORED;

	/**
	 * Constructor without arguments.
	 */
	public ZipPostStep() {
		super();
		// Register valid parameters.
		registerParameterName("filePattern");
		registerParameterName("outputFile");
		registerParameterName("method");
		registerParameterName("deleteOriginal");
	}

	/**
	 * The file pattern.
	 */
	private String filePattern;

	/**
	 * The outputFile.
	 */
	private String outputFile;

	/**
	 * The compression method.
	 */
	protected int method = METHOD_DEFLATED;

	/**
	 * The indicator to delete original uncompressed files.
	 */
	private boolean deleteOriginal = true;

	/**
	 * Get the file pattern.
	 * 
	 * @return
	 */
	public String getFilePattern() {
		return filePattern;
	}

	/**
	 * Set the file pattern.
	 * 
	 * @param filePattern
	 */
	public void setFilePattern(String filePattern) {
		this.filePattern = filePattern;
	}

	/**
	 * Get the output file.
	 * 
	 * @return
	 */
	public String getOutputFile() {
		return outputFile;
	}

	/**
	 * Set the output file.
	 * 
	 * @param outputFile
	 */
	public void setOutputFile(String outputFile) {
		this.outputFile = outputFile;
	}

	/**
	 * Get the compression method.
	 * 
	 * @return
	 */
	public int getMethod() {
		return method;
	}

	/**
	 * Set the compression method (METHOD_DEFLATED or METHOD_STORED).
	 * 
	 * @param method
	 */
	public void setMethod(int method) {
		if (method != METHOD_DEFLATED && method != METHOD_STORED) {
			throw new IllegalArgumentException("Invalid method.");
		}
		this.method = method;
	}

	/**
	 * Get the delete original flag.
	 * 
	 * @return
	 */
	public boolean isDeleteOriginal() {
		return deleteOriginal;
	}

	/**
	 * Set the delete original flag.
	 * 
	 * @param deleteOriginal
	 */
	public void setDeleteOriginal(boolean deleteOriginal) {
		this.deleteOriginal = deleteOriginal;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "ZipPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing zip the workflow document or other files corresponding to a pattern.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "ZipPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// First, get the basic supported pipeline for a post-processing step
		// (this handles the constraints on the last step type)
		SupportedPipeline supportedPipeline = getPostProcessSupportPipeline();

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		filePattern = parameters.getProperty("filePattern");
		outputFile = parameters.getProperty("outputFile");

		String method = parameters.getProperty("method", "Deflated");
		if ("Deflated".equalsIgnoreCase(method)) {
			setMethod(METHOD_DEFLATED);
		} else if ("Stored".equalsIgnoreCase(method)) {
			setMethod(METHOD_STORED);
		} else {
			throw new IllegalArgumentException(
					"The 'method' parameter must be Deflated or Stored.");
		}

		deleteOriginal = Boolean.valueOf(
				parameters.getProperty("deleteOriginal", "true"))
				.booleanValue();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// If not active, return immediately.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		try {
			// Get the current workflow file.
			Message responseMessage = pipeline.getResponseMessage();

			// Prepare the zip file.
			File zipFile = null;
			if (this.outputFile != null) {
				// If the outputFile is specified, use it.
				zipFile = new File(this.outputFile);

			} else {
				// Otherwise, create a temporary file.
				zipFile = createTempFile();
			}

			// Execute.
			if (filePattern == null || filePattern.trim().length() == 0) {
				// Execute the step with the current workflow file.
				zip(new File(responseMessage.getCurrentFile()), zipFile, false);
			} else {
				// Execute the step with the given file pattern.
				zip(new File(this.filePattern), zipFile, deleteOriginal);
			}

			// If the outputFile is null, set the zip file as the new current
			// file in the pipeline.
			if (this.outputFile == null) {
				responseMessage.setCurrentFile(zipFile.getAbsolutePath());
				responseMessage.setFormat(MessageFormat.OTHER);
			}

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Zip the files matching the file pattern into the given zip file.
	 * 
	 * @param pattern
	 *            The file pattern to select folder or files.
	 * @param zipFile
	 *            The resulting zip file.
	 * @param deleteOriginal
	 *            Delete the original files after compression.
	 * @throws IOException
	 */
	protected void zip(File pattern, File zipFile, boolean deleteOriginal)
			throws IOException {

		// Get the files and folders to process.
		File[] files = FileUtils.getMatchingFiles(pattern);

		ZipOutputStream os = null;
		try {
			// Prepare the output stream.
			os = new ZipOutputStream(new BufferedOutputStream(
					new FileOutputStream(zipFile)));
			os.setMethod(this.getMethod());

			// Get the base directory.
			File base = pattern.getParentFile();

			// Zip the files.
			for (int i = 0; i < files.length; i++) {
				zip(files[i], base, os);
			}

			// Delete original files.
			if (deleteOriginal) {
				for (int i = 0; i < files.length; i++) {
					org.apache.commons.io.FileUtils.forceDelete(files[i]);
				}
			}

		} finally {
			if (os != null) {
				os.close();
			}
		}
	}

	/**
	 * Zip a file or folder in the given output stream.
	 * 
	 * @param file
	 *            The file or folder name.
	 * @param base
	 *            Base file to relativize filenames in the Zip.
	 * @param os
	 *            The output stream.
	 */
	protected void zip(File file, File base, ZipOutputStream os)
			throws IOException {
		InputStream is = null;

		// Process folders recursively.
		if (file.isDirectory()) {
			File[] files = file.listFiles();
			for (int i = 0; i < files.length; i++) {
				zip(files[i], base, os);
			}
		} else {
			String relative = base.toURI().relativize(file.toURI()).getPath();
			log.info("Compressing file: '" + relative + "'");

			// Zip the file.
			try {
				is = new BufferedInputStream(new FileInputStream(file));
				ZipEntry entry = new ZipEntry(relative);
				os.putNextEntry(entry);
				IOUtils.copy(is, os);
			} finally {
				if (is != null) {
					is.close();
				}
			}
		}
	}

}
