package com.taleo.integration.client.customstep.control;

import java.io.File;
import java.io.InputStream;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.csv.BaseCustomCsvStep;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Processing step to break workflow execution.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>exitCode</b>: Optional - The exit code to be set on the Java process
 * as a result of this step. (Default: The value set in the system property
 * named <code>com.taleo.client.customstep.BreakStep.ExitCode</code>, or 0 if
 * absent.)</li>
 * <br>
 * <li><b>filename</b>: <i>@deprecated - Use
 * {@link com.taleo.integration.client.customstep.file.WriteFileStep}
 * instead.</i> Optional - Full path (absolute or relative) to store the content
 * of the current file from the pipeline in an uncompressed format. If empty,
 * the file is not stored.</li>
 * </ul>
 * <p>
 * This step is typically used with a variable setting the <code>active</code>
 * parameter, to get control over when it is executed.
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class BreakStep extends BaseCustomCsvStep {

	private static Logger log = Logger.getLogger(BreakStep.class);

	/**
	 * Constant for the ExitCode system property.
	 */
	private static final String PROP_BREAK_EXIT_CODE = "com.taleo.client.customstep.BreakStep.ExitCode";

	/**
	 * Default value for com.taleo.client.customstep.BreakStep.ExitCode system
	 * property.
	 */
	private static final int DEFAULT_BREAK_EXIT_CODE = 0;

	/**
	 * The exit code.
	 */
	private int exitCode = DEFAULT_BREAK_EXIT_CODE;

	/**
	 * The filename.
	 * 
	 * @deprecated
	 */
	private String filename;

	/**
	 * Constructor without arguments.
	 */
	public BreakStep() {
		super();
		registerParameterName("exitCode");
		registerParameterName("filename");

		// Set the default value for exit code.
		String prop = System.getProperty(PROP_BREAK_EXIT_CODE);
		if (StringUtils.isNotBlank(prop)) {
			try {
				// Read the system property for exit code.
				exitCode = Integer.parseInt(System
						.getProperty(PROP_BREAK_EXIT_CODE));
			} catch (NumberFormatException e) {
				// Invalid property value: keep default value.
				log.warn("Invalid value '" + prop + "' for system property "
						+ PROP_BREAK_EXIT_CODE);
			}
		}
	}

	/**
	 * Get the exit code.
	 * 
	 * @return
	 */
	public int getExitCode() {
		return exitCode;
	}

	/**
	 * Set the exit code.
	 * 
	 * @param exitCode
	 */
	public void setExitCode(int exitCode) {
		this.exitCode = exitCode;
	}

	/**
	 * Get the filename.
	 * 
	 * @deprecated
	 * @return
	 */
	public String getFilename() {
		return filename;
	}

	/**
	 * Set the filename.
	 * 
	 * @deprecated
	 * @param filename
	 */
	public void setFilename(String filename) {
		this.filename = filename;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "BreakStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Processing step to break TCC workflow execution.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "BreakStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		try {
			setExitCode(Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("exitCode"),
					Integer.toString(this.exitCode))));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'exitCode' parameter must have a numeric value.");
		}

		setFilename(parameters.getProperty("filename"));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			// Store current file content if needed.
			if (StringUtils.isNotBlank(this.filename)) {
				File file = null;
				Message responseMessage = pipeline.getResponseMessage();
				Message requestMessage = pipeline.getRequestMessage();
				if (responseMessage.isSetCurrentFile()) {
					file = new File(responseMessage.getCurrentFile());
				} else {
					if (requestMessage.isSetCurrentFile()) {
						file = new File(requestMessage.getCurrentFile());
					} else {
						file = new File(requestMessage.getOriginalFile());
					}
				}
				log.info("Creating file: " + filename + ".");
				InputStream is = IOHelper.getInputStream(file);
				try {
					IOHelper.transferStreamToFile(is, new File(filename));
				} finally {
					is.close();
				}
			}

			// Note: TCC currently does not provide any means to abort
			// execution. The only way is to kill the process.
			log.info("The run was aborted with code: " + exitCode + "\n");
			System.exit(exitCode);

		} catch (Exception e) {
			throw new StepException(e);
		}
	}
}
