package com.taleo.integration.client.customstep.control;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.regex.Pattern;

import javax.xml.namespace.NamespaceContext;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.Constants;
import com.taleo.integration.client.customstep.util.xml.XPathHelper;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.integration.client.workflow.WorkflowException;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Post-processing step to filter export errors matching a pattern and exit the
 * process with a given code.
 * <p>
 * This step is useful to allow interpreting errors in the calling job using a
 * specific error code.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>pattern</b>: RegEx pattern to match the error reason. If the reason
 * matches the pattern, the process ends with the exit code mentioned below. If
 * it does not match, it continues normally. If the pattern is an empty string,
 * all messages will match.</li>
 * <br>
 * <li><b>exitCode</b>: Optional - The exit code to be set on the Java process
 * as a result of this step. (Default: 0.)</li>
 * </ul>
 * <p>
 * Many of those steps can be chained in sequence to catch different errors and
 * exit with different codes. Typically, this step or step chain will be ended
 * with a {@link RaiseExportErrorPostStep}, so that if none of the filters
 * match, an exception will be raised.
 * <p>
 * When setting the pattern, make sure to follow RegEx syntax, taking into
 * consideration that some characters may need to be escaped. Also, multiple
 * patterns can be separated with the | character meaning OR in RegEx. If an
 * error has multiple reasons, the pattern will be tested with each of them for
 * the first one that matches.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/FilterExportErrorPostStep-1.png" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class FilterExportErrorPostStep extends BaseCustomStep {

	private static Logger log = Logger
			.getLogger(FilterExportErrorPostStep.class);

	/**
	 * Constructor without arguments.
	 */
	public FilterExportErrorPostStep() {
		super();
		registerParameterName("pattern");
		registerParameterName("exitCode");
	}

	/**
	 * The pattern.
	 */
	private String pattern;

	/**
	 * The exit code.
	 */
	private int exitCode = 0;

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "FilterExportErrorPostStep";
	}

	/**
	 * Get the pattern.
	 * 
	 * @return
	 */
	public String getPattern() {
		return pattern;
	}

	/**
	 * Set the pattern.
	 * 
	 * @param pattern
	 */
	public void setPattern(String pattern) {
		this.pattern = pattern;
	}

	/**
	 * Get the exit code.
	 * 
	 * @return
	 */
	public int getExitCode() {
		return exitCode;
	}

	/**
	 * Set the exit code.
	 * 
	 * @param exitCode
	 */
	public void setExitCode(int exitCode) {
		this.exitCode = exitCode;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to filter export errors.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "FilterExportErrorPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		setPattern(parameters.getProperty("pattern"));

		try {
			setExitCode(Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("exitCode"),
					Integer.toString(this.exitCode))));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'exitCode' parameter must have a numeric value.");
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// If not active, return immediately.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		try {
			// Get the results file.
			Message responseMessage = pipeline.getResponseMessage();
			File file = new File(responseMessage.getCurrentFile());

			// Execute the step.
			execute(file);

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Execute the step by matching error message.
	 * 
	 * @param file
	 *            The result file being processed.
	 * @throws IOException
	 * @throws WorkflowException
	 */
	public void execute(File file) throws IOException, WorkflowException {

		// Error files are always small (1-2K), for optimization detect large
		// files immediately as not containing errors.
		if (file.length() > 5000) {
			return;
		}

		// Errors are always reported in XML format (even for CSV format), so if
		// not a well formed XML, quit immediately.
		// if (!XmlUtils.isWellFormed(file)) {
		// return;
		// }

		// Build the namespace context.
		NamespaceContext nsContext = new NamespaceContext() {
			public String getNamespaceURI(String prefix) {
				if ("itk".equals(prefix)) {
					return Constants.ITK_NAMESPACE;
				} else {
					return "";
				}
			}

			// This method isn't necessary for XPath processing.
			public String getPrefix(String uri) {
				throw new UnsupportedOperationException();
			}

			// This method isn't necessary for XPath processing.
			public Iterator getPrefixes(String uri) {
				throw new UnsupportedOperationException();
			}
		};

		// Compile the RegEx pattern.
		Pattern pattern = null;
		if (StringUtils.isNotBlank(this.pattern)) {
			pattern = Pattern.compile(this.pattern, Pattern.DOTALL);
		}

		try {
			// Find the errors with XPath.
			InputStream xmlStream = IOHelper.getInputStream(file);
			final String xPath = "//itk:ExportErrors/itk:Error/itk:Reason";
			XPathHelper helper = new XPathHelper(nsContext);
			NodeList reasons = (NodeList) helper.evaluate(xmlStream, xPath,
					XPathConstants.NODESET);

			// If errors were found, try to match them.
			for (int i = 0; i < reasons.getLength(); i++) {
				Node reason = reasons.item(i);
				if (reason.getFirstChild() != null) {
					String reasonText = reason.getFirstChild().getNodeValue()
							.trim();
					if (StringUtils.isBlank(this.pattern)
							|| pattern.matcher(reasonText).find()) {

						// Note: TCC currently does not provide any means to
						// abort execution. The only way is to kill the process.
						log.error(reasonText);
						log.info("The run was aborted with code: " + exitCode
								+ "\n");
						System.exit(exitCode);
					}
				}
			}

		} catch (XPathExpressionException e) {
			throw new RuntimeException(e);
		} catch (ParserConfigurationException e) {
			throw new RuntimeException(e);
		} catch (SAXException e) {
			throw new RuntimeException(e);
		}
	}

}
