package com.taleo.integration.client.customstep.control;

import java.util.Enumeration;

import org.apache.log4j.MDC;

import com.taleo.integration.client.customstep.csv.BaseCustomCsvStep;
import com.taleo.ws.integration.client.GlobalConfig;

/**
 * Processing step to set logging context variables.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvStep} and the
 * following additional parameters:
 * <ul>
 * <li>All other parameters are added to the Mapped Diagnostic Context (<a href=
 * "http://logging.apache.org/log4j/1.2/apidocs/org/apache/log4j/PatternLayout.html"
 * >MDC</a>) and which can be used in the PatternLayout within the log
 * configuration file (log4j.xml). Those variables can be retrieved using %X{*},
 * where * is the variable name.
 * <p>
 * For example, by setting a parameter named <code>TouchpointID</code> to
 * something identifying the touchpoint, %X{TouchpointID} can be used in the
 * PatternLayout so that each line in the log identifies the touchpoint. This
 * improves readability, especially when multiple touchpoints are run
 * simultaneously.</li>
 * </ul>
 * 
 * @author Romain Guay, Oracle Corporation
 */
public class LogContextStep extends BaseCustomCsvStep {

	/**
	 * Validate the parameters names.
	 * <p>
	 * Override the default method to allow any parameter.
	 */
	protected void validateParameterNames() {
	}

	/**
	 * Get an MDC variable by its name.
	 * 
	 * @param name
	 * 
	 * @return
	 */
	public String getMDC(String name) {
		return MDC.get(name).toString();
	}

	/**
	 * Set an MDC variable by its name.
	 * 
	 * @param name
	 * @param value
	 */
	public void setMDC(String name, String value) {
		MDC.put(name, value);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "LogContextStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Processing step to set logging context variables.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "LogContextStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the MDC variables.
		Enumeration keys = parameters.keys();
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			if (!parameterNames.contains(key)) {
				setMDC(key, parameters.get(key).toString());
			}
		}
	}
}
