package com.taleo.integration.client.customstep.control;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;

import javax.xml.namespace.NamespaceContext;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.apache.log4j.Logger;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.Constants;
import com.taleo.integration.client.customstep.util.xml.XPathHelper;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.integration.client.workflow.WorkflowException;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Post-processing step to raise an exception when an export error is detected
 * in the results file. It works as well for CSV or XML format.
 * <p>
 * This step is useful because the behavior of TCC is not to raise an exception
 * when an error is reported inside the results file. For example, when the
 * volume limit has been reached. This is a problem because in this case the
 * monitoring does not report the error and the LRD is updated as normal.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep}.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * It should be defined first in the post processing steps. Here is an example
 * of how to define this custom step in the TCC configuration file:
 * <p>
 * <img src="doc-files/RaiseExportErrorPostStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class RaiseExportErrorPostStep extends BaseCustomStep {

	private static Logger log = Logger
			.getLogger(RaiseExportErrorPostStep.class);

	/**
	 * Constructor without arguments.
	 */
	public RaiseExportErrorPostStep() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "RaiseExportErrorPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to raise an exception when an export error is detected in the results file.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "RaiseExportErrorPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// If not active, return immediately.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		try {
			// Get the results file.
			Message responseMessage = pipeline.getResponseMessage();
			File file = new File(responseMessage.getCurrentFile());

			// Execute the step.
			execute(file);

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Execute the step by detecting and raising errors in the file.
	 * 
	 * @param file
	 *            The result file being processed.
	 * @throws IOException
	 * @throws WorkflowException
	 */
	public void execute(File file) throws IOException, WorkflowException {

		// Error files are always small (1-2K), for optimization detect large
		// files immediately as not containing errors.
		if (file.length() > 5000) {
			return;
		}

		// Errors are always reported in XML format (even for CSV format), so if
		// not a well formed XML, quit immediately.
		// if (!XmlUtils.isWellFormed(file)) {
		// return;
		// }

		// Build the namespace context.
		NamespaceContext nsContext = new NamespaceContext() {
			public String getNamespaceURI(String prefix) {
				if ("itk".equals(prefix)) {
					return Constants.ITK_NAMESPACE;
				} else {
					return "";
				}
			}

			// This method isn't necessary for XPath processing.
			public String getPrefix(String uri) {
				throw new UnsupportedOperationException();
			}

			// This method isn't necessary for XPath processing.
			public Iterator getPrefixes(String uri) {
				throw new UnsupportedOperationException();
			}
		};

		// Find the errors with XPath.
		try {
			InputStream xmlStream = IOHelper.getInputStream(file);
			final String xPath = "//itk:ExportErrors/itk:Error/itk:Reason";
			XPathHelper helper = new XPathHelper(nsContext);
			NodeList reasons = (NodeList) helper.evaluate(xmlStream, xPath,
					XPathConstants.NODESET);

			// If errors were found, raise an exception.
			if (reasons.getLength() > 0) {
				throw new WorkflowException(buildMessage(reasons));
			}
		} catch (XPathExpressionException e) {
			throw new RuntimeException(e);
		} catch (ParserConfigurationException e) {
			throw new RuntimeException(e);
		} catch (SAXException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Builds an error message with the reasons list.
	 * 
	 * @param reasons
	 *            The reasons node list.
	 * @return An error message.
	 */
	private String buildMessage(NodeList reasons) {
		StringBuffer message = new StringBuffer();
		for (int i = 0; i < reasons.getLength(); i++) {
			Node reason = reasons.item(i);
			if (i > 0) {
				message.append("; ");
			}
			if (reason.getFirstChild() != null) {
				message.append(reason.getFirstChild().getNodeValue().trim());
			}
		}
		return message.toString();
	}

}
