package com.taleo.integration.client.customstep.count;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.integration.client.util.csv.CSVConstant;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Pre-processing step to convert any query to a count query with the result in
 * CSV format.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameter:
 * <ul>
 * <li><b>csvDelimiter</b>: Optional - The CSV delimiter. Default is a comma
 * (,).</li><br>
 * <br>
 * <li><b>csvHeaderPresent</b>: Optional - Indicates if a CSV header is to be
 * displayed in CSV mode. Default is <code>false</code>.</li><br>
 * <br>
 * <li><b>entityName</b>: Optional - Name of the entity to display along with
 * the count. If absent the result will not contain that column.</li>
 * </ul>
 * This class applies to pretty much any query, but there are cases where we
 * want or need to exclude a projection from the count. To do this, we can
 * simply add a tag "exclude-from-count" to the &lt;quer:projection&gt;, for
 * example:
 * 
 * <pre>
 * 	&lt;quer:projection tag="exclude-from-count"&gt;
 * 	  &lt;quer:field path="AttachedFiles,Content"/&gt;
 * 	&lt;/quer:projection&gt;
 * </pre>
 * 
 * This step must be defined before the 'Prepare Export' step. <br>
 * 
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * <br>
 * <br>
 * Here is an example of how to define this custom step in the TCC configuration
 * file;
 * <p>
 * <img src="doc-files/ConvertToCountPreStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class ConvertToCountPreStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(ConvertToCountPreStep.class);

	/** The XSL transformation file. */
	private final static String XSL_TRANSFORM = "xsl/ConvertToCount.xsl";

	/**
	 * Constructor without arguments.
	 */
	public ConvertToCountPreStep() {
		super();
		// Register valid parameters.
		registerParameterName("csvDelimiter");
		registerParameterName("csvHeaderPresent");
		registerParameterName("entityName");
	}

	/**
	 * The CSV delimiter.
	 */
	private char csvDelimiter = CSVConstant.DEFAULT_DELIMITER;

	/**
	 * CSV header present indicator.
	 */
	private boolean csvHeaderPresent = false;

	/**
	 * CSV header present indicator.
	 */
	private String entityName;

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "ConvertToCountPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Pre-processing step to convert any query into a count query.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "ConvertToCountPreStep";
	}

	/**
	 * @return
	 */
	public char getCsvDelimiter() {
		return csvDelimiter;
	}

	/**
	 * @param csvDelimiter
	 */
	public void setCsvDelimiter(char csvDelimiter) {
		this.csvDelimiter = csvDelimiter;
	}

	/**
	 * Get the CSV header present indicator.
	 * 
	 * @return
	 */
	public boolean isCsvHeaderPresent() {
		return csvHeaderPresent;
	}

	/**
	 * Set the CSV header present indicator.
	 * 
	 * @param csvHeaderPresent
	 */
	public void setCsvHeaderPresent(boolean csvHeaderPresent) {
		this.csvHeaderPresent = csvHeaderPresent;
	}

	/**
	 * Get the entity name.
	 * 
	 * @return
	 */
	public String getEntityName() {
		return entityName;
	}

	/**
	 * Set the entity name.
	 * 
	 * @param entityName
	 */
	public void setEntityName(String entityName) {
		this.entityName = entityName;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// Get the supported pipeline.
		SupportedPipeline supportedPipeline = getPreProcessSupportPipeline();

		// Return the supported pipeline.
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		String csvDelimiter = (parameters.getProperty("csvDelimiter"));
		if (StringUtils.isNotBlank(csvDelimiter)) {
			char[] chars = csvDelimiter.toCharArray();
			if (chars.length > 1) {
				throw new IllegalArgumentException(
						"The 'csvDelimiter' parameter must be a single character.");
			}
			setCsvDelimiter(chars[0]);
		}

		String csvHeaderPresent = (parameters.getProperty("csvHeaderPresent"));
		if (StringUtils.isNotBlank(csvHeaderPresent)) {
			setCsvHeaderPresent(Boolean.valueOf(csvHeaderPresent)
					.booleanValue());
		}

		setEntityName(parameters.getProperty("entityName"));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the request info.
			Message requestInfo = pipeline.getRequestMessage();
			String queryFile = requestInfo.getCurrentFile();
			if (queryFile == null) {
				queryFile = requestInfo.getOriginalFile();
			}
			File inFile = new File(queryFile);

			// Transform the query file.
			File outFile = createTempFile();
			transform(inFile, outFile);

			// Set the transformed query in the request.
			requestInfo.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Transform the query file containing into a count query.
	 * 
	 * @param queryFile
	 *            The query file.
	 * @param outFile
	 *            The output file.
	 * @throws TransformerException
	 * @throws IOException
	 */
	protected void transform(File queryFile, File outFile)
			throws TransformerException, IOException {
		log.info("Transforming the query into a count query.");

		// Create an XSL transformer.
		InputStream xsl = this.getClass().getClassLoader()
				.getResourceAsStream(XSL_TRANSFORM);
		TransformerFactory tFactory = TransformerFactory.newInstance();
		Transformer transformer = tFactory
				.newTransformer(new StreamSource(xsl));

		// Set parameters.
		transformer.setParameter("csvdelimiter",
				Character.toString(csvDelimiter));
		transformer.setParameter("csvheader",
				Boolean.toString(csvHeaderPresent));
		transformer.setParameter("entityName", entityName);

		// Prepare streams.
		InputStream is = IOHelper.getInputStream(queryFile);
		OutputStream os = IOHelper.getOutputStream(
				new FileOutputStream(outFile), getClass().getName());

		// Prepare source and result.
		Source source = new StreamSource(is);
		Result result = new StreamResult(os);

		// Apply the transformation.
		transformer.transform(source, result);

		// Close.
		is.close();
		os.close();
	}

}
