package com.taleo.integration.client.customstep.count;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.TimeZone;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.time.DateUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.Constants;

/**
 * Manage the record count.
 * <p>
 * Count will be reset, overwriting the daily count file, each day at 00:00:00
 * in the "America/New_York" time zone. This time zone adapts to daylight
 * savings and is the one all Taleo servers are set and should not need to
 * change. However, if needed, the time zone can be set differently using the
 * System property called
 * "com.taleo.client.customstep.CountManager.ResetTimeZone".
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public class CountManager {

	private static Logger log = Logger.getLogger(CountManager.class);

	/** Key for the CountManager in the pipeline. */
	public static final Object PIPELINE_KEY = "CountManager";

	/** Date format to parse dates. */
	private static final DateFormat dateFormat = new SimpleDateFormat(
			"yyyy-MM-dd HH:mm:ss");
	private static final DateFormat dateFormatFilename = new SimpleDateFormat(
			"yyyy-MM-dd");

	/**
	 * Constant for the ResetTimeZone system property.
	 */
	private static final String PROP_RESET_TIME_ZONE = "com.taleo.client.customstep.CountManager.ResetTimeZone";

	/**
	 * Default value for com.taleo.client.customstep.CountManager.ResetTimeZone
	 * system property.
	 */
	public static final String DEFAULT_RESET_TIME_ZONE = "America/New_York";

	/**
	 * The time zone used for resetting the daily count.
	 */
	private static TimeZone resetTimeZone;
	static {
		String tz = DEFAULT_RESET_TIME_ZONE;
		String prop = System.getProperty(PROP_RESET_TIME_ZONE);
		if (StringUtils.isNotBlank(prop)) {
			tz = prop;
		}
		resetTimeZone = TimeZone.getTimeZone(tz);
	}

	/**
	 * The list of runs.
	 */
	private List runs = new ArrayList();

	/**
	 * The current run.
	 */
	private Run currentRun;

	/**
	 * The file to store the count state.
	 */
	private File file;

	/**
	 * Constructor without arguments.
	 */
	public CountManager() {
		// Add an empty current run.
		this.currentRun = new Run();
		this.runs.add(currentRun);
	}

	/**
	 * Constructor with the file to store the count.
	 * 
	 * @param file
	 *            The file to store the count state.
	 */
	public CountManager(File file) throws IOException {
		if (file == null) {
			throw new IllegalArgumentException(
					"Parameter 'file' cannot be null");
		}
		this.file = file;

		boolean reset = false;
		Date lastModified = null;
		if (file.exists()) {

			// Determine if the file must be reset.
			Calendar cal = Calendar.getInstance();
			Date now = cal.getTime();
			cal.setTimeInMillis(file.lastModified());
			lastModified = cal.getTime();
			reset = mustReset(now, lastModified);

			if (reset) {
				// Rename the count file with its last modified date.
				String filename = file.getPath() + "."
						+ dateFormatFilename.format(lastModified);
				log.info("Renaming daily count file to '" + filename + "'.");
				file.renameTo(new File(filename));

			} else {
				// Load the file.
				BufferedReader reader = null;
				try {
					reader = new BufferedReader(new InputStreamReader(
							new FileInputStream(file), Constants.UTF8));
					// Skip header.
					reader.readLine();

					// Read each line.
					String line;
					while ((line = reader.readLine()) != null) {
						if (line.trim().length() > 0) {
							Run run = new Run(line);
							this.runs.add(run);
						}
					}
				} finally {
					if (reader != null) {
						reader.close();
					}
				}
			}
		}

		// Add an empty current run.
		this.currentRun = new Run();
		this.runs.add(currentRun);
	}

	/**
	 * Determine if a count file with the given last modified date must be
	 * reset. It will be if it is in a previous day in the specified reset time
	 * zone.
	 * 
	 * @param now
	 * @param lastModifiedDate
	 * @return
	 */
	protected boolean mustReset(Date now, Date lastModifiedDate) {
		Calendar _now = Calendar.getInstance(resetTimeZone);
		_now.setTime(now);
		Calendar _lastModifiedDate = Calendar.getInstance(resetTimeZone);
		_lastModifiedDate.setTime(lastModifiedDate);
		return (!DateUtils.isSameDay(_now, _lastModifiedDate));
	}

	/**
	 * Get the next reset time in the local time zone.
	 * 
	 * @return
	 */
	public Date getNextResetTime() {
		Calendar cal = Calendar.getInstance();
		return getNextResetTime(cal.getTime());
	}

	/**
	 * Get the next reset time in the local time zone.
	 * 
	 * @param now
	 * @return
	 */
	protected Date getNextResetTime(Date now) {
		Calendar reset = Calendar.getInstance(resetTimeZone);
		reset.setTime(now);
		reset.set(Calendar.HOUR_OF_DAY, 24);
		reset.set(Calendar.MINUTE, 0);
		reset.set(Calendar.SECOND, 0);

		Calendar local = Calendar.getInstance();
		local.setTime(reset.getTime());
		return local.getTime();
	}

	/**
	 * Set the current extract Id.
	 */
	public void setCurrentExtractId(String extractId) {
		currentRun.extractId = extractId;
	}

	/**
	 * Get the current extract Id.
	 * 
	 * @return
	 */
	public String getCurrentExtractId() {
		return currentRun.extractId;
	}

	/**
	 * Set the current entity.
	 */
	public void setCurrentEntity(String entity) {
		currentRun.entity = entity;
	}

	/**
	 * Get the current entity.
	 * 
	 * @return
	 */
	public String getCurrentEntity() {
		return currentRun.entity;
	}

	/**
	 * Set the current execution date.
	 */
	public void setCurrentDate(Date date) {
		currentRun.date = date;
	}

	/**
	 * Get the current execution date.
	 * 
	 * @return
	 */
	public Date getCurrentDate() {
		return currentRun.date;
	}

	/**
	 * Set the current record count.
	 */
	public void setCurrentRecordCount(int recordCount) {
		currentRun.recordCount = recordCount;
	}

	/**
	 * Get the current record count.
	 * 
	 * @return The current record count or -1 if there is no current record
	 *         count.
	 */
	public int getCurrentRecordCount() {
		return currentRun.recordCount;
	}

	/**
	 * Get the daily completed run count (excluding the current run).
	 * 
	 * @return
	 */
	public int getDailyRunCount() {
		return runs.size() - 1;
	}

	/**
	 * Get the daily count.
	 * 
	 * @return
	 */
	public int getDailyRecordCount() {
		int count = 0;
		for (Iterator iterator = runs.iterator(); iterator.hasNext();) {
			Run run = (Run) iterator.next();
			if (run.recordCount >= 0) {
				count += run.recordCount;
			}
		}
		return count;
	}

	/**
	 * Get the file where the count state is stored.
	 * 
	 * @return
	 */
	public File getFile() {
		return this.file;
	}

	/**
	 * Save the count state on disk.
	 * 
	 * @throws IOException
	 */
	public void save() throws IOException {
		log.info("Saving daily count file '" + this.file.getAbsolutePath()
				+ "'.");

		BufferedWriter writer = null;
		try {
			writer = new BufferedWriter(new FileWriter(this.file));
			// Write header.
			writer.write("extractId|entity|date|recordCount");
			writer.newLine();

			// Write each run.
			for (Iterator i = runs.iterator(); i.hasNext();) {
				Run run = (Run) i.next();
				writer.write(run.toString());
				writer.newLine();
			}
			writer.flush();
		} finally {
			if (writer != null) {
				writer.close();
			}
		}
	}

	/**
	 * Class to hold info for each run.
	 */
	private class Run {

		/**
		 * The extract Id.
		 */
		public String extractId = "";

		/**
		 * The entity name.
		 */
		public String entity = "";

		/**
		 * The execution date.
		 */
		public Date date = new Date();

		/**
		 * Record count.
		 */
		public int recordCount = -1;

		/**
		 * Constructor without arguments.
		 */
		public Run() {
		}

		/**
		 * Constructor with a line from the file.
		 * 
		 * @param line
		 */
		private Run(String line) {
			try {
				String[] arr = line.split("\\|");
				this.extractId = arr[0];
				this.entity = arr[1];
				this.date = dateFormat.parse(arr[2]);
				this.recordCount = Integer.parseInt(arr[3]);

			} catch (Exception e) {
				log.warn("Invalid line: '" + line
						+ "' is causing this exception: " + e.toString());
			}
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#toString()
		 */
		public String toString() {
			return extractId + "|" + entity + "|" + dateFormat.format(date)
					+ "|" + recordCount;
		}
	}

}
