package com.taleo.integration.client.customstep.count;

import java.io.File;
import java.io.IOException;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.workflow.WorkflowHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedPipeline;
import com.taleo.ws.integration.client.WorkflowError;
import com.taleo.ws.integration.client.WorkflowErrorCode;

/**
 * Pre-processing step interrupt execution when a given maximum of records has
 * been extracted on the current day.
 * <p>
 * The record count is not global to the zone, it is specific to this instance
 * and the count filename provided as parameter.
 * <p>
 * This pre-processing step works in pair with {@link ExportCountPostStep}. It
 * uses the count that this step has stored on disk to determine if the maximum
 * has been reached.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>countFilename</b>: Mandatory - Name of the file where the daily count
 * is stored. Can be an absolute or relative path or only a file name. If only a
 * file name, the file is searched for in the workflow temporary folder.</li>
 * <br>
 * <br>
 * <li><b>maximumCount</b>: Optional - The maximum number of records to extract
 * on a given day. If absent or 0, no maximum applies.</li>
 * </ul>
 * <p>
 * When the maximum is reached, the process exits abruptly with code 1. All
 * subsequent steps are then ignored. The exit code can be set differently using
 * the system property named
 * <code>com.taleo.client.customstep.DailyCountPreStep.MaximumExitCode</code>.
 * <p>
 * <b>Note:</b> Exiting with a specific code is currently the only option to end
 * TCC on demand. The down side is that the monitoring file remains incomplete
 * and some temporary files for that run are not deleted.
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/DailyCountPreStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 * 
 */
public class DailyCountPreStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(DailyCountPreStep.class);

	/**
	 * Constant for the Maximum Exit Code system property.
	 */
	private static final String PROP_MAXIMUM_EXIT_CODE = "com.taleo.client.customstep.DailyCountPreStep.MaximumExitCode";

	/**
	 * Default value for
	 * com.taleo.client.customstep.DailyCountPreStep.MaximumExitCode system
	 * property.
	 */
	private static final int DEFAULT_MAXIMUM_EXIT_CODE = 1;

	/**
	 * Constructor without arguments.
	 */
	public DailyCountPreStep() {
		super();
		// Register valid parameters.
		registerParameterName("countFilename");
		registerParameterName("maximumCount");
	}

	/**
	 * The maximum count.
	 */
	protected int maximumCount = 0;

	/**
	 * The count filename.
	 */
	private String countFilename;

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "DailyCountPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Pre-processing step for setting a maximum number of records daily.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "DailyCountPreStep";
	}

	/**
	 * Get the maximum count.
	 * 
	 * @return
	 */
	public int getMaximumCount() {
		return maximumCount;
	}

	/**
	 * Set the maximum count.
	 * 
	 * @return
	 */
	public void setMaximumCount(int maximumCount) {
		if (maximumCount < 0) {
			throw new IllegalArgumentException(
					"The 'maximumCount' parameter must be >= 0");
		}
		this.maximumCount = maximumCount;
	}

	/**
	 * Get the count filename.
	 * 
	 * @return
	 */
	public String getCountFilename() {
		return countFilename;
	}

	/**
	 * Set the count filename.
	 * 
	 * @param countFilename
	 */
	public void setCountFilename(String countFilename) {
		this.countFilename = countFilename;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// Get the supported pipeline.
		SupportedPipeline supportedPipeline = getPreProcessSupportPipeline();

		// Return the supported pipeline.
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);
		
		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		try {
			setMaximumCount(Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("maximumCount"), "0")));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The maximumCount parameter must be numeric.");
		}

		setCountFilename(parameters.getProperty("countFilename"));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Quit if maximum count is 0.
			if (this.maximumCount == 0) {
				log.info("No check of daily count since maximumCount is not defined.");
				return;
			}

			// Create a CountManager.
			if (this.countFilename == null) {
				throw new StepException(
						"Parameter 'countFilename' is mandatory.");
			}
			CountManager countManager = null;
			File countFile = new File(this.countFilename);
			if (countFile.getParent() == null) {
				countManager = new CountManager(new File(getTempFolder(),
						this.countFilename));
			} else {
				countManager = new CountManager(countFile);
			}

			log.info("Using the following count file: '"
					+ countFile.getAbsolutePath() + "'.");

			log.info("Daily record count: "
					+ countManager.getDailyRecordCount() + ".");

			// Exit if maximum count has been reached.
			if (countManager.getDailyRecordCount() >= this.maximumCount) {

				// Note: TCC currently does not provide any means to abort
				// execution. The only way is to kill the process.
				int exitCode = DEFAULT_MAXIMUM_EXIT_CODE;
				String prop = System.getProperty(PROP_MAXIMUM_EXIT_CODE);
				if (StringUtils.isNotBlank(prop)) {
					try {
						exitCode = Integer.parseInt(prop);
					} catch (NumberFormatException e) {
						// Invalid property value: keep default value.
						log.warn("Invalid value '" + prop
								+ "' for system property "
								+ PROP_MAXIMUM_EXIT_CODE);
					}
				}
				log.info("The run was aborted since "
						+ countManager.getDailyRecordCount()
						+ " records were extracted today and the maximum is set to "
						+ this.maximumCount
						+ ". Next reset of the counter will occur "
						+ countManager.getNextResetTime() + ". "
						+ "Exiting with code: " + exitCode + "\n");
				System.exit(exitCode);
			}

		} catch (IOException e) {
			WorkflowError error = WorkflowHelper.createWorkflowError(
					WorkflowErrorCode.IO, e);
			throw new StepException(error);

		} catch (Exception e) {
			throw new StepException(e);
		}
	}
}
