package com.taleo.integration.client.customstep.count;

import java.io.File;
import java.io.IOException;
import java.util.Date;
import java.util.Iterator;
import java.util.Properties;

import javax.xml.transform.TransformerConfigurationException;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.ElementHandler;
import org.dom4j.ElementPath;
import org.dom4j.io.SAXReader;
import org.xml.sax.SAXException;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedMessage;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Post processing step to count the records coming out directly from the zone,
 * before any client transformation.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>countFilename</b>: Optional - Name of the file to store the daily
 * count for subsequent runs. Can be an absolute or relative path or only a file
 * name. If only a file name, the file is searched for in the workflow temporary
 * folder. If absent, the count will not be saved and only the current count
 * will be available to the workflow.</li>
 * <br>
 * <br>
 * <li><b>extractId</b>: Optional - ID of the current extract, will display in
 * the daily count file.</li>
 * </ul>
 * <p>
 * This step takes only T-SOAP format and gets the count for XML or CSV content.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://dom4j.org/dom4j-1.6.1/">dom4j library</a>
 * (dom4j-1.6.1.jar)</li>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/ExportCountPostStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 * 
 */
public class ExportCountPostStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(ExportCountPostStep.class);

	/**
	 * Constructor without arguments.
	 */
	public ExportCountPostStep() {
		super();
		// Register valid parameters.
		registerParameterName("countFilename");
		registerParameterName("extractId");
	}

	/**
	 * The count filename.
	 */
	private String countFilename;

	/**
	 * The extract Id.
	 */
	private String extractId;

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "ExportCountPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step for counting records resulting from an export.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "ExportCountPostStep";
	}

	/**
	 * Get the count filename.
	 * 
	 * @return
	 */
	public String getCountFilename() {
		return countFilename;
	}

	/**
	 * Set the count filename.
	 * 
	 * @param countFilename
	 */
	public void setCountFilename(String countFilename) {
		this.countFilename = countFilename;
	}

	/**
	 * Get the extract Id.
	 * 
	 * @return
	 */
	public String getExtractId() {
		return extractId;
	}

	/**
	 * Set the extract Id.
	 * 
	 * @param extractId
	 */
	public void setExtractId(String extractId) {
		this.extractId = extractId;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// Get the supported pipeline.
		SupportedPipeline supportedPipeline = getPostProcessSupportPipeline();

		// Add the supported request message format.
		SupportedMessage supportedMessage = supportedPipeline
				.addNewSupportedRequestMessage();
		supportedMessage.addNewSupportedFormats();
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.T_SOAP);

		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		setCountFilename(parameters.getProperty("countFilename"));
		setExtractId(parameters.getProperty("extractId"));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// If not active, return immediately.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Create the CountManager and store it in the pipeline.
			CountManager countManager = null;
			if (StringUtils.isBlank(this.countFilename)) {
				countManager = new CountManager();
			} else {
				File countFile = new File(this.countFilename);
				if (countFile.getParent() == null) {
					countManager = new CountManager(new File(getTempFolder(),
							this.countFilename));
				} else {
					countManager = new CountManager(countFile);
				}
			}
			pipeline.documentProperties().put(CountManager.PIPELINE_KEY,
					countManager);

			// Get the input file.
			Message responseMessage = pipeline.getResponseMessage();
			File inFile = new File(responseMessage.getCurrentFile());

			// Attributes from T-SOAP document.
			Properties attributes = new Properties();

			// Execute.
			execute(inFile, attributes);

			// Add attributes to the count manager.
			countManager.setCurrentDate(new Date());
			countManager.setCurrentExtractId(extractId);
			countManager.setCurrentEntity((String) attributes.get("entity"));
			countManager.setCurrentRecordCount(Integer
					.parseInt((String) attributes.get("count")));

			log.info("Current record count: "
					+ countManager.getCurrentRecordCount());

			// Save the count manager.
			if (countManager.getFile() != null) {
				countManager.save();
			}

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Execute over the input file.
	 * 
	 * @param inputFile
	 *            The input file.
	 * @param attributes
	 *            The attributes from T-SOAP document.
	 * @throws TransformerConfigurationException
	 * @throws DocumentException
	 * @throws IOException
	 */
	protected void execute(File inputFile, Properties attributes)
			throws TransformerConfigurationException, DocumentException,
			IOException {

		// Create a SAXReader with a handler on the attributes.
		SAXReader reader = new SAXReader();
		reader.addHandler(
				"/Envelope/Body/getDocumentByKeyResponse/Document/Attributes",
				new ExportCountElementHandler(attributes));

		// Read the input file.
		try {
			reader.read(IOHelper.getInputStream(inputFile), getEncoding());
		} catch (DocumentException e) {
			if (e.getNestedException().getCause() instanceof EndOfProcessingException) {
				// Do nothing.
			} else {
				throw e;
			}
		}
	}

	/**
	 * Exception to end processing file as soon as the count is found.
	 */
	private class EndOfProcessingException extends SAXException {
		private static final long serialVersionUID = 1L;

		public EndOfProcessingException() {
			super();
		}
	}

	/**
	 * Class to get the count from the T-SOAP result.
	 */
	private class ExportCountElementHandler implements ElementHandler {

		private Properties attributes;

		/**
		 * Constructor
		 * 
		 * @throws TransformerConfigurationException
		 */
		ExportCountElementHandler(Properties attributes)
				throws TransformerConfigurationException {
			this.attributes = attributes;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.dom4j.ElementHandler#onStart(org.dom4j.ElementPath)
		 */
		public void onStart(ElementPath path) {
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.dom4j.ElementHandler#onEnd(org.dom4j.ElementPath)
		 */
		public void onEnd(ElementPath path) {

			try {
				// Retrieve the current element.
				Element element = path.getCurrent();
				if (log.isDebugEnabled()) {
					log.debug("Reading: " + element.asXML());
				}

				// Extract attribute values.
				for (Iterator i = element.elementIterator(); i.hasNext();) {
					Element attribute = (Element) i.next();
					attributes.put(attribute.attributeValue("name"),
							attribute.getText());
				}

				// Free memory.
				element.detach();

				// Throw an exception to terminate parsing.
				throw new EndOfProcessingException();

			} catch (EndOfProcessingException e) {
				throw new RuntimeException(e);
			} catch (Exception e) {
				// Log the error, because details are not carried by SAX
				// parsing.
				log.error("", e);
				throw new RuntimeException(e);
			}
		}
	}

}
