package com.taleo.integration.client.customstep.csv;

import java.io.File;
import java.io.IOException;
import java.util.Comparator;

import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.customstep.util.csv.MergeSort;
import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Abstract processing step to sort CSV data.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvStep}.
 * <p>
 * 
 * @author Romain Guay, Taleo Corporation
 */
public abstract class AbstractSortStep extends BaseCustomCsvStep {

	private static Logger log = Logger.getLogger(AbstractSortStep.class);

	/**
	 * Constructor without arguments.
	 */
	public AbstractSortStep() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			boolean isPreProcess = true;

			// Get the input file.
			File inFile = null;
			Message responseMessage = pipeline.getResponseMessage();
			Message requestMessage = pipeline.getRequestMessage();
			if (responseMessage.isSetCurrentFile()) {
				inFile = new File(responseMessage.getCurrentFile());
				isPreProcess = false;
			} else {
				if (requestMessage.isSetCurrentFile()) {
					inFile = new File(requestMessage.getCurrentFile());
				} else {
					inFile = new File(requestMessage.getOriginalFile());
				}
			}

			// Create a temporary file for the output.
			File outFile = createTempFile();

			// Apply the transformation.
			execute(inFile, outFile);

			// Set the new current file.
			if (isPreProcess) {
				requestMessage.setCurrentFile(outFile.getAbsolutePath());
			} else {
				responseMessage.setCurrentFile(outFile.getAbsolutePath());
			}

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Copy the input file to the output file while applying the required
	 * sorting.
	 * 
	 * @param inFile
	 *            The input file
	 * @param outFile
	 *            The output file
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	public void execute(File inFile, File outFile) throws IOException,
			InvalidCSVFileFormat {

		MergeSort sort = new MergeSort();
		sort.setCsvHeaderPresent(isCsvHeaderPresent());
		sort.setCsvDelimiter(getCsvDelimiter());
		sort.setCsvQuoteCharacter(getCsvQuoteCharacter());
		sort.setRemoveDuplicates(false);
		sort.setEncoding(getEncoding());
		sort.setWorkingFolder(tempFolder);

		// Do the sort.
		sort.sort(inFile, getComparator(inFile), outFile);
	}

	/**
	 * Get the comparator for the CSV rows.
	 * <p>
	 * The comparator must take String[] or List as arguments, each representing
	 * a CSV row.
	 * 
	 * @param file
	 *            The CSV file.
	 * @return
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	protected abstract Comparator getComparator(File file) throws IOException,
			InvalidCSVFileFormat;

}
