package com.taleo.integration.client.customstep.csv;

import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;

import org.apache.commons.lang.StringUtils;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.csv.CsvUtils;
import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.integration.client.util.QueryRuntimeHelper;
import com.taleo.integration.client.util.csv.CSVConstant;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.query.QueryDocument.Query;

/**
 * Abstract custom step for all processing of CSV files.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>csvDelimiter</b>: Optional - The CSV delimiter. Default is a comma
 * (,).</li>
 * <br>
 * <br>
 * <li><b>csvQuoteCharacter</b>: Optional - The CSV quote character. Default is
 * a double quote (").</li>
 * <br>
 * <br>
 * <li><b>csvHeaderPresent</b>: Optional - The CSV header presence indicator.
 * Default is <code>false</code>.</li>
 * </ul>
 * <p>
 * In the case of an export, those parameters can be read from the query file
 * which is accessible from the custom steps. But there are cases where we need
 * to specify them here. For example, if the extract is done in XML and
 * transformed in CSV after, the query will not contain those values. If values
 * are specified in the query and as parameters, the parameters have precedence.
 * Ultimately, if they are not specified at all, the defaults apply.
 * <p>
 * In the case of an import, the load specification is not accessible from
 * within the custom steps. Those parameters must then be provided otherwise the
 * defaults will be used.
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public abstract class BaseCustomCsvStep extends BaseCustomStep {

	/**
	 * Constructor without arguments.
	 */
	public BaseCustomCsvStep() {
		super();
		// Register valid parameters.
		registerParameterName("csvDelimiter");
		registerParameterName("csvQuoteCharacter");
		registerParameterName("csvHeaderPresent");
	}

	/**
	 * The global config.
	 */
	private GlobalConfig config;

	/**
	 * The CSV delimiter.
	 */
	private char csvDelimiter = CSVConstant.DEFAULT_DELIMITER;

	/**
	 * The CSV quote character.
	 */
	private char csvQuoteCharacter = CSVConstant.DEFAULT_QUOTE_CHARACTER;

	/**
	 * The CSV header presence indicator.
	 */
	private boolean csvHeaderPresent = false;

	/**
	 * @return
	 */
	public char getCsvDelimiter() {
		return csvDelimiter;
	}

	/**
	 * @param csvDelimiter
	 */
	public void setCsvDelimiter(char csvDelimiter) {
		this.csvDelimiter = csvDelimiter;
	}

	/**
	 * @return
	 */
	public boolean isCsvHeaderPresent() {
		return csvHeaderPresent;
	}

	/**
	 * @param csvHeaderPresent
	 */
	public void setCsvHeaderPresent(boolean csvHeaderPresent) {
		this.csvHeaderPresent = csvHeaderPresent;
	}

	/**
	 * @return
	 */
	public char getCsvQuoteCharacter() {
		return csvQuoteCharacter;
	}

	/**
	 * @param csvQuoteCharacter
	 */
	public void setCsvQuoteCharacter(char csvQuoteCharacter) {
		this.csvQuoteCharacter = csvQuoteCharacter;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);
		this.config = config;

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		String csvHeaderPresent = (parameters.getProperty("csvHeaderPresent"));
		if (StringUtils.isNotBlank(csvHeaderPresent)) {
			setCsvHeaderPresent(Boolean.valueOf(csvHeaderPresent)
					.booleanValue());
		}

		String csvDelimiter = (parameters.getProperty("csvDelimiter"));
		if (StringUtils.isNotBlank(csvDelimiter)) {
			char[] chars = csvDelimiter.toCharArray();
			if (chars.length > 1) {
				throw new IllegalArgumentException(
						"The 'csvDelimiter' parameter must be a single character.");
			}
			setCsvDelimiter(chars[0]);
		}

		String csvQuoteCharacter = (parameters.getProperty("csvQuoteCharacter"));
		if (StringUtils.isNotBlank(csvQuoteCharacter)) {
			char[] chars = csvQuoteCharacter.toCharArray();
			if (chars.length > 1) {
				throw new IllegalArgumentException(
						"The 'csvQuoteCharacter' parameter must be a single character.");
			}
			setCsvQuoteCharacter(chars[0]);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Get the request file format.
		MessageFormat.Enum requestFormat = config.getGeneral()
				.getRequestMessage().getFormat();

		// Get the response file format.
		MessageFormat.Enum responseFormat = config.getGeneral()
				.getResponseMessage().getFormat();

		// Get the CSV parameters in the query in the case of a CSV export.
		if (requestFormat.equals(MessageFormat.SQ_XML)
				&& responseFormat.equals(MessageFormat.CSV)) {
			File requestFile = new File(pipeline.getRequestMessage()
					.getOriginalFile());
			Query query = QueryRuntimeHelper.loadQueryInStep(requestFile,
					productCode);

			setCsvHeaderPresent(query.getCsvheader());
			if (query.getCsvdelimiter() != null) {
				setCsvDelimiter(query.getCsvdelimiter().charAt(0));
			}
			if (query.getCsvquote() != null) {
				setCsvQuoteCharacter(query.getCsvquote().charAt(0));
			}
		}

		// Override with the parameters if provided.
		String csvHeaderPresent = (parameters.getProperty("csvHeaderPresent"));
		if (StringUtils.isNotBlank(csvHeaderPresent)) {
			setCsvHeaderPresent(Boolean.valueOf(csvHeaderPresent)
					.booleanValue());
		}

		String csvDelimiter = (parameters.getProperty("csvDelimiter"));
		if (StringUtils.isNotBlank(csvDelimiter)) {
			char[] chars = csvDelimiter.toCharArray();
			if (chars.length > 1) {
				throw new IllegalArgumentException(
						"The 'csvDelimiter' parameter must be a single character.");
			}
			setCsvDelimiter(chars[0]);
		}

		String csvQuoteCharacter = (parameters.getProperty("csvQuoteCharacter"));
		if (StringUtils.isNotBlank(csvQuoteCharacter)) {
			char[] chars = csvQuoteCharacter.toCharArray();
			if (chars.length > 1) {
				throw new IllegalArgumentException(
						"The 'csvQuoteCharacter' parameter must be a single character.");
			}
			setCsvQuoteCharacter(chars[0]);
		}
	}

	/**
	 * Get the index of the specified column in the file. The column may already
	 * contain the index number or may contain the name of the column.
	 * 
	 * @param file
	 *            The CSV file
	 * @param column
	 *            The name or index (0 based) of the column
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	protected int getColumnIndex(File file, String column) throws IOException,
			InvalidCSVFileFormat {
		Reader reader = null;
		String[] header = null;
		try {
			if (isCsvHeaderPresent()) {
				// Get the file header.
				reader = new InputStreamReader(IOHelper.getInputStream(file),
						getEncoding());
				header = CsvUtils.readFirstLine(reader, getCsvDelimiter(),
						getCsvQuoteCharacter());
			}

			return CsvUtils.getColumnIndex(column, header);
		} finally {
			if (reader != null) {
				reader.close();
			}
		}
	}

	/**
	 * Get the name of the specified column in the file. The column may already
	 * contain the name of may also contain the index of the column.
	 * 
	 * @param file
	 *            The CSV file
	 * @param column
	 *            The name or index (0 based) of the column
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	protected String getColumnName(File file, String column)
			throws IOException, InvalidCSVFileFormat {

		if (!isCsvHeaderPresent()) {
			// There is no header, hence no column name.
			return null;

		} else {

			try {
				// Try to parse the column as a number.
				int index = Integer.parseInt(column);

				// It's a number, so get the column name.
				Reader reader = null;
				try {
					// Get the file header.
					reader = new InputStreamReader(
							IOHelper.getInputStream(file), getEncoding());
					String[] header = CsvUtils.readFirstLine(reader,
							getCsvDelimiter(), getCsvQuoteCharacter());

					// Get the column name.
					try {
						return (String) header[index];
					} catch (IndexOutOfBoundsException e) {
						throw new IllegalArgumentException("The column index '"
								+ column + "' is out of bounds.");
					}
				} finally {
					if (reader != null) {
						reader.close();
					}
				}

			} catch (NumberFormatException e) {
				// Return the column name as is.
				return column;
			}
		}
	}

}
