package com.taleo.integration.client.customstep.csv;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.StringReader;
import java.util.Properties;

import org.apache.commons.lang.StringUtils;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.runtime.RuntimeConstants;
import org.apache.velocity.runtime.log.NullLogChute;
import org.apache.velocity.runtime.resource.loader.FileResourceLoader;

import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;

/**
 * Abstract post-processing step to add a header and/or footer to a CSV file.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvPostStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>headerTemplateFile</b>: The name of a template file with the contents
 * of the header.</li>
 * <br>
 * <br>
 * <li><b>headerTemplate</b>: The template string for the header. This may be
 * used in replacement of a <code>headerTemplateFile</code> if the template is
 * very short. If both are present, this one has priority.</li>
 * <br>
 * <br>
 * <li><b>footerTemplateFile</b>: The name of a template file with the contents
 * of the footer.</li>
 * <br>
 * <br>
 * <li><b>footerTemplate</b>: The template string for the footer. This may be
 * used in replacement of a <code>footerTemplateFile</code> if the template is
 * very short. If both are present, this one has priority.</li>
 * </ul>
 * <p>
 * Details of the variables available to write the template are specific to
 * concrete subclasses of this one.
 * <p>
 * The files can be specified as an absolute path or relative to the working
 * directory.
 * <p>
 * This class and its subclasses are dependent on the <a
 * href="http://velocity.apache.org/">Apache Velocity</a> library version 1.5.
 * 
 * @author Romain Guay, Taleo Corporation
 */
public abstract class BaseHeaderFooterPostStep extends BaseCustomCsvPostStep {

	/**
	 * Constructor without arguments.
	 */
	public BaseHeaderFooterPostStep() {
		super();
		// Register valid parameters.
		registerParameterName("headerTemplateFile");
		registerParameterName("headerTemplate");
		registerParameterName("footerTemplateFile");
		registerParameterName("footerTemplate");

		// Create the velocity engine.
		this.velocityEngine = new VelocityEngine();
		Properties props = new Properties();
		// Disable default logging.
		props.put(RuntimeConstants.RUNTIME_LOG_LOGSYSTEM_CLASS,
				NullLogChute.class.getName());
		props.put("resource.loader", "file");
		props.put("file.resource.loader.class",
				FileResourceLoader.class.getName());
		// Empty loader path allows specifying templates with an
		// absolute path.
		props.put("file.resource.loader.path", "");
		props.put("file.resource.loader.cache", "false");
		try {
			velocityEngine.init(props);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * The header template file.
	 */
	private String headerTemplateFile;

	/**
	 * The header template string.
	 */
	private String headerTemplate;

	/**
	 * The footer template file.
	 */
	private String footerTemplateFile;

	/**
	 * The footer template string.
	 */
	private String footerTemplate;

	/**
	 * The Velocity engine.
	 */
	private VelocityEngine velocityEngine;

	/**
	 * @return
	 */
	public String getHeaderTemplateFile() {
		return headerTemplateFile;
	}

	/**
	 * @param templateFile
	 */
	public void setHeaderTemplateFile(String templateFile) {
		this.headerTemplateFile = templateFile;
	}

	/**
	 * @return
	 */
	public String getHeaderTemplate() {
		return headerTemplate;
	}

	/**
	 * @param template
	 */
	public void setHeaderTemplate(String template) {
		this.headerTemplate = template;
	}

	/**
	 * @return
	 */
	public String getFooterTemplateFile() {
		return footerTemplateFile;
	}

	/**
	 * @param templateFile
	 */
	public void setFooterTemplateFile(String templateFile) {
		this.footerTemplateFile = templateFile;
	}

	/**
	 * @return
	 */
	public String getFooterTemplate() {
		return footerTemplate;
	}

	/**
	 * @param template
	 */
	public void setFooterTemplate(String template) {
		this.footerTemplate = template;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		headerTemplateFile = parameters.getProperty("headerTemplateFile");

		headerTemplate = parameters.getProperty("headerTemplate");

		footerTemplateFile = parameters.getProperty("footerTemplateFile");

		footerTemplate = parameters.getProperty("footerTemplate");
	}

	/**
	 * Copy the input file to the output file while adding the header and/or
	 * footer.
	 * 
	 * @param inFile
	 *            The input file
	 * @param outFile
	 *            The output file
	 * @throws Exception
	 */
	public void execute(File inFile, File outFile) throws Exception {
		// Consider the outFile as target.
		execute(inFile, outFile, outFile);
	}

	/**
	 * Copy the input file to the output file while adding the header and/or
	 * footer.
	 * 
	 * @param inFile
	 *            The input file
	 * @param outFile
	 *            The output file
	 * @param targetFile
	 *            The target results file for the integration. This is used to
	 *            allow using the file name and path as variables in the
	 *            Velocity templates.
	 * @throws Exception
	 */
	public void execute(File inFile, File outFile, File targetFile)
			throws Exception {
		// Create the Velocity context.
		VelocityContext context = createContext(inFile, targetFile);

		// Write the outFile.
		BufferedReader reader = null;
		BufferedWriter writer = null;
		try {
			// Get a reader.
			reader = new BufferedReader(new InputStreamReader(
					IOHelper.getInputStream(inFile), getEncoding()));

			// Get a writer.
			writer = new BufferedWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(new FileOutputStream(outFile),
							getClass().getName()), getEncoding()));

			// Write the header.
			writeHeader(velocityEngine, context, writer);

			// Write the content.
			String line;
			while ((line = reader.readLine()) != null) {
				if (line.trim().length() > 0) {
					writer.write(line);
					writer.newLine();
				}
				writer.flush();
			}

			// Write the footer.
			writeFooter(velocityEngine, context, writer);

		} finally {
			if (reader != null) {
				reader.close();
			}
			if (writer != null) {
				writer.close();
			}
		}
	}

	/**
	 * Create the Velocity context with the given file and target file. This
	 * method must be implemented by subclasses.
	 * 
	 * @param file
	 *            The file containing the data before adding the header and
	 *            footer.
	 * @param targetFile
	 *            The target results file for this integration.
	 * @return The Velocity context.
	 * @throws IOException
	 */
	protected abstract VelocityContext createContext(File file, File targetFile)
			throws IOException;

	/**
	 * Write the header in the given writer.
	 * 
	 * @param velocity
	 *            The Velocity engine.
	 * @param context
	 *            The Velocity context.
	 * @param writer
	 *            The writer.
	 * @throws IOException
	 */
	protected void writeHeader(VelocityEngine velocity,
			VelocityContext context, BufferedWriter writer) throws IOException {
		if (StringUtils.isBlank(headerTemplateFile)
				&& StringUtils.isBlank(headerTemplate)) {
			return;
		}

		// Render the template.
		try {
			if (StringUtils.isNotBlank(headerTemplate)) {
				velocity.evaluate(context, writer, "headerTemplate",
						new StringReader(headerTemplate));
			} else {
				File template = new File(headerTemplateFile);
				velocity.mergeTemplate(template.getAbsolutePath(),
						getEncoding(), context, writer);
			}
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		// Add a new line after the header.
		writer.newLine();

		// Flush the writer.
		writer.flush();
	}

	/**
	 * Write the footer in the given writer.
	 * 
	 * @param velocity
	 *            The Velocity engine.
	 * @param context
	 *            The Velocity context.
	 * @param writer
	 *            The writer.
	 * @throws IOException
	 */
	protected void writeFooter(VelocityEngine velocity,
			VelocityContext context, BufferedWriter writer) throws IOException {
		if (StringUtils.isBlank(footerTemplateFile)
				&& StringUtils.isBlank(footerTemplate)) {
			return;
		}

		// Render the template.
		try {
			if (StringUtils.isNotBlank(footerTemplate)) {
				velocity.evaluate(context, writer, "footerTemplate",
						new StringReader(footerTemplate));
			} else {
				File template = new File(footerTemplateFile);
				velocity.mergeTemplate(template.getAbsolutePath(),
						getEncoding(), context, writer);
			}
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		// Flush the writer.
		writer.flush();
	}

}
