package com.taleo.integration.client.customstep.csv;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.csv.CSVReader;
import com.taleo.integration.client.customstep.util.csv.CsvWriter;
import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Abstract post-processing step to transform a column in a CSV file containing
 * the results of an export.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvPostStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>column</b>: The name or index (0 based) of the source column. The
 * column name can be used only if the results file contains a CSV header row.</li>
 * <br>
 * <br>
 * <li><b>overwrite</b>: Indicates if the value in the source column is
 * overwritten (<code>true</code> or <code>false</code>, default:
 * <code>true</code>).</li>
 * <br>
 * <br>
 * <li><b>targetIndex</b>: The index (0 based) of the target column. The default
 * value is the index of the source column.
 * <ul>
 * <li>If overwrite = <code>true</code>, will overwrite the content of the
 * target column.</li>
 * <li>If overwrite = <code>false</code>, will insert a new column at the given
 * index.</li>
 * </ul>
 * </li>
 * <br>
 * <li><b>targetName</b>: The name of the target column. Default: The name of
 * the source column. If the results file has no CSV header row, this has no
 * effect.</li>
 * </ul>
 * 
 * @author Romain Guay, Taleo Corporation
 */
public abstract class BaseTransformPostStep extends BaseCustomCsvPostStep {

	private static Logger log = Logger.getLogger(BaseTransformPostStep.class);

	/**
	 * Constructor without arguments.
	 */
	public BaseTransformPostStep() {
		super();
		// Register valid parameters.
		registerParameterName("column");
		registerParameterName("overwrite");
		registerParameterName("targetIndex");
		registerParameterName("targetName");
	}

	/**
	 * The source column name or index.
	 */
	private String column;

	/**
	 * The overwrite flag.
	 */
	private boolean overwrite;

	/**
	 * The target column index.
	 */
	private int targetIndex = -1;

	/**
	 * The target column name.
	 */
	private String targetName;

	/**
	 * @return
	 */
	public String getColumn() {
		return column;
	}

	/**
	 * @param column
	 */
	public void setColumn(String column) {
		this.column = column;
	}

	/**
	 * @return
	 */
	public boolean isOverwrite() {
		return overwrite;
	}

	/**
	 * @param overwrite
	 */
	public void setOverwrite(boolean overwrite) {
		this.overwrite = overwrite;
	}

	/**
	 * @return
	 */
	public int getTargetIndex() {
		return targetIndex;
	}

	/**
	 * @param targetIndex
	 */
	public void setTargetIndex(int targetIndex) {
		this.targetIndex = targetIndex;
	}

	/**
	 * @return
	 */
	public String getTargetName() {
		return targetName;
	}

	/**
	 * @param targetName
	 */
	public void setTargetName(String targetName) {
		this.targetName = targetName;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		column = parameters.getProperty("column");
		if (StringUtils.isBlank(column)) {
			throw new IllegalArgumentException(
					"The 'column' parameter is mandatory.");
		}

		overwrite = Boolean
				.valueOf(parameters.getProperty("overwrite", "true"))
				.booleanValue();

		try {
			targetIndex = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("targetIndex"),
					Integer.toString(targetIndex)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'targetIndex' parameter must have a numeric value.");
		}

		targetName = parameters.getProperty("targetName");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			// Get the input file.
			Message responseMessage = pipeline.getResponseMessage();
			File inFile = new File(responseMessage.getCurrentFile());

			// Create a temporary file for the output.
			File outFile = createTempFile();

			// Apply the transformation.
			execute(inFile, outFile);

			// Set the new current file.
			responseMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Copy the input file to the output file while applying the required
	 * transformation.
	 * 
	 * @param inFile
	 *            The input file
	 * @param outFile
	 *            The output file
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	public void execute(File inFile, File outFile) throws IOException,
			InvalidCSVFileFormat {

		// Get the source column index.
		int sourceIndex = getColumnIndex(inFile, column);

		// Get the source column name.
		String sourceName = getColumnName(inFile, column);

		// Set the default value for the target index.
		if (targetIndex == -1) {
			targetIndex = sourceIndex;
		}

		// Set the default value for the target name.
		if (targetName == null) {
			targetName = sourceName;
		}

		CSVReader reader = null;
		CsvWriter writer = null;
		try {
			// Initialize a CSV reader.
			reader = new CSVReader(new InputStreamReader(
					IOHelper.getInputStream(inFile), getEncoding()), false,
					getCsvDelimiter(), getCsvQuoteCharacter());

			// Initialize a CSV writer.
			writer = new CsvWriter(new BufferedWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(new FileOutputStream(outFile),
							getClass().getName()), getEncoding())),
					getCsvDelimiter(), getCsvQuoteCharacter());

			List values = new ArrayList();
			List newValues;

			// Copy the header if present.
			if (isCsvHeaderPresent()) {
				reader.readLine(values);
				newValues = new ArrayList(values);
				String value = (String) values.get(targetIndex);
				newValues.set(targetIndex, targetName);
				if (!overwrite) {
					newValues.add(targetIndex + 1, value);
				}
				writer.writeLine(newValues);
			}

			// Copy the contents while applying the transformation.
			while (reader.readLine(values)) {
				newValues = new ArrayList(values);
				String value = (String) values.get(targetIndex);
				newValues.set(targetIndex,
						transform((String) values.get(sourceIndex)));
				if (!overwrite) {
					newValues.add(targetIndex + 1, value);
				}
				writer.writeLine(newValues);
			}
		} finally {
			if (reader != null) {
				reader.close();
			}
			if (writer != null) {
				writer.close();
			}
		}
	}

	/**
	 * Transform the given value.
	 * 
	 * @param value
	 *            The value to transform.
	 * @return The transformed value.
	 */
	protected abstract String transform(String value);

}
