package com.taleo.integration.client.customstep.csv;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.commons.lang.StringUtils;
import org.dom4j.io.XMLWriter;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

import com.taleo.integration.client.customstep.util.csv.CSVReader;
import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Pre-processing step to create an XML file from a CSV. The source CSV file is
 * the current file in the pipeline.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvPreStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>fields</b>: Optional - The comma delimited list of CSV field names or
 * indexes (0 based) to use in building the XML. If absent, all the fields are
 * translated into the XML result.</li>
 * <br>
 * <li><b>xmlFields</b>: Optional - The comma delimited list of XML field names
 * for the corresponding CSV fields. If a field is not mentioned here, it will
 * keep its original name or be named FIELD_0, FIELD_1, according to its
 * position in the file if there is no CSV header. Names should be valid XML
 * names.</li>
 * <br>
 * <li><b>outputFile</b>: Optional - The complete path of the resulting XML. If
 * absent the output becomes the current document in the execution pipeline.</li>
 * </ul>
 * <p>
 * The resulting XML is simply structured as follows:
 * 
 * <pre>
 * &lt;XML&gt;
 *    &lt;RECORD&gt;
 *       &lt;FIELDNAME1&gt;xxxxxxxxxx&lt;/FIELDNAME1&gt;
 *       &lt;FIELDNAME2&gt;xxxxxxxxxx&lt;/FIELDNAME2&gt;
 *       ...
 *    &lt;/RECORD&gt;
 *    ...
 * &lt;/XML&gt;
 * </pre>
 * 
 * If the CSV has no header row, the fields are named "FIELD_0", "FIELD_1",
 * etc., by their index in the CSV. This also applies if a given header has no
 * value.<br>
 * This simple format can further be transformed with XSL in subsequent steps if
 * needed. <br>
 * This step is also tolerant with regards to incomplete CSV rows. It will
 * simply set blank values to the missing fields.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://dom4j.org/dom4j-1.6.1/">dom4j library</a>
 * (dom4j-1.6.1.jar)</li>
 * <li><a href="http://jaxen.codehaus.org/">jaxen library</a> (jaxen-1.1.jar)</li>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/CsvToXmlPreStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class CsvToXmlPreStep extends BaseCustomCsvPreStep {

	/**
	 * Constructor without arguments.
	 */
	public CsvToXmlPreStep() {
		super();
		// Register valid parameters.
		registerParameterName("fields");
		registerParameterName("xmlFields");
		registerParameterName("outputFile");
	}

	/**
	 * The fields list.
	 */
	private String[] fields = new String[0];

	/**
	 * The XML fields list.
	 */
	private String[] xmlFields = new String[0];

	/**
	 * The output file name.
	 */
	private String outputFile;

	/**
	 * @return
	 */
	public String[] getFields() {
		return fields;
	}

	/**
	 * @param strings
	 */
	public void setFields(String[] fields) {
		for (int i = 0; i < fields.length; i++) {
			if (StringUtils.isBlank(fields[i])) {
				throw new IllegalArgumentException("Argument fields[" + i
						+ "] cannot be null or empty.");
			}
		}
		this.fields = fields;
	}

	/**
	 * @return
	 */
	public String[] getXmlFields() {
		return xmlFields;
	}

	/**
	 * @param strings
	 */
	public void setXmlFields(String[] xmlFields) {
		this.xmlFields = xmlFields;
	}

	/**
	 * Get the output file.
	 * 
	 * @return
	 */
	public String getOutputFile() {
		return outputFile;
	}

	/**
	 * Set the output file.
	 * 
	 * @param outputFile
	 */
	public void setOutputFile(String outputFile) {
		this.outputFile = outputFile;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "CsvToXmlPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Pre-processing step to convert a CSV file to XML.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "CsvToXmlPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		String fields = StringEscapeUtils.unescapeJava(parameters
				.getProperty("fields"));
		if (!StringUtils.isBlank(fields)) {
			setFields(StringUtils.splitPreserveAllTokens(fields, ','));
		}

		String xmlFields = StringEscapeUtils.unescapeJava(parameters
				.getProperty("xmlFields"));
		if (!StringUtils.isBlank(xmlFields)) {
			setXmlFields(StringUtils.splitPreserveAllTokens(xmlFields, ','));
		}

		setOutputFile(parameters.getProperty("outputFile"));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			return;
		}
		super.execute(pipeline);

		try {
			// Get the input file.
			Message requestMessage = pipeline.getRequestMessage();
			File inFile = null;
			if (requestMessage.isSetCurrentFile()) {
				inFile = new File(requestMessage.getCurrentFile());
			} else {
				inFile = new File(requestMessage.getOriginalFile());
			}

			// Execute.
			File outFile = createTempFile();
			execute(inFile, outFile);

			if (this.outputFile != null) {
				// If outputFile is set, decompress the result into it.
				com.taleo.integration.client.customstep.util.IOHelper
						.getDecompressedFile(outFile, new File(outputFile));
			} else {
				// Otherwise, set it as the new current file in the pipeline.
				requestMessage.setCurrentFile(outFile.getAbsolutePath());
			}

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Create the XML file.
	 * 
	 * @param inFile
	 *            The CSV input file
	 * @param outFile
	 *            The XML output file
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 * @throws SAXException
	 */
	public void execute(File inFile, File outFile) throws IOException,
			InvalidCSVFileFormat, SAXException {

		CSVReader reader = null;
		XMLWriter writer = null;
		try {
			// Initialize a CSV reader.
			reader = new CSVReader(new InputStreamReader(
					IOHelper.getInputStream(inFile), getEncoding()), false,
					getCsvDelimiter(), getCsvQuoteCharacter());

			// Prepare the list of CSV headers.
			List headers = new ArrayList();
			if (isCsvHeaderPresent()) {
				reader.readLine(headers);
			}

			// Get the indexes of requested fields.
			List indexes = new ArrayList();
			for (int i = 0; i < fields.length; i++) {
				String field = fields[i].trim();
				int index = -1;
				if (StringUtils.isNumeric(field)) {
					index = Integer.parseInt(field);
				} else {
					if (!isCsvHeaderPresent()) {
						throw new IllegalArgumentException(
								"Argument fields["
										+ i
										+ "] must be numeric if there is no CSV header.");
					}
					index = headers.indexOf(field);
				}
				if (index > -1) {
					indexes.add(new Integer(index));
				} else {
					if (StringUtils.isNumeric(field)) {
						throw new IndexOutOfBoundsException("Field '" + field
								+ "' is not a valid index.");
					} else {
						throw new RuntimeException("Field '" + field
								+ "' is not found in the CSV file.");
					}
				}
			}

			// Ensure output folder exists.
			org.apache.commons.io.FileUtils.forceMkdir(outFile.getParentFile());

			// Initialize an XML writer.
			writer = new XMLWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(new FileOutputStream(outFile),
							getClass().getName()), getEncoding()));

			// Write XML declaration and root element.
			writer.startDocument();
			writer.startElement(null, "XML", "XML", new AttributesImpl());

			// Create each record.
			List values = new ArrayList();
			while (reader.readLine(values)) {
				if (fields.length == 0 && indexes.size() != values.size()) {
					// If not set explicitly, prepare indexes for all fields of
					// that specific row.
					indexes.clear();
					for (int index = 0; index < values.size(); index++) {
						indexes.add(new Integer(index));
					}
				}
				// Write record and fields.
				writer.startElement(null, "RECORD", "RECORD",
						new AttributesImpl());
				for (int i = 0; i < indexes.size(); i++) {
					int index = ((Integer) indexes.get(i)).intValue();
					String xmlField = null;
					if (i < xmlFields.length) {
						xmlField = xmlFields[i];
					}
					if (StringUtils.isBlank(xmlField) && index < headers.size()) {
						xmlField = headers.get(index).toString();
					}
					if (StringUtils.isBlank(xmlField)) {
						xmlField = "FIELD_" + index;
					}
					writer.startElement(null, xmlField, xmlField,
							new AttributesImpl());
					if (index < values.size()) {
						writer.write(values.get(index));
					}
					writer.endElement(null, xmlField, xmlField);
				}
				writer.endElement(null, "RECORD", "RECORD");
				writer.flush();
			}

			// End the root element and the document.
			writer.endElement(null, "XML", "XML");
			writer.endDocument();

		} finally {
			if (reader != null) {
				reader.close();
			}
			if (writer != null) {
				writer.close();
			}
		}
	}
}
