package com.taleo.integration.client.customstep.csv;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import org.apache.commons.lang.StringUtils;

import com.taleo.ws.integration.client.GlobalConfig;

/**
 * Post-processing step to decode a column in a CSV file containing the results
 * of an export.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseTransformPostStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>mappingFile</b>: The name of a properties file containing the mappings
 * for the decode in the following format: actualValue=newValue, for example:<br>
 * A=1<br>
 * B=2<br>
 * C= <br>
 * ... <br>
 * The file can be specified as an absolute path or relative to the working
 * directory.</li>
 * <br>
 * <br>
 * <li><b>isDefaultEmpty</b>: Indicates if the default value is an empty string
 * when a value is not found in the mapping file (<code>true</code> or
 * <code>false</code>, default: <code>true</code>). If <code>false</code>, the
 * default is the original value unchanged.</li>
 * <br>
 * <br>
 * <li><b>caseSensitive</b>: Indicates if the actual values are compared in a
 * case sensitive manner (<code>true</code> or <code>false</code>, default:
 * <code>false</code>).</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/DecodePostStep-1.jpg" />
 * 
 * @see BaseTransformPostStep
 * @author Romain Guay, Taleo Corporation
 */
public class DecodePostStep extends BaseTransformPostStep {

	private static final String EMPTY_STRING = "";

	/**
	 * Constructor without arguments.
	 */
	public DecodePostStep() {
		super();
		// Register valid parameters.
		registerParameterName("mappingFile");
		registerParameterName("isDefaultEmpty");
		registerParameterName("caseSensitive");
	}

	/**
	 * The mapping for the decode.
	 */
	private Map mapping;

	/**
	 * The isDefaultEmpty flag.
	 */
	private boolean isDefaultEmpty = true;

	/**
	 * The caseSensitive flag.
	 */
	private boolean caseSensitive = false;

	/**
	 * @return
	 */
	protected Map getMapping() {
		return mapping;
	}

	/**
	 * @param mapping
	 */
	protected void setMapping(Map mapping) {
		this.mapping = mapping;
	}

	/**
	 * @return
	 */
	public boolean isDefaultEmpty() {
		return isDefaultEmpty;
	}

	/**
	 * @param isDefaultEmpty
	 */
	public void setDefaultEmpty(boolean isDefaultEmpty) {
		this.isDefaultEmpty = isDefaultEmpty;
	}

	/**
	 * @return
	 */
	public boolean isCaseSensitive() {
		return caseSensitive;
	}

	/**
	 * @param caseSensitive
	 */
	public void setCaseSensitive(boolean caseSensitive) {
		this.caseSensitive = caseSensitive;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "DecodePostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to apply a decode to a CSV column.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "DecodePostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);
		
		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		String mappingFile = parameters.getProperty("mappingFile");
		if (StringUtils.isBlank(mappingFile)) {
			throw new IllegalArgumentException(
					"The 'mappingFile' parameter is mandatory.");
		}

		isDefaultEmpty = Boolean.valueOf(
				parameters.getProperty("isDefaultEmpty", "true"))
				.booleanValue();

		caseSensitive = Boolean.valueOf(
				parameters.getProperty("caseSensitive", "false"))
				.booleanValue();

		// Load the mapping.
		try {
			if (isActive()) {
				mapping = loadMapping(mappingFile);
			}
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Load the decode mappings from the specified file.
	 * 
	 * @param file
	 * @return
	 * @throws IOException
	 */
	private Map loadMapping(String file) throws IOException {
		BufferedReader in = null;
		Map mapping = new Hashtable();
		try {
			in = new BufferedReader(new InputStreamReader(new FileInputStream(
					file), getEncoding()));
			String str;
			while ((str = in.readLine()) != null) {
				if (str.indexOf("=") >= 0) {
					String[] split = str.split("=");
					if (split.length == 1) {
						mapping.put(split[0].trim(), "");
					} else if (split.length == 2) {
						mapping.put(split[0].trim(), split[1].trim());
					} else {
						mapping.put(split[0].trim(), str.substring(
								str.indexOf("=") + 1).trim());
					}
				}
			}
		} catch (IOException e) {
			throw new RuntimeException(e);
		} finally {
			if (in != null) {
				in.close();
			}
		}
		return mapping;
	}

	/**
	 * Transform the given value by applying a decode.
	 * 
	 * @param value
	 *            The value to transform.
	 * @return The transformed value.
	 */
	protected String transform(String value) {
		Object mappedValue = null;
		for (Iterator iter = mapping.keySet().iterator(); iter.hasNext();) {
			String key = (String) iter.next();
			boolean equals = false;
			if (caseSensitive) {
				equals = key.equals(value);
			} else {
				equals = key.equalsIgnoreCase(value);
			}
			if (equals) {
				mappedValue = mapping.get(key);
				break;
			}
		}
		if (mappedValue != null) {
			return mappedValue.toString();
		} else {
			if (isDefaultEmpty) {
				return EMPTY_STRING;
			} else {
				return value;
			}
		}
	}

}
