package com.taleo.integration.client.customstep.csv;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringReader;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.csv.CsvUtils;
import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.customstep.util.csv.MergeSort;
import com.taleo.integration.client.customstep.util.csv.RowComparator;
import com.taleo.integration.client.customstep.util.csv.SortCriterion;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Post-processing step to calculate the difference between a file (the target)
 * and the current file in the pipeline (the source). Consequence is to remove
 * matching records contained in the current pipeline file from the target. The
 * result will become the new current file in the pipeline.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvPostStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>keys</b>: Mandatory - The delimited list of key fields. Alphanumeric
 * values will be considered field names. Numeric values will be considered
 * column index (0 based). This list is delimited as the CSV file itself. <br>
 * Source and target files must have key columns in the same position. However,
 * files do not need to have the same total number of columns, i.e the source
 * can have only up to the keys and no more.</li>
 * <br>
 * <br>
 * <li><b>targetFile</b>: Optional - The path to the target file from which
 * records will be removed. Can be specified as an absolute or relative path. If
 * absent, the current result filename itself is used, assuming it is there
 * already from previous runs.</li>
 * </ul>
 * <p>
 * <b>Note</b>: This operation requires sorting the file according to key
 * fields. This is an unavoidable side effect of this step.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class DifferencePostStep extends BaseCustomCsvPostStep {

	private static Logger log = Logger.getLogger(DifferencePostStep.class);

	/**
	 * Constructor without arguments.
	 */
	public DifferencePostStep() {
		super();
		// Register valid parameters.
		registerParameterName("targetFile");
		registerParameterName("keys");
	}

	/**
	 * The target file.
	 */
	private String targetFile = null;

	/**
	 * The keys list.
	 */
	private String[] keys = new String[0];

	/**
	 * Get the target file.
	 * 
	 * @return
	 */
	public String getTargetFile() {
		return targetFile;
	}

	/**
	 * Set the target file.
	 * 
	 * @param targetFile
	 */
	public void setTargetFile(String targetFile) {
		this.targetFile = targetFile;
	}

	/**
	 * Get keys.
	 * 
	 * @return
	 */
	public String[] getKeys() {
		return keys;
	}

	/**
	 * Set keys.
	 * 
	 * @param keys
	 */
	public void setKeys(String[] keys) {
		for (int i = 0; i < keys.length; i++) {
			if (StringUtils.isBlank(keys[i])) {
				throw new IllegalArgumentException("Argument keys[" + i
						+ "] cannot be empty or null.");
			}
		}
		this.keys = keys;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "DifferencePostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to calculate the difference between files.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "DifferencePostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Read parameters.
		targetFile = parameters.getProperty("targetFile");

		// Note: Keys must be set in the execute method because they depend on
		// the CSV settings.
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			// Set the keys parameter.
			setKeys(CsvUtils.readFirstLine(new StringReader(parameters
					.getProperty("keys", "").trim()), getCsvDelimiter(),
					getCsvQuoteCharacter()));

			// Get the source file.
			Message responseMessage = pipeline.getResponseMessage();
			File sourceFile = new File(responseMessage.getCurrentFile());

			// Get the target file.
			String targetFile = this.targetFile;
			if (StringUtils.isBlank(this.targetFile)) {
				targetFile = responseMessage.getOriginalFile();
			}

			// Execute the difference operation.
			File outFile = difference(sourceFile, new File(targetFile));

			// Set the new current file.
			responseMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Calculate the difference, i.e. remove matching records contained in
	 * source file from target file.
	 * 
	 * @param sourceFile
	 *            The file containing records to be removed from target.
	 * @param targetFile
	 *            The file in which matching records will be removed.
	 * @return The result file.
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	protected File difference(File sourceFile, File targetFile)
			throws IOException, InvalidCSVFileFormat {
		BufferedReader reader = null;
		BufferedWriter writer = null;
		File mergedFile = null;

		// Read the CSV header if present.
		String[] header = null;
		if (isCsvHeaderPresent()) {
			Reader headerReader = new InputStreamReader(
					IOHelper.getInputStream(sourceFile), getEncoding());
			header = CsvUtils.readLine(headerReader, getCsvDelimiter(),
					getCsvQuoteCharacter(), 1);
			headerReader.close();
		}

		// Get the key column indexes.
		int[] keyIndexes = CsvUtils.getColumnIndexes(getKeys(), header);

		if (keyIndexes.length > 0) {
			log.info("Using columns " + ArrayUtils.toString(keyIndexes)
					+ " as key");
		} else {
			throw new IllegalStateException("Keys parameter is mandatory.");
		}

		try {
			boolean headerWritten = false;

			// Create a temp file for merging.
			mergedFile = createTempFile();
			mergedFile.deleteOnExit();

			// Create an array of files to be merged.
			// Source file is added twice to force each of its records to be
			// duplicated. This way, the sort process will remove them no matter
			// if they matched the target or not.
			File[] files = new File[] { targetFile, sourceFile, sourceFile };

			// Initialize a writer.
			writer = new BufferedWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(new FileOutputStream(mergedFile,
							true), getClass().getName()), getEncoding()));

			// Process each file.
			for (int i = 0; i < files.length; i++) {
				File file = files[i];
				log.info("Merging file " + file.getAbsolutePath());

				try {
					// Initialize a reader.
					reader = new BufferedReader(new InputStreamReader(
							IOHelper.getInputStream(file), getEncoding()));
					String line;
					int j = 0;
					while ((line = reader.readLine()) != null) {

						// Write each non empty line.
						if (line.trim().length() > 0) {

							// Skip the CSV header row if present and it was
							// written already.
							if (j++ == 0 && isCsvHeaderPresent()
									&& headerWritten) {
								continue;
							} else {
								headerWritten = true;
							}

							// Write the line.
							writer.write(line);
							writer.newLine();
						}
						writer.flush();
					}

				} finally {
					if (reader != null) {
						reader.close();
					}
				}
			}
			writer.close();

			// Sort and remove all duplicates.
			sort(mergedFile, mergedFile, keyIndexes);

			// Return the merged file.
			return mergedFile;

		} finally {
			if (writer != null) {
				writer.close();
			}
		}
	}

	/**
	 * Sort the inFile and write result to the outFile, removing all duplicates.
	 * 
	 * @param inFile
	 *            The input file
	 * @param outFile
	 *            The output file
	 * @param indexes
	 *            The sort column indexes
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	protected void sort(File inFile, File outFile, int[] indexes)
			throws IOException, InvalidCSVFileFormat {

		// Build an array of SortCriterion.
		SortCriterion[] sortCriteria = new SortCriterion[indexes.length];
		for (int i = 0; i < indexes.length; i++) {
			sortCriteria[i] = new SortCriterion(indexes[i]);
		}

		if (sortCriteria.length == 0) {
			log.info("No difference will be calculated because no key is defined.");
		} else {
			// Otherwise, do the sort.
			if (log.isDebugEnabled()) {
				log.debug("Sorting " + inFile.getAbsolutePath()
						+ " on columns: " + ArrayUtils.toString(indexes)
						+ " and removing all duplicates.");
			}
			MergeSort sort = new MergeSort();
			sort.setCsvHeaderPresent(isCsvHeaderPresent());
			sort.setCsvDelimiter(getCsvDelimiter());
			sort.setCsvQuoteCharacter(getCsvQuoteCharacter());
			sort.setRemoveAllDuplicates(true);
			sort.setEncoding(getEncoding());
			sort.setWorkingFolder(tempFolder);
			sort.sort(inFile, new RowComparator(sortCriteria), outFile);
		}
	}
}
