package com.taleo.integration.client.customstep.csv;

import org.apache.commons.codec.digest.DigestUtils;

import com.taleo.ws.integration.client.GlobalConfig;

/**
 * Post-processing step to apply a digest algorithm (one-way hash function) to a
 * column in a CSV file containing the results of an export.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseTransformPostStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>algorithm</b>: Optional - Name of the digest algorithm to use.
 * Currently only MD5 and SHA-1 are supported. The default is MD5.</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/codec/">Apache Commons Codec</a>
 * (commons-codec-1.3.jar - provided in TCC)</li>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/DigestPostStep-1.jpg" />
 * 
 * @see BaseTransformPostStep
 * @author Romain Guay, Taleo Corporation
 */
public class DigestPostStep extends BaseTransformPostStep {

	public final static String MD5 = "MD5";
	public final static String SHA_1 = "SHA-1";

	/**
	 * Constructor without arguments.
	 */
	public DigestPostStep() {
		super();
		// Register valid parameters.
		registerParameterName("algorithm");
	}

	/**
	 * The isDefaultEmpty flag.
	 */
	private String algorithm = MD5;

	/**
	 * Get the digest algorithm.
	 * 
	 * @return
	 */
	public String getAlgorithm() {
		return algorithm;
	}

	/**
	 * Set the digest algorithm.
	 * 
	 * @param algorithm
	 */
	public void setAlgorithm(String algorithm) {
		this.algorithm = algorithm;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "DigestPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to apply a digest algorithm to a CSV column.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "DigestPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);
		
		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		algorithm = parameters.getProperty("algorithm", algorithm);
	}

	/**
	 * Transform the given value by applying the digest algorithm.
	 * 
	 * @param value
	 *            The value to transform.
	 * @return The transformed value.
	 */
	protected String transform(String value) {
		if (MD5.equals(algorithm)) {
			return DigestUtils.md5Hex(value);
		} else if (SHA_1.equals(algorithm)) {
			return DigestUtils.shaHex(value);
		} else {
			throw new RuntimeException("Unsupported digest algorithm: "
					+ algorithm + ".");
		}
	}

}
