package com.taleo.integration.client.customstep.csv;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.security.GeneralSecurityException;
import java.util.zip.GZIPOutputStream;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.UsernamePasswordCredentials;
import org.apache.commons.httpclient.auth.BasicScheme;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPReply;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.HttpUtils;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.PasswordUtils;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedMessage;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Pre-processing step to replace a URL to a document into its actual content in
 * Base64GZip format. Supports URLs using http, https and ftp protocols.
 * Supports proxy on HTTP connection as set in System properties, but currently
 * no proxy on FTP connections.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseTransformPostStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>prefix</b>: Optional - The prefix to add to the column value to build
 * the URL.</li>
 * <br>
 * <br>
 * <li><b>username</b>: Optional - The username for the URL connection.</li>
 * <br>
 * <br>
 * <li><b>password</b>: Optional - The AES128 encrypted or unencrypted password.
 * Encrypted passwords are recognized by their ending with "=".</li>
 * <br>
 * <br>
 * <li><b>failOnError</b>: Optional - Indicate if the process fails if an error
 * occurs while retrieving a document (default: <code>false</code>). If
 * <code>false</code>, an error is written to the log and the document content
 * is left empty).</li>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/DocumentFromURLPreStep-1.png" />
 * <p>
 * <img src="doc-files/DocumentFromURLPreStep-2.png" />
 * 
 * @author Romain Guay, Oracle Corporation
 */
public class DocumentFromURLPreStep extends BaseTransformPostStep {

	private static Logger log = Logger.getLogger(DocumentFromURLPreStep.class);

	private HttpClient httpClient = null;

	/**
	 * Constructor without arguments.
	 */
	public DocumentFromURLPreStep() {
		super();
		// Register valid parameters.
		registerParameterName("prefix");
		registerParameterName("username");
		registerParameterName("password");
		registerParameterName("failOnError");
	}

	/**
	 * The prefix.
	 */
	private String prefix;

	/**
	 * The username.
	 */
	private String username;

	/**
	 * The password.
	 */
	private String password;

	/**
	 * The password.
	 */
	private boolean failOnError = false;

	/**
	 * @return
	 */
	public String getPrefix() {
		return prefix;
	}

	/**
	 * @param prefix
	 */
	public void setPrefix(String prefix) {
		this.prefix = prefix;
	}

	/**
	 * @return
	 */
	public String getUsername() {
		return username;
	}

	/**
	 * @param username
	 */
	public void setUsername(String username) {
		this.username = username;
	}

	/**
	 * @return
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * @param password
	 */
	public void setPassword(String password) {
		this.password = password;
	}

	/**
	 * @return
	 */
	public boolean isFailOnError() {
		return failOnError;
	}

	/**
	 * @param failOnError
	 */
	public void setFailOnError(boolean failOnError) {
		this.failOnError = failOnError;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "DocumentFromURLPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Pre-processing step to replace a URL to a document into its actual content in Base64GZip format.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "DocumentFromURLPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		SupportedPipeline supportedPipeline = getPreProcessSupportPipeline();

		// Add the supported response message information
		SupportedMessage supportedMessage = supportedPipeline
				.addNewSupportedRequestMessage();

		// Add the supported request message format
		supportedMessage.addNewSupportedFormats();
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.CSV);

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		prefix = parameters.getProperty("prefix");
		username = parameters.getProperty("username");
		password = parameters.getProperty("password");
		failOnError = Boolean.valueOf(
				parameters.getProperty("failOnError", "false")).booleanValue();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.customstep.csv.BaseTransformPostStep#execute
	 * (com.taleo.ws.integration.client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the input file.
			Message requestMessage = pipeline.getRequestMessage();
			File inFile = null;
			if (requestMessage.isSetCurrentFile()) {
				inFile = new File(requestMessage.getCurrentFile());
			} else {
				inFile = new File(requestMessage.getOriginalFile());
			}

			// Create a temporary file for the output.
			File outFile = createTempFile();

			// Apply the transformation.
			execute(inFile, outFile);

			// Set the new current file.
			requestMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Transform the given value by getting the document content and rendering
	 * it as Base64Gzip.
	 * 
	 * @param value
	 *            The value to transform.
	 * @return The transformed value.
	 */
	protected String transform(String value) {
		// Skip if value is blank.
		if (StringUtils.isBlank(value)) {
			return value;
		}

		String urlString = value;
		if (!StringUtils.isBlank(prefix)) {
			urlString = prefix + urlString;
		}

		String content = null;
		try {
			// Build the URL.
			URL url = new URL(urlString);

			// Execute according to the protocol.
			String protocol = url.getProtocol();
			if ("http".equals(protocol) || "https".equals(protocol)) {
				content = transform_http(url);
			} else if ("ftp".equals(protocol)) {
				content = transform_ftp(url);
			} else {
				throw new RuntimeException("Unsupported protocol: " + protocol
						+ ".");
			}

		} catch (IOException e) {
			log.error("Document '" + urlString + "' cannot be retrieved. "
					+ e.getMessage());
			if (failOnError) {
				throw new RuntimeException(e);
			}
		}
		return content;
	}

	/**
	 * Transform the given value by getting the document content from the given
	 * HTTP or HTTPS URL and rendering it as Base64Gzip.
	 * 
	 * @param url
	 * @return The transformed value.
	 * @throws IOException
	 */
	private String transform_http(URL url) throws IOException {
		// Create httpClient the first time.
		if (httpClient == null) {
			httpClient = HttpUtils.getHttpClient();
		}

		// Create a GetMethod on the URL.
		GetMethod get = new GetMethod(url.toString());

		// Set authentication.
		if (username != null) {
			UsernamePasswordCredentials creds = new UsernamePasswordCredentials(
					username, getUnencryptedPassword(password));
			Header authorizationHeader = new Header("authorization",
					BasicScheme.authenticate(creds, "ISO-8859-1"));
			get.setRequestHeader(authorizationHeader);
		}

		InputStream is = null;
		try {
			// Get the response.
			int status = httpClient.executeMethod(get);

			// Check for error.
			if (status != 200) {
				throw new IOException("Server returned HTTP response code: "
						+ status + " for URL: " + url);
			}

			// Get response stream.
			is = get.getResponseBodyAsStream();

			// Return document content as Base64GZip string.
			return new String(Base64.encodeBase64(gzip(is)));

		} finally {
			if (is != null) {
				is.close();
			}
			get.releaseConnection();
		}
	}

	/**
	 * Transform the given value by getting the document content from the given
	 * FTP URL and rendering it as Base64Gzip.
	 * 
	 * @param url
	 * @return The transformed value.
	 * @throws IOException
	 */
	private String transform_ftp(URL url) throws IOException {
		// Create FTPClient.
		FTPClient ftpClient = getFtpClient(url);

		// Get the response stream.
		InputStream is = null;
		try {
			// Get the response stream.
			is = ftpClient.retrieveFileStream(url.getPath());

			// Return document content as Base64GZip string.
			return new String(Base64.encodeBase64(gzip(is)));

		} finally {
			if (is != null) {
				is.close();
			}
			// Logout from FTP client.
			if (ftpClient != null && ftpClient.isConnected()) {
				try {
					ftpClient.logout();
					ftpClient.disconnect();
				} catch (Exception e) {
					throw new RuntimeException(e);
				}
			}
		}
	}

	/**
	 * Get the FTP Client.
	 * <p>
	 * Note: Must create a new client each time because the URL may be
	 * different.
	 * 
	 * @param url
	 * @return
	 * @throws IOException
	 */
	protected FTPClient getFtpClient(URL url) throws IOException {
		// Get FTP port.
		int ftpPort = url.getPort();
		if (ftpPort == -1) {
			ftpPort = 21;
		}

		// Create FTPClient.
		FTPClient ftpCient = new FTPClient();
		log.debug("Connected to " + url.getHost() + ":" + ftpPort + " as "
				+ username);

		// Connect and login.
		ftpCient.connect(url.getHost(), ftpPort);
		ftpCient.login(username, getUnencryptedPassword(password));
		if (!FTPReply.isPositiveCompletion(ftpCient.getReplyCode())) {
			throw new IOException("FTP Error: " + ftpCient.getReplyString());
		}

		return ftpCient;
	}

	/**
	 * Get the unencrypted password.
	 * 
	 * @param password
	 * @return
	 * @throws IOException
	 */
	private String getUnencryptedPassword(String password) throws IOException {
		if (StringUtils.isBlank(password)) {
			return password;
		}
		try {
			if (password.endsWith("=")) {
				return PasswordUtils.decryptAES128(password);
			} else {
				return password;
			}
		} catch (GeneralSecurityException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Compress a stream in GZip format.
	 * 
	 * @param is
	 *            The stream to compress.
	 * @return <code>byte[]</code>
	 * @throws IOException
	 */
	private static byte[] gzip(InputStream is) throws IOException {
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		OutputStream gzipStream = new GZIPOutputStream(out);
		IOUtils.copy(is, gzipStream);
		gzipStream.close();
		return out.toByteArray();
	}
}
