package com.taleo.integration.client.customstep.csv;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.csv.CSVReader;
import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Post-processing step to transform a CSV file to a fixed length file.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvPostStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>columnsFile</b>: The name of a properties file containing the columns
 * width specification in one of the the following formats:
 * <ul>
 * <li>columnIndex=rightPad(width)</li>
 * <li>columnIndex=rightPad(width, padString)</li>
 * <li>columnIndex=leftPad(width)</li>
 * <li>columnIndex=leftPad(width, padString)</li>
 * </ul>
 * For example:
 * <ul>
 * <li>0=rightPad(10)</li>
 * <li>1=leftPad(15, 0)</li>
 * <li>...</li>
 * </ul>
 * If a column is absent from the list, an exception is thrown. <br>
 * The file can be specified as an absolute path or relative to the working
 * directory.</li>
 * <br>
 * <br>
 * <li><b>truncate</b>: Indicates if values that exceed the column width are
 * truncated (<code>true</code> or <code>false</code>, default:
 * <code>true</code>). If <code>false</code>, an exception is thrown when a
 * value is too large.</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/FixedLengthPostStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class FixedLengthPostStep extends BaseCustomCsvPostStep {

	private static Logger log = Logger.getLogger(FixedLengthPostStep.class);

	/**
	 * Constructor without arguments.
	 */
	public FixedLengthPostStep() {
		super();
		// Register valid parameters.
		registerParameterName("columnsFile");
		registerParameterName("truncate");
	}

	/**
	 * The list containing the columns width specification.
	 */
	private List columns;

	/**
	 * Indicates if values that are too large are truncated.
	 */
	private boolean truncate = true;

	/**
	 * @return
	 */
	public List getColumns() {
		return columns;
	}

	/**
	 * @param columns
	 */
	public void setColumns(List columns) {
		this.columns = columns;
	}

	/**
	 * @return
	 */
	public boolean isTruncate() {
		return truncate;
	}

	/**
	 * @param truncate
	 */
	public void setTruncate(boolean truncate) {
		this.truncate = truncate;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "FixedLengthPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to convert to a fixed length format.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "FixedLengthPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		String columnsFile = parameters.getProperty("columnsFile");
		if (StringUtils.isBlank(columnsFile)) {
			throw new IllegalArgumentException(
					"The 'columnsFile' parameter is mandatory.");
		}
		try {
			if (isActive()) {
				columns = getColumns(columnsFile);
			}
		} catch (IOException e) {
			throw new RuntimeException(e);
		}

		this.truncate = Boolean.valueOf(
				parameters.getProperty("truncate", "true")).booleanValue();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			// Get the input file.
			Message responseMessage = pipeline.getResponseMessage();
			File inFile = new File(responseMessage.getCurrentFile());

			// Create a temporary file for the output.
			File outFile = createTempFile();

			// Execute the transformation.
			execute(inFile, outFile);

			// Set the new current file.
			responseMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Copy the input file to the output file while applying the columns
	 * formatting.
	 * 
	 * @param inFile
	 *            The input file
	 * @param outFile
	 *            The output file
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	public void execute(File inFile, File outFile) throws IOException,
			InvalidCSVFileFormat {

		CSVReader reader = null;
		Writer writer = null;
		try {
			// Initialize a CSV reader.
			reader = new CSVReader(new InputStreamReader(
					IOHelper.getInputStream(inFile), getEncoding()), false,
					getCsvDelimiter(), getCsvQuoteCharacter());

			// Initialize a writer.
			writer = new BufferedWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(new FileOutputStream(outFile),
							getClass().getName()), getEncoding()));

			// Build the formatters list.
			List formatters = new ArrayList();
			for (Iterator iter = columns.iterator(); iter.hasNext();) {
				String column = (String) iter.next();
				formatters.add(new Formatter(column));
			}

			// Write contents formatted to the specified width.
			List values = new ArrayList();
			int row = 0;
			while (reader.readLine(values)) {
				row++;
				for (int i = 0; i < values.size(); i++) {
					// Get the value.
					String value = (String) values.get(i);

					// Format the value.
					Formatter formatter = null;
					try {
						formatter = (Formatter) formatters.get(i);
						value = formatter.format(value, truncate);
					} catch (IndexOutOfBoundsException e) {
						throw new IllegalArgumentException(
								"No pattern was specified for column " + i
										+ ".");
					} catch (ValueTooLargeException e) {
						throw new RuntimeException(
								"The value '"
										+ value
										+ "' found in row "
										+ row
										+ " at column "
										+ i
										+ " is too large to fit in the specified width: "
										+ formatter.width + ".");
					}

					// Write the value.
					writer.write(value);
				}
				writer.write("\n");
			}
		} finally {
			if (reader != null) {
				reader.close();
			}
			if (writer != null) {
				writer.close();
			}
		}
	}

	/**
	 * Get the list of columns width specification based on the specified
	 * columns file.
	 * 
	 * @param columnsFile
	 *            The columns file.
	 * @return The list of specifications.
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	protected List getColumns(String columnsFile) throws FileNotFoundException,
			IOException {

		// Load the columns file.
		Properties props = new Properties();
		props.load(new FileInputStream(columnsFile));

		// Build the list.
		Set keys = props.keySet();
		List columns = new ArrayList(keys.size());
		for (int i = 0; i < keys.size(); i++) {
			columns.add(null);
		}
		for (Iterator iter = keys.iterator(); iter.hasNext();) {
			String key = (String) iter.next();
			int index = -1;
			try {
				index = Integer.parseInt(key);
			} catch (NumberFormatException e) {
				throw new IllegalArgumentException("The column index: " + key
						+ " is not numeric.");
			}
			columns.set(index, (String) props.get(key));
		}
		return columns;
	}

	/**
	 * Private class to format values according to a given pattern.
	 * <p>
	 * The pattern has one of the following values:
	 * <ul>
	 * <li>rightPad(size)</li>
	 * <li>rightPad(size, padChar)</li>
	 * <li>leftPad(size)</li>
	 * <li>leftPad(size, padChar)</li>
	 * </ul>
	 */
	private class Formatter {

		private String operation;

		private int width;

		private String padString = " ";

		/**
		 * A regEx pattern matching: operation(width, padString)
		 */
		private final Pattern regEx1 = Pattern
				.compile("([^\\(]*)\\((\\d*),\\s?\"?([^\"]*)\"?\\)");

		/**
		 * A regEx pattern matching: operation(width)
		 */
		private final Pattern regEx2 = Pattern.compile("([^\\(]*)\\((\\d*)\\)");

		/**
		 * Constructor with the pattern.
		 * 
		 * @param pattern
		 */
		public Formatter(String pattern) {

			// Try to match with a pad string.
			Matcher matcher = regEx1.matcher(pattern);
			boolean matchFound = matcher.find();
			if (matchFound) {
				this.operation = matcher.group(1);
				this.width = Integer.parseInt(matcher.group(2));
				this.padString = matcher.group(3);

			} else {
				// Try to match without a pad string.
				matcher = regEx2.matcher(pattern);
				matchFound = matcher.find();
				if (matchFound) {
					this.operation = matcher.group(1);
					this.width = Integer.parseInt(matcher.group(2));
				} else {
					throw new IllegalArgumentException("Invalid pattern: "
							+ pattern);
				}
			}

			// Validate the operation.
			if (!operation.equals("rightPad") && !operation.equals("leftPad")) {
				throw new IllegalArgumentException("Invalid pattern: "
						+ pattern
						+ ". The operation must be rightPad or leftPad.");
			}
		}

		/**
		 * Format the provided value.
		 * 
		 * @param value
		 *            The value to format.
		 * @param truncate
		 *            Indicates if the value is truncated if too large.
		 * @return
		 * @throws ValueTooLargeException
		 *             If the value is too large to fit in the width.
		 */
		public String format(String value, boolean truncate)
				throws ValueTooLargeException {
			// Test for length.
			if (value.length() > width) {
				if (truncate) {
					value = value.substring(0, width);
				} else {
					throw new ValueTooLargeException();
				}
			}
			// Pad the value.
			if (operation.equals("rightPad")) {
				return StringUtils.rightPad(value, width, padString);
			} else {
				return StringUtils.leftPad(value, width, padString);
			}
		}
	}

	/**
	 * Private exception to signal a value that is too large to fit in the
	 * specified width.
	 */
	private class ValueTooLargeException extends Exception {
		private static final long serialVersionUID = 1L;
	}
}
