package com.taleo.integration.client.customstep.csv;

import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

import org.apache.commons.lang.StringUtils;

import com.taleo.integration.client.customstep.util.csv.CSVReader;
import com.taleo.integration.client.customstep.util.csv.HierarchicalRowComparator;
import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;

/**
 * Processing step to sort CSV data based on parent/child relationship.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.AbstractSortStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>keyField</b>: Mandatory - The index (0 based) or name of the CSV field
 * representing the key of the record.</li>
 * <li><b>parentKeyField</b>: Mandatory - The index (0 based) or name of the CSV
 * field representing the parent key of the record.</li>
 * <br>
 * <br>
 * <li><b>parentKeyField</b>: Mandatory - The index (0 based) or name of the CSV
 * field representing the parent key of the record.</li>
 * <br>
 * <br>
 * <li><b>topDown</b>: Optional - Sort in top down (parent first) order
 * (default: true).</li>
 * </ul>
 * <p>
 * Record key does not have to be unique, many rows with the same key will be
 * sorted equal. However, two rows with the same key but different parents, will
 * cause an error.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/HierarchicalSortStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class HierarchicalSortStep extends AbstractSortStep {

	/**
	 * Constructor without arguments.
	 */
	public HierarchicalSortStep() {
		super();
		// Register valid parameters.
		registerParameterName("keyField");
		registerParameterName("parentKeyField");
		registerParameterName("topDown");
	}

	/**
	 * The key field.
	 */
	private String keyField = null;

	/**
	 * The parent key field.
	 */
	private String parentKeyField = null;

	/**
	 * The topDown flag.
	 */
	private boolean topDown = true;

	/**
	 * Get the key field.
	 * 
	 * @return
	 */
	public String getKeyField() {
		return keyField;
	}

	/**
	 * Set the key field.
	 * 
	 * @param keyField
	 */
	public void setKeyField(String keyField) {
		this.keyField = keyField;
	}

	/**
	 * Get the parent key field.
	 * 
	 * @return
	 */
	public String getParentKeyField() {
		return parentKeyField;
	}

	/**
	 * Set the parent key field.
	 * 
	 * @param parentKeyField
	 */
	public void setParentKeyField(String parentKeyField) {
		this.parentKeyField = parentKeyField;
	}

	/**
	 * Get the topDown flag.
	 * 
	 * @return
	 */
	public boolean isTopDown() {
		return topDown;
	}

	/**
	 * Set the topDown flag.
	 * 
	 * @param topDown
	 */
	public void setTopDown(boolean topDown) {
		this.topDown = topDown;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "HierarchicalSortStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Processing step to sort a CSV file based on parent/child relationship.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "HierarchicalSortStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Read parameters.
		keyField = parameters.getProperty("keyField");
		if (StringUtils.isBlank(keyField)) {
			throw new IllegalArgumentException(
					"The 'keyField' parameter is mandatory.");
		}

		parentKeyField = parameters.getProperty("parentKeyField");
		if (StringUtils.isBlank(parentKeyField)) {
			throw new IllegalArgumentException(
					"The 'parentKeyField' parameter is mandatory.");
		}

		topDown = Boolean.valueOf(
				StringUtils.defaultIfEmpty(parameters.getProperty("topDown"),
						"true")).booleanValue();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.customstep.csv.AbstractSortStep#getComparator
	 * (java.io.File)
	 */
	protected Comparator getComparator(File file) throws IOException,
			InvalidCSVFileFormat {

		// Get the key and parent key indexes.
		int keyIndex = getColumnIndex(file, keyField);
		int parentKeyIndex = getColumnIndex(file, parentKeyField);

		// Create a hierarchical comparator.
		HierarchicalRowComparator comparator = new HierarchicalRowComparator(
				keyIndex, topDown);

		// Prepare the comparator.
		CSVReader reader = null;
		try {
			// Initialize a CSV reader.
			reader = new CSVReader(new InputStreamReader(
					IOHelper.getInputStream(file), getEncoding()),
					isCsvHeaderPresent(), getCsvDelimiter(),
					getCsvQuoteCharacter());

			List values = new ArrayList();
			while (reader.readLine(values)) {
				String key = (String) values.get(keyIndex);
				String parentKey = (String) values.get(parentKeyIndex);
				comparator.setParent(key, parentKey);
			}

			return comparator;

		} finally {
			if (reader != null) {
				reader.close();
			}
		}
	}
}
