package com.taleo.integration.client.customstep.csv;

import java.io.File;
import java.io.IOException;

import org.apache.log4j.Logger;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.tools.generic.DateTool;
import org.apache.velocity.tools.generic.MathTool;

import com.taleo.integration.client.customstep.velocity.CsvImportFileTool;
import com.taleo.integration.client.customstep.velocity.StringTool;
import com.taleo.integration.client.customstep.velocity.SystemTool;
import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Post-processing step to add a header and/or footer to a CSV file containing
 * the results of an import.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseHeaderFooterPostStep}.
 * <p>
 * The following variables are made available to build header/footer templates:
 * <ul>
 * <li><b>$date</b>: A date tool instance (
 * {@link org.apache.velocity.tools.generic.DateTool}) to access various date
 * functions.</li>
 * <br>
 * <br>
 * <li><b>$math</b>: A Math tool instance (
 * {@link org.apache.velocity.tools.generic.MathTool}) to access various 
 * mathematical functions.</li>
 * <br>
 * <br>
 * <li><b>$string</b>: A String tool instance (
 * {@link com.taleo.integration.client.customstep.velocity.StringTool}) to
 * access various string functions.</li>
 * <br>
 * <br>
 * <li><b>$file</b>: A File tool instance (
 * {@link com.taleo.integration.client.customstep.velocity.CsvImportFileTool})
 * to access file properties and content.</li>
 * <br>
 * <br>
 * <li><b>$system</b>: A System tool instance (
 * {@link com.taleo.integration.client.customstep.velocity.SystemTool}) to
 * access system properties.</li>
 * </ul>
 * <p>
 * Example uses of these variables in the templates:
 * <ul>
 * <li>$date.get("yyyy-MM-dd")</li>
 * <li>$date.getYear()</li>
 * <li>$date.getMonth()</li>
 * <li>$date.getDay()</li>
 * <li>$math.add($date.getYear(), 1)</li>
 * <li>$string.rightPad("123", 10)</li>
 * <li>$string.rightPad("123", 10, "_")</li>
 * <li>$string.leftPad("123", 10)</li>
 * <li>$string.leftPad("123", 10, "0")</li>
 * <li>$string.substring("123456", 2, 3)</li>
 * <li>$string.substringAfter("123456", "3")</li>
 * <li>$file.name()</li>
 * <li>$file.path()</li>
 * <li>$file.lastModified()</li>
 * <li>$file.size()</li>
 * <li>$file.recordCount()</li>
 * <li>$file.lineCount()</li>
 * <li>$file.successCount()</li>
 * <li>$file.createCount()</li>
 * <li>$file.updateCount()</li>
 * <li>$file.deleteCount()</li>
 * <li>$file.errorCount()</li>
 * <li>$system.getProperty("user.name")</li>
 * <li>$system.getProperty("myProperty")</li>
 * </ul>
 * <p>
 * There are also other Velocity constructs that can be used. For reference, see
 * <a href=
 * "http://velocity.apache.org/engine/releases/velocity-1.4/user-guide.html"
 * >Velocity User Guide</a>.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://velocity.apache.org/index.html">Apache Velocity
 * library</a> (velocity-dep-1.5.jar)</li>
 * <li><a href="http://velocity.apache.org/tools/devel/generic.html">Apache
 * Velocity Tools Generic library</a> (velocity-tools-generic-1.4.jar)</li>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/ImportHeaderFooterPostStep-1.jpg" />
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public class ImportHeaderFooterPostStep extends BaseHeaderFooterPostStep {

	private static Logger log = Logger
			.getLogger(ImportHeaderFooterPostStep.class);

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Write a header and/or a footer for an import result in CSV format.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "CsvImportHeaderFooterPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "CsvImportHeaderFooterPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			// TODO: Get the delimiter, quoteCharacter and csvHeaderPresent from
			// the loadFile. Currently, the API does not allow this.

			// Get the input file.
			Message responseMessage = pipeline.getResponseMessage();
			File inFile = new File(responseMessage.getCurrentFile());

			// Create a temporary file for the output.
			File outFile = createTempFile();

			// Get the target output file.
			File targetFile = new File(responseMessage.getOriginalFile());

			// Execute by adding the header and/or footer.
			execute(inFile, outFile, targetFile);

			// Set the new current file.
			responseMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Create the Velocity context with the given file and the target results
	 * file.
	 * 
	 * @param file
	 *            The file containing the data before adding the header and
	 *            footer.
	 * @param targetFile
	 *            The target results file for this integration.
	 * @throws IOException
	 */
	protected VelocityContext createContext(File file, File targetFile)
			throws IOException {
		VelocityContext context = new VelocityContext();

		// Add the Date tool.
		context.put("date", new DateTool());

		// Add the Math tool.
		context.put("math", new MathTool());

		// Add the String tool.
		context.put("string", new StringTool());

		// Add the CsvImportFileTool tool.
		context.put("file", new CsvImportFileTool(file, isCsvHeaderPresent(),
				getCsvDelimiter(), getCsvQuoteCharacter(), getEncoding(),
				targetFile));

		// Add the System tool.
		context.put("system", new SystemTool());

		return context;
	}
}
