package com.taleo.integration.client.customstep.csv;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.integration.client.customstep.util.csv.CSVReader;
import com.taleo.integration.client.customstep.util.csv.CsvUtils;
import com.taleo.integration.client.customstep.util.csv.CsvWriter;
import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.customstep.util.csv.MergeSort;
import com.taleo.integration.client.customstep.util.csv.RowComparator;
import com.taleo.integration.client.customstep.util.csv.SortCriterion;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Post-processing step to merge result CSV files. Files corresponding to a
 * pattern will be merged with the current result and be set as the current
 * document in the execution pipeline. This step also allows removing duplicates
 * while merging.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvPostStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>filePattern</b>: Optional - The pattern for selecting the files to
 * merge with the current result. Can be specified as an absolute or relative
 * path and may use wildcards as defined in
 * {@link org.apache.commons.io.filefilter.WildcardFileFilter}, in the file
 * names only. Pattern is case sensitive. If the pattern includes multiple
 * files, all of them will be merged in sequence. If absent, the current result
 * filename itself is used, so instead of overwriting the existing result file,
 * it will be merged with the current one.
 * <p>
 * Multiple patterns can be set by separating them with a pipe (|). The special
 * dot (.) pattern represents the current result filename.</li>
 * <br>
 * <br>
 * <li><b>deleteMergedFiles</b>: Optional - Delete the merged files after
 * processing (Default: <code>true</code>).</li>
 * <br>
 * <br>
 * <li><b>sort</b>: Optional - Sort merged records. (Default: <code>false</code>
 * ).<br>
 * Sorting occurs if keys are present. Key fields are specified in the
 * <code>keys</code> parameter or by the "key" value being present in the tag
 * attribute of projections in the SQ-XML query.<br>
 * For example: &lt;quer:projection alias="Number" tag="key"&gt;<br>
 * The <code>keys</code> parameter has priority over the "key" tag and if no key
 * is specified, no sorting occurs.</li>
 * <br>
 * <br>
 * <li><b>removeDuplicates</b>: Optional - Remove duplicate records. If
 * duplicate records are found, the last one in file processing order is kept.
 * (Default: <code>true</code>).<br>
 * Duplicates are identified with the same key fields as for sorting. In fact,
 * removing duplicates forces sorting the records, no matter what is set in the
 * <code>sort</code> parameter.</li><br>
 * <br>
 * Duplicates can also be identified at the parent level, when merging child
 * tables. Which means that the whole set of children for a given parent found
 * in the same file will be replaced when a duplicate parent is found after
 * merging files. Duplicate parents are identified either by specifying parent
 * key fields in the <code>parentKeys</code> parameter or by the "parentKey"
 * value being present in the tag attribute of projections in the SQ-XML query.<br>
 * For example: &lt;quer:projection alias="CandidateNumber"
 * tag="key parentKey"&gt;<br>
 * The <code>parentKeys</code> parameter has priority over the "parentKey" tag.
 * Also, if the <code>keys</code> parameter is present, "parentKey" tag is
 * ignored.<br>
 * <br>
 * Note: If duplicates are to be removed while there are possibly many files
 * matching the <code>filePattern</code>, those files should be named so that
 * sorting them alphabetically also means chronologically, using a time stamp
 * like "yyyyMMdd HHmmss". This way the newer records will be further down in
 * the merged file and if duplicates are removed the most recent records will be
 * kept. The current result is always the most recent so will be necessarily at
 * the end, replacing any previous duplicate. <br>
 * <br>
 * <li><b>keys</b>: Optional - The delimited list of key fields. Alphanumeric
 * values will be considered field names. Numeric values will be considered
 * column index (0 based). This list is delimited as the CSV file itself. <br>
 * Also see <code>removeDuplicates</code> parameter.</li>
 * <br>
 * <br>
 * <li><b>parentKeys</b>: Optional - The delimited list of parent key fields.
 * Alphanumeric values will be considered field names. Numeric values will be
 * considered column index (0 based). This list is delimited as the CSV file
 * itself. <br>
 * Also see <code>removeDuplicates</code> parameter.</li>
 * <br>
 * <br>
 * <li><b>forceSortMatchingFiles</b>: Optional - For performance reasons, this
 * step assumes that the files matching the <code>filePattern</code> have been
 * produced by this same step and are thus sorted already. If this is not the
 * case, then set this parameter to <code>true</code> to force sorting them
 * appropriately. (Default: <code>false</code>).<br>
 * <br>
 * <br>
 * <li><b>extraHeaderRows</b>: Optional - Number of extra header rows included
 * in the files to be merged. Those rows will be stripped when merging files.
 * (Default: 0). <br>
 * <b>IMPORTANT</b>: This does not apply to the current file in the pipeline
 * which is considered not having extra header rows.<br>
 * <br>
 * <br>
 * <li><b>extraFooterRows</b>: Optional - Number of extra footer rows included
 * in the files to be merged. Those rows will be stripped when merging files.
 * (Default: 0). <br>
 * <b>IMPORTANT</b>: This does not apply to the current file in the pipeline
 * which is considered not having extra footer rows.<br>
 * </ul>
 * <p>
 * <b>Note</b>: Removing duplicates requires sorting the file according to key
 * fields. This is an unavoidable side effect of this step.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class MergePostStep extends BaseCustomCsvPostStep {

	private static Logger log = Logger.getLogger(MergePostStep.class);

	/** The XSL transformation file. */
	private final static String XSL_TRANSFORM = "xsl/GetProjectionIndexesWithTag.xsl";

	/**
	 * Constructor without arguments.
	 */
	public MergePostStep() {
		super();
		// Register valid parameters.
		registerParameterName("filePattern");
		registerParameterName("deleteMergedFiles");
		registerParameterName("sort");
		registerParameterName("removeDuplicates");
		registerParameterName("keys");
		registerParameterName("parentKeys");
		registerParameterName("forceSortMatchingFiles");
		registerParameterName("extraHeaderRows");
		registerParameterName("extraFooterRows");
	}

	/**
	 * The file pattern.
	 */
	private String filePattern = null;

	/**
	 * The deleteMergedFiles flag.
	 */
	private boolean deleteMergedFiles = true;

	/**
	 * The sort flag.
	 */
	private boolean sort = false;

	/**
	 * The removeDuplicates flag.
	 */
	private boolean removeDuplicates = true;

	/**
	 * The keys list.
	 */
	private String[] keys = new String[0];

	/**
	 * The parent keys list.
	 */
	private String[] parentKeys = new String[0];

	/**
	 * The forceSortMatchingFiles flag.
	 */
	private boolean forceSortMatchingFiles = false;

	/**
	 * The extraHeaderRows count.
	 */
	private int extraHeaderRows = 0;

	/**
	 * The extraFooterRows count.
	 */
	private int extraFooterRows = 0;

	/**
	 * Get the file pattern.
	 * 
	 * @return
	 */
	public String getFilePattern() {
		return filePattern;
	}

	/**
	 * Set the file pattern.
	 * 
	 * @param filePattern
	 */
	public void setFilePattern(String filePattern) {
		this.filePattern = filePattern;
	}

	/**
	 * Get the deleteMergedFiles flag.
	 * 
	 * @return
	 */
	public boolean isDeleteMergedFiles() {
		return deleteMergedFiles;
	}

	/**
	 * Set the deleteMergedFiles flag.
	 * 
	 * @param deleteMergedFiles
	 */
	public void setDeleteMergedFiles(boolean deleteMergedFiles) {
		this.deleteMergedFiles = deleteMergedFiles;
	}

	/**
	 * Get the sort flag.
	 * 
	 * @return
	 */
	public boolean isSort() {
		return sort;
	}

	/**
	 * Set the sort flag.
	 * 
	 * @param sort
	 */
	public void setSort(boolean sort) {
		this.sort = sort;
	}

	/**
	 * Get the removeDuplicates flag.
	 * 
	 * @return
	 */
	public boolean isRemoveDuplicates() {
		return removeDuplicates;
	}

	/**
	 * Set the removeDuplicates flag.
	 * 
	 * @param removeDuplicates
	 */
	public void setRemoveDuplicates(boolean removeDuplicates) {
		this.removeDuplicates = removeDuplicates;
	}

	/**
	 * Get keys.
	 * 
	 * @return
	 */
	public String[] getKeys() {
		return keys;
	}

	/**
	 * Set keys.
	 * 
	 * @param keys
	 */
	public void setKeys(String[] keys) {
		for (int i = 0; i < keys.length; i++) {
			if (StringUtils.isBlank(keys[i])) {
				throw new IllegalArgumentException("Argument keys[" + i
						+ "] cannot be empty or null.");
			}
		}
		this.keys = keys;
	}

	/**
	 * Get parent keys.
	 * 
	 * @return
	 */
	public String[] getParentKeys() {
		return parentKeys;
	}

	/**
	 * Set parent keys.
	 * 
	 * @param parentKeys
	 */
	public void setParentKeys(String[] parentKeys) {
		for (int i = 0; i < parentKeys.length; i++) {
			if (StringUtils.isBlank(parentKeys[i])) {
				throw new IllegalArgumentException("Argument parentKeys[" + i
						+ "] cannot be empty or null.");
			}
		}
		this.parentKeys = parentKeys;
	}

	/**
	 * Get the forceSortMatchingFiles flag.
	 * 
	 * @return
	 */
	public boolean isForceSortMatchingFiles() {
		return forceSortMatchingFiles;
	}

	/**
	 * Set the forceSortMatchingFiles flag.
	 * 
	 * @param forceSortMatchingFiles
	 */
	public void setForceSortMatchingFiles(boolean forceSortMatchingFiles) {
		this.forceSortMatchingFiles = forceSortMatchingFiles;
	}

	/**
	 * Get the extra header rows count.
	 * 
	 * @return
	 */
	public int getExtraHeaderRows() {
		return extraHeaderRows;
	}

	/**
	 * Set the extra header rows count.
	 * 
	 * @param extraHeaderRows
	 */
	public void setExtraHeaderRows(int extraHeaderRows) {
		if (extraHeaderRows < 0) {
			throw new IllegalArgumentException(
					"The 'extraHeaderRows' value must be positive.");
		}
		this.extraHeaderRows = extraHeaderRows;
	}

	/**
	 * Get the extra footer rows count.
	 * 
	 * @return
	 */
	public int getExtraFooterRows() {
		return extraFooterRows;
	}

	/**
	 * Set the extra footer rows count.
	 * 
	 * @param extraFooterRows
	 */
	public void setExtraFooterRows(int extraFooterRows) {
		if (extraFooterRows < 0) {
			throw new IllegalArgumentException(
					"The 'extraFooterRows' value must be positive.");
		}
		this.extraFooterRows = extraFooterRows;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "MergePostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to merge result CSV files.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "MergePostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Read parameters.
		filePattern = parameters.getProperty("filePattern");

		deleteMergedFiles = Boolean.valueOf(
				StringUtils.defaultIfEmpty(
						parameters.getProperty("deleteMergedFiles"), "true"))
				.booleanValue();

		sort = Boolean.valueOf(
				StringUtils.defaultIfEmpty(parameters.getProperty("sort"),
						"false")).booleanValue();

		removeDuplicates = Boolean.valueOf(
				StringUtils.defaultIfEmpty(
						parameters.getProperty("removeDuplicates"), "true"))
				.booleanValue();

		forceSortMatchingFiles = Boolean.valueOf(
				StringUtils.defaultIfEmpty(
						parameters.getProperty("forceSortMatchingFiles"),
						"false")).booleanValue();

		try {
			setExtraHeaderRows(Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("extraHeaderRows"), "0")));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'extraHeaderRows' parameter must have a numeric value.");
		}

		try {
			setExtraFooterRows(Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("extraFooterRows"), "0")));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'extraFooterRows' parameter must have a numeric value.");
		}

		// Note: Keys must be set in the execute method because they depend on
		// the CSV settings.
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			// Set the keys parameter.
			setKeys(CsvUtils.readFirstLine(new StringReader(parameters
					.getProperty("keys", "").trim()), getCsvDelimiter(),
					getCsvQuoteCharacter()));

			// Set the parentKeys parameter.
			setParentKeys(CsvUtils.readFirstLine(new StringReader(parameters
					.getProperty("parentKeys", "").trim()), getCsvDelimiter(),
					getCsvQuoteCharacter()));

			// Load the query file.
			File queryFile = new File(pipeline.getRequestMessage()
					.getOriginalFile());

			// If the keys are empty, read them from the query.
			if (getKeys().length == 0) {
				setKeys(getKeysFromTag(queryFile, "key"));
			}

			// If the keys and parentKeys are empty, read them from the query.
			if (getKeys().length == 0 && getParentKeys().length == 0) {
				setParentKeys(getKeysFromTag(queryFile, "parentKey"));
			}

			// Get the input file.
			Message responseMessage = pipeline.getResponseMessage();
			File inFile = new File(responseMessage.getCurrentFile());

			// Set the default file pattern.
			if (StringUtils.isBlank(this.filePattern)) {
				this.filePattern = responseMessage.getOriginalFile();
			}

			// Replace the "." pattern.
			String[] patterns = StringUtils.split(filePattern, "|");
			for (int i = 0; i < patterns.length; i++) {
				if (patterns[i].trim().equals(".")) {
					patterns[i] = responseMessage.getOriginalFile();
				}
			}
			this.filePattern = StringUtils.join(patterns, "|");

			// Manage the correct merge process depending if parent keys are
			// present or not.
			File outFile = null;
			if (getParentKeys().length > 0) {
				outFile = childMerge(inFile);
			} else {
				outFile = merge(inFile);
			}

			// Set the new current file.
			responseMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Merge the files corresponding to the filePattern with the provided
	 * inFile.
	 * 
	 * @param inFile
	 *            The input file
	 * @return The merged file.
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	protected File merge(File inFile) throws IOException, InvalidCSVFileFormat {
		BufferedReader reader = null;
		BufferedWriter writer = null;
		File mergedFile = null;

		// Read the CSV header if present.
		String[] header = null;
		if (isCsvHeaderPresent()) {
			header = CsvUtils.readLine(
					new InputStreamReader(IOHelper.getInputStream(inFile),
							getEncoding()), getCsvDelimiter(),
					getCsvQuoteCharacter(), getExtraHeaderRows() + 1);
		}

		// Get the key column indexes.
		int[] keyIndexes = CsvUtils.getColumnIndexes(getKeys(), header);

		if (keyIndexes.length > 0) {
			log.info("Merge using columns "
					+ ArrayUtils.toString(keyIndexes)
					+ " as key"
					+ (isRemoveDuplicates() ? " and removing duplicates."
							: (isSort() ? " and sorting." : ".")));
		}

		try {
			boolean headerWritten = false;
			boolean hasExtraHeaderFooter = (getExtraHeaderRows()
					+ getExtraFooterRows() > 0);

			// Get files matching the pattern.
			File[] files = getMatchingFiles(filePattern);
			log.info(files.length + " file(s) match the pattern: "
					+ filePattern + ".");

			// Create a temp file for merging.
			mergedFile = createTempFile();
			mergedFile.deleteOnExit();
			log.info("Merging into " + mergedFile.getAbsolutePath());

			// Add inFile to the array of files to be merged.
			files = (File[]) ArrayUtils.add(files, inFile);

			// Initialize a writer.
			writer = new BufferedWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(new FileOutputStream(mergedFile,
							true), getClass().getName()), getEncoding()));

			// Process each file.
			for (int i = 0; i < files.length; i++) {
				File file = files[i];
				log.info("Merging file " + file.getAbsolutePath());

				// The inFile has no extra header footer.
				if (file == inFile) {
					hasExtraHeaderFooter = false;
				}

				// Get the line count.
				int lineCount = 0;
				if (hasExtraHeaderFooter) {
					lineCount = FileUtils.getLineCount(file);
				}

				try {
					// Initialize a reader.
					reader = new BufferedReader(new InputStreamReader(
							IOHelper.getInputStream(file), getEncoding()));
					String line;
					int index = 0;
					int j = 0;
					while ((line = reader.readLine()) != null) {
						index++;

						// Write each non empty line.
						if (line.trim().length() > 0) {

							// Skip extra header and footer rows.
							if (hasExtraHeaderFooter) {
								if (index <= getExtraHeaderRows()) {
									if (log.isDebugEnabled()) {
										log.debug("Skipping extra header row: "
												+ line);
									}
									continue;
								}
								if (index > lineCount - getExtraFooterRows()) {
									if (log.isDebugEnabled()) {
										log.debug("Skipping extra footer row: "
												+ line);
									}
									continue;
								}
							}

							// Skip the CSV header row if present and it was
							// written already.
							if (j++ == 0 && isCsvHeaderPresent()
									&& headerWritten) {
								continue;
							} else {
								headerWritten = true;
							}

							// Write the line.
							writer.write(line);
							writer.newLine();
						}
						writer.flush();
					}

				} finally {
					if (reader != null) {
						reader.close();
					}
				}
			}
			writer.close();

			// Sort if duplicates must be removed or if requested explicitly.
			if (isRemoveDuplicates() || isSort()) {
				sort(mergedFile, mergedFile, keyIndexes, isRemoveDuplicates());
			}

			// Delete merged files (except the inFile) if everything went fine.
			if (deleteMergedFiles) {
				for (int i = 0; i < files.length - 1; i++) {
					File file = files[i];
					if (file != inFile) {
						if (log.isDebugEnabled()) {
							log.debug("Deleting file " + file.getAbsolutePath()
									+ ".");
						}
						file.delete();
					}
				}
			}

			// Return the merged file.
			return mergedFile;

		} finally {
			if (writer != null) {
				writer.close();
			}
		}
	}

	/**
	 * Merge the files corresponding to the filePattern with the provided
	 * inFile. This method applies the merge considering the parent keys. To do
	 * so, it prepares the outFile by making one row per parent and all its
	 * children in a single Base64 encoded field. It then sorts on the parent
	 * keys and finally expands back the children in the output file.
	 * <p>
	 * Note: This strategy is relevant only if duplicates must be removed. If it
	 * is not so, a simple merge is performed.
	 * 
	 * @param inFile
	 *            The input file
	 * @return The merged file.
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	protected File childMerge(File inFile) throws IOException,
			InvalidCSVFileFormat {
		CSVReader reader = null;
		CsvWriter writer = null;
		File mergedFile = null;

		// If duplicates are not removed, do a simple merge.
		if (!isRemoveDuplicates()) {
			return merge(inFile);
		}

		// Read the CSV header if present.
		String[] header = null;
		if (isCsvHeaderPresent()) {
			Reader headerReader = new InputStreamReader(
					IOHelper.getInputStream(inFile), getEncoding());
			header = CsvUtils.readLine(headerReader, getCsvDelimiter(),
					getCsvQuoteCharacter(), getExtraHeaderRows() + 1);
			headerReader.close();
		}

		// Get the key column indexes.
		int[] keyIndexes = CsvUtils.getColumnIndexes(getKeys(), header);
		int[] parentKeyIndexes = CsvUtils.getColumnIndexes(getParentKeys(),
				header);

		log.info("Child merge using columns "
				+ ArrayUtils.toString(keyIndexes)
				+ " as key and "
				+ ArrayUtils.toString(parentKeyIndexes)
				+ " as parent key"
				+ (isRemoveDuplicates() ? " and removing duplicates."
						: (isSort() ? " and sorting." : ".")));

		try {
			// Get files matching the pattern.
			File[] files = getMatchingFiles(filePattern);
			log.info(files.length + " file(s) match the pattern: "
					+ filePattern + ".");

			// Add inFile to the array of files to be merged.
			files = (File[]) ArrayUtils.add(files, inFile);

			// ### 1. Merge files aggregating child rows on a single row with
			// their parent keys. ###
			{
				// Prepare a row comparator on parent keys.
				SortCriterion[] sortCriteria = new SortCriterion[parentKeyIndexes.length];
				for (int i = 0; i < parentKeyIndexes.length; i++) {
					sortCriteria[i] = new SortCriterion(parentKeyIndexes[i]);
				}
				RowComparator comparator = new RowComparator(sortCriteria);

				// Create a temporary file for merging.
				mergedFile = createTempFile();
				mergedFile.deleteOnExit();

				// Initialize a CSV writer.
				writer = new CsvWriter(new BufferedWriter(
						new OutputStreamWriter(IOHelper.getOutputStream(
								new FileOutputStream(mergedFile, true),
								getClass().getName()), getEncoding())),
						getCsvDelimiter(), getCsvQuoteCharacter());

				// Process each file.
				boolean headerWritten = false;
				boolean hasExtraHeaderFooter = (getExtraHeaderRows()
						+ getExtraFooterRows() > 0);
				for (int i = 0; i < files.length; i++) {
					File file = files[i];
					log.info("Merging file " + file.getAbsolutePath());

					// Sort the last file (inFile) and force sorting the other
					// files only if required.
					if (file == inFile || forceSortMatchingFiles) {
						File tempFile = createTempFile();
						tempFile.deleteOnExit();

						// Make sure no file has extra header and footer, for
						// sorting.
						if (file != inFile) {
							copyWithoutExtraHeaderFooter(file, tempFile);
						} else {
							org.apache.commons.io.FileUtils.copyFile(file,
									tempFile);
						}
						hasExtraHeaderFooter = false;

						// Sort on keys and remove duplicates.
						sort(tempFile, tempFile, keyIndexes, true);

						// Sort file on parent keys, NOT removing duplicates.
						sort(tempFile, tempFile, parentKeyIndexes, false);

						// Reassign the file.
						file = tempFile;
					}

					// Get the line count.
					int lineCount = 0;
					if (hasExtraHeaderFooter) {
						lineCount = FileUtils.getLineCount(file);
					}

					String[] parentRow = null;
					String[] previousRow = null;
					String[] currentRow = null;
					List childRows = new ArrayList();
					int index = 0;
					int j = 0;
					boolean eof = false;
					try {
						// Initialize a CSV reader.
						reader = new CSVReader(new InputStreamReader(
								IOHelper.getInputStream(file), getEncoding()),
								false, getCsvDelimiter(),
								getCsvQuoteCharacter());

						List line = new ArrayList();
						do {
							eof = !reader.readLine(line);
							index++;
							if (!eof) {

								// Skip extra header and footer rows.
								if (hasExtraHeaderFooter) {
									if (index <= getExtraHeaderRows()) {
										if (log.isDebugEnabled()) {
											log.debug("Skipping extra header row: "
													+ line);
										}
										continue;
									}
									if (index > lineCount
											- getExtraFooterRows()) {
										if (log.isDebugEnabled()) {
											log.debug("Skipping extra footer row: "
													+ line);
										}
										continue;
									}
								}

								// Get the current row.
								currentRow = (String[]) line
										.toArray(new String[line.size()]);

								// Write the header row if present and it was
								// not written already.
								if (j++ == 0 && isCsvHeaderPresent()) {
									if (!headerWritten) {
										writer.writeLine(currentRow);
										writer.flush();
										headerWritten = true;
									}
									continue;
								}
							}

							// If starting a new set of child rows.
							if (previousRow == null
									|| comparator.compare(previousRow,
											currentRow) != 0 || eof) {

								// Process the previous set of child rows.
								if (childRows.size() > 0) {

									// Create a parent row with the parent keys.
									parentRow = new String[parentKeyIndexes.length + 1];
									for (int k = 0; k < parentKeyIndexes.length; k++) {
										parentRow[k] = previousRow[parentKeyIndexes[k]];
									}

									// Serialize the child rows and add to
									// parent row.
									parentRow[parentKeyIndexes.length] = CsvUtils
											.serializeBase64(childRows);

									// Write to file.
									writer.writeLine(parentRow);
									writer.flush();
								}

								// Reset child rows.
								childRows.clear();
							}

							// Substitute previous row and store in child rows.
							previousRow = currentRow;
							childRows.add(currentRow);
						} while (!eof);

					} finally {
						if (reader != null) {
							reader.close();
						}
					}
				}
				writer.close();
			}

			// ### 2. Sort on parent keys and remove duplicates. ###
			{
				// Since parent indexes are now the first columns in the
				// aggregated file, calculate their indexes for sorting.
				int[] indexes = new int[parentKeyIndexes.length];
				for (int i = 0; i < parentKeyIndexes.length; i++) {
					indexes[i] = i;
				}
				sort(mergedFile, mergedFile, indexes, true);
			}

			// Create a temporary file for output.
			File outFile = createTempFile();
			outFile.deleteOnExit();

			// ### 3. Expand the child rows. ###
			{
				// Initialize a CSV writer on the outFile.
				writer = new CsvWriter(new BufferedWriter(
						new OutputStreamWriter(IOHelper.getOutputStream(
								new FileOutputStream(outFile, true), getClass()
										.getName()), getEncoding())),
						getCsvDelimiter(), getCsvQuoteCharacter());

				try {
					// Initialize a CSV reader.
					reader = new CSVReader(
							new InputStreamReader(
									IOHelper.getInputStream(mergedFile),
									getEncoding()), false, getCsvDelimiter(),
							getCsvQuoteCharacter());

					List row = new ArrayList();
					List childRows = null;
					int j = 0;
					while (reader.readLine(row)) {
						// Write the header row (if present), only for the first
						// file.
						if (isCsvHeaderPresent() && j++ == 0) {
							writer.writeLine(row);
							writer.flush();
							continue;
						}

						// Deserialize child rows.
						try {
							childRows = (List) CsvUtils
									.deserializeBase64((String) row.get(row
											.size() - 1));
						} catch (ClassNotFoundException e) {
							// Can't happen.
						}

						// Write each child row.
						for (Iterator iterator = childRows.iterator(); iterator
								.hasNext();) {
							writer.writeLine((String[]) iterator.next());
							writer.flush();
						}
					}
				} finally {
					if (reader != null) {
						reader.close();
					}
				}
				if (writer != null) {
					writer.close();
				}
			}

			// Delete merged files (except the inFile) if everything went fine.
			if (deleteMergedFiles) {
				for (int i = 0; i < files.length; i++) {
					File file = files[i];
					if (file != inFile) {
						if (log.isDebugEnabled()) {
							log.debug("Deleting file " + file.getAbsolutePath()
									+ ".");
						}
						file.delete();
					}
				}
			}

			// Return the output file.
			return outFile;

		} finally {
			if (writer != null) {
				writer.close();
			}
		}
	}

	/**
	 * Get the list of files matching the pattern.
	 * 
	 * @param pattern
	 * @return
	 * @throws IOException
	 */
	private File[] getMatchingFiles(String pattern) throws IOException {
		File[] files = new File[0];
		String[] patterns = StringUtils.split(pattern, "|");
		for (int i = 0; i < patterns.length; i++) {
			File[] tempFiles = FileUtils
					.getMatchingFiles(new File(patterns[i]));
			FileUtils.sortFilesByNameAscending(tempFiles, true);
			files = (File[]) ArrayUtils.addAll(files, tempFiles);
		}
		return files;
	}

	/**
	 * Sort the inFile and write result to the outFile, removing duplicates if
	 * required.
	 * 
	 * @param inFile
	 *            The input file
	 * @param outFile
	 *            The output file
	 * @param indexes
	 *            The sort column indexes
	 * @param removeDuplicates
	 *            Remove duplicates
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	protected void sort(File inFile, File outFile, int[] indexes,
			boolean removeDuplicates) throws IOException, InvalidCSVFileFormat {

		// Build an array of SortCriterion.
		SortCriterion[] sortCriteria = new SortCriterion[indexes.length];
		for (int i = 0; i < indexes.length; i++) {
			sortCriteria[i] = new SortCriterion(indexes[i]);
		}

		if (sortCriteria.length == 0) {
			// Do nothing if there are no sort criteria.
			if (sort) {
				log.info("No sorting will occur because no key is defined.");
			}
			if (removeDuplicates) {
				log.info("Duplicates won't be removed because no key is defined.");
			}
		} else {
			// Otherwise, do the sort.
			if (log.isDebugEnabled()) {
				log.debug("Sorting "
						+ inFile.getAbsolutePath()
						+ " on columns: "
						+ ArrayUtils.toString(indexes)
						+ (removeDuplicates ? " and removing duplicates." : "."));
			}
			MergeSort sort = new MergeSort();
			sort.setCsvHeaderPresent(isCsvHeaderPresent());
			sort.setCsvDelimiter(getCsvDelimiter());
			sort.setCsvQuoteCharacter(getCsvQuoteCharacter());
			sort.setRemoveDuplicates(removeDuplicates);
			sort.setEncoding(getEncoding());
			sort.setWorkingFolder(tempFolder);
			sort.sort(inFile, new RowComparator(sortCriteria), outFile);
		}
	}

	/**
	 * Get the key indexes from tag attribute of projections in the SQ-XML
	 * query.
	 * 
	 * @param queryFile
	 *            The query file.
	 * @param tagName
	 *            The tag name identifying the keys.
	 * @return
	 * @throws TransformerException
	 * @throws IOException
	 */
	protected String[] getKeysFromTag(File queryFile, String tagName)
			throws TransformerException, IOException {

		// Create an XSL transformer.
		InputStream xsl = this.getClass().getClassLoader()
				.getResourceAsStream(XSL_TRANSFORM);
		TransformerFactory tFactory = TransformerFactory.newInstance();
		Transformer transformer = tFactory
				.newTransformer(new StreamSource(xsl));

		// Set parameters.
		transformer.setParameter("tagName", tagName);

		// Prepare streams.
		Reader reader = new InputStreamReader(
				IOHelper.getInputStream(queryFile), getEncoding());
		Writer writer = new StringWriter();

		// Prepare source and result.
		Source source = new StreamSource(reader);
		Result result = new StreamResult(writer);

		// Apply the transformation.
		transformer.transform(source, result);

		reader.close();
		return StringUtils.split(writer.toString(), ',');
	}

	/**
	 * Copy a file without its extra header and footer rows.
	 * 
	 * @param inFile
	 * @param outFile
	 * @return
	 * @throws IOException
	 */
	private void copyWithoutExtraHeaderFooter(File inFile, File outFile)
			throws IOException {

		// Get the line count.
		int lineCount = FileUtils.getLineCount(inFile);

		BufferedReader reader = null;
		BufferedWriter writer = null;
		int index = 0;
		try {
			reader = new BufferedReader(new InputStreamReader(
					IOHelper.getInputStream(inFile), getEncoding()));
			writer = new BufferedWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(
							new FileOutputStream(outFile, true), getClass()
									.getName()), getEncoding()));

			// Skip extra header rows.
			String line;
			for (int i = 0; i < getExtraHeaderRows(); i++) {
				line = reader.readLine();
				index++;

				if (log.isDebugEnabled()) {
					log.debug("Skipping header row: " + line);
				}
			}

			// Copy content until the extra footer rows are reached.
			while ((line = reader.readLine()) != null) {
				index++;
				if (index <= lineCount - getExtraFooterRows()) {
					writer.write(line);
					writer.newLine();
				} else {
					if (log.isDebugEnabled()) {
						log.debug("Skipping footer row: " + line);
					}
				}
			}

		} finally {
			if (reader != null) {
				reader.close();
			}
			if (writer != null) {
				writer.close();
			}
		}
	}
}
