package com.taleo.integration.client.customstep.csv;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;

import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedMessage;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Post-processing step to remove empty lines in a CSV file.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep}.
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/RemoveEmptyLinesPostStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class RemoveEmptyLinesPostStep extends BaseCustomCsvStep {

	private static Logger log = Logger
			.getLogger(RemoveEmptyLinesPostStep.class);

	/**
	 * Constructor without arguments.
	 */
	public RemoveEmptyLinesPostStep() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "RemoveEmptyLinesPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to remove empty lines from a CSV file.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "RemoveEmptyLinesPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// First, get the basic supported pipeline for a post-processing step
		// (this handles the constraints on the last step type)
		SupportedPipeline supportedPipeline = getPostProcessSupportPipeline();

		// Add the supported response message information
		SupportedMessage supportedMessage = supportedPipeline
				.addNewSupportedResponseMessage();

		// Add the supported response message format
		supportedMessage.addNewSupportedFormats();
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.CSV);

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the input file.
			Message responseMessage = pipeline.getResponseMessage();
			File inFile = new File(responseMessage.getCurrentFile());

			// Create a temporary file for the output.
			File outFile = createTempFile();

			// Execute.
			execute(inFile, outFile);

			// Set the new current file.
			responseMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Copy the input file to the output file while removing empty lines.
	 * 
	 * @param inFile
	 *            The input file
	 * @param outFile
	 *            The output file
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	public void execute(File inFile, File outFile) throws IOException,
			InvalidCSVFileFormat {

		BufferedReader reader = null;
		BufferedWriter writer = null;
		try {
			// Initialize a reader.
			reader = new BufferedReader(new InputStreamReader(
					IOHelper.getInputStream(inFile), getEncoding()));

			// Initialize a writer.
			writer = new BufferedWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(new FileOutputStream(outFile),
							getClass().getName()), getEncoding()));

			String str;
			int i = 0;
			while ((str = reader.readLine()) != null) {
				if (str.trim().length() > 0) {
					if (i++ > 0) {
						writer.newLine();
					}
					writer.write(str);
				}
				writer.flush();
			}

		} finally {
			if (reader != null) {
				reader.close();
			}
			if (writer != null) {
				writer.close();
			}
		}
	}
}
