package com.taleo.integration.client.customstep.csv;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.csv.CSVReader;
import com.taleo.integration.client.customstep.util.csv.CsvWriter;
import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.integration.client.util.QueryRuntimeHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.query.QueryDocument.Query;

/**
 * Post-processing step to replace strings in a CSV file containing the results
 * of an export. Strings are replaced in every column of the file. The CSV
 * delimiter and quote character are not affected.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvPostStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>strings</b>: The delimited list of strings. Does not allow empty
 * string.</li>
 * <br>
 * <br>
 * <li><b>replaceStrings</b>: The delimited list of replacement strings. Can be
 * of a smaller size than <b>strings</b> in which case extra characters will be
 * replaced with an empty string.</li>
 * <br>
 * <br>
 * <li><b>stringsDelimiter</b>: The delimiter for the lists of strings. The
 * default is a comma.</li>
 * </ul>
 * <p>
 * Note: The string lists must be strictly delimited, all extra spaces will be
 * maintained as is. The following special characters may be used:
 * <ul>
 * <li>\n : line feed</li>
 * <li>\r : carriage return</li>
 * <li>\t : tab</li>
 * <li>\' : quote</li>
 * <li>\" : double quote</li>
 * <li>\\ : backslash</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/ReplaceAllPostStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class ReplaceAllPostStep extends BaseCustomCsvPostStep {

	private static Logger log = Logger.getLogger(ReplaceAllPostStep.class);

	/**
	 * The default string list delimiter.
	 */
	public static final char DEFAULT_STRING_DELIMITER = ',';

	/**
	 * Constructor without arguments.
	 */
	public ReplaceAllPostStep() {
		super();
		// Register valid parameters.
		registerParameterName("strings");
		registerParameterName("replaceStrings");
		registerParameterName("stringsDelimiter");
	}

	/**
	 * The strings list.
	 */
	private String[] strings = new String[0];

	/**
	 * The replacement strings list.
	 */
	private String[] replaceStrings = new String[0];

	/**
	 * @return
	 */
	public String[] getStrings() {
		return strings;
	}

	/**
	 * @param strings
	 */
	public void setStrings(String[] strings) {
		for (int i = 0; i < strings.length; i++) {
			if (strings[i] == null || "".equals(strings[i])) {
				throw new IllegalArgumentException("Argument strings[" + i
						+ "] cannot be empty or null.");
			}
		}
		this.strings = strings;
	}

	/**
	 * @return
	 */
	public String[] getReplaceStrings() {
		return replaceStrings;
	}

	/**
	 * @param replaceStrings
	 */
	public void setReplaceStrings(String[] replaceStrings) {
		for (int i = 0; i < replaceStrings.length; i++) {
			if (replaceStrings[i] == null) {
				throw new IllegalArgumentException("Argument replaceStrings["
						+ i + "] cannot be null.");
			}
		}
		this.replaceStrings = replaceStrings;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "ReplaceAllPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to replace strings all over in a CSV file.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "ReplaceAllPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		char stringsDelimiter = DEFAULT_STRING_DELIMITER;
		String delimiterString = parameters.getProperty("stringsDelimiter");
		if (!StringUtils.isBlank(delimiterString)) {
			char[] chars = delimiterString.toCharArray();
			if (chars.length > 1) {
				throw new IllegalArgumentException(
						"The 'stringsDelimiter' parameter must be a single character.");
			}
			stringsDelimiter = chars[0];
		}

		String strings = StringEscapeUtils.unescapeJava(parameters
				.getProperty("strings"));
		if (!StringUtils.isBlank(strings)) {
			setStrings(StringUtils.splitPreserveAllTokens(strings,
					stringsDelimiter));
		}

		String replaceStrings = StringEscapeUtils.unescapeJava(parameters
				.getProperty("replaceStrings"));
		if (!StringUtils.isBlank(replaceStrings)) {
			setReplaceStrings(StringUtils.splitPreserveAllTokens(
					replaceStrings, stringsDelimiter));
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			// Load the query file.
			File requestFile = new File(pipeline.getRequestMessage()
					.getOriginalFile());
			Query query = QueryRuntimeHelper.loadQueryInStep(requestFile,
					productCode);

			// Get the CSV delimiter.
			if (query.getCsvdelimiter() != null) {
				setCsvDelimiter(query.getCsvdelimiter().charAt(0));
			}

			// Get the CSV quote character.
			if (query.getCsvquote() != null) {
				setCsvQuoteCharacter(query.getCsvquote().charAt(0));
			}

			// Get the input file.
			Message responseMessage = pipeline.getResponseMessage();
			File inFile = null;
			if (responseMessage.isSetCurrentFile()) {
				inFile = new File(responseMessage.getCurrentFile());
			} else {
				inFile = new File(responseMessage.getOriginalFile());
			}

			// Create a temporary file for the output.
			File outFile = createTempFile();

			// Apply the replacement.
			execute(inFile, outFile);

			// Set the new current file.
			responseMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Copy the input file to the output file while applying the required
	 * replacements.
	 * 
	 * @param inFile
	 *            The input file
	 * @param outFile
	 *            The output file
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	public void execute(File inFile, File outFile) throws IOException,
			InvalidCSVFileFormat {

		CSVReader reader = null;
		CsvWriter writer = null;
		try {
			// Initialize a CSV reader.
			reader = new CSVReader(new InputStreamReader(
					IOHelper.getInputStream(inFile), getEncoding()), false,
					getCsvDelimiter(), getCsvQuoteCharacter());

			// Initialize a CSV writer.
			writer = new CsvWriter(new BufferedWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(new FileOutputStream(outFile),
							getClass().getName()), getEncoding())),
					getCsvDelimiter(), getCsvQuoteCharacter());

			// Copy the contents while applying the replacements.
			List values = new ArrayList();
			while (reader.readLine(values)) {
				for (int i = 0; i < values.size(); i++) {
					values.set(i, replace((String) values.get(i)));
				}
				writer.writeLine(values);
				writer.flush();
			}

		} finally {
			if (reader != null) {
				reader.close();
			}
			if (writer != null) {
				writer.close();
			}
		}
	}

	/**
	 * Execute the replacements on the specified value.
	 * 
	 * @param value
	 *            The value.
	 * @return The value after the replacements.
	 */
	protected String replace(String value) {
		for (int i = 0; i < strings.length; i++) {
			String replace = "";
			if (i < replaceStrings.length) {
				replace = replaceStrings[i];
			}
			value = StringUtils.replace(value, strings[i], replace);
		}
		return value;
	}
}
