package com.taleo.integration.client.customstep.csv;

import java.io.File;

import org.apache.log4j.Logger;

import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedMessage;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Post-processing step to replace strings in a CSV file containing the results
 * of an export. Strings are replaced in every column of the file. The CSV
 * delimiter and quote character are not affected.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.ReplaceAllPostStep}.
 * <p>
 * In the case of an import, the load specification is not accessible from
 * within the custom steps. Consequently, <b>csvDelimiter</b>,
 * <b>csvQuoteCharacter</b> and <b>csvHeaderPresent</b> must be provided
 * otherwise the defaults will be used, respectively: comma (,), double quote
 * (") and <code>false</code>.
 * <p>
 * See {@link com.taleo.integration.client.customstep.csv.ReplaceAllPostStep}
 * for details about usage.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class ReplaceAllPreStep extends ReplaceAllPostStep {

	private static Logger log = Logger.getLogger(ReplaceAllPreStep.class);

	/**
	 * Constructor without arguments.
	 */
	public ReplaceAllPreStep() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "ReplaceAllPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Pre-processing step to replace strings all over in a CSV file.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "ReplaceAllPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		SupportedPipeline supportedPipeline = getPreProcessSupportPipeline();

		// Add the supported response message information
		SupportedMessage supportedMessage = supportedPipeline
				.addNewSupportedRequestMessage();

		// Add the supported request message format
		supportedMessage.addNewSupportedFormats();
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.CSV);

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the input file.
			Message requestMessage = pipeline.getRequestMessage();
			File inFile = null;
			if (requestMessage.isSetCurrentFile()) {
				inFile = new File(requestMessage.getCurrentFile());
			} else {
				inFile = new File(requestMessage.getOriginalFile());
			}

			// Create a temporary file for the output.
			File outFile = createTempFile();

			// Apply the transformation.
			execute(inFile, outFile);

			// Set the new current file.
			requestMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

}
