package com.taleo.integration.client.customstep.csv;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;

import com.taleo.ws.integration.client.GlobalConfig;

/**
 * Post-processing step to replace values based on Regular Expression matching
 * in a CSV file containing the results of an export.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseTransformPostStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>pattern</b>: The Regular Expression matching pattern as defined in
 * {@link java.util.regex.Pattern}.</li>
 * <br>
 * <br>
 * <li><b>replace</b>: The Regular Expression replacement string as defined in
 * {@link java.util.regex.Matcher}.</li>
 * <br>
 * <br>
 * <li><b>replaceAll</b>: Indicates that every substring matching the pattern is
 * replaced with the replacement string (<code>true</code> or <code>false</code>
 * , default: <code>true</code>). If <code>false</code>, only the first matching
 * substring is replaced.</li>
 * </ul>
 * A typical transformation using this step would define a matching pattern
 * containing a group and a replacement string that references this group. For
 * example, the pattern '(\d{3})(.*)' defines a first group with 3 consecutive
 * digits followed by a second group with any number of characters. A
 * replacement string could contain the variable '$1' to output the value of the
 * first matching group and the variable '$2' to ouput the second group.
 * <p>
 * For an online Regular Expression tester, visit <a
 * href="http://www.fileformat.info/tool/regex.htm">Regex tester</a>.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/ReplacePostStep-1.jpg" />
 * 
 * @see java.util.regex.Pattern
 * @see java.util.regex.Matcher
 * @author Romain Guay, Taleo Corporation
 */
public class ReplacePostStep extends BaseTransformPostStep {

	/**
	 * Constructor without arguments.
	 */
	public ReplacePostStep() {
		super();
		// Register valid parameters.
		registerParameterName("pattern");
		registerParameterName("replace");
		registerParameterName("replaceAll");
	}

	/**
	 * The compiled Regex pattern.
	 */
	private Pattern compiledPattern = null;

	/**
	 * The regex matching pattern.
	 */
	private String pattern;

	/**
	 * The regex replacement string.
	 */
	private String replace;

	/**
	 * The indicator to replace all matching substrings.
	 */
	private boolean replaceAll = true;

	/**
	 * @return
	 */
	public String getPattern() {
		return pattern;
	}

	/**
	 * @param pattern
	 */
	public void setPattern(String pattern) {
		this.pattern = pattern;
		compiledPattern = Pattern.compile(pattern);
	}

	/**
	 * @return
	 */
	public String getReplace() {
		return replace;
	}

	/**
	 * @param replace
	 */
	public void setReplace(String replace) {
		this.replace = replace;
	}

	/**
	 * @return
	 */
	public boolean isReplaceAll() {
		return replaceAll;
	}

	/**
	 * @param replaceAll
	 */
	public void setReplaceAll(boolean replaceAll) {
		this.replaceAll = replaceAll;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "ReplacePostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to apply Regular Expression replacement to a CSV column.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "ReplacePostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);
		
		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		pattern = parameters.getProperty("pattern");
		if (StringUtils.isBlank(pattern)) {
			throw new IllegalArgumentException(
					"The 'pattern' parameter is mandatory.");
		}
		compiledPattern = Pattern.compile(pattern);

		replace = parameters.getProperty("replace");

		replaceAll = Boolean.valueOf(
				parameters.getProperty("replaceAll", "true")).booleanValue();
	}

	/**
	 * Transform the given value by applying the Regex matching and replacement.
	 * <p>
	 * If no matching substring is found, an empty string is returned.
	 * 
	 * @param value
	 *            The value to transform.
	 * @return The transformed value.
	 */
	protected String transform(String value) {
		Matcher matcher = compiledPattern.matcher(value);
		if (replaceAll) {
			return matcher.replaceAll(replace);
		} else {
			return matcher.replaceFirst(replace);
		}
	}

}
