package com.taleo.integration.client.customstep.csv;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Pre-processing step to strip header and footer rows from a CSV input file.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvPreStep} and
 * the following additional parameters:
 * <ul>
 * <li><b>headerRows</b>: Optional - Number of header rows to be stripped.
 * (Default: 0). <br>
 * <br>
 * <li><b>footerRows</b>: Optional - Number of footer rows to be stripped.
 * (Default: 0). <br>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/StripHeaderFooterPreStep-1.png" />
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public class StripHeaderFooterPreStep extends BaseCustomCsvPreStep {

	private static Logger log = Logger
			.getLogger(StripHeaderFooterPreStep.class);

	/**
	 * Constructor without arguments.
	 */
	public StripHeaderFooterPreStep() {
		super();
		// Register valid parameters.
		registerParameterName("headerRows");
		registerParameterName("footerRows");
	}

	/**
	 * The headerRows count.
	 */
	private int headerRows = 0;

	/**
	 * The footerRows count.
	 */
	private int footerRows = 0;

	/**
	 * Get the header rows count.
	 * 
	 * @return
	 */
	public int getHeaderRows() {
		return headerRows;
	}

	/**
	 * Set the header rows count.
	 * 
	 * @param headerRows
	 */
	public void setHeaderRows(int headerRows) {
		if (headerRows < 0) {
			throw new IllegalArgumentException(
					"The 'headerRows' value must be positive.");
		}
		this.headerRows = headerRows;
	}

	/**
	 * Get the footer rows count.
	 * 
	 * @return
	 */
	public int getFooterRows() {
		return footerRows;
	}

	/**
	 * Set the footer rows count.
	 * 
	 * @param footerRows
	 */
	public void setFooterRows(int footerRows) {
		if (footerRows < 0) {
			throw new IllegalArgumentException(
					"The 'footerRows' value must be positive.");
		}
		this.footerRows = footerRows;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Strip header and/or footer rows from a CSV input file.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "StripHeaderFooterPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "StripHeaderFooterPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		try {
			setHeaderRows(Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("headerRows"), "0")));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'headerRows' parameter must have a numeric value.");
		}

		try {
			setFooterRows(Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("footerRows"), "0")));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'footerRows' parameter must have a numeric value.");
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			// Get the input file.
			Message requestMessage = pipeline.getRequestMessage();
			File inFile = null;
			if (requestMessage.isSetCurrentFile()) {
				inFile = new File(requestMessage.getCurrentFile());
			} else {
				inFile = new File(requestMessage.getOriginalFile());
			}

			// Create a temporary file for the output.
			File outFile = createTempFile();

			// Execute by adding the header and/or footer.
			execute(inFile, outFile);

			// Set the new current file.
			requestMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * @param inFile
	 *            The input file.
	 * @param outFile
	 *            The output file.
	 * @throws IOException
	 */
	protected void execute(File inFile, File outFile) throws IOException {
		BufferedReader reader = null;
		BufferedWriter writer = null;
		try {
			// Initialize a reader.
			reader = new BufferedReader(new InputStreamReader(
					IOHelper.getInputStream(inFile), getEncoding()));

			// Initialize a writer.
			writer = new BufferedWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(new FileOutputStream(outFile),
							getClass().getName()), getEncoding()));

			// Get the line count.
			int lineCount = FileUtils.getLineCount(inFile);

			String line;
			int index = 0;
			while ((line = reader.readLine()) != null) {
				index++;

				// Write each non empty line.
				if (line.trim().length() > 0) {

					// Skip extra header and footer rows.
					if (index <= getHeaderRows()) {
						if (log.isDebugEnabled()) {
							log.debug("Stripping header row: " + line);
						}
						continue;
					}
					if (index > lineCount - getFooterRows()) {
						if (log.isDebugEnabled()) {
							log.debug("Stripping footer row: " + line);
						}
						continue;
					}

					// Write the line.
					writer.write(line);
					writer.newLine();
				}
				writer.flush();
			}

		} finally {
			if (reader != null) {
				reader.close();
			}
			if (writer != null) {
				writer.close();
			}
		}
	}
}
