package com.taleo.integration.client.customstep.db;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Properties;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.csv.CsvWriter;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.integration.client.util.PasswordUtils;
import com.taleo.integration.client.util.csv.CSVConstant;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * This step executes a Select query on a database and sets the result as input
 * to the TCC import process.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>driverClassName</b>: The JDBC driver class name (e.g.
 * oracle.jdbc.driver.OracleDriver).</li>
 * <br>
 * <li><b>url</b>: The database URL (e.g.
 * jdbc:oracle:thin:@servername:PORT:DATABASE).</li>
 * <br>
 * <li><b>username</b>: The database username.</li>
 * <br>
 * <li><b>password</b>: The AES128 encrypted database password.</li>
 * <br>
 * <li><b>query</b>: The SQL Select query. This parameter may be used in
 * replacement of a <code>queryFile</code> if the query is very short. If both
 * are present, this one has priority.</li>
 * <br>
 * <li><b>queryFile</b>: The file containing the SQL Select query.</li>
 * <br>
 * <li><b>csvDelimiter</b>: The CSV delimiter of the resulting file. (Default:
 * comma ',')</li>
 * <br>
 * <li><b>csvQuoteCharacter</b>: The CSV quote character of the resulting file.
 * (Default: double quote '"')</li>
 * <br>
 * <li><b>csvHeaderPresent</b>: The CSV header presence indicator of the
 * resulting file. (Default: false)</li>
 * <br>
 * <li>All other parameters are passed to the SQL query using the named
 * parameters convention (:NAME). All parameters are passed as String values, so
 * the query must do any necessary conversion (e.g. using to_number() or
 * to_date()). The named parameters conversion is managed prior to sending the
 * query, so is independent of the database implementation. Here are some sample
 * queries:
 * <ul>
 * <li>Select * from MYTABLE where USERID = :USERID</li>
 * <li>Select * from MYTABLE where PRICE >= TO_NUMBER(:MIN_PRICE)</li>
 * </ul>
 * </li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * <li>The JDBC driver library used for connecting to the database.</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file.
 * <p>
 * <img src="doc-files/SelectPreStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class SelectPreStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(SelectPreStep.class);

	/**
	 * Constructor without arguments.
	 */
	public SelectPreStep() {
		super();
		// Register valid parameters.
		registerParameterName("driverClassName");
		registerParameterName("url");
		registerParameterName("username");
		registerParameterName("password");
		registerParameterName("query");
		registerParameterName("queryFile");
		registerParameterName("csvDelimiter");
		registerParameterName("csvQuoteCharacter");
		registerParameterName("csvHeaderPresent");
	}

	/**
	 * Validate the parameters names.
	 * <p>
	 * Override the default method to allow for extra parameters to be passed to
	 * the SQL query.
	 */
	protected void validateParameterNames() {
		Enumeration keys = parameters.keys();
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			if (!parameterNames.contains(key)) {
				log.info("Using SQL parameter " + key + "="
						+ parameters.getProperty(key) + ".");
			}
		}
	}

	/**
	 * The driver class name to connect to the database.
	 */
	private String driverClassName = null;

	/**
	 * The URL to connect to the database.
	 */
	private String url = null;

	/**
	 * The username to connect to the database.
	 */
	private String username = null;

	/**
	 * The password to connect to the database.
	 */
	private String password = null;

	/**
	 * The SQL query.
	 */
	private String query = null;

	/**
	 * The SQL query file.
	 */
	private String queryFile = null;

	/**
	 * The CSV delimiter.
	 */
	private char csvDelimiter = CSVConstant.DEFAULT_DELIMITER;

	/**
	 * The CSV quote character.
	 */
	private char csvQuoteCharacter = CSVConstant.DEFAULT_QUOTE_CHARACTER;

	/**
	 * The CSV header presence indicator.
	 */
	private boolean csvHeaderPresent = false;

	/**
	 * The SQL parameters.
	 */
	private Properties sqlParameters = new Properties();

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return getClass().getName();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Pre-processing step to extract data from a database.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return getClass().getName();
	}

	/**
	 * Get the SQL query.
	 * 
	 * @return
	 */
	public String getQuery() {
		return query;
	}

	/**
	 * Set the SQL query.
	 * 
	 * @param query
	 */
	public void setQuery(String query) {
		this.query = query;
	}

	/**
	 * Get the SQL query file.
	 * 
	 * @return
	 */
	public String getQueryFile() {
		return queryFile;
	}

	/**
	 * Set the SQL query file.
	 * 
	 * @param queryFile
	 */
	public void setQueryFile(String queryFile) {
		this.queryFile = queryFile;
	}

	/**
	 * Get the driver class name.
	 * 
	 * @return
	 */
	public String getDriverClassName() {
		return driverClassName;
	}

	/**
	 * Set the driver class name.
	 * 
	 * @param driverClassName
	 */
	public void setDriverClassName(String driverClassName) {
		this.driverClassName = driverClassName;
	}

	/**
	 * Get the database URL.
	 * 
	 * @return
	 */
	public String getUrl() {
		return url;
	}

	/**
	 * Set the database URL.
	 * 
	 * @param url
	 */
	public void setUrl(String url) {
		this.url = url;
	}

	/**
	 * Get the database user name.
	 * 
	 * @return
	 */
	public String getUsername() {
		return username;
	}

	/**
	 * Set the database user name.
	 * 
	 * @param username
	 */
	public void setUsername(String username) {
		this.username = username;
	}

	/**
	 * Get the database password.
	 * 
	 * @return
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * Set the database password.
	 * 
	 * @param password
	 */
	public void setPassword(String password) {
		this.password = password;
	}

	/**
	 * Get the CSV delimiter.
	 * 
	 * @return
	 */
	public char getCsvDelimiter() {
		return csvDelimiter;
	}

	/**
	 * Set the CSV delimiter.
	 * 
	 * @param csvDelimiter
	 */
	public void setCsvDelimiter(char csvDelimiter) {
		this.csvDelimiter = csvDelimiter;
	}

	/**
	 * Indicate if the CSV header is to be present.
	 * 
	 * @return
	 */
	public boolean isCsvHeaderPresent() {
		return csvHeaderPresent;
	}

	/**
	 * Set if the CSV header is to be present.
	 * 
	 * @param csvHeaderPresent
	 */
	public void setCsvHeaderPresent(boolean csvHeaderPresent) {
		this.csvHeaderPresent = csvHeaderPresent;
	}

	/**
	 * Get the CSV quote character.
	 * 
	 * @return
	 */
	public char getCsvQuoteCharacter() {
		return csvQuoteCharacter;
	}

	/**
	 * Set the CSV quote character.
	 * 
	 * @param csvQuoteCharacter
	 */
	public void setCsvQuoteCharacter(char csvQuoteCharacter) {
		this.csvQuoteCharacter = csvQuoteCharacter;
	}

	/**
	 * Get the SQL parameter by its name.
	 * 
	 * @param name
	 * 
	 * @return
	 */
	public String getSqlParameter(String name) {
		return sqlParameters.getProperty(name);
	}

	/**
	 * Set the SQL parameter by its name.
	 * 
	 * @param name
	 * @param value
	 */
	public void setSqlParameter(String name, String value) {
		this.sqlParameters.put(name, value);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get parameters.
		driverClassName = parameters.getProperty("driverClassName");
		url = parameters.getProperty("url");
		username = parameters.getProperty("username");
		password = parameters.getProperty("password");

		query = parameters.getProperty("query");
		queryFile = parameters.getProperty("queryFile");
		if (StringUtils.isBlank(query) && StringUtils.isBlank(queryFile)) {
			throw new IllegalArgumentException(
					"Either 'query' or 'queryFile' parameter must be provided.");
		}

		String csvHeaderPresent = (parameters.getProperty("csvHeaderPresent"));
		if (StringUtils.isNotBlank(csvHeaderPresent)) {
			setCsvHeaderPresent(Boolean.valueOf(csvHeaderPresent)
					.booleanValue());
		}

		String csvDelimiter = (parameters.getProperty("csvDelimiter"));
		if (StringUtils.isNotBlank(csvDelimiter)) {
			char[] chars = csvDelimiter.toCharArray();
			if (chars.length > 1) {
				throw new IllegalArgumentException(
						"The 'csvDelimiter' parameter must be a single character.");
			}
			setCsvDelimiter(chars[0]);
		}

		String csvQuoteCharacter = (parameters.getProperty("csvQuoteCharacter"));
		if (StringUtils.isNotBlank(csvQuoteCharacter)) {
			char[] chars = csvQuoteCharacter.toCharArray();
			if (chars.length > 1) {
				throw new IllegalArgumentException(
						"The 'csvQuoteCharacter' parameter must be a single character.");
			}
			setCsvQuoteCharacter(chars[0]);
		}

		// Get the remaining parameters as SQL parameters.
		Enumeration keys = parameters.keys();
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			if (!parameterNames.contains(key)) {
				setSqlParameter(key, parameters.get(key).toString());
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Create a temporary file for the output.
			File outFile = createTempFile();

			// Execute the step.
			execute(outFile);

			// Get the request message.
			Message requestMessage = pipeline.getRequestMessage();

			// Set the original and current file.
			requestMessage.setOriginalFile(outFile.getAbsolutePath());
			requestMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Execute the step and write result in the given file.
	 * 
	 * @throws Exception
	 */
	public void execute(File outFile) throws Exception {
		Connection conn = null;
		Writer writer = null;
		try {
			// Get and prepare the query string.
			String query = this.query;
			if (StringUtils.isBlank(query)) {
				query = org.apache.commons.io.FileUtils.readFileToString(
						new File(queryFile), getEncoding());
			}
			query = prepareQuery(query, sqlParameters);

			if (log.isDebugEnabled()) {
				log.debug("driverClassName: " + driverClassName);
				log.debug("url: " + url);
				log.debug("username: " + username);
				log.debug("password: " + password);
			}

			// Create a connection.
			Class.forName(driverClassName).newInstance();
			conn = DriverManager.getConnection(url, username,
					PasswordUtils.decryptAES128(password));

			// Build a statement.
			conn.setReadOnly(true);
			Statement stmt = conn.createStatement(ResultSet.TYPE_FORWARD_ONLY,
					ResultSet.CONCUR_READ_ONLY);

			// Run the statement.
			if (log.isDebugEnabled()) {
				log.debug("Executing SQL query: " + query);
			}
			ResultSet resultSet = stmt.executeQuery(query);

			// Trace query results.
			if (log.isTraceEnabled()) {
				ResultSetMetaData metadata = resultSet.getMetaData();
				for (int i = 0; i < resultSet.getMetaData().getColumnCount(); i++) {
					log.trace(metadata.getColumnName(i + 1) + ".type = "
							+ metadata.getColumnClassName(i + 1));
				}
			}

			// Write to CSV file.
			log.info("Writing database query results to '"
					+ outFile.getAbsolutePath() + "'.");
			writer = new OutputStreamWriter(IOHelper.getOutputStream(
					new FileOutputStream(outFile), getClass().getName()),
					getEncoding());
			writeCsv(resultSet, writer);
			resultSet.close();
			stmt.close();

		} finally {
			if (writer != null) {
				writer.close();
			}
			if (conn != null) {
				conn.close();
			}
		}
	}

	/**
	 * Prepare the query by stripping any ";" ending character and resolving
	 * named parameters.
	 * 
	 * @param query
	 * @return
	 */
	protected String prepareQuery(String query, Properties parameters) {

		// Strip the ending ";"
		query = StringUtils.stripEnd(query.trim(), ";");

		// Resolve parameters.
		for (Iterator iterator = sqlParameters.entrySet().iterator(); iterator
				.hasNext();) {
			Entry entry = (Entry) iterator.next();
			query = StringUtils.replace(query, ":" + entry.getKey().toString(),
					"'" + entry.getValue().toString() + "'");
		}
		return query;
	}

	/**
	 * Write a resultSet in a writer in CSV format.
	 * 
	 * @param resultSet
	 *            The resultSet.
	 * @param writer
	 *            The writer.
	 * @throws SQLException
	 * @throws IOException
	 */
	private void writeCsv(ResultSet resultSet, Writer writer)
			throws SQLException, IOException {

		CsvWriter csvWriter = new CsvWriter(writer);
		ResultSetMetaData metadata = resultSet.getMetaData();
		String[] values = new String[metadata.getColumnCount()];

		// Write CSV header.
		if (csvHeaderPresent) {
			for (int i = 0; i < metadata.getColumnCount(); i++) {
				values[i] = metadata.getColumnName(i + 1);
			}
			csvWriter.writeLine(values);
			csvWriter.flush();
		}

		// Write each row.
		int rowIndex = 0;
		while (resultSet.next()) {
			rowIndex++;
			try {
				for (int i = 0; i < values.length; i++) {
					values[i] = resultSet.getString(i + 1);
				}
				csvWriter.writeLine(values);
				writer.flush();

			} catch (IOException e) {
				throw e;
			} catch (Exception e) {
				throw new RuntimeException("Error processing row: " + rowIndex,
						e);
			}
		}

		if (log.isDebugEnabled()) {
			log.debug("Extracted " + rowIndex + " rows.");
		}
	}

}
