package com.taleo.integration.client.customstep.file;

import java.io.File;
import java.io.IOException;

import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.csv.BaseCustomCsvStep;
import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Processing step to delete files from disk.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>filePattern</b>: Mandatory - The pattern for selecting files to
 * delete. Can be specified as an absolute or relative path and may use
 * wildcards as defined in
 * {@link org.apache.commons.io.filefilter.WildcardFileFilter}, in the file
 * names only. Pattern is case sensitive.</li>
 * </ul>
 * <p>
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class DeleteFileStep extends BaseCustomCsvStep {

	private static Logger log = Logger.getLogger(DeleteFileStep.class);

	/**
	 * The file pattern.
	 */
	protected String filePattern = null;

	/**
	 * Constructor without arguments.
	 */
	public DeleteFileStep() {
		super();
		registerParameterName("filePattern");
	}

	/**
	 * Get the file pattern.
	 * 
	 * @return
	 */
	public String getFilePattern() {
		return filePattern;
	}

	/**
	 * Set the file pattern.
	 * 
	 * @param filePattern
	 */
	public void setFilePattern(String filePattern) {
		this.filePattern = filePattern;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "DeleteFileStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Processing step to delete files.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "DeleteFileStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		filePattern = parameters.getProperty("filePattern");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			delete();
		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Delete files matching the pattern.
	 * 
	 * @throws IOException
	 */
	public void delete() throws IOException {
		// Delete files matching the pattern.
		File[] files = FileUtils.getMatchingFiles(new File(filePattern));
		log.info(files.length + " files are matching the pattern '" + filePattern + "'.");
		for (int i = 0; i < files.length; i++) {
			log.info("Deleting file " + files[i]);
			org.apache.commons.io.FileUtils.forceDelete(files[i]);
		}
	}
}
