package com.taleo.integration.client.customstep.file;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.csv.BaseCustomCsvStep;
import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Processing step to move files on disk.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>filePattern</b>: Mandatory - The pattern for selecting files to move.
 * Can be specified as an absolute or relative path and may use wildcards as
 * defined in {@link org.apache.commons.io.filefilter.WildcardFileFilter}, in
 * the file names only. Pattern is case sensitive.</li>
 * <li><b>moveDirectory</b>: Mandatory - The directory to move files to.</li>
 * <br>
 * <li><b>overwrite</b>: Optional - Overwrite existing files with the same name
 * (<code>true</code> or <code>false</code>, default: <code>false</code>). If
 * the destination file already exist and overwrite is false, the file is not
 * moved and a warning is written in the log.</li>
 * </ul>
 * <p>
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class MoveFileStep extends BaseCustomCsvStep {

	private static Logger log = Logger.getLogger(MoveFileStep.class);

	/**
	 * The file pattern.
	 */
	protected String filePattern = null;

	/**
	 * The directory to move files to.
	 */
	protected String moveDirectory = null;

	/**
	 * The overwrite flag.
	 */
	private boolean overwrite = false;

	/**
	 * Constructor without arguments.
	 */
	public MoveFileStep() {
		super();
		registerParameterName("filePattern");
		registerParameterName("moveDirectory");
		registerParameterName("overwrite");
	}

	/**
	 * Get the file pattern.
	 * 
	 * @return
	 */
	public String getFilePattern() {
		return filePattern;
	}

	/**
	 * Set the file pattern.
	 * 
	 * @param filePattern
	 */
	public void setFilePattern(String filePattern) {
		this.filePattern = filePattern;
	}

	/**
	 * Get the move directory.
	 * 
	 * @return
	 */
	public String getMoveDirectory() {
		return moveDirectory;
	}

	/**
	 * Set the move directory.
	 * 
	 * @param moveDirectory
	 */
	public void setMoveDirectory(String moveDirectory) {
		this.moveDirectory = moveDirectory;
	}

	/**
	 * Get the overwrite flag.
	 * 
	 * @return
	 */
	public boolean isOverwrite() {
		return overwrite;
	}

	/**
	 * Set the overwrite flag.
	 * 
	 * @param overwrite
	 */
	public void setOverwrite(boolean overwrite) {
		this.overwrite = overwrite;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "MoveFileStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Processing step to move files.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "MoveFileStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		filePattern = parameters.getProperty("filePattern");
		moveDirectory = parameters.getProperty("moveDirectory");
		overwrite = Boolean.valueOf(
				StringUtils.defaultIfEmpty(parameters.getProperty("overwrite"),
						"false")).booleanValue();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			move();
		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Move files matching the pattern to the move directory.
	 * 
	 * @throws IOException
	 */
	public void move() throws IOException {

		// Force creating move directory.
		org.apache.commons.io.FileUtils.forceMkdir(new File(moveDirectory));

		// Move files.
		File[] files = FileUtils.getMatchingFiles(new File(filePattern));
		for (int i = 0; i < files.length; i++) {
			log.info("Moving file " + files[i] + " to " + moveDirectory);
			File destFile = new File(moveDirectory, files[i].getName());

			// Check overwrite.
			if (destFile.exists() && !overwrite) {
				log.warn("Could not move file " + files[i] + " to "
						+ moveDirectory + ", because it already exist.");
				continue;
			}

			// Copy to destination.
			FileInputStream is = null;
			FileOutputStream os = null;
			try {
				is = new FileInputStream(files[i]);
				os = new FileOutputStream(destFile, false);
				IOUtils.copy(is, os);
			} finally {
				os.close();
				is.close();
			}

			// Delete the original.
			files[i].delete();
		}
	}

}
