package com.taleo.integration.client.customstep.file;

import java.io.File;
import java.io.InputStream;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.count.CountManager;
import com.taleo.integration.client.customstep.count.ExportCountPostStep;
import com.taleo.integration.client.customstep.csv.BaseCustomCsvStep;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Processing step to write a copy of the current file on disk.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.csv.BaseCustomCsvStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>filename</b>: Optional - Full path (absolute or relative) to store the
 * content of the current file from the pipeline in an uncompressed format. If
 * absent, the response filename set in the configuration is used.</li>
 * <li><b>skipIfNoRecords</b>: Optional, default <code>false</code> - Applies on
 * export results only. Skip this step if the record count is 0. Gets the record
 * count from the {@link CountManager} that must have has been set in the
 * pipeline by {@link ExportCountPostStep}.</li>
 * </ul>
 * <p>
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class WriteFileStep extends BaseCustomCsvStep {

	private static Logger log = Logger.getLogger(WriteFileStep.class);

	/**
	 * The filename.
	 */
	private String filename;

	/**
	 * The skipIfNoRecords flag.
	 */
	private boolean skipIfNoRecords = false;

	/**
	 * Constructor without arguments.
	 */
	public WriteFileStep() {
		super();
		registerParameterName("filename");
		registerParameterName("skipIfNoRecords");
	}

	/**
	 * Get the filename.
	 * 
	 * @return
	 */
	public String getFilename() {
		return filename;
	}

	/**
	 * Set the filename.
	 * 
	 * @param filename
	 */
	public void setFilename(String filename) {
		this.filename = filename;
	}

	/**
	 * Get the SkipIfNoRecords flag.
	 * 
	 * @return
	 */
	public boolean isSkipIfNoRecords() {
		return skipIfNoRecords;
	}

	/**
	 * Set the SkipIfNoRecords flag.
	 * 
	 * @param skipIfNoRecords
	 */
	public void setSkipIfNoRecords(boolean skipIfNoRecords) {
		this.skipIfNoRecords = skipIfNoRecords;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "WriteFileStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Processing step to write a copy of the current file.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "WriteFileStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		setFilename(parameters.getProperty("filename"));

		skipIfNoRecords = Boolean.valueOf(
				StringUtils.defaultIfEmpty(
						parameters.getProperty("skipIfNoRecords"), "false"))
				.booleanValue();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		super.execute(pipeline);

		try {
			if (skipIfNoRecords) {
				// Get the CountManager from the pipeline.
				CountManager countManager = (CountManager) pipeline
						.documentProperties().get(CountManager.PIPELINE_KEY);
				if (countManager == null) {
					throw new StepException(
							"No CountManager was found in the pipeline.");
				}
				// Exit if count = 0.
				if (countManager.getCurrentRecordCount() == 0) {
					log.info("No result file was written because it contained no records.");
					return;
				}
			}

			// Get the current file.
			File file = null;
			Message responseMessage = pipeline.getResponseMessage();
			Message requestMessage = pipeline.getRequestMessage();
			if (responseMessage.isSetCurrentFile()) {
				file = new File(responseMessage.getCurrentFile());
			} else {
				if (requestMessage.isSetCurrentFile()) {
					file = new File(requestMessage.getCurrentFile());
				} else {
					file = new File(requestMessage.getOriginalFile());
				}
			}

			// Set the default filename.
			if (StringUtils.isBlank(filename)) {
				filename = responseMessage.getOriginalFile();
			}

			// Copy file, uncompressing it if required.
			log.info("Creating file: " + filename + ".");
			InputStream is = IOHelper.getInputStream(file);
			try {
				IOHelper.transferStreamToFile(is, new File(filename));
			} finally {
				is.close();
			}

		} catch (Exception e) {
			throw new StepException(e);
		}
	}
}
