package com.taleo.integration.client.customstep.ftp;

import java.io.IOException;
import java.security.GeneralSecurityException;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPHTTPClient;
import org.apache.commons.net.ftp.FTPReply;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.RetryHelper;
import com.taleo.integration.client.util.PasswordUtils;
import com.taleo.ws.integration.client.GlobalConfig;

/**
 * Abstract class for FTP steps.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>host</b>: The FTP host.</li>
 * <br>
 * <br>
 * <li><b>port</b>: Optional - The FTP port (default: 21).</li>
 * <br>
 * <br>
 * <li><b>username</b>: The username.</li>
 * <br>
 * <br>
 * <li><b>password</b>: The AES128 encrypted or clear password. Encrypted
 * passwords are recognized by their ending with "=".</li>
 * <br>
 * <br>
 * <li><b>remoteDirectory</b>: The remote directory on the FTP server.</li>
 * <br>
 * <br>
 * <li><b>nbRetries</b>: Optional - Represents the number of retries that will
 * be done if the operation fails (default: 0).</li>
 * <br>
 * <br>
 * <li><b>retryInterval</b>: Optional - Represents the number of seconds to wait
 * between the retry attempts (default: 0).</li>
 * <br>
 * <br>
 * <li><b>retryFactor</b>: Optional - This factor is used as a multiplier on the
 * <code>retryInterval</code> between each attempt. It multiplies the last time
 * waited by this amount. For example. if <code>retryInterval</code> is 10
 * seconds and the retry factor is 2, wait time will be successively 10, 20, 40
 * seconds and so on, until the maximum number of retries has been reached
 * (default: 1).</li>
 * <br>
 * <br>
 * <li><b>transferType</b>: Optional - Transfer type: Binary or ASCII (default:
 * Binary).</li>
 * <br>
 * <br>
 * <li><b>transferMode</b>: Optional - Transfer mode: Active or Passive
 * (default: Passive).</li>
 * <br>
 * <br>
 * <li><b>proxyType</b>: Optional - The proxy type: NONE, FTP or HTTP (default
 * NONE).</li>
 * <br>
 * <br>
 * <li><b>proxyHost</b>: Optional - The proxy host.</li>
 * <br>
 * <br>
 * <li><b>proxyPort</b>: Optional - The proxy port (default: 80).</li>
 * <br>
 * <br>
 * <li><b>proxyUsername</b>: Optional - The proxy username.</li>
 * <br>
 * <br>
 * <li><b>proxyPassword</b>: Optional - The AES128 encrypted or clear proxy
 * password. Encrypted passwords are recognized by their ending with "=".</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * <li><a href="http://commons.apache.org/net/">Apache Commons Net</a>
 * (commons-net-3.1.jar)</li>
 * </ul>
 * 
 * @author Romain Guay, Oracle Corporation
 */
public abstract class BaseFtpStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(BaseFtpStep.class);

	/**
	 * The default port.
	 */
	private final int DEFAULT_PORT = 21;

	/**
	 * The transfer types.
	 */
	public final int TRANSFER_TYPE_BINARY = FTPClient.BINARY_FILE_TYPE;
	public final int TRANSFER_TYPE_ASCII = FTPClient.ASCII_FILE_TYPE;

	/**
	 * The transfer modes.
	 */
	public final String TRANSFER_MODE_ACTIVE = "ACTIVE";
	public final String TRANSFER_MODE_PASSIVE = "PASSIVE";

	/**
	 * The proxy types.
	 */
	public final String PROXY_TYPE_NONE = "NONE";
	public final String PROXY_TYPE_FTP = "FTP";
	public final String PROXY_TYPE_HTTP = "HTTP";

	/**
	 * Constructor without arguments.
	 */
	public BaseFtpStep() {
		super();
		// Register valid parameters.
		registerParameterName("host");
		registerParameterName("port");
		registerParameterName("username");
		registerParameterName("password");
		registerParameterName("remoteDirectory");
		registerParameterName("nbRetries");
		registerParameterName("retryInterval");
		registerParameterName("retryFactor");
		registerParameterName("transferType");
		registerParameterName("transferMode");
		registerParameterName("proxyType");
		registerParameterName("proxyHost");
		registerParameterName("proxyPort");
		registerParameterName("proxyUsername");
		registerParameterName("proxyPassword");
	}

	/**
	 * The FTP host.
	 */
	protected String host;

	/**
	 * The FTP port.
	 */
	protected int port = DEFAULT_PORT;

	/**
	 * The FTP username.
	 */
	protected String username;

	/**
	 * The FTP password (AES128 encrypted or clear).
	 */
	protected String password;

	/**
	 * The FTP remote directory.
	 */
	protected String remoteDirectory;

	/**
	 * The number of retries.
	 */
	protected int nbRetries = RetryHelper.NB_RETRIES_DEFAULT;

	/**
	 * The interval between retries, in seconds.
	 */
	protected int retryInterval = RetryHelper.RETRY_INTERVAL_DEFAULT;

	/**
	 * The retry factor.
	 */
	protected int retryFactor = RetryHelper.RETRY_FACTOR_DEFAULT;

	/**
	 * The transfer type.
	 */
	protected int transferType = TRANSFER_TYPE_BINARY;

	/**
	 * The transfer mode.
	 */
	protected String transferMode = TRANSFER_MODE_PASSIVE;

	/**
	 * The proxy type.
	 */
	protected String proxyType = PROXY_TYPE_NONE;

	/**
	 * The proxy host.
	 */
	protected String proxyHost;

	/**
	 * The proxy port.
	 */
	protected int proxyPort = 80;

	/**
	 * The proxy username.
	 */
	protected String proxyUsername;

	/**
	 * The proxy password (AES128 encrypted or clear).
	 */
	protected String proxyPassword;

	/**
	 * The FTP client.
	 */
	private FTPClient ftp = null;

	/**
	 * Get the FTP host.
	 * 
	 * @return
	 */
	public String getHost() {
		return host;
	}

	/**
	 * Set the FTP host.
	 * 
	 * @param host
	 */
	public void setHost(String host) {
		this.host = host;
	}

	/**
	 * Get the FTP password.
	 * 
	 * @return
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * Set the FTP password.
	 * 
	 * @param password
	 */
	public void setPassword(String password) {
		this.password = password;
	}

	/**
	 * Get the FTP port.
	 * 
	 * @return
	 */
	public int getPort() {
		return port;
	}

	/**
	 * Set the FTP port.
	 * 
	 * @param port
	 */
	public void setPort(int port) {
		this.port = port;
	}

	/**
	 * Get the FTP remote directory.
	 * 
	 * @return
	 */
	public String getRemoteDirectory() {
		return remoteDirectory;
	}

	/**
	 * Set the FTP remote directory.
	 * 
	 * @param remoteDirectory
	 */
	public void setRemoteDirectory(String remoteDirectory) {
		this.remoteDirectory = remoteDirectory;
	}

	/**
	 * Get the FTP username.
	 * 
	 * @return
	 */
	public String getUsername() {
		return username;
	}

	/**
	 * Set the FTP username.
	 * 
	 * @param username
	 */
	public void setUsername(String username) {
		this.username = username;
	}

	/**
	 * Get the number of retries.
	 * 
	 * @return
	 */
	public int getNbRetries() {
		return nbRetries;
	}

	/**
	 * Set the number of retries.
	 * 
	 * @param nbRetries
	 */
	public void setNbRetries(int nbRetries) {
		this.nbRetries = nbRetries;
	}

	/**
	 * Get the retry interval, in seconds.
	 * 
	 * @return
	 */
	public int getRetryInterval() {
		return retryInterval;
	}

	/**
	 * Set the retry interval, in seconds.
	 * 
	 * @param retryInterval
	 */
	public void setRetryInterval(int retryInterval) {
		this.retryInterval = retryInterval;
	}

	/**
	 * Get the retry factor.
	 * 
	 * @return
	 */
	public int getRetryFactor() {
		return retryFactor;
	}

	/**
	 * Set the retry factor.
	 * 
	 * @param retryFactor
	 */
	public void setRetryFactor(int retryFactor) {
		this.retryFactor = retryFactor;
	}

	/**
	 * Get the transfer type.
	 * 
	 * @return
	 */
	public int getTransferType() {
		return transferType;
	}

	/**
	 * Set the transfer type (TRANSFER_TYPE_BINARY or TRANSFER_TYPE_ASCII).
	 * 
	 * @param transferType
	 */
	public void setTransferType(int transferType) {
		if (transferType != TRANSFER_TYPE_BINARY
				&& transferType != TRANSFER_TYPE_ASCII) {
			throw new IllegalArgumentException("Invalid transfer type.");
		}
		this.transferType = transferType;
	}

	/**
	 * Get the transfer mode.
	 * 
	 * @return
	 */
	public String getTransferMode() {
		return transferMode;
	}

	/**
	 * Set the transfer mode (TRANSFER_MODE_ACTIVE or TRANSFER_MODE_PASSIVE).
	 * 
	 * @param transferMode
	 */
	public void setTransferMode(String transferMode) {
		if (!(TRANSFER_MODE_ACTIVE.equals(transferMode) || TRANSFER_MODE_PASSIVE
				.equals(transferMode))) {
			throw new IllegalArgumentException("Invalid transferMode ("
					+ transferMode + ").");
		}
		this.transferMode = transferMode;
	}

	/**
	 * Get the proxy type.
	 * 
	 * @return
	 */
	public String getProxyType() {
		return proxyType;
	}

	/**
	 * Set the proxy type.
	 * 
	 * @param proxyType
	 */
	public void setProxyType(String proxyType) {
		if (!(PROXY_TYPE_NONE.equals(proxyType)
				|| PROXY_TYPE_FTP.equals(proxyType) || PROXY_TYPE_HTTP
					.equals(proxyType))) {
			throw new IllegalArgumentException("Invalid proxyType ("
					+ proxyType + ").");
		}
		this.proxyType = proxyType;
	}

	/**
	 * Get the proxy host.
	 * 
	 * @return
	 */
	public String getProxyHost() {
		return proxyHost;
	}

	/**
	 * Set the proxy host.
	 * 
	 * @param proxyHost
	 */
	public void setProxyHost(String proxyHost) {
		this.proxyHost = proxyHost;
	}

	/**
	 * Get the proxy port.
	 * 
	 * @return
	 */
	public int getProxyPort() {
		return proxyPort;
	}

	/**
	 * Set the proxy port.
	 * 
	 * @param proxyPort
	 */
	public void setProxyPort(int proxyPort) {
		this.proxyPort = proxyPort;
	}

	/**
	 * Get the proxy username.
	 * 
	 * @return
	 */
	public String getProxyUsername() {
		return proxyUsername;
	}

	/**
	 * Set the proxy username.
	 * 
	 * @param proxyUsername
	 */
	public void setProxyUsername(String proxyUsername) {
		this.proxyUsername = proxyUsername;
	}

	/**
	 * Get the proxy password.
	 * 
	 * @return
	 */
	public String getProxyPassword() {
		return proxyPassword;
	}

	/**
	 * Set the proxy password.
	 * 
	 * @param proxyPassword
	 */
	public void setProxyPassword(String proxyPassword) {
		this.proxyPassword = proxyPassword;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		host = parameters.getProperty("host");
		if (host == null || host.trim().length() == 0) {
			throw new IllegalArgumentException(
					"The 'host' parameter is mandatory.");
		}

		try {
			port = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("port"), Integer.toString(port)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'port' parameter must have a numeric value.");
		}

		username = parameters.getProperty("username");

		password = parameters.getProperty("password");

		remoteDirectory = parameters.getProperty("remoteDirectory");

		try {
			nbRetries = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("nbRetries"),
					Integer.toString(nbRetries)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'nbRetries' parameter must have a numeric value.");
		}

		try {
			retryInterval = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("retryInterval"),
					Integer.toString(retryInterval)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'retryInterval' parameter must have a numeric value.");
		}

		try {
			retryFactor = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("retryFactor"),
					Integer.toString(retryFactor)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'retryFactor' parameter must have a numeric value.");
		}

		String transferType = parameters.getProperty("transferType", "Binary");
		if ("Binary".equalsIgnoreCase(transferType)) {
			setTransferType(TRANSFER_TYPE_BINARY);
		} else if ("ASCII".equalsIgnoreCase(transferType)) {
			setTransferType(TRANSFER_TYPE_ASCII);
		} else {
			throw new IllegalArgumentException(
					"The 'transferType' parameter must be Binary or ASCII.");
		}

		String transferMode = parameters.getProperty("transferMode",
				this.transferMode);
		if (TRANSFER_MODE_ACTIVE.equalsIgnoreCase(transferMode)) {
			setTransferMode(TRANSFER_MODE_ACTIVE);
		} else if (TRANSFER_MODE_PASSIVE.equalsIgnoreCase(transferMode)) {
			setTransferMode(TRANSFER_MODE_PASSIVE);
		} else {
			throw new IllegalArgumentException(
					"The 'transferMode' parameter must be Active or Passive.");
		}

		setProxyType(StringUtils.defaultIfEmpty(
				parameters.getProperty("proxyType"), this.proxyType));

		proxyHost = parameters.getProperty("proxyHost");

		try {
			proxyPort = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("proxyPort"),
					Integer.toString(proxyPort)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'proxyPort' parameter must have a numeric value.");
		}

		proxyUsername = parameters.getProperty("proxyUsername");

		proxyPassword = parameters.getProperty("proxyPassword");
	}

	/**
	 * Get the FTP Client.
	 * 
	 * @return The open FTP client.
	 * @throws IOException
	 * @throws FTPException
	 * @throws FTPIllegalReplyException
	 */
	protected FTPClient getFtpClient() throws IOException {

		// Return current FTPClient immediately if already connected.
		if (ftp != null && ftp.isConnected()) {
			return ftp;
		}

		// Prepare the RetryHelper.
		RetryHelper retry = new RetryHelper(nbRetries, retryFactor,
				retryInterval);

		// Decrypt the passwords.
		String password = this.password;
		if (password != null && password.endsWith("=")) {
			try {
				password = PasswordUtils.decryptAES128(password);
			} catch (GeneralSecurityException e) {
				throw new RuntimeException(e);
			}
		}
		String proxyPassword = this.proxyPassword;
		if (proxyPassword != null && proxyPassword.endsWith("=")) {
			try {
				proxyPassword = PasswordUtils.decryptAES128(proxyPassword);
			} catch (GeneralSecurityException e) {
				throw new RuntimeException(e);
			}
		}

		boolean completed = false;
		while (!completed) {
			try {
				// Create the ftp client.
				ftp = new FTPClient();

				log.info("Opening FTP channel on " + host + " - Attempt "
						+ retry.getCurrentAttempt() + ".");

				// Connect and login.
				if (proxyType.equals(PROXY_TYPE_NONE)) {
					// Direct connection.
					ftp.connect(host, port);
					log.info("Connected to " + host + ":" + port);

					ftp.login(username, password);
					if (!FTPReply.isPositiveCompletion(ftp.getReplyCode())) {
						throw new IOException("FTP Error: "
								+ ftp.getReplyString());
					}
					log.info("Logged in as " + username);

				} else if (proxyType.equals(PROXY_TYPE_HTTP)) {
					// HTTP proxy connection.
					log.info("Using HTTP proxy server: " + proxyHost);
					ftp = new FTPHTTPClient(proxyHost, proxyPort,
							proxyUsername, proxyPassword);

					ftp.connect(host, port);
					log.info("Connected to " + host + ":" + port);

					ftp.login(username, password);
					if (!FTPReply.isPositiveCompletion(ftp.getReplyCode())) {
						throw new IOException("FTP Error: "
								+ ftp.getReplyString());
					}
					log.info("Logged in as " + username);

				} else if (proxyType.equals(PROXY_TYPE_FTP)) {
					// FTP proxy connection.
					log.info("Using FTP proxy server: " + proxyHost);
					ftp.connect(proxyHost, proxyPort);
					log.info("Connected to " + proxyHost + ":" + proxyPort);

					ftp.login(proxyUsername, proxyPassword);
					if (!FTPReply.isPositiveCompletion(ftp.getReplyCode())) {
						throw new IOException("FTP Error: "
								+ ftp.getReplyString());
					}
					log.info("Logged in as " + proxyUsername);

					// TODO: Could not login when setting the port explicitly:
					// username + "@" + host + ":" + port
					ftp.login(username + "@" + host, password);
					if (!FTPReply.isPositiveCompletion(ftp.getReplyCode())) {
						throw new IOException("FTP Error: "
								+ ftp.getReplyString());
					}
					log.info("Logged in as " + username + "@" + host);
				}

				// Set the transfer type.
				ftp.setFileType(getTransferType());

				// Set the transfer mode.
				if (getTransferMode() == TRANSFER_MODE_PASSIVE) {
					ftp.enterLocalPassiveMode();
				} else {
					ftp.enterLocalActiveMode();
				}

				// Mark as complete.
				completed = true;

			} catch (IOException e) {
				// If there are retry attempts left, retry.
				if (retry.getCurrentAttempt() < retry.getMaxAttempts()) {
					retry.waitForNextRetry();
				} else {
					throw e;
				}
			}
		}
		return ftp;
	}

	/**
	 * Log out the FTP Client.
	 */
	protected void logout() {
		if (ftp != null && ftp.isConnected()) {
			try {
				ftp.logout();
				ftp.disconnect();
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	}

}
