package com.taleo.integration.client.customstep.ftp;

import java.io.IOException;
import java.security.GeneralSecurityException;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.ProxyHTTP;
import com.jcraft.jsch.Session;
import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.RetryHelper;
import com.taleo.integration.client.util.PasswordUtils;
import com.taleo.ws.integration.client.GlobalConfig;

/**
 * Abstract class for SFTP steps.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>host</b>: The SFTP host.</li>
 * <br>
 * <br>
 * <li><b>port</b>: The SFTP port (default: 22).</li>
 * <br>
 * <br>
 * <li><b>username</b>: The username.</li>
 * <br>
 * <br>
 * <li><b>password</b>: The AES128 encrypted or clear password. Encrypted
 * passwords are recognized by their ending with "=".</li>
 * <br>
 * <br>
 * <li><b>remoteDirectory</b>: The remote directory on the FTP server.</li>
 * <br>
 * <br>
 * <li><b>nbRetries</b>: Optional - Represents the number of retries that will
 * be done if the operation fails (default: 0).</li>
 * <br>
 * <br>
 * <li><b>retryInterval</b>: Optional - Represents the number of seconds to wait
 * between the retry attempts (default: 0).</li>
 * <br>
 * <br>
 * <li><b>retryFactor</b>: Optional - This factor is used as a multiplier on the
 * <code>retryInterval</code> between each attempt. It multiplies the last time
 * waited by this amount. For example. if <code>retryInterval</code> is 10
 * seconds and the retry factor is 2, wait time will be successively 10, 20, 40
 * seconds and so on, until the maximum number of retries has been reached
 * (default: 1).</li>
 * <br>
 * <br>
 * <li><b>privateKey</b>: Optional - Absolute or relative path to the private
 * RSA key file. If present, it will be used instead of the password for
 * authentication.</li>
 * <br>
 * <br>
 * <li><b>passphrase</b>: Optional - The passphrase used to encrypt/decrypt the
 * private key. The passphrase itself may be AES128 encrypted or clear.
 * Encrypted passphrases are recognized by their ending with "="</li>
 * <br>
 * <br>
 * <li><b>proxyHost</b>: Optional - The HTTP proxy host.</li>
 * <br>
 * <br>
 * <li><b>proxyPort</b>: Optional - The HTTP proxy port (default: 80).</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * <li><a href="http://www.jcraft.com/jsch">JCraft - JSch</a> (jsch-0.1.37.jar)</li>
 * </ul>
 * 
 * @author Romain Guay, Oracle Corporation
 */
public abstract class BaseSftpStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(BaseSftpStep.class);

	/**
	 * The default port.
	 */
	private final int DEFAULT_PORT = 22;

	/**
	 * Constructor without arguments.
	 */
	public BaseSftpStep() {
		super();
		// Register valid parameters.
		registerParameterName("host");
		registerParameterName("port");
		registerParameterName("username");
		registerParameterName("password");
		registerParameterName("remoteDirectory");
		registerParameterName("nbRetries");
		registerParameterName("retryInterval");
		registerParameterName("retryFactor");
		registerParameterName("privateKey");
		registerParameterName("passphrase");
		registerParameterName("proxyHost");
		registerParameterName("proxyPort");
	}

	/**
	 * The SFTP host.
	 */
	protected String host;

	/**
	 * The SFTP port.
	 */
	protected int port = DEFAULT_PORT;

	/**
	 * The SFTP username.
	 */
	protected String username;

	/**
	 * The SFTP password (AES128 encrypted or clear).
	 */
	protected String password;

	/**
	 * The SFTP remote directory.
	 */
	protected String remoteDirectory;

	/**
	 * The number of retries.
	 */
	protected int nbRetries = RetryHelper.NB_RETRIES_DEFAULT;

	/**
	 * The interval between retries, in seconds.
	 */
	protected int retryInterval = RetryHelper.RETRY_INTERVAL_DEFAULT;

	/**
	 * The retry factor.
	 */
	protected int retryFactor = RetryHelper.RETRY_FACTOR_DEFAULT;

	/**
	 * Path to the private RSA key file.
	 */
	private String privateKey;

	/**
	 * Passphrase used to encrypt/decrypt the key.
	 */
	private String passphrase;

	/**
	 * The proxy host.
	 */
	protected String proxyHost;

	/**
	 * The proxy port.
	 */
	protected int proxyPort = 80;

	/**
	 * The SFTP Channel.
	 */
	private ChannelSftp sftp = null;

	/**
	 * Get the SFTP host.
	 * 
	 * @return
	 */
	public String getHost() {
		return host;
	}

	/**
	 * Set the SFTP host.
	 * 
	 * @param host
	 */
	public void setHost(String host) {
		this.host = host;
	}

	/**
	 * Get the SFTP password.
	 * 
	 * @return
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * Set the SFTP password.
	 * 
	 * @param password
	 */
	public void setPassword(String password) {
		this.password = password;
	}

	/**
	 * Get the SFTP port.
	 * 
	 * @return
	 */
	public int getPort() {
		return port;
	}

	/**
	 * Set the SFTP port.
	 * 
	 * @param port
	 */
	public void setPort(int port) {
		this.port = port;
	}

	/**
	 * Get the SFTP remote directory.
	 * 
	 * @return
	 */
	public String getRemoteDirectory() {
		return remoteDirectory;
	}

	/**
	 * Set the SFTP remote directory.
	 * 
	 * @param remoteDirectory
	 */
	public void setRemoteDirectory(String remoteDirectory) {
		this.remoteDirectory = remoteDirectory;
	}

	/**
	 * Get the SFTP username.
	 * 
	 * @return
	 */
	public String getUsername() {
		return username;
	}

	/**
	 * Set the SFTP username.
	 * 
	 * @param username
	 */
	public void setUsername(String username) {
		this.username = username;
	}

	/**
	 * Get the number of retries.
	 * 
	 * @return
	 */
	public int getNbRetries() {
		return nbRetries;
	}

	/**
	 * Set the number of retries.
	 * 
	 * @param nbRetries
	 */
	public void setNbRetries(int nbRetries) {
		this.nbRetries = nbRetries;
	}

	/**
	 * Get the retry interval, in seconds.
	 * 
	 * @return
	 */
	public int getRetryInterval() {
		return retryInterval;
	}

	/**
	 * Set the retry interval, in seconds.
	 * 
	 * @param retryInterval
	 */
	public void setRetryInterval(int retryInterval) {
		this.retryInterval = retryInterval;
	}

	/**
	 * Get the retry factor.
	 * 
	 * @return
	 */
	public int getRetryFactor() {
		return retryFactor;
	}

	/**
	 * Set the retry factor.
	 * 
	 * @param retryFactor
	 */
	public void setRetryFactor(int retryFactor) {
		this.retryFactor = retryFactor;
	}

	/**
	 * Get the private key.
	 * 
	 * @return
	 */
	public String getPrivateKey() {
		return privateKey;
	}

	/**
	 * Set the private key.
	 * 
	 * @param privateKey
	 */
	public void setPrivateKey(String privateKey) {
		this.privateKey = privateKey;
	}

	/**
	 * Get the passphrase.
	 * 
	 * @return
	 */
	public String getPassphrase() {
		return passphrase;
	}

	/**
	 * Set the passphrase.
	 * 
	 * @param passphrase
	 */
	public void setPassphrase(String passphrase) {
		this.passphrase = passphrase;
	}

	/**
	 * Get the proxy host.
	 * 
	 * @return
	 */
	public String getProxyHost() {
		return proxyHost;
	}

	/**
	 * Set the proxy host.
	 * 
	 * @param proxyHost
	 */
	public void setProxyHost(String proxyHost) {
		this.proxyHost = proxyHost;
	}

	/**
	 * Get the proxy port.
	 * 
	 * @return
	 */
	public int getProxyPort() {
		return proxyPort;
	}

	/**
	 * Set the proxy port.
	 * 
	 * @param proxyPort
	 */
	public void setProxyPort(int proxyPort) {
		this.proxyPort = proxyPort;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		host = parameters.getProperty("host");
		if (host == null || host.trim().length() == 0) {
			throw new IllegalArgumentException(
					"The 'host' parameter is mandatory.");
		}

		try {
			port = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("port"), Integer.toString(port)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'port' parameter must have a numeric value.");
		}

		username = parameters.getProperty("username");

		password = parameters.getProperty("password");

		remoteDirectory = parameters.getProperty("remoteDirectory");

		try {
			nbRetries = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("nbRetries"),
					Integer.toString(nbRetries)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'nbRetries' parameter must have a numeric value.");
		}

		try {
			retryInterval = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("retryInterval"),
					Integer.toString(retryInterval)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'retryInterval' parameter must have a numeric value.");
		}

		try {
			retryFactor = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("retryFactor"),
					Integer.toString(retryFactor)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'retryFactor' parameter must have a numeric value.");
		}

		privateKey = parameters.getProperty("privateKey");

		passphrase = parameters.getProperty("passphrase");

		proxyHost = parameters.getProperty("proxyHost");

		try {
			proxyPort = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("proxyPort"),
					Integer.toString(proxyPort)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'proxyPort' parameter must have a numeric value.");
		}
	}

	/**
	 * Get the SFTP channel.
	 * 
	 * @return The open SFTP channel.
	 * @throws IOException
	 * @throws JSchException
	 */
	protected ChannelSftp getSftpChannel() throws JSchException, IOException {
		// Return current ChannelSftp immediately if already connected.
		if (sftp != null && sftp.isConnected()) {
			return sftp;
		}

		// Decrypt the password.
		if (password != null && password.endsWith("=")) {
			try {
				password = PasswordUtils.decryptAES128(password);
			} catch (GeneralSecurityException e) {
				throw new RuntimeException(e);
			}
		}

		// Decrypt the passphrase.
		if (passphrase != null && passphrase.endsWith("=")) {
			try {
				passphrase = PasswordUtils.decryptAES128(passphrase);
			} catch (GeneralSecurityException e) {
				throw new RuntimeException(e);
			}
		}

		// Prepare the RetryHelper.
		RetryHelper retry = new RetryHelper(nbRetries, retryFactor,
				retryInterval);

		boolean completed = false;
		while (!completed) {
			try {
				// Create a session.
				JSch jsch = new JSch();
				if (StringUtils.isNotBlank(privateKey)) {
					jsch.addIdentity(privateKey, passphrase);
				}
				Session session = jsch.getSession(username, host, port);
				if (StringUtils.isNotBlank(password)) {
					session.setPassword(password);
				}

				// Setup configuration options.
				java.util.Properties config = new java.util.Properties();
				config.put("StrictHostKeyChecking", "no");
				session.setConfig(config);

				// Proxy settings.
				if (StringUtils.isNotBlank(proxyHost)) {
					log.info("Using HTTP proxy server: " + proxyHost);
					session.setProxy(new ProxyHTTP(proxyHost, proxyPort));
				}

				// Connect to the SFTP Server.
				log.info("Opening SFTP channel on " + host + " - Attempt "
						+ retry.getCurrentAttempt() + ".");
				session.connect();

				// Open the SFTP channel.
				sftp = (ChannelSftp) session.openChannel("sftp");
				sftp.connect();

				// Mark as complete.
				completed = true;

			} catch (JSchException e) {
				// If there are retry attempts left, retry.
				if (retry.getCurrentAttempt() < retry.getMaxAttempts()) {
					retry.waitForNextRetry();
				} else {
					throw e;
				}
			}
		}
		return sftp;
	}

	/**
	 * Log out the SFTP Channel.
	 */
	protected void logout() {
		if (sftp != null && sftp.isConnected()) {
			sftp.disconnect();
			sftp.getSession().disconnect();
		}
	}

}
