package com.taleo.integration.client.customstep.ftp;

import java.io.IOException;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.RetryHelper;
import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Processing step to delete a file from an FTP server.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseFtpStep} and the following
 * additional parameters:
 * <ul>
 * <li><b>filename</b>: The filename to delete from the FTP server.</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * <li><a href="http://commons.apache.org/net/">Apache Commons Net</a>
 * (commons-net-3.1.jar)</li>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/DeleteFtpStep-1.jpg" />
 * 
 * @author Romain Guay, Oracle Corporation
 */
public class DeleteFtpStep extends BaseFtpStep {

	private static Logger log = Logger.getLogger(DeleteFtpStep.class);

	/**
	 * Constructor without arguments.
	 */
	public DeleteFtpStep() {
		super();
		// Register valid parameters.
		registerParameterName("filename");
	}

	/**
	 * The filename.
	 */
	protected String filename;

	/**
	 * Get the filename.
	 * 
	 * @return
	 */
	public String getFilename() {
		return filename;
	}

	/**
	 * Set the filename.
	 * 
	 * @param filename
	 */
	public void setFilename(String filename) {
		this.filename = filename;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "DeleteFtpStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Processing step to delete a file from an FTP server.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "DeleteFtpStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		filename = parameters.getProperty("filename");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// If not active, return immediately.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		try {
			// Execute the step.
			delete();

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Delete the file from the FTP server.
	 * 
	 * @throws Exception
	 */
	public synchronized void delete() throws Exception {

		// Create the ftp client.
		FTPClient ftp = getFtpClient();

		// Prepare the RetryHelper.
		RetryHelper retry = new RetryHelper(nbRetries, retryFactor,
				retryInterval);

		boolean completed = false;
		boolean success = false;
		boolean cwd = false;
		while (!completed) {
			try {
				// Change the remote directory.
				remoteDirectory = remoteDirectory.trim().replace("\\", "/");
				if (StringUtils.isNotBlank(remoteDirectory)) {
					log.info("Changing to remote directory: " + remoteDirectory);
					success = ftp.changeWorkingDirectory(remoteDirectory);
					if (!success) {
						throw new IOException("FTP Error: "
								+ ftp.getReplyString());
					}
					cwd = true;
				}

				// Delete the file.
				log.info("Deleting file: " + getFilename() + " - Attempt "
						+ retry.getCurrentAttempt() + ".");
				ftp.deleteFile(getFilename());

				log.info("Complete");
				completed = true;

			} catch (Exception e) {
				// If there are retry attempts left, retry.
				if (retry.getCurrentAttempt() < retry.getMaxAttempts()) {
					retry.waitForNextRetry();
				} else {
					throw e;
				}
			} finally {
				// Move back to root directory.
				if (cwd) {
					for (int i = 0; i < remoteDirectory.split("/").length; i++) {
						ftp.changeToParentDirectory();
					}
					cwd = false;
				}
				logout();
			}
		}
	}
}
