package com.taleo.integration.client.customstep.ftp;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.jcraft.jsch.ChannelSftp;
import com.taleo.integration.client.customstep.util.RetryHelper;
import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Processing step to delete a file from an SFTP server.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.ftp.BaseSftpStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>filename</b>: The filename to delete from the FTP server.</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * <li><a href="http://www.jcraft.com/jsch">JCraft - JSch</a> (jsch-0.1.37.jar)</li>
 * </ul>
 * <p>
 * Here are two examples of how to define this custom step in the TCC
 * configuration file:
 * <p>
 * <img src="doc-files/DeleteSftpStep-1.jpg" />
 * <p>
 * <img src="doc-files/DeleteSftpStep-2.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class DeleteSftpStep extends BaseSftpStep {

	private static Logger log = Logger.getLogger(DeleteSftpStep.class);

	/**
	 * Constructor without arguments.
	 */
	public DeleteSftpStep() {
		super();
		// Register valid parameters.
		registerParameterName("filename");
	}

	/**
	 * The filename.
	 */
	protected String filename;

	/**
	 * Get the filename.
	 * 
	 * @return
	 */
	public String getFilename() {
		return filename;
	}

	/**
	 * Set the filename.
	 * 
	 * @param filename
	 */
	public void setFilename(String filename) {
		this.filename = filename;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "DeleteSftpStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Processing step to delete a file from an SFTP server.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "DeleteSftpStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.customstep.ftp.DeleteFtpStep#init(com
	 * .taleo.ws.integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		filename = parameters.getProperty("filename");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// If not active, return immediately.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		try {
			// Execute the step.
			delete();

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Execute the step by deleting the file from the SFTP server.
	 * 
	 * @throws Exception
	 */
	public synchronized void delete() throws Exception {

		// Get the SFTP Channel.
		ChannelSftp sftp = getSftpChannel();

		// Prepare the RetryHelper.
		RetryHelper retry = new RetryHelper(nbRetries, retryFactor,
				retryInterval);

		boolean completed = false;
		boolean cwd = false;
		while (!completed) {
			try {
				// Change the remote directory.
				remoteDirectory = remoteDirectory.trim().replace("\\", "/");
				if (StringUtils.isNotBlank(remoteDirectory)) {
					log.info("Changing to remote dir: " + remoteDirectory);
					sftp.cd(remoteDirectory);
					cwd = true;
				}

				// Send the file.
				log.info("Deleting file: " + getFilename() + " - Attempt "
						+ retry.getCurrentAttempt() + ".");
				sftp.rm(getFilename());

				log.info("Complete");
				completed = true;

			} catch (Exception e) {
				// If there are retry attempts left, retry.
				if (retry.getCurrentAttempt() < retry.getMaxAttempts()) {
					retry.waitForNextRetry();
				} else {
					throw e;
				}
			} finally {
				// Move back to root directory.
				if (cwd) {
					for (int i = 0; i < remoteDirectory.split("/").length; i++) {
						sftp.cd("..");
					}
					cwd = false;
				}
				logout();
			}
		}
	}
}
