package com.taleo.integration.client.customstep.ftp;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.RetryHelper;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Pre-Processing step to download a file from an FTP server and use it as
 * request data file for import.
 * <p>
 * When using this step, the request file does not have to be specified when
 * calling TCC. In fact, if one is specified, it will be ignored.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseFtpStep} and the following
 * additional parameter:
 * <ul>
 * <li><b>filename</b>: The filename to get from the FTP server.</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * <li><a href="http://commons.apache.org/net/">Apache Commons Net</a>
 * (commons-net-3.1.jar)</li>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/DownloadFtpPreStep-1.jpg" />
 * 
 * @author Romain Guay, Oracle Corporation
 */
public class DownloadFtpPreStep extends BaseFtpStep {

	private static Logger log = Logger.getLogger(DownloadFtpPreStep.class);

	/**
	 * Constructor without arguments.
	 */
	public DownloadFtpPreStep() {
		super();
		// Register valid parameters.
		registerParameterName("filename");
	}

	/**
	 * The filename.
	 */
	protected String filename;

	/**
	 * Get the filename.
	 * 
	 * @return
	 */
	public String getFilename() {
		return filename;
	}

	/**
	 * Set the filename.
	 * 
	 * @param filename
	 */
	public void setFilename(String filename) {
		this.filename = filename;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "DownloadFtpPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Pre-Processing step to download a file from an FTP server and use it as request data file for import.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "DownloadFtpPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		filename = parameters.getProperty("filename");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		SupportedPipeline supportedPipeline = getPreProcessSupportPipeline();

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// If not active, return immediately.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Create a local temp file to store the downloaded data.
			File localFile = createTempFile();

			// Download the file.
			download(localFile);

			// Set the file as the import request.
			Message requestMessage = pipeline.getRequestMessage();
			requestMessage.setOriginalFile(localFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Download the file from the FTP server and put the result in the specified
	 * local file.
	 * 
	 * @param localFile
	 *            The local file to contain the downloaded data.
	 * @throws Exception
	 */
	public synchronized void download(File localFile) throws Exception {

		// Create the FTP client.
		FTPClient ftp = getFtpClient();

		// Prepare the RetryHelper.
		RetryHelper retry = new RetryHelper(nbRetries, retryFactor,
				retryInterval);

		boolean completed = false;
		boolean success = false;
		boolean cwd = false;
		while (!completed) {
			try {
				// Change the remote directory.
				remoteDirectory = remoteDirectory.trim().replace("\\", "/");
				if (StringUtils.isNotBlank(remoteDirectory)) {
					log.info("Changing to remote directory: " + remoteDirectory);
					success = ftp.changeWorkingDirectory(remoteDirectory);
					if (!success) {
						throw new IOException("FTP Error: "
								+ ftp.getReplyString());
					}
					cwd = true;
				}

				// Download the file.
				log.info("Getting file: " + filename + " into: "
						+ localFile.getAbsolutePath() + " - Attempt "
						+ retry.getCurrentAttempt() + ".");

				OutputStream os = IOHelper.getOutputStream(
						new FileOutputStream(localFile), getClass().getName());
				success = ftp.retrieveFile(filename, os);
				os.close();

				if (!success) {
					throw new IOException("FTP Error: " + ftp.getReplyString());
				}

				log.info("Complete");
				completed = true;

			} catch (Exception e) {
				// If there are retry attempts left, retry.
				if (retry.getCurrentAttempt() < retry.getMaxAttempts()) {
					retry.waitForNextRetry();
				} else {
					throw e;
				}
			} finally {
				// Move back to root directory.
				if (cwd) {
					for (int i = 0; i < remoteDirectory.split("/").length; i++) {
						ftp.changeToParentDirectory();
					}
					cwd = false;
				}
				logout();
			}
		}
	}
}
