package com.taleo.integration.client.customstep.ftp;

import java.io.File;
import java.io.IOException;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;

import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Utility class for shared logic between FTP steps.
 * 
 * @author Romain Guay, Oracle Corporation
 * 
 */
abstract class FtpUtil {

	/**
	 * Get the list of files matching the pattern and their corresponding remote
	 * directory in a two-dimensional array. String[0] contains filenames,
	 * String[1] contains remote directories and String[2] contains remote
	 * filenames.
	 * 
	 * @param pipeline
	 * @param filename
	 * @param filePattern
	 * @param remoteDirectory
	 * @return
	 * @throws IOException
	 */
	public static String[][] getMatchingFiles(Pipeline pipeline,
			String filename, String filePattern, String remoteDirectory)
			throws IOException {

		// Get the workflow file to upload.
		File workflowFile = null;
		String defaultFilename = null;
		Message responseMessage = pipeline.getResponseMessage();
		Message requestMessage = pipeline.getRequestMessage();
		if (responseMessage.isSetCurrentFile()) {
			workflowFile = new File(responseMessage.getCurrentFile());
			if (responseMessage.getOriginalFile() != null) {
				defaultFilename = new File(responseMessage.getOriginalFile())
						.getName();
			}
		} else {
			if (requestMessage.isSetCurrentFile()) {
				workflowFile = new File(requestMessage.getCurrentFile());
			} else {
				workflowFile = new File(requestMessage.getOriginalFile());
			}
			if (requestMessage.getOriginalFile() != null) {
				defaultFilename = new File(requestMessage.getOriginalFile())
						.getName();
			}
		}

		// Set the workflow file as default filename.
		if (StringUtils.isBlank(filename)) {
			filename = defaultFilename;
		}

		// Set the default filePattern.
		if (StringUtils.isBlank(filePattern)) {
			filePattern = ".";
		}

		// Replace the "." filePattern.
		boolean filenameRequired = false;
		String[] patterns = StringUtils.split(filePattern, "|");
		for (int i = 0; i < patterns.length; i++) {
			if (patterns[i].trim().equals(".")) {
				patterns[i] = workflowFile.getPath();
				filenameRequired = true;
			}
		}
		filePattern = StringUtils.join(patterns, "|");

		// Validate workflow filename.
		if (StringUtils.isBlank(filename) && filenameRequired) {
			throw new IllegalArgumentException(
					"A value must be provided for the filename parameter.");
		}

		// Get files and their corresponding remote directory in a
		// 2 dimensional array. File[0] contains filenames, File[1] contains
		// remote directories and File[2] contains remote filenames.
		String[][] files = new String[3][0];
		String[] directories = StringUtils.splitPreserveAllTokens(
				remoteDirectory, "|");
		for (int i = 0; i < patterns.length; i++) {

			// Add matching files.
			File[] tempFiles = FileUtils
					.getMatchingFiles(new File(patterns[i]));
			FileUtils.sortFilesByNameAscending(tempFiles, true);
			for (int j = 0; j < tempFiles.length; j++) {
				files[0] = (String[]) ArrayUtils.add(files[0],
						tempFiles[j].getPath());
			}

			// Add directories.
			String directory = "";
			if (i < directories.length) {
				directory = directories[i];
			}
			for (int j = 0; j < tempFiles.length; j++) {
				files[1] = (String[]) ArrayUtils.add(files[1], directory);
			}
		}

		// Add remote filenames.
		files[2] = new String[files[0].length];
		String remoteFile = null;
		for (int i = 0; i < files[2].length; i++) {
			if (files[0][i].equals(workflowFile.getPath())) {
				remoteFile = filename;
			} else {
				remoteFile = new File(files[0][i]).getName();
			}
			files[2][i] = remoteFile;
		}

		// Return the files array.
		return files;
	}
}
