package com.taleo.integration.client.customstep.ftp;

import java.io.File;
import java.io.InputStream;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.jcraft.jsch.ChannelSftp;
import com.taleo.integration.client.customstep.util.RetryHelper;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Processing step to upload the current workflow file on an SFTP server.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.ftp.BaseSftpStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>filePattern</b>: Optional - The pattern for selecting files to upload.
 * Can be specified as an absolute or relative path and may use wildcards as
 * defined in {@link org.apache.commons.io.filefilter.WildcardFileFilter}, in
 * the file names only. Pattern is case sensitive. If the pattern includes
 * multiple files, all of them will be uploaded in sequence. If absent, only the
 * current workflow file will be uploaded.
 * <p>
 * Multiple patterns can be set by separating them with a pipe (|). The special
 * dot (.) pattern represents the current workflow file.</li>
 * <br>
 * <br>
 * <li><b>filename</b>: Optional - The filename to set on the FTP server. If
 * absent, the original filename from the response will be used. If there is
 * none, the step will fail. This property only applies to the current workflow
 * file, other files identified by the <b>filePattern</b> will maintain their
 * original name.</li>
 * <br>
 * <br>
 * <li><b>remoteDirectory</b>: The remote directory on the FTP server.
 * <p>
 * If multiple file patterns are provided, correspondingly multiple remote
 * directories cans be set too, using the same pipe (|) delimiter. If the number
 * of patterns is larger than the number of remote directories, the last
 * directory will be used for remaining patterns.</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * <li><a href="http://www.jcraft.com/jsch">JCraft - JSch</a> (jsch-0.1.37.jar)</li>
 * </ul>
 * <p>
 * Here are two examples of how to define this custom step in the TCC
 * configuration file:
 * <p>
 * <img src="doc-files/UploadSftpStep-1.jpg" />
 * <p>
 * <img src="doc-files/UploadSftpStep-2.jpg" />
 * <p>
 * <img src="doc-files/UploadSftpStep-3.png" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class UploadSftpStep extends BaseSftpStep {

	private static Logger log = Logger.getLogger(UploadSftpStep.class);

	/**
	 * Constructor without arguments.
	 */
	public UploadSftpStep() {
		super();
		// Register valid parameters.
		registerParameterName("filename");
		registerParameterName("filePattern");
	}

	/**
	 * The filename.
	 */
	protected String filename;

	/**
	 * The file pattern.
	 */
	protected String filePattern;

	/**
	 * Get the filename.
	 * 
	 * @return
	 */
	public String getFilename() {
		return filename;
	}

	/**
	 * Set the filename.
	 * 
	 * @param filename
	 */
	public void setFilename(String filename) {
		this.filename = filename;
	}

	/**
	 * Get the file pattern.
	 * 
	 * @return
	 */
	public String getFilePattern() {
		return filePattern;
	}

	/**
	 * Set the file pattern.
	 * 
	 * @param filePattern
	 */
	public void setFilePattern(String filePattern) {
		this.filePattern = filePattern;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "UploadSftpStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Processing step to upload the current file to an SFTP server.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "UploadSftpStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.customstep.ftp.UploadFtpStep#init(com
	 * .taleo.ws.integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		filename = parameters.getProperty("filename");
		filePattern = parameters.getProperty("filePattern");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// If not active, return immediately.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the list of files.
			String[][] files = FtpUtil.getMatchingFiles(pipeline, filename,
					filePattern, remoteDirectory);

			// Get the SFTP Channel.
			ChannelSftp sftp = getSftpChannel();

			// Execute the step.
			for (int i = 0; i < files[0].length; i++) {
				doUpload(sftp, new File(files[0][i]), files[1][i], files[2][i]);
			}

		} catch (Exception e) {
			throw new StepException(e);
		} finally {
			logout();
		}
	}

	/**
	 * Upload a file to the FTP server in the given remote directory and with
	 * the provided name.
	 * 
	 * @param file
	 *            The file to upload.
	 * @param remoteDirectory
	 *            The remote directory.
	 * @param remoteFilename
	 *            The remote file name.
	 * @throws Exception
	 */
	public void upload(File file, String remoteDirectory, String remoteFilename)
			throws Exception {

		// Get the SFTP Channel.
		ChannelSftp sftp = getSftpChannel();

		try {
			doUpload(sftp, file, remoteDirectory, remoteFilename);
		} finally {
			logout();
		}
	}

	/**
	 * Private method to upload the file to the SFTP server. Does not logout,
	 * calling method must do it.
	 * 
	 * @param sftp
	 *            The SFTP channel.
	 * @param file
	 *            The file to upload.
	 * @param remoteDirectory
	 *            The remote directory.
	 * @param remoteFilename
	 *            The remote file name.
	 * @throws Exception
	 */
	protected synchronized void doUpload(ChannelSftp sftp, File file,
			String remoteDirectory, String remoteFilename) throws Exception {

		// Prepare the RetryHelper.
		RetryHelper retry = new RetryHelper(nbRetries, retryFactor,
				retryInterval);

		InputStream is = null;
		boolean completed = false;
		boolean cwd = false;
		while (!completed) {
			try {
				// Change the remote directory.
				remoteDirectory = remoteDirectory.trim().replace("\\", "/");
				if (StringUtils.isNotBlank(remoteDirectory)) {
					log.info("Changing to remote dir: " + remoteDirectory);
					sftp.cd(remoteDirectory);
					cwd = true;
				}

				// Upload the file.
				if (file.getName().equals(remoteFilename)) {
					log.info("Sending file: " + file.getName() + " - Attempt "
							+ retry.getCurrentAttempt() + ".");
				} else {
					log.info("Sending file: " + file.getName()
							+ " with remote name " + remoteFilename
							+ " - Attempt " + retry.getCurrentAttempt() + ".");
				}
				is = IOHelper.getInputStream(file);
				sftp.put(is, remoteFilename.trim());

				log.info("Complete");
				completed = true;

			} catch (Exception e) {
				// If there are retry attempts left, retry.
				if (retry.getCurrentAttempt() < retry.getMaxAttempts()) {
					retry.waitForNextRetry();
				} else {
					throw e;
				}
			} finally {
				// Move back to root directory.
				if (cwd) {
					for (int i = 0; i < remoteDirectory.split("/").length; i++) {
						sftp.cd("..");
					}
					cwd = false;
				}
				if (is != null) {
					is.close();
				}
			}
		}
	}

}
