package com.taleo.integration.client.customstep.http;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.security.GeneralSecurityException;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.UsernamePasswordCredentials;
import org.apache.commons.httpclient.auth.BasicScheme;
import org.apache.commons.httpclient.methods.FileRequestEntity;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.integration.client.customstep.util.HttpUtils;
import com.taleo.integration.client.customstep.util.RetryHelper;
import com.taleo.integration.client.customstep.util.xml.XmlUtils;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.integration.client.util.PasswordUtils;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Post-processing step to HTTP Post the current workflow file or other files
 * corresponding to a pattern.
 * <p>
 * The result of this step is a file containing each HTTP Post response, in the
 * following structure:
 * 
 * <pre>
 * &lt;HTTPPosts&gt;
 *   &lt;HTTPPost filename="" url=""&gt;
 *     &lt;Request&gt;
 *       &lt;Headers&gt;
 *         &lt;Header name="" value=""/&gt;
 *         ...
 *       &lt;/Headers&gt;
 *       &lt;Body&gt;
 *       ...
 *       &lt;/Body&gt;
 *     &lt;/Request&gt;
 *     &lt;Response status=""&gt;
 *       &lt;Headers&gt;
 *         &lt;Header name="" value=""/&gt;
 *         ...
 *       &lt;/Headers&gt;
 *       &lt;Body&gt;
 *       ...
 *       &lt;/Body&gt;
 *     &lt;/Response&gt;
 *   &lt;/HTTPPost&gt;
 *   ...
 * &lt;/HTTPPosts&gt;
 * </pre>
 * 
 * Response status corresponds to <a
 * href="http://www.w3.org/Protocols/HTTP/HTRESP.html">HTTP status codes</a>.
 * <p>
 * This file becomes current in the pipeline at the end of the step.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>url</b>: The URL to post to.</li>
 * <br>
 * <br>
 * <li><b>username</b>: The username.</li>
 * <br>
 * <br>
 * <li><b>password</b>: The AES128 encrypted or unencrypted password. Encrypted
 * passwords are recognized by their ending with "=".</li>
 * <br>
 * <br>
 * <li><b>filePattern</b>: Optional - The pattern for selecting the files to
 * HTTP Post. Can be specified as an absolute or relative path and may use
 * wildcards as defined in
 * {@link org.apache.commons.io.filefilter.WildcardFileFilter} in the file names
 * only. Pattern is case sensitive. If the pattern includes multiple files, they
 * will be processed in ascending name order. If absent, the current workflow
 * document is posted.</li>
 * <br>
 * <br>
 * <li><b>moveFolder</b>: Optional - Folder to move files after successful
 * posting. If absent, files are deleted. If the folder is the same as the file
 * itself, leave the file there.</li>
 * <br>
 * <br>
 * <li><b>nbRetries</b>: Optional - Represents the number of retries that will
 * be done if the operation fails (default: 0).</li>
 * <br>
 * <br>
 * <li><b>retryInterval</b>: Optional - Represents the number of seconds to wait
 * between the retry attempts (default: 0).</li>
 * <br>
 * <br>
 * <li><b>retryFactor</b>: Optional - This factor is used as a multiplier on the
 * <code>retryInterval</code> between each attempt. It multiplies the last time
 * waited by this amount. For example. if <code>retryInterval</code> is 10
 * seconds and the retry factor is 2, wait time will be successively 10, 20, 40
 * seconds and so on, until the maximum number of retries has been reached
 * (default: 1).</li>
 * <br>
 * <br>
 * <li><b>HTTP Headers</b>: HTTP headers are supported through the use of
 * advanced parameters having the prefix "Header:" followed by the name of the
 * header. For example, the Content-type parameter name will be
 * Header:Content-type.</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * <li><a href="http://hc.apache.org/httpclient-3.x/">Apache Commons
 * HTTPClient</a> (commons-httpclient-3.1.jar) provided in TCC/lib.</li>
 * </ul>
 * <p>
 * Here are two examples of how to define this custom step in the TCC
 * configuration file:
 * <p>
 * <img src="doc-files/HttpPostPostStep-1.jpg" />
 * <p>
 * <img src="doc-files/HttpPostPostStep-2.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class HttpPostPostStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(HttpPostPostStep.class);

	private HttpClient httpClient = HttpUtils.getHttpClient();

	/**
	 * Constructor without arguments.
	 */
	public HttpPostPostStep() {
		super();
		// Register valid parameters.
		registerParameterName("url");
		registerParameterName("username");
		registerParameterName("password");
		registerParameterName("filePattern");
		registerParameterName("moveFolder");
		registerParameterName("nbRetries");
		registerParameterName("retryInterval");
		registerParameterName("retryFactor");
	}

	/**
	 * Validate the parameters names.
	 * <p>
	 * Override the default method to allow for HTTP headers to be passed.
	 */
	protected void validateParameterNames() {
		Enumeration keys = parameters.keys();
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			if (!parameterNames.contains(key) && !key.startsWith("Header:")) {
				throw new IllegalArgumentException("Unknown parameter " + key
						+ ".");
			}
		}
	}

	/**
	 * The url.
	 */
	private String url;

	/**
	 * The FTP username.
	 */
	private String username;

	/**
	 * The FTP password (AES128 encrypted or unencrypted).
	 */
	private String password;

	/**
	 * The file pattern.
	 */
	private String filePattern;

	/**
	 * The moveFolder.
	 */
	private String moveFolder;

	/**
	 * The number of retries.
	 */
	private int nbRetries = RetryHelper.NB_RETRIES_DEFAULT;

	/**
	 * The interval between retries, in seconds.
	 */
	private int retryInterval = RetryHelper.RETRY_INTERVAL_DEFAULT;

	/**
	 * The retry factor.
	 */
	private int retryFactor = RetryHelper.RETRY_FACTOR_DEFAULT;

	/**
	 * The headers.
	 */
	private Properties headers = new Properties();

	/**
	 * Get the URL.
	 * 
	 * @return
	 */
	public String getURL() {
		return url;
	}

	/**
	 * Set the URL.
	 * 
	 * @param url
	 */
	public void setURL(String url) {
		this.url = url;
	}

	/**
	 * Get the HTTP password.
	 * 
	 * @return
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * Set the HTTP password.
	 * 
	 * @param password
	 */
	public void setPassword(String password) {
		this.password = password;
	}

	/**
	 * Get the file pattern.
	 * 
	 * @return
	 */
	public String getFilePattern() {
		return filePattern;
	}

	/**
	 * Set the file pattern.
	 * 
	 * @param filePattern
	 */
	public void setFilePattern(String filePattern) {
		this.filePattern = filePattern;
	}

	/**
	 * Get the move folder.
	 * 
	 * @return
	 */
	public String getMoveFolder() {
		return moveFolder;
	}

	/**
	 * Set the move folder.
	 * 
	 * @param moveFolder
	 */
	public void setMoveFolder(String moveFolder) {
		this.moveFolder = moveFolder;
	}

	/**
	 * Get the HTTP username.
	 * 
	 * @return
	 */
	public String getUsername() {
		return username;
	}

	/**
	 * Set the HTTP username.
	 * 
	 * @param username
	 */
	public void setUsername(String username) {
		this.username = username;
	}

	/**
	 * Get the number of retries.
	 * 
	 * @return
	 */
	public int getNbRetries() {
		return nbRetries;
	}

	/**
	 * Set the number of retries.
	 * 
	 * @param nbRetries
	 */
	public void setNbRetries(int nbRetries) {
		this.nbRetries = nbRetries;
	}

	/**
	 * Get the retry interval, in seconds.
	 * 
	 * @return
	 */
	public int getRetryInterval() {
		return retryInterval;
	}

	/**
	 * Set the retry interval, in seconds.
	 * 
	 * @param retryInterval
	 */
	public void setRetryInterval(int retryInterval) {
		this.retryInterval = retryInterval;
	}

	/**
	 * Get the retry factor.
	 * 
	 * @return
	 */
	public int getRetryFactor() {
		return retryFactor;
	}

	/**
	 * Set the retry factor.
	 * 
	 * @param retryFactor
	 */
	public void setRetryFactor(int retryFactor) {
		this.retryFactor = retryFactor;
	}

	/**
	 * Get the header by its name.
	 * 
	 * @param name
	 * 
	 * @return
	 */
	public String getHeader(String name) {
		return headers.getProperty(name);
	}

	/**
	 * Set the header by its name.
	 * 
	 * @param name
	 * @param value
	 */
	public void setHeader(String name, String value) {
		this.headers.put(name, value);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "HttpPostPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to HTTP Post the workflow document or other files corresponding to a pattern.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "HttpPostPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// First, get the basic supported pipeline for a post-processing step
		// (this handles the constraints on the last step type)
		SupportedPipeline supportedPipeline = getPostProcessSupportPipeline();

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		url = parameters.getProperty("url");
		if (url == null || url.trim().length() == 0) {
			throw new IllegalArgumentException(
					"The 'url' parameter is mandatory.");
		}

		username = parameters.getProperty("username");

		password = parameters.getProperty("password");

		filePattern = parameters.getProperty("filePattern");

		moveFolder = parameters.getProperty("moveFolder");

		try {
			nbRetries = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("nbRetries"),
					Integer.toString(nbRetries)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'nbRetries' parameter must have a numeric value.");
		}

		try {
			retryInterval = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("retryInterval"),
					Integer.toString(retryInterval)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'retryInterval' parameter must have a numeric value.");
		}

		try {
			retryFactor = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("retryFactor"),
					Integer.toString(retryFactor)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'retryFactor' parameter must have a numeric value.");
		}

		// Get the HTTP Headers.
		Enumeration keys = parameters.keys();
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			if (!parameterNames.contains(key) && key.startsWith("Header:")) {
				setHeader(key.substring(7), parameters.get(key).toString());
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// If not active, return immediately.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		// Get the file pattern.
		String filePattern = this.filePattern;
		Message responseMessage = pipeline.getResponseMessage();
		if (filePattern == null || filePattern.trim().length() == 0) {
			if (responseMessage.isSetCurrentFile()) {
				filePattern = responseMessage.getCurrentFile();
			} else {
				filePattern = responseMessage.getOriginalFile();
			}
		}

		try {
			// Execute the step.
			Document document = httpPost(new File(filePattern));

			// Store the document in a temp file.
			File outFile = createTempFile();
			OutputStream os = IOHelper.getOutputStream(new FileOutputStream(
					outFile), getClass().getName());
			XmlUtils.writeXml(document, os);
			os.close();

			// Set the new current file.
			responseMessage.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * HTTP Post the files corresponding to the file pattern.
	 * 
	 * @param pattern
	 *            The file pattern.
	 * @return A DOM Document containing the requests and responses, with the
	 *         following structure:
	 * 
	 *         <pre>
	 * &lt;HTTPPosts&gt;
	 *   &lt;HTTPPost&gt;
	 *   	...
	 *   &lt;/HTTPPost&gt;
	 *   ...
	 * &lt;/HTTPPosts&gt;
	 * </pre>
	 * @throws IOException
	 */
	public Document httpPost(File pattern) throws IOException {
		// Get the files to process.
		File[] files = FileUtils.getMatchingFiles(pattern);

		FileUtils.sortFilesByNameAscending(files, true);
		log.info(files.length + " file(s) match the pattern.");

		// Create a document to store the result.
		Document document = XmlUtils.createDomDocument();
		Element httpPosts = document.createElement("HTTPPosts");
		document.appendChild(httpPosts);

		// Process each file.
		for (int i = 0; i < files.length; i++) {
			File file = files[i];

			// Post the file.
			Document httpPost = doHttpPost(file, getURL());

			// Append to the result.
			Node node = document.adoptNode(httpPost.getDocumentElement());
			httpPosts.appendChild(node);

			// Move file to target folder.
			if (moveFolder != null) {
				if (new File(moveFolder).equals(file.getParentFile())) {
					// Do nothing.
				} else if (moveFolder.trim().length() > 0) {
					File moveFile = new File(moveFolder, file.getName());
					if (moveFile.exists()) {
						moveFile.delete();
					}
					org.apache.commons.io.FileUtils.moveFileToDirectory(file,
							new File(moveFolder), true);
				}
			} else {
				file.delete();
			}
		}
		return document;
	}

	/**
	 * HTTP Post the file.
	 * 
	 * @param file
	 *            The file to post.
	 * @param url
	 *            The URL to post to.
	 * @return A DOM Document containing the request and response, with the
	 *         following structure:
	 * 
	 *         <pre>
	 *   &lt;HTTPPost&gt;
	 *   	...
	 *   &lt;/HTTPPost&gt;
	 * </pre>
	 * @throws IOException
	 */
	protected Document doHttpPost(File file, String url) throws IOException {
		// Prepare HTTP post.
		PostMethod post = new PostMethod(url);

		// Prepare authorization header.
		String password = this.password;
		if (password != null && password.endsWith("=")) {
			try {
				password = PasswordUtils.decryptAES128(password);
			} catch (GeneralSecurityException e) {
				throw new RuntimeException(e);
			}
		}
		if (username != null) {
			UsernamePasswordCredentials creds = new UsernamePasswordCredentials(
					username, password);
			Header authorizationHeader = new Header("authorization",
					BasicScheme.authenticate(creds, "ISO-8859-1"));
			post.setRequestHeader(authorizationHeader);
		}

		// Set the default content type.
		String contentType = getHeader("Content-Type");
		if (contentType == null) {
			contentType = "text/xml; charset=" + getEncoding();
		}

		// Set request content.
		post.setRequestEntity(new FileRequestEntity(file, contentType));

		// Add headers.
		Iterator it = headers.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry entry = (Map.Entry) it.next();
			post.addRequestHeader((String) entry.getKey(),
					(String) entry.getValue());
		}

		// Create a document to store the result.
		Document document = XmlUtils.createDomDocument();
		Element httpPost = document.createElement("HTTPPost");
		document.appendChild(httpPost);
		httpPost.setAttribute("filename", file.getName());
		httpPost.setAttribute("url", url);
		Element request = document.createElement("Request");
		httpPost.appendChild(request);
		Element response = document.createElement("Response");
		httpPost.appendChild(response);

		// Add request headers to the result document.
		Element requestHeaders = document.createElement("Headers");
		request.appendChild(requestHeaders);
		for (int i = 0; i < post.getRequestHeaders().length; i++) {
			Header header = post.getRequestHeaders()[i];
			Element element = document.createElement("Header");
			requestHeaders.appendChild(element);
			element.setAttribute("name", header.getName());
			element.setAttribute("value", header.getValue());
		}

		// Add request body to the result document.
		String body = null;
		if (post.getRequestEntity() != null) {
			OutputStream baos = new ByteArrayOutputStream();
			post.getRequestEntity().writeRequest(baos);
			body = baos.toString();
		}
		if (body != null) {
			Element requestBody = document.createElement("Body");
			request.appendChild(requestBody);
			if (XmlUtils.isWellFormed(body)) {
				// Add request XML content.
				Document bodyDocument;
				try {
					bodyDocument = XmlUtils.createDomDocument(body);
				} catch (SAXException e) {
					throw new RuntimeException(e);
				}
				Node node = document.adoptNode(bodyDocument
						.getDocumentElement());
				requestBody.appendChild(node);
			} else {
				// If not a valid XML request, load it as text.
				requestBody.setTextContent(body);
			}
		}

		// Prepare the RetryHelper.
		RetryHelper retry = new RetryHelper(nbRetries, retryFactor,
				retryInterval);

		// Execute
		boolean completed = false;
		while (!completed) {
			try {
				log.info("HTTP posting file: '" + file.getName() + "'.");
				log.info("Opening connection on " + url + " - Attempt "
						+ retry.getCurrentAttempt() + ".");

				// Execute the request.
				int statusCode = httpClient.executeMethod(post);

				// Add response headers to the result document.
				Element responseHeaders = document.createElement("Headers");
				response.appendChild(responseHeaders);
				response.setAttribute("status", Integer.toString(statusCode));
				for (int i = 0; i < post.getResponseHeaders().length; i++) {
					Header header = post.getResponseHeaders()[i];
					Element element = document.createElement("Header");
					responseHeaders.appendChild(element);
					element.setAttribute("name", header.getName());
					element.setAttribute("value", header.getValue());
				}

				// Add response body to the result document.
				Element responseBody = document.createElement("Body");
				response.appendChild(responseBody);
				String responseString = post.getResponseBodyAsString();
				if (XmlUtils.isWellFormed(responseString)) {
					// Add response XML content.
					Document bodyDocument;
					try {
						bodyDocument = XmlUtils
								.createDomDocument(responseString);
					} catch (SAXException e) {
						throw new RuntimeException(e);
					}
					Node node = document.adoptNode(bodyDocument
							.getDocumentElement());
					responseBody.appendChild(node);
				} else {
					// If not a valid XML response, load it as text.
					responseBody.setTextContent(responseString);
				}

				log.info("Complete");
				completed = true;

			} catch (IOException e) {
				// If there are retry attempts left, retry.
				if (retry.getCurrentAttempt() < retry.getMaxAttempts()) {
					retry.waitForNextRetry();
				} else {
					throw e;
				}
			} finally {
				// Release connection.
				post.releaseConnection();
			}
		}

		return document;
	}
}
