package com.taleo.integration.client.customstep.lrd;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.ParseException;
import java.util.Date;

import org.apache.commons.io.FileUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.DateUtils;

/**
 * Date provider that uses a date stored on file. The file must contain the date
 * in ISO-8601 format.
 * 
 * @author Romain Guay - Taleo
 * 
 */
public class FileDateProvider extends AbstractDateProvider {

	private static Logger log = Logger.getLogger(FileDateProvider.class);

	/**
	 * The date.
	 */
	private Date date;

	/**
	 * The file.
	 */
	private File file;

	/**
	 * Constructor with the file. Date will be null if the file does not exist
	 * or is empty.
	 * 
	 * @param file
	 *            The file to get the date from.
	 */
	public FileDateProvider(File file) {
		this(file, null);
	}

	/**
	 * Constructor with the file and base provider.
	 * 
	 * @param file
	 *            The file to get the date from.
	 * @param baseProvider
	 *            The DateProvider to use if the file is empty or absent.
	 */
	public FileDateProvider(File file, DateProvider baseProvider) {
		this(file, baseProvider, false);
	}

	/**
	 * Constructor with the file, base provider and reset flag.
	 * 
	 * @param file
	 *            The file to get the date from.
	 * @param baseProvider
	 *            The DateProvider to use if the file is empty or absent.
	 * @param reset
	 *            Reset the date, forcing to read date from the baseProvider.
	 */
	public FileDateProvider(File file, DateProvider baseProvider, boolean reset) {
		if (file == null) {
			throw new IllegalArgumentException("Argument 'file' is mandatory.");
		}
		Date date = null;

		// Look if the file exists or reset is true.
		if (file == null || !file.exists() || reset) {
			if (baseProvider != null) {
				date = baseProvider.getDate();
			}

		} else {
			try {
				// Read the file.
				BufferedReader in = new BufferedReader(new FileReader(file));
				String str = in.readLine();
				if (str == null) {
					if (baseProvider != null) {
						date = baseProvider.getDate();
					}
				} else {
					try {
						date = DateUtils.parseDateFromISO8601(str);
					} catch (ParseException e) {
						throw new RuntimeException(e);
					}
				}
				in.close();
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
		}
		this.date = date;
		this.file = file;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.customstep.lrd.DateProvider#getDate()
	 */
	public Date getDate() {
		return date;
	}

	/**
	 * Save the date in the file or clear the file if date is null.
	 * 
	 * @throws IOException
	 */
	public void saveDate() throws IOException {
		// Create the file if necessary.
		if (file.getParentFile() != null) {
			FileUtils.forceMkdir(file.getParentFile());
		}
		file.createNewFile();

		// Write in the file.
		BufferedWriter out = new BufferedWriter(new FileWriter(file));
		if (date == null) {
			log.info("Deleting date from " + file.getPath());
			out.write("");
		} else {
			String formattedDate = DateUtils.formatDateAsISO8601(date);
			log.info("Saving date: " + formattedDate + " to " + file.getPath());
			out.write(formattedDate);
		}
		out.close();
	}

}
