package com.taleo.integration.client.customstep.lrd;

import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.paging.PagingManager;
import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Post-processing step to write the Last Run Date at the end of the execution.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep}.
 * <p>
 * This post-processing step works in pair with {@link LRDPreStep}. It can be
 * safely applied to a query that has no LRD filter, in which case it does
 * nothing.
 * <p>
 * It also works in pair with
 * {@link com.taleo.integration.client.customstep.paging.PagingPreStep} if it is
 * present. If this is the case, it will not save the LRD until the paging is
 * complete.
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file. Note that this step must be defined as the last one.
 * <p>
 * <img src="doc-files/LRDPostStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class LRDPostStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(LRDPostStep.class);

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "LRDPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step for LRD filter";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "LRDPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the RunDateManager that was put in the pipeline by the
			// pre-processing step.
			RunDateManager runDateManager = (RunDateManager) pipeline
					.documentProperties().get(RunDateManager.PIPELINE_KEY);

			// If the RunDateManager is null, quit immediately.
			if (runDateManager == null) {
				return;
			}

			// Get the PagingManager that was put in the pipeline by the
			// PagingPreStep.
			PagingManager pagingManager = (PagingManager) pipeline
					.documentProperties().get(PagingManager.PIPELINE_KEY);

			// If the PagingManager is incomplete, return immediately without
			// saving the LRD.
			if (pagingManager != null && !pagingManager.isComplete()) {
				return;
			}

			// Save the new last run date.
			runDateManager.saveLastRunDate();

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

}
