package com.taleo.integration.client.customstep.lrd;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.Duration;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.DateUtils;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.integration.client.util.InstallHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Pre-processing step to substitute a "Last Run Date (LRD)" filter at runtime
 * with a "Between" filter in which the current and last run dates have been
 * modified with a mask that allows to synchronize multiple extracts on a
 * specific datetime in the past.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>dateTimeMask</b>: Optional - This mask is a date format that must
 * respect the following pattern: "yyyy-MM-dd HH:mm:ss"
 * {@link java.text.SimpleDateFormat} and that is applied to the current run
 * date to make sure multiple successive extracts are synchronized on the same
 * current date. The default is set to "yyyy-MM-dd HH:mm:ss", which does no
 * synchronization. <br>
 * A typical mask is "yyyy-MM-dd 00:00:00", which would synchronize current run
 * dates on midnight. Another example could be "yyyy-MM-01 00:00:00", which
 * would synchronize on the first day of the month. Also, "yyyy-MM-dd HH:00:00"
 * would synchronize on the beginning of the hour. This strategy must be set in
 * consideration of the frequency that the integration is running.<br>
 * Note that the time part of the mask is interpreted in the local time zone.</li>
 * <br>
 * <br>
 * <li><b>abortOnRerun</b>: Optional - Indicates if the run is aborted when the
 * last run date and the effective current run date are equal. (
 * <code>true</code> or <code>false</code>, default: <code>false</code>). <br>
 * This is a useful strategy when running in a loop over a time increment (see
 * <code>maxExtractDays</code>). Setting this to <code>true</code> will then be
 * used as a signal to end looping. It is also useful when rerunning a series of
 * extracts while skipping the ones that already ran successfully. The strategy
 * applies no matter if the query contains a Last Run Date filter or not.<br>
 * When <code>true</code>, a rerun will cause aborting the Java process with the
 * value 0 for exit code. The exit code can be set differently using the system
 * property named
 * <code>com.taleo.client.customstep.LRDPreStep.RerunExitCode</code>.
 * <p>
 * <b>Note:</b> Exiting with a specific code is currently the only option to end
 * TCC on demand. The down side is that the monitoring file remains incomplete.</li>
 * <br>
 * <br>
 * <li><b>dateTimeSource</b>: <i>@deprecated - This parameter is no more
 * supported and will be ignored. Date/time is now always read from the local
 * server.</i></li>
 * <br>
 * <br>
 * <li><b>currentRunDate</b>: Optional - Specific current run date to use. If
 * present, this value overrides the actual datetime read from the local server
 * (see {@link LocalDateProvider}). This parameter must match one of the
 * following formats: "yyyy-MM-dd", "yyyy-MMM-dd", "yyyy-MMM-dd HH:mm:ss" or
 * "yyyy-MM-dd HH:mm:ss" and will be interpreted in the local time zone.
 * <p>
 * This date can be in the future only if <b>abortOnRerun</b> is false.
 * Otherwise the process would not abort until the clock reaches that future
 * date.</li>
 * <br>
 * <br>
 * <li><b>maxExtractDays</b>: Optional - Maximum number of days (integer or
 * decimal) to be extracted on a given run. Zero (0) or null value means no
 * limit.</li>
 * <br>
 * <br>
 * <li><b>maxExtractDuration</b>: <i>@deprecated - This parameter is no more
 * used since maxExtractDays supports decimal values.</i> - Maximum duration <a
 * href="http://www.w3.org/TR/xmlschema-2/#duration">XML Schema 1.0 section
 * 3.2.6.1</a> to be extracted on a given run. Empty value or 0 means no limit.
 * <br>
 * If both <code>maxExtractDays</code> and <code>maxExtractDuration</code> are
 * defined, <code>maxExtractDays</code> will prevail.</li>
 * <br>
 * <br>
 * <li><b>configFile</b>: <i>@deprecated - This parameter is no more used since
 * dateTimeSource was @deprecated.</i> The filename of the configuration file.
 * This value may be supplied as a relative or absolute path. It is required if
 * we use the <code>zone</code> dateTimeSource, because the config contains the
 * zone information (URL, username, password) to which the TCC API does not give
 * access directly.</li>
 * <br>
 * <br>
 * <li><b>lrdNaming</b>: <i>@deprecated - Use the lrdFilename instead.</i> The
 * LRD naming strategy (<code>config</code> or <code>query</code>, default:
 * <code>config</code>).
 * <ul>
 * <li><code>config</code>: Choose this option if the LRD is stored by the
 * configuration file name, as with the standard TCC strategy.</li>
 * <li><code>query</code>: Choose this option if the LRD is stored by the SQ-XML
 * query file name. This allows to run multiple extracts that use the LRD from
 * the same configuration.</li>
 * </ul>
 * </li>
 * <li><b>lrdFilename</b>: Optional - The name of the file to read and store the
 * LRD. The ".lrd" extension will be added if not present in the filename. If
 * absent, the configuration identifier will be used.</li>
 * <br>
 * <br>
 * <li><b>crdFilename</b>: Optional - The name of the file to read and store the
 * Current Run Date (CRD). The ".crd" extension will be added if not present in
 * the filename and the file will be stored in the temporary folder. If absent,
 * the CRD is simply not stored. It is useful to store the CRD when an
 * integration based on LRD filtering is performed in multiple extracts and each
 * extract must use the same LRD and CRD. It can also be used to force a
 * specific CRD. If a crdFilename is provided, the CRD is written to the file at
 * the end of the LRDPreStep.
 * <p>
 * To use that strategy, the first extract in the series will reset the CRD (see
 * <b>resetCrd</b> parameter) and the other ones will read it from that same
 * file. Only the last extract in the series will contain the LRDPostStep.</li>
 * <br>
 * <br>
 * <li><b>resetCrd</b>: Optional - Reset the CRD, forcing to read it afresh from
 * the server clock (<code>true</code> or <code>false</code>, default:
 * <code>false</code>).</li>
 * <br>
 * <br>
 * <li><b>createOnly</b>: Optional - Transform LRD filters, limiting result to
 * only records that were created (not updated) in the time interval (
 * <code>true</code> or <code>false</code>, default: <code>false</code>).<br>
 * This applies to queries that have multiple LRD filters. In those cases, it is
 * possible to identify which one is pulling created records by adding a tag
 * "create" to the &lt;quer:lrd&gt;, for example:
 * 
 * <pre>
 * 		&lt;quer:lrd tag="create"&gt; 
 * 		  &lt;quer:field path="CreationDate"/&gt; 
 * 		&lt;/quer:lrd&gt;
 * </pre>
 * 
 * Setting this to <code>true</code> will turn off those other filters that are
 * not tagged as "create". If the query has no LRD filter tagged as "create",
 * this option does nothing.</li>
 * </ul>
 * <p>
 * This pre-processing step must be defined before the 'Prepare Export' step. It
 * works in pair with {@link LRDPostStep}. It can be safely applied to a query
 * that has no LRD filter, in which case it has no effect.
 * <p>
 * Here are some examples of how to define this custom step in the TCC
 * configuration file;
 * <p>
 * <img src="doc-files/LRDPreStep-1.jpg" />
 * <p>
 * <img src="doc-files/LRDPreStep-2.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class LRDPreStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(LRDPreStep.class);

	/**
	 * Constant for the RerunExitCode system property.
	 */
	private static final String PROP_RERUN_EXIT_CODE = "com.taleo.client.customstep.LRDPreStep.RerunExitCode";

	/**
	 * Default value for com.taleo.client.customstep.LRDPreStep.RerunExitCode
	 * system property.
	 */
	private static final int DEFAULT_RERUN_EXIT_CODE = 0;

	/**
	 * Data type factory for Duration objects.
	 */
	static DatatypeFactory dtFactory = null;
	static {
		try {
			dtFactory = DatatypeFactory.newInstance();
		} catch (DatatypeConfigurationException e) {
			throw new RuntimeException(e);
		}
	}

	/** Default datetime source. */
	public static final String DEFAULT_DATETIME_SOURCE = "local";

	/** Default lRD naming strategy. */
	public static final String DEFAULT_LRD_NAMING = "config";

	/** The XSL transformation file for CreateOnly. */
	private final static String XSL_TRANSFORM_CREATE_ONLY = "xsl/LRDFilterCreateOnly.xsl";

	/** The XSL transformation file for LRD. */
	private final static String XSL_TRANSFORM_LRD = "xsl/ReplaceLRDFilter.xsl";

	/** Date formats to read and write dates */
	private final DateFormat dateFormat1 = new SimpleDateFormat("yyyy-MM-dd");
	private final DateFormat dateFormat2 = new SimpleDateFormat("yyyy-MMM-dd");
	private final DateFormat datetimeFormat1 = new SimpleDateFormat(
			"yyyy-MM-dd HH:mm:ss");
	private final DateFormat datetimeFormat2 = new SimpleDateFormat(
			"yyyy-MMM-dd HH:mm:ss");

	/** Default datetime mask. */
	public final static String DEFAULT_DATETIME_MASK = "yyyy-MM-dd HH:mm:ss";

	/**
	 * Constructor without arguments.
	 */
	public LRDPreStep() {
		super();
		// Register valid parameters.
		registerParameterName("dateTimeMask");
		registerParameterName("abortOnRerun");
		registerParameterName("dateTimeSource"); // @deprecated
		registerParameterName("configFile"); // @deprecated
		registerParameterName("lrdNaming");
		registerParameterName("currentRunDate");
		registerParameterName("maxExtractDays");
		registerParameterName("maxExtractDuration");
		registerParameterName("lrdFilename");
		registerParameterName("crdFilename");
		registerParameterName("resetCrd");
		registerParameterName("createOnly");
	}

	/**
	 * The datetime mask.
	 */
	private String dateTimeMask = DEFAULT_DATETIME_MASK;

	/**
	 * The abort on rerun flag.
	 */
	private boolean abortOnRerun = false;

	/**
	 * The datetime source.
	 * 
	 * @deprecated
	 */
	private String dateTimeSource = DEFAULT_DATETIME_SOURCE;

	/**
	 * The LRD naming strategy.
	 * 
	 * @deprecated
	 */
	private String lrdNaming = DEFAULT_LRD_NAMING;

	/**
	 * The LRD filename.
	 */
	private String lrdFilename;

	/**
	 * The CRD filename.
	 */
	private String crdFilename;

	/**
	 * The reset CRD flag.
	 */
	private boolean resetCrd = false;

	/**
	 * The configuration file name.
	 * 
	 * @deprecated
	 */
	private String configFile;

	/**
	 * The global config.
	 */
	private GlobalConfig globalConfig;

	/**
	 * The current run date.
	 */
	private Date currentRunDate;

	/**
	 * The maximum number of days to extract.
	 */
	private double maxExtractDays = 0d;

	/**
	 * The createOnly flag.
	 */
	private boolean createOnly = false;

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "LRDPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Pre-processing step for LRD filter";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "LRDPreStep";
	}

	/**
	 * Get the datetime mask.
	 * 
	 * @return
	 */
	public String getDateTimeMask() {
		return dateTimeMask;
	}

	/**
	 * Set the datetime mask.
	 * 
	 * @param dateTimeMask
	 */
	public void setDateTimeMask(String dateTimeMask) {
		this.dateTimeMask = dateTimeMask;
	}

	/**
	 * Get the abort on rerun flag.
	 * 
	 * @return
	 */
	public boolean isAbortOnRerun() {
		return abortOnRerun;
	}

	/**
	 * Set the abort on rerun flag.
	 * 
	 * @param abortOnRerun
	 */
	public void setAbortOnRerun(boolean abortOnRerun) {
		this.abortOnRerun = abortOnRerun;
	}

	/**
	 * Get the datetime source.
	 * 
	 * @deprecated
	 * @return
	 */
	public String getDateTimeSource() {
		return dateTimeSource;
	}

	/**
	 * Set the datetime source.
	 * 
	 * @deprecated
	 * @param dateTimeSource
	 */
	public void setDateTimeSource(String dateTimeSource) {
		if (dateTimeSource != null) {
			if (!(dateTimeSource.equals("local") || dateTimeSource
					.equals("zone"))) {
				throw new IllegalArgumentException(
						"The 'dateTimeSource' parameter must be either 'local' or 'zone'.");
			}
			this.dateTimeSource = dateTimeSource;
		}
	}

	/**
	 * Get the config file.
	 * 
	 * @deprecated
	 * @return
	 */
	public String getConfigFile() {
		return configFile;
	}

	/**
	 * Set the config file.
	 * 
	 * @deprecated
	 * @param configFile
	 */
	public void setConfigFile(String configFile) {
		this.configFile = configFile;
	}

	/**
	 * Get the LRD naming strategy.
	 * 
	 * @return
	 * @deprecated
	 */
	public String getLrdNaming() {
		return lrdNaming;
	}

	/**
	 * Set the LRD naming strategy.
	 * 
	 * @param lrdNaming
	 * @deprecated
	 */
	public void setLrdNaming(String lrdNaming) {
		if (lrdNaming != null) {
			if (!(lrdNaming.equals("config") || lrdNaming.equals("query"))) {
				throw new IllegalArgumentException(
						"The 'lrdNaming' parameter must be either 'config' or 'query'.");
			}
			this.lrdNaming = lrdNaming;
		}
	}

	/**
	 * Get the LRD filename.
	 * 
	 * @return
	 */
	public String getLrdFilename() {
		return lrdFilename;
	}

	/**
	 * Set the LRD filename.
	 * 
	 * @param lrdFilename
	 */
	public void setLrdFilename(String lrdFilename) {
		this.lrdFilename = lrdFilename;
	}

	/**
	 * Get the CRD filename.
	 * 
	 * @return
	 */
	public String getCrdFilename() {
		return crdFilename;
	}

	/**
	 * Set the CRD filename.
	 * 
	 * @param crdFilename
	 */
	public void setCrdFilename(String crdFilename) {
		this.crdFilename = crdFilename;
	}

	/**
	 * Get the resetCrd flag.
	 * 
	 * @return
	 */
	public boolean isResetCrd() {
		return resetCrd;
	}

	/**
	 * Set the resetCrd flag.
	 * 
	 * @param resetCrd
	 */
	public void setResetCrd(boolean resetCrd) {
		this.resetCrd = resetCrd;
	}

	/**
	 * Get the current run date.
	 * 
	 * @return
	 */
	public Date getCurrentRunDate() {
		return currentRunDate;
	}

	/**
	 * Set the current run date.
	 * 
	 * @param currentRunDate
	 */
	public void setCurrentRunDate(Date currentRunDate) {
		this.currentRunDate = currentRunDate;
	}

	/**
	 * Get the maximum number of days to extract.
	 * 
	 * @return
	 */
	public double getMaxExtractDays() {
		return maxExtractDays;
	}

	/**
	 * Set the maximum number of days to extract.
	 * 
	 * @param maxExtractDays
	 */
	public void setMaxExtractDays(double maxExtractDays) {
		if (maxExtractDays < 0) {
			throw new IllegalArgumentException(
					"The 'maxExtractDays' parameter must be >= 0");
		}
		this.maxExtractDays = maxExtractDays;
	}

	/**
	 * Get the maximum duration to extract.
	 * 
	 * @deprecated
	 * @return The duration or null if none was set.
	 */
	public Duration getMaxExtractDuration() {
		return dtFactory.newDuration(Math.round(maxExtractDays * 24d * 60d
				* 60d * 1000d));
	}

	/**
	 * Set the maximum duration to extract.
	 * 
	 * @deprecated
	 * @param maxExtractDuration
	 */
	public void setMaxExtractDuration(Duration maxExtractDuration) {
		if (maxExtractDuration == null) {
			throw new IllegalArgumentException(
					"The 'maxExtractDuration' parameter must not be null.");
		}
		if (maxExtractDuration.getSign() < 0) {
			throw new IllegalArgumentException(
					"The 'maxExtractDuration' parameter must be positive.");
		}
		this.maxExtractDays = maxExtractDuration.getTimeInMillis(new Date())
				/ 1000d / 60d / 60d / 24d;
	}

	/**
	 * Get the createOnly flag.
	 * 
	 * @return
	 */
	public boolean isCreateOnly() {
		return createOnly;
	}

	/**
	 * Set the createOnly flag.
	 * 
	 * @param createOnly
	 */
	public void setCreateOnly(boolean createOnly) {
		this.createOnly = createOnly;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);
		this.globalConfig = config;

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		String dateTimeMask = parameters.getProperty("dateTimeMask");
		if (StringUtils.isNotBlank(dateTimeMask)) {
			setDateTimeMask(dateTimeMask);
		}

		abortOnRerun = Boolean.valueOf(
				parameters.getProperty("abortOnRerun", "false")).booleanValue();

		setDateTimeSource(parameters.getProperty("dateTimeSource",
				DEFAULT_DATETIME_SOURCE));

		setConfigFile(parameters.getProperty("configFile"));
		if (dateTimeSource.equals("zone") && configFile == null) {
			throw new IllegalArgumentException(
					"The 'configFile' parameter is required if 'dateTimeSource' is equal to 'zone'.");
		}

		setLrdNaming(parameters.getProperty("lrdNaming", DEFAULT_LRD_NAMING));

		setLrdFilename(parameters.getProperty("lrdFilename"));

		setCrdFilename(parameters.getProperty("crdFilename"));

		resetCrd = Boolean.valueOf(parameters.getProperty("resetCrd", "false"))
				.booleanValue();

		String crd = parameters.getProperty("currentRunDate");
		if (StringUtils.isNotBlank(crd)) {
			try {
				setCurrentRunDate(datetimeFormat1.parse(crd));
			} catch (ParseException e1) {
				try {
					setCurrentRunDate(datetimeFormat2.parse(crd));
				} catch (ParseException e2) {
					try {
						setCurrentRunDate(dateFormat1.parse(crd));
					} catch (ParseException e3) {
						try {
							setCurrentRunDate(dateFormat2.parse(crd));
						} catch (ParseException e4) {
							throw new IllegalArgumentException(
									"The currentRunDate ("
											+ crd
											+ ") must match either the 'yyyy-MM-dd', 'yyyy-MMM-dd', 'yyyy-MM-dd HH:mm:ss' or the 'yyyy-MMM-dd HH:mm:ss' format.");
						}
					}
				}
			}
		}

		try {
			setMaxExtractDays(Double.parseDouble(StringUtils.defaultIfEmpty(
					parameters.getProperty("maxExtractDays"), "0")));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The maxExtractDays must be numeric.");
		}

		String duration = parameters.getProperty("maxExtractDuration");
		if (StringUtils.isNotBlank(duration)) {
			try {
				setMaxExtractDays(Double.parseDouble(StringUtils
						.defaultIfEmpty(duration, "0")));
			} catch (NumberFormatException e) {
				setMaxExtractDuration(dtFactory.newDuration(duration));
			}
		}

		createOnly = Boolean.valueOf(
				parameters.getProperty("createOnly", "false")).booleanValue();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the request info.
			Message requestInfo = pipeline.getRequestMessage();
			String queryFilename = requestInfo.getCurrentFile();
			if (queryFilename == null) {
				queryFilename = requestInfo.getOriginalFile();
			}
			File queryFile = new File(queryFilename);

			// Get the last run date filename.
			String lrd;
			if (lrdFilename != null) {
				lrd = lrdFilename;
			} else {
				lrd = globalConfig.getGeneral().getConfigurationIdentifier();
			}
			if (!RunDateManager.LRD_EXTENSION.equals(FilenameUtils
					.getExtension(lrd))) {
				lrd += "." + RunDateManager.LRD_EXTENSION;
			}

			// Create the LRD file in the LRD folder.
			String lrdFolder = InstallHelper.getLRDFolder().getAbsolutePath();
			File lrdFile = new File(lrdFolder, lrd);
			log.info("Using the following LRD file: '"
					+ lrdFile.getAbsolutePath() + "'.");

			// Default current run date.
			Date currentRunDate = this.currentRunDate;

			// Get the current run date file.
			File crdFile = null;
			if (crdFilename != null) {
				String crd = crdFilename;

				if (!RunDateManager.CRD_EXTENSION.equals(FilenameUtils
						.getExtension(crd))) {
					crd += "." + RunDateManager.CRD_EXTENSION;
				}

				// Create the CRD file in the temp folder.
				crdFile = new File(getTempFolder(), crd);
				log.info("Using the following CRD file: '"
						+ crdFile.getAbsolutePath() + "'.");
			}

			// Create the date provider for the current run date.
			DateProvider dateProvider = null;
			if (currentRunDate != null) {
				// Use a DateProvider on the given current run date.
				// Do not allow a future date if abortOnRerun is true.
				dateProvider = new FixedDateProvider(currentRunDate,
						!abortOnRerun);

			} else {
				// Use the local DateProvider.
				dateProvider = new LocalDateProvider();
			}

			// Wrap it into a FileDateProvider with the crdFile.
			if (crdFile != null) {
				dateProvider = new FileDateProvider(crdFile, dateProvider,
						resetCrd);
			}

			// Create a RunDateManager and store it in the pipeline for
			// retrieval at the end of the workflow.
			RunDateManager runDateManager = new RunDateManager(lrdFile,
					dateProvider, getMaxExtractDays(), getDateTimeMask());
			pipeline.documentProperties().put(RunDateManager.PIPELINE_KEY,
					runDateManager);

			// Get the begin and end dates.
			Date begin = runDateManager.getLastRunDate();
			Date end = runDateManager.getEffectiveRunDate();

			if (getMaxExtractDays() > 0) {
				log.info("Using a time increment of " + getMaxExtractDays()
						+ " days.");
			} else {
				log.info("No time increment set.");
			}

			// Abort on rerun if required.
			if (abortOnRerun && (begin.equals(end) || begin.after(end))) {
				// Note: TCC currently does not provide any means to abort
				// execution. The only way is to kill the process.
				int exitCode = DEFAULT_RERUN_EXIT_CODE;
				String prop = System.getProperty(PROP_RERUN_EXIT_CODE);
				if (StringUtils.isNotBlank(prop)) {
					try {
						// Read the system property for exit code.
						exitCode = Integer.parseInt(System
								.getProperty(PROP_RERUN_EXIT_CODE));
					} catch (NumberFormatException e) {
						// Invalid property value: keep default value.
						log.warn("Invalid value '" + prop
								+ "' for system property "
								+ PROP_RERUN_EXIT_CODE);
					}
				}
				log.info("The run was aborted since the last run date ("
						+ DateUtils.formatDateAsISO8601(begin)
						+ ") is greater than or equal to the current run date ("
						+ DateUtils.formatDateAsISO8601(end)
						+ "). Exiting with code: " + exitCode + "\n");
				System.exit(exitCode);
			}

			// Transform the LRD filters for createOnly option.
			if (isCreateOnly()) {
				File outFile = createTempFile();
				transform_CreateOnly(queryFile, outFile);
				queryFile = outFile;
			}

			// Transform the LRD filters to Between filters.
			File outFile = createTempFile();
			transform_LRD(queryFile, outFile, begin, end);

			// Set the transformed query in the request.
			requestInfo.setCurrentFile(outFile.getAbsolutePath());

			// Save the current run date.
			if (crdFile != null) {
				((FileDateProvider) dateProvider).saveDate();
			}

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Transform LRD filters to manage createOnly option.
	 * 
	 * @param queryFile
	 *            The query file.
	 * @param outFile
	 *            The output file.
	 * @throws TransformerException
	 * @throws IOException
	 */
	protected void transform_CreateOnly(File queryFile, File outFile)
			throws TransformerException, IOException {
		log.info("Transforming the query to filter on created records only, if applicable.");

		// Create an XSL transformer.
		InputStream xsl = this.getClass().getClassLoader()
				.getResourceAsStream(XSL_TRANSFORM_CREATE_ONLY);
		TransformerFactory tFactory = TransformerFactory.newInstance();
		Transformer transformer = tFactory
				.newTransformer(new StreamSource(xsl));

		// Prepare streams.
		InputStream is = IOHelper.getInputStream(queryFile);
		OutputStream os = IOHelper.getOutputStream(
				new FileOutputStream(outFile), outFile.getName());

		// Prepare source and result.
		Source source = new StreamSource(is);
		Result result = new StreamResult(os);

		// Apply the transformation.
		transformer.transform(source, result);

		// Close.
		is.close();
		os.close();
	}

	/**
	 * Transform LRD filters into Between filters using the specified begin and
	 * end dates.
	 * 
	 * @param queryFile
	 *            The query file.
	 * @param outFile
	 *            The output file.
	 * @param beginDate
	 *            The filter begin date.
	 * @param endDate
	 *            The filter end date.
	 * @throws TransformerException
	 * @throws IOException
	 */
	protected void transform_LRD(File queryFile, File outFile, Date beginDate,
			Date endDate) throws TransformerException, IOException {

		// Build the dates for the filter and set them as parameters of the
		// transformation.
		String begin = DateUtils.formatDateAsISO8601(beginDate);
		String end = DateUtils.formatDateAsISO8601(endDate);

		log.info("Transforming the LRD filter into: Between " + begin + " And "
				+ end);

		// Create an XSL transformer.
		InputStream xsl = this.getClass().getClassLoader()
				.getResourceAsStream(XSL_TRANSFORM_LRD);
		TransformerFactory tFactory = TransformerFactory.newInstance();
		Transformer transformer = tFactory
				.newTransformer(new StreamSource(xsl));

		// Set parameters.
		transformer.setParameter("beginDate", begin);
		transformer.setParameter("endDate", end);

		// Prepare streams.
		InputStream is = IOHelper.getInputStream(queryFile);
		OutputStream os = IOHelper.getOutputStream(
				new FileOutputStream(outFile), outFile.getName());

		// Prepare source and result.
		Source source = new StreamSource(is);
		Result result = new StreamResult(os);

		// Apply the transformation.
		transformer.transform(source, result);

		// Close.
		is.close();
		os.close();
	}

}
