package com.taleo.integration.client.customstep.lrd;

import java.io.File;
import java.io.IOException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Properties;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;

/**
 * RunDateManager that saves the last run dates on the file system.
 * <p>
 * It also exposes the current run date that is by default the local system date
 * but can be taken from another date provider.
 * <p>
 * This class can be called from within a Java process or run in standalone mode
 * to get/set the last run date in a file.
 * <p>
 * If the last run date is null, it is considered 1 day before now.
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public class RunDateManager {

	/** The LRD file extension. */
	final static String LRD_EXTENSION = "lrd";

	/** The CRD file extension. */
	final static String CRD_EXTENSION = "crd";

	/** Key for the RunDateManager in the pipeline. */
	public final static String PIPELINE_KEY = "RunDateManager";

	/** Key for the properties file in System properties. */
	private static final String PROPERTIES_FILE_KEY = RunDateManager.class
			.getName() + ".properties";

	/** Date format to parse masked dates. */
	private static final DateFormat maskParser = new SimpleDateFormat(
			"yyyy-MM-dd HH:mm:ss");

	/**
	 * The current run date.
	 */
	private Date runDate = null;

	/**
	 * Data type factory for Duration objects.
	 */
	static DatatypeFactory dtFactory = null;
	static {
		try {
			dtFactory = DatatypeFactory.newInstance();
		} catch (DatatypeConfigurationException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * The effective run date, considering the maximum extract days and the
	 * dateTime mask.
	 */
	private Date effectiveRunDate = null;

	/**
	 * The file to store and retrieve last run date.
	 */
	private File lastRunDateFile = null;

	/**
	 * A cache for last run date.
	 */
	private Date lastRunDateCache = null;

	/**
	 * The date/time mask matching the following pattern: "yyyy-MM-dd HH:mm:ss".
	 */
	private String dateTimeMask = null;

	/**
	 * Execute to get and/or set the last run date.
	 * <p>
	 * Use the -help option to get usage description.
	 * <p>
	 * Exits with 0 on success and 1 in case of error.
	 * 
	 * @param args
	 *            Arguments
	 */
	public static void main(String[] args) {
		try {
			final String DATE_FORMAT = "yyyy-MM-dd HH:mm:ss";
			DateFormat dateFormat = new SimpleDateFormat(DATE_FORMAT);

			// Parse the arguments.
			Date lastRunDate = null;
			String property = null;
			String filename = null;
			int i = 0;
			int j = 0;
			if (args.length == 0) {
				System.err.println(usage());
				System.exit(1);
			}
			while (i < args.length) {
				String arg = args[i++];

				if (arg.startsWith("-")) {

					if (arg.equals("-help")) {
						System.out.println(usage());
						System.exit(0);

					} else if (arg.equals("-lastRunDate")) {
						if (i < args.length) {
							String date = args[i++];
							try {
								lastRunDate = dateFormat.parse(date);
							} catch (ParseException e) {
								throw new IllegalArgumentException(
										e.getMessage());
							}
						} else {
							throw new IllegalArgumentException(
									"-lastRunDate requires a value");
						}
					}

				} else {
					if (j == 0) {
						property = arg;
						j++;
					} else {
						throw new IllegalArgumentException(
								"Unrecognized argument: " + arg);
					}
				}
			}
			if (property == null) {
				throw new IllegalArgumentException(
						"filename/property is required");
			}

			// Try to resolve the property or fallback on the filename.
			String propertiesFile = System.getProperty(PROPERTIES_FILE_KEY);
			if (propertiesFile != null) {
				Properties properties = new Properties();
				properties.load(RunDateManager.class.getClassLoader()
						.getResourceAsStream(propertiesFile));
				filename = (String) properties.get(property);
				if (filename == null) {
					filename = property;
				}
			} else {
				filename = property;
			}

			// Instantiate the manager.
			RunDateManager manager = new RunDateManager(new File(filename));

			// Save the specified last run date, if present.
			if (lastRunDate != null) {
				manager.saveLastRunDate(lastRunDate);
			} else {
				// Otherwise, simply write the last run date.
				System.out.println(dateFormat.format(manager.getLastRunDate()));
			}

		} catch (IllegalArgumentException e) {
			System.err.println("#ERROR#: " + e.getMessage());
			System.err.println();
			System.err.println(usage());
			System.exit(1);
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * Build a usage description.
	 * 
	 * @return
	 */
	private static String usage() {
		String usage = "Manages last run dates stored on the file system. \n"
				+ "Usage:\n"
				+ "RunDateManager <filename/property> [-lastRunDate <date>] \n"
				+ "\n"
				+ "     <filename/property>: The filename or the property that contains the filename in the properties file. The property is searched first. \n"
				+ "                          The filename may be defined as absolute or relative to the current directory. \n"
				+ "     [-lastRunDate <date>]: Optional - Value of the last run date in the following format (yyyy-MM-dd HH:mm:ss).\n"
				+ "                            If absent, the actual last run date is returned. The local time zone is considered. \n"
				+ "\n"
				+ "System properties: \n"
				+ "-D"
				+ PROPERTIES_FILE_KEY
				+ ": The properties file. Must be reachable from the classpath.\n"
				+ "\n"
				+ "Examples: \n"
				+ "java.exe com.taleo.integration.client.customstep.lrd.RunDateManager myLastRunDate.lrd -lastRunDate \"2006-12-01 12:00:00\" \n"
				+ "java.exe -Dcom.taleo.integration.client.customstep.lrd.RunDateManager.properties=\"myProperties.properties\" com.taleo.integration.client.customstep.lrd.RunDateManager lastRunDate.filename -lastRunDate \"2006-12-01 12:00:00\" \n";
		return usage;
	}

	/**
	 * Constructor with the file to store the last run date.
	 * 
	 * @param file
	 *            The file to store the last run date.
	 */
	public RunDateManager(File file) {
		this(file, null);
	}

	/**
	 * Constructor with the file to store the last run date and the date
	 * provider.
	 * 
	 * @param file
	 *            The file to store the last run date.
	 * @param dateProvider
	 *            The date provider for the current date.
	 */
	public RunDateManager(File file, DateProvider dateProvider) {
		this(file, dateProvider, 0);
	}

	/**
	 * Constructor with the file to store the last run date, the date provider
	 * and the maximum number of days to extract.
	 * 
	 * @param file
	 *            The file to store the last run date.
	 * @param dateProvider
	 *            The date provider for the current date.
	 * @param maxExtractDays
	 *            The maximum number of days to extract (0 means no maximum).
	 */
	public RunDateManager(File file, DateProvider dateProvider,
			double maxExtractDays) {
		this(file, dateProvider, maxExtractDays, null);
	}

	/**
	 * Constructor with the file to store the last run date, the date provider
	 * and the maximum number of days to extract.
	 * 
	 * @param file
	 *            The file to store the last run date.
	 * @param dateProvider
	 *            The date provider for the current date.
	 * @param maxExtractDays
	 *            The maximum number of days to extract (0 means no maximum).
	 * @param dateTimeMask
	 *            The date/time mask to apply to dates.
	 */
	public RunDateManager(File file, DateProvider dateProvider,
			double maxExtractDays, String dateTimeMask) {
		if (file == null) {
			throw new IllegalArgumentException(
					"The 'file' parameter must not be null.");
		}
		if (maxExtractDays < 0) {
			throw new IllegalArgumentException(
					"The 'maxExtractDays' parameter must be positive.");
		}
		if (dateTimeMask != null) {
			Pattern pattern = Pattern
					.compile("[y|\\d]{4}-[M|\\d]{2}-[d|\\d]{2} [H|\\d]{2}:[m|\\d]{2}:[s|\\d]{2}");
			Matcher matcher = pattern.matcher(dateTimeMask);
			if (!matcher.matches()) {
				throw new IllegalArgumentException(
						"The 'dateTimeMask' parameter must match the following pattern: 'yyyy-MM-dd HH:mm:ss'");
			}
		}

		// Set the default date provider.
		if (dateProvider == null) {
			dateProvider = new LocalDateProvider();
		}

		// Store properties.
		this.lastRunDateFile = file;
		this.runDate = dateProvider.getDate();
		this.dateTimeMask = dateTimeMask;

		// The effectiveRunDate defaults to the masked current run date.
		if (this.effectiveRunDate == null) {
			this.effectiveRunDate = maskDate(this.runDate);
		}

		// Calculate the adjusted run date by adding the maximum extract
		// days (if > 0) to the last run date.
		if (maxExtractDays > 0) {
			Calendar current = GregorianCalendar.getInstance();
			current.setTime(this.effectiveRunDate);
			Calendar adjusted = GregorianCalendar.getInstance();
			adjusted.setTime(this.getLastRunDate());
			adjusted.add(Calendar.SECOND,
					(int) Math.round(maxExtractDays * 24d * 60d * 60d));

			// If the adjusted run date is prior to the current run date,
			// keep it as the effective run date.
			if (adjusted.before(current)) {
				this.effectiveRunDate = adjusted.getTime();
			}
		}
	}

	/**
	 * Return the file where the last run date is stored.
	 * 
	 * @return
	 */
	public File getFile() {
		return lastRunDateFile;
	}

	/**
	 * Get the current run date.
	 * 
	 * @return
	 */
	public Date getRunDate() {
		return runDate;
	}

	/**
	 * Get the last run date.
	 * 
	 * @return
	 * @throws IOException
	 */
	public Date getLastRunDate() {
		// First look in the cache.
		if (lastRunDateCache != null) {
			return lastRunDateCache;
		}

		// Get from file.
		DateProvider provider = new FileDateProvider(lastRunDateFile);
		Date lrd = provider.getDate();

		// Use default if null.
		if (lrd == null) {
			lrd = getDefaultLastRunDate();
		}

		// Write to cache and return;
		lastRunDateCache = lrd;
		return lrd;
	}

	/**
	 * Get the effective run date used for filtering.
	 * 
	 * @return
	 */
	public Date getEffectiveRunDate() {
		return effectiveRunDate;
	}

	/**
	 * Get the effective last run date used for filtering.
	 * 
	 * @return
	 * @throws IOException
	 * @deprecated Use getLastRunDate() instead.
	 */
	public Date getEffectiveLastRunDate() {
		return getLastRunDate();
	}

	/**
	 * Get the default last run date, which is 1 day before the current run
	 * date.
	 * 
	 * @return
	 */
	public Date getDefaultLastRunDate() {
		Calendar cal = Calendar.getInstance();
		cal.setTime(getRunDate());
		cal.add(Calendar.DATE, -1);
		return cal.getTime();
	}

	/**
	 * Save the effective run date as the new last run date.
	 * 
	 * @return The newly saved last run date.
	 * @throws IOException
	 */
	public Date saveLastRunDate() throws IOException {
		return saveLastRunDate(getEffectiveRunDate());
	}

	/**
	 * Save the given date as the new last run date.
	 * 
	 * @return The newly saved last run date.
	 * @throws IOException
	 */
	public Date saveLastRunDate(Date date) throws IOException {
		// Save the last run date.
		FixedDateProvider fixed = new FixedDateProvider(date);
		FileDateProvider provider = new FileDateProvider(lastRunDateFile,
				fixed, true);
		provider.saveDate();

		// Reset the cache and return.
		lastRunDateCache = null;
		return date;
	}

	/**
	 * Mask the given date.
	 * 
	 * @param date
	 *            The date to be masked.
	 * @return The masked date.
	 */
	protected Date maskDate(Date date) {
		if (dateTimeMask == null) {
			return date;
		}

		// Build a date format for masking.
		DateFormat maskFormatter = new SimpleDateFormat(dateTimeMask);

		// Make sure the time zone for the maskParser is set to the default.
		// The default time zone may have changed for testing or other reasons.
		maskParser.setTimeZone(TimeZone.getDefault());

		try {
			// Return the masked date.
			return maskParser.parse(maskFormatter.format(date));

		} catch (ParseException e) {
			throw new RuntimeException(e);
		}
	}

}
