package com.taleo.integration.client.customstep.mail;

import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.activation.DataHandler;
import javax.activation.DataSource;
import javax.activation.FileDataSource;
import javax.mail.BodyPart;
import javax.mail.MessagingException;
import javax.mail.Multipart;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.runtime.RuntimeConstants;
import org.apache.velocity.runtime.log.NullLogChute;
import org.apache.velocity.runtime.resource.loader.FileResourceLoader;
import org.apache.velocity.tools.generic.DateTool;
import org.apache.velocity.tools.generic.MathTool;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.velocity.CsvExportFileTool;
import com.taleo.integration.client.customstep.velocity.CsvImportFileTool;
import com.taleo.integration.client.customstep.velocity.LogicTool;
import com.taleo.integration.client.customstep.velocity.StringTool;
import com.taleo.integration.client.customstep.velocity.WorkflowTool;
import com.taleo.integration.client.customstep.velocity.XmlExportFileTool;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.integration.client.util.QueryRuntimeHelper;
import com.taleo.integration.client.util.csv.CSVConstant;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.query.QueryDocument.Query;

/**
 * Post-processing step to send the current workflow file by email.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>SMTPHost</b>: Mandatory - The SMTP host.</li>
 * <br>
 * <br>
 * <li><b>port</b>: Optional - The SMTP port (default: 25).</li>
 * <br>
 * <br>
 * <li><b>from</b>: Optional - The sender of the email (default: Taleo Connect
 * Client <donotreply@invalidemail.com>).</li>
 * <br>
 * <br>
 * <li><b>to</b>: Mandatory - The receiver list separated by comma.</li>
 * <br>
 * <br>
 * <li><b>cc</b>: Optional - The cc receiver list separated by comma.</li>
 * <br>
 * <br>
 * <li><b>subject</b>: Mandatory - The subject of the email.</li>
 * <br>
 * <br>
 * <li><b>sendAttachment</b>: Optional - Indicates if the file is to be attached
 * to the email. (<code>true</code> or <code>false</code>, default:
 * <code>true</code>).</li>
 * <br>
 * <br>
 * <li><b>filename</b>: Optional - The filename for the attachment if
 * "sendAttachment" is <code>true</code>. If absent, the original filename from
 * the response will be used. If there is none, the step will fail.</li>
 * <br>
 * <br>
 * <li><b>messageTemplateFile</b>: The name of a template file that contains the
 * email message body. Can be specified as an absolute path or relative to the
 * working directory. If absent, the body of the email is empty.</li>
 * <br>
 * <br>
 * <li><b>messageTemplate</b>: The template string that contains the email
 * message body. This may be used in replacement of a
 * <code>messageTemplateFile</code> if the template is very short. If both are
 * present, this one has priority.</li>
 * <br>
 * <br>
 * <li><b>mimeType</b>: Optional - The Mime type of the email body (
 * <code>text/plain</code> or <code>text/html</code>). Other Mime types have not
 * been tested. Default: <code>text/plain</code></li>
 * <br>
 * <br>
 * <li><b>failOnError</b>: Optional - Fail if the email sending errors out
 * (Default: true). If false, only an error log is written.</li>
 * <br>
 * <br>
 * <li><b>triggeringRule</b>: Optional - Triggering rule for sending of email
 * (Default: true). The rule must return a boolean value and typically uses the
 * $logic variable along with other variables described below.</li>
 * </ul>
 * <p>
 * The following additional parameters are required in the case of an import
 * integration if the extra counters (success, create, update, delete, error)
 * are to be used in the message template.
 * <ul>
 * <li><b>csvHeaderPresent</b>: Indicates if a CSV header is present in the file
 * (<code>true</code> or <code>false</code>, default: <code>false</code>).</li>
 * <br>
 * <br>
 * <li><b>csvDelimiter</b>: The CSV delimiter (default: ,).</li>
 * <br>
 * <br>
 * <li><b>csvQuoteCharacter</b>: The CSV quote character (default: ").</li>
 * </ul>
 * <p>
 * The following variables are made available to build the email template and
 * triggering rule:
 * <ul>
 * <li><b>$date</b>: A Date tool instance (
 * {@link org.apache.velocity.tools.generic.DateTool}) to access various date
 * functions.</li>
 * <br>
 * <br>
 * <li><b>$math</b>: A Math tool instance (
 * {@link org.apache.velocity.tools.generic.MathTool}) to access various
 * mathematical functions.</li>
 * <br>
 * <br>
 * <li><b>$string</b>: A string tool instance (
 * {@link com.taleo.integration.client.customstep.velocity.StringTool}) to
 * access various string functions.</li>
 * <br>
 * <br>
 * <li><b>$file</b>: A file tool instance<br>
 * ({@link com.taleo.integration.client.customstep.velocity.CsvExportFileTool})
 * to access CSV export file properties and content.</li> <br>
 * ({@link com.taleo.integration.client.customstep.velocity.XmlExportFileTool})
 * to access XML export file properties and content.</li> <br>
 * ({@link com.taleo.integration.client.customstep.velocity.CsvImportFileTool})
 * to access CSV import file properties and content.</li> <br>
 * <br>
 * <li><b>$workflow</b>: A workflow tool instance (
 * {@link com.taleo.integration.client.customstep.velocity.WorkflowTool}) to
 * access various workflow variables.</li>
 * <br>
 * </ul>
 * <p>
 * Example uses of these variables in the templates:
 * <ul>
 * <li>$date.get("yyyy-MM-dd")</li>
 * <li>$date.getYear()</li>
 * <li>$date.getMonth()</li>
 * <li>$date.getDay()</li>
 * <li>$math.add($date.getYear(), 1)</li>
 * <li>$string.rightPad("123", 10)</li>
 * <li>$string.rightPad("123", 10, "_")</li>
 * <li>$string.leftPad("123", 10)</li>
 * <li>$string.leftPad("123", 10, "0")</li>
 * <li>$string.substring("123456", 2, 3)</li>
 * <li>$string.substringAfter("123456", "3")</li>
 * <li>$file.name()</li>
 * <li>$file.path()</li>
 * <li>$file.lastModified()</li>
 * <li>$file.size()</li>
 * <li>$file.content()</li>
 * <li>$file.recordCount()</li>
 * <li>$file.lineCount()</li>
 * <li>$workflow.getRunDate()</li>
 * <li>$workflow.getLastRunDate()</li>
 * <li>$workflow.getEffectiveRunDate()</li>
 * <li>$workflow.getEffectiveLastRunDate()</li>
 * <li>$date.format("yyyy-MM-dd", $workflow.getRunDate())</li>
 * </ul>
 * <p>
 * For import results, there is also:
 * <ul>
 * <li>$file.successCount()</li>
 * <li>$file.createCount()</li>
 * <li>$file.updateCount()</li>
 * <li>$file.deleteCount()</li>
 * <li>$file.errorCount()</li>
 * <li>$file.errorCount("excludePattern")</li>
 * <li>$file.errorContent()</li>
 * <li>$file.errorContent("excludePattern")</li>
 * </ul>
 * <p>
 * For triggering rule, the $logic variable (
 * {@link com.taleo.integration.client.customstep.velocity.LogicTool}) may be
 * used:
 * <ul>
 * <li>$logic.greaterThan($file.errorCount(), 0)</li>
 * </ul>
 * <p>
 * There are also other Velocity constructs that can be used. For reference, see
 * <a href=
 * "http://velocity.apache.org/engine/releases/velocity-1.4/user-guide.html"
 * >Velocity User Guide</a>.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://velocity.apache.org/index.html">Apache Velocity
 * library</a> (velocity-dep-1.5.jar)</li>
 * <li><a href="http://velocity.apache.org/tools/devel/generic.html">Apache
 * Velocity Tools Generic library</a> (velocity-tools-generic-1.4.jar)</li>
 * <li><a href="http://commons.apache.org/net/">Java mail library</a> (mail.jar
 * - provided with TCC)</li>
 * </ul>
 * <p>
 * Here are some examples of how to define this custom step in the TCC
 * configuration file:
 * <p>
 * <img src="doc-files/SendEmailPostStep-1.jpg" />
 * <p>
 * <img src="doc-files/SendEmailPostStep-2.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class SendEmailPostStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(SendEmailPostStep.class);

	private static final String DEFAULT_SENDER = "Taleo Connect Client <donotreply@invalidemail.com>";

	/**
	 * Constructor without arguments.
	 */
	public SendEmailPostStep() {
		super();
		// Register valid parameters.
		registerParameterName("SMTPHost");
		registerParameterName("port");
		registerParameterName("from");
		registerParameterName("to");
		registerParameterName("cc");
		registerParameterName("subject");
		registerParameterName("sendAttachment");
		registerParameterName("filename");
		registerParameterName("messageTemplateFile");
		registerParameterName("messageTemplate");
		registerParameterName("mimeType");
		registerParameterName("csvHeaderPresent");
		registerParameterName("csvDelimiter");
		registerParameterName("csvQuoteCharacter");
		registerParameterName("failOnError");
		registerParameterName("triggeringRule");

		// Create the velocity engine.
		this.velocity = new VelocityEngine();
		Properties props = new Properties();
		// Disable default logging.
		props.put(RuntimeConstants.RUNTIME_LOG_LOGSYSTEM_CLASS,
				NullLogChute.class.getName());
		props.put("resource.loader", "file");
		props.put("file.resource.loader.class",
				FileResourceLoader.class.getName());
		// Empty loader path allows specifying templates with an
		// absolute path.
		props.put("file.resource.loader.path", "");
		props.put("file.resource.loader.cache", "false");
		try {
			velocity.init(props);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * The global config.
	 */
	private GlobalConfig config;

	/**
	 * The default port.
	 */
	private final int DEFAULT_PORT = 25;

	/**
	 * The SMTP host.
	 */
	private String SmtpHost;

	/**
	 * The SMTP port.
	 */
	private int port = DEFAULT_PORT;

	/**
	 * The sender email.
	 */
	private String from = DEFAULT_SENDER;

	/**
	 * The receivers list of emails separated by comma.
	 */
	private String to = "";

	/**
	 * The cc receivers list of emails separated by comma.
	 */
	private String cc = "";

	/**
	 * The subject of the email.
	 */
	private String subject = "";

	/**
	 * Indicates if the results file is sent as an attachment.
	 */
	private boolean sendAttachment = true;

	/**
	 * The filename.
	 */
	private String filename;

	/**
	 * The template file to build the email message.
	 */
	private String messageTemplateFile;

	/**
	 * The template string to build the email message.
	 */
	private String messageTemplate;

	/**
	 * The Mime type of the email message.
	 */
	private String mimeType = "text/plain";

	/**
	 * Indicates if a CSV header is present in the file.
	 */
	private boolean csvHeaderPresent = false;

	/**
	 * The CSV delimiter.
	 */
	private char csvDelimiter = CSVConstant.DEFAULT_DELIMITER;

	/**
	 * The CSV quote character.
	 */
	private char csvQuoteCharacter = CSVConstant.DEFAULT_QUOTE_CHARACTER;

	/**
	 * The failOnError flag.
	 */
	private boolean failOnError = true;

	/**
	 * The triggeringRule.
	 */
	private String triggeringRule = "true";

	/**
	 * The Velocity engine.
	 */
	private VelocityEngine velocity = null;

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "SendEmailPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to send an email with the results of a TCC export or import.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "SendEmailPostStep";
	}

	/**
	 * @return
	 */
	public String getCc() {
		return cc;
	}

	/**
	 * @param cc
	 */
	public void setCc(String cc) {
		if (cc == null) {
			this.cc = "";
		} else {
			this.cc = cc;
		}
	}

	/**
	 * @return
	 */
	public String getFrom() {
		return from;
	}

	/**
	 * @param from
	 */
	public void setFrom(String from) {
		if (from == null) {
			this.from = "";
		} else {
			this.from = from;
		}
	}

	/**
	 * @return
	 */
	public int getPort() {
		return port;
	}

	/**
	 * @param port
	 */
	public void setPort(int port) {
		this.port = port;
	}

	/**
	 * @return
	 */
	public boolean isSendAttachment() {
		return sendAttachment;
	}

	/**
	 * @param sendAttachment
	 */
	public void setSendAttachment(boolean sendAttachment) {
		this.sendAttachment = sendAttachment;
	}

	/**
	 * Get the filename.
	 * 
	 * @return
	 */
	public String getFilename() {
		return filename;
	}

	/**
	 * Set the filename.
	 * 
	 * @param filename
	 */
	public void setFilename(String filename) {
		this.filename = filename;
	}

	/**
	 * @return
	 */
	public String getSmtpHost() {
		return SmtpHost;
	}

	/**
	 * @param smtpHost
	 */
	public void setSmtpHost(String smtpHost) {
		SmtpHost = smtpHost;
	}

	/**
	 * @return
	 */
	public String getSubject() {
		return subject;
	}

	/**
	 * @param subject
	 */
	public void setSubject(String subject) {
		if (subject == null) {
			this.subject = "";
		} else {
			this.subject = subject;
		}
	}

	/**
	 * @return
	 */
	public String getMessageTemplateFile() {
		return messageTemplateFile;
	}

	/**
	 * @param messageTemplateFile
	 */
	public void setMessageTemplateFile(String messageTemplateFile) {
		this.messageTemplateFile = messageTemplateFile;
	}

	/**
	 * @return
	 */
	public String getMessageTemplate() {
		return messageTemplate;
	}

	/**
	 * @param messageTemplate
	 */
	public void setMessageTemplate(String messageTemplate) {
		this.messageTemplate = messageTemplate;
	}

	/**
	 * Get the Mime type of the email.
	 * 
	 * @return
	 */
	public String getMimeType() {
		return mimeType;
	}

	/**
	 * Set the Mime type os the email.
	 * 
	 * @param mimeType
	 */
	public void setMimeType(String mimeType) {
		this.mimeType = mimeType;
	}

	/**
	 * @return
	 */
	public String getTo() {
		return to;
	}

	/**
	 * @param to
	 */
	public void setTo(String to) {
		if (to == null) {
			this.to = "";
		} else {
			this.to = to;
		}
	}

	/**
	 * @return
	 */
	public boolean isCsvHeaderPresent() {
		return csvHeaderPresent;
	}

	/**
	 * @param csvHeaderPresent
	 */
	public void setCsvHeaderPresent(boolean csvHeaderPresent) {
		this.csvHeaderPresent = csvHeaderPresent;
	}

	/**
	 * @return
	 */
	public char getCsvDelimiter() {
		return csvDelimiter;
	}

	/**
	 * @param csvDelimiter
	 */
	public void setCsvDelimiter(char csvDelimiter) {
		this.csvDelimiter = csvDelimiter;
	}

	/**
	 * @return
	 */
	public char getCsvQuoteCharacter() {
		return csvQuoteCharacter;
	}

	/**
	 * @param csvQuoteCharacter
	 */
	public void setCsvQuoteCharacter(char csvQuoteCharacter) {
		this.csvQuoteCharacter = csvQuoteCharacter;
	}

	/**
	 * Get the FailOnError flag.
	 * 
	 * @return
	 */
	public boolean isFailOnError() {
		return failOnError;
	}

	/**
	 * Set the FailOnError flag.
	 * 
	 * @param failOnError
	 */
	public void setFailOnError(boolean failOnError) {
		this.failOnError = failOnError;
	}

	/**
	 * Get the triggering rule.
	 * 
	 * @return
	 */
	public String getTriggeringRule() {
		return triggeringRule;
	}

	/**
	 * Set the triggering rule.
	 * 
	 * @param triggeringRule
	 */
	public void setTriggeringRule(String triggeringRule) {
		this.triggeringRule = triggeringRule.trim();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);
		this.config = config;

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		setSmtpHost(parameters.getProperty("SMTPHost"));
		if (StringUtils.isBlank(SmtpHost)) {
			throw new IllegalArgumentException(
					"The 'SmtpHost' parameter is mandatory.");
		}

		try {
			port = Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("port"), Integer.toString(port)));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'port' parameter must have a numeric value.");
		}

		String from = parameters.getProperty("from");
		if (!StringUtils.isBlank(from)) {
			setFrom(from);
		}

		setTo(parameters.getProperty("to"));
		if (StringUtils.isBlank(to)) {
			throw new IllegalArgumentException(
					"The 'to' parameter is mandatory.");
		}

		setCc(parameters.getProperty("cc"));

		setSubject(parameters.getProperty("subject"));
		if (StringUtils.isBlank(subject)) {
			throw new IllegalArgumentException(
					"The 'subject' parameter is mandatory.");
		}

		setSendAttachment(Boolean.valueOf(
				parameters.getProperty("sendAttachment", "true"))
				.booleanValue());

		filename = parameters.getProperty("filename");

		setMessageTemplateFile(parameters.getProperty("messageTemplateFile"));

		setMessageTemplate(parameters.getProperty("messageTemplate"));

		setMimeType(parameters.getProperty("mimeType", mimeType));

		setCsvHeaderPresent(Boolean.valueOf(
				parameters.getProperty("csvHeaderPresent", "false"))
				.booleanValue());

		String csvDelimiter = (parameters.getProperty("csvDelimiter"));
		if (StringUtils.isNotBlank(csvDelimiter)) {
			char[] chars = csvDelimiter.toCharArray();
			if (chars.length > 1) {
				throw new IllegalArgumentException(
						"The 'csvDelimiter' parameter must be a single character.");
			}
			setCsvDelimiter(chars[0]);
		}

		String csvQuoteCharacter = (parameters.getProperty("csvQuoteCharacter"));
		if (StringUtils.isNotBlank(csvQuoteCharacter)) {
			char[] chars = csvQuoteCharacter.toCharArray();
			if (chars.length > 1) {
				throw new IllegalArgumentException(
						"The 'csvQuoteCharacter' parameter must be a single character.");
			}
			setCsvQuoteCharacter(chars[0]);
		}

		setFailOnError(Boolean.valueOf(
				parameters.getProperty("failOnError", "true")).booleanValue());

		setTriggeringRule(parameters.getProperty("triggeringRule", "true"));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// If not active, return immediately.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		try {
			// Get the request file format.
			MessageFormat.Enum requestFormat = config.getGeneral()
					.getRequestMessage().getFormat();

			// Get the response file format.
			MessageFormat.Enum responseFormat = config.getGeneral()
					.getResponseMessage().getFormat();

			// Get the results file and the target filename.
			Message responseMessage = pipeline.getResponseMessage();
			File file = new File(responseMessage.getCurrentFile());
			String filename = this.filename;
			if (sendAttachment & StringUtils.isBlank(filename)) {
				if (responseMessage.getOriginalFile() != null) {
					filename = new File(responseMessage.getOriginalFile())
							.getName();
				} else {
					throw new StepException(
							"A value must be provided for the filename parameter.");
				}
			}

			// Get the CSV header present in the case of a CSV export.
			if (requestFormat.equals(MessageFormat.SQ_XML)
					&& responseFormat.equals(MessageFormat.CSV)) {
				File requestFile = new File(pipeline.getRequestMessage()
						.getOriginalFile());
				Query query = QueryRuntimeHelper.loadQueryInStep(requestFile,
						productCode);
				setCsvHeaderPresent(query.getCsvheader());
			}

			// Build Velocity context.
			VelocityContext velocityContext = buildVelocityContext(
					requestFormat, responseFormat, file, pipeline, this.config);

			// It the triggering rule evaluates to false, return immediately.
			if (!evaluateTriggeringRule(velocityContext)) {
				log.info("No email sent because triggering rule returned 'false'.");
				return;
			}

			// Execute the step.
			try {
				String message = prepareMessage(velocityContext);
				execute(message, file, filename);

			} catch (MessagingException e) {
				if (failOnError) {
					throw new StepException(e);
				} else {
					// Log an error, but don't propagate.
					log.error("Error while sending file by email.", e);
				}
			}

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Execute the step.
	 * 
	 * @param message
	 *            The email message.
	 * @param file
	 *            The import or export results file.
	 * @param filename
	 *            The target filename for the attachment.
	 * @throws MessagingException
	 */
	public void execute(String message, File file, String filename)
			throws MessagingException {

		// Get mail session.
		Properties props = new Properties();
		props.put("mail.smtp.host", SmtpHost);
		props.put("mail.smtp.port", Integer.toString(port));
		Session session = Session.getInstance(props);

		// Define message properties.
		MimeMessage mimeMessage = new MimeMessage(session);
		mimeMessage.setFrom(new InternetAddress(from));
		mimeMessage.setRecipients(MimeMessage.RecipientType.TO,
				InternetAddress.parse(to, false));
		mimeMessage.setRecipients(MimeMessage.RecipientType.CC,
				InternetAddress.parse(cc, false));
		mimeMessage.setSubject(subject);

		// Create the message multipart.
		Multipart multipart = new MimeMultipart();
		BodyPart messageBodyPart = null;

		// Prepare part one: the message body.
		messageBodyPart = new MimeBodyPart();
		messageBodyPart.setContent(message, mimeType);
		multipart.addBodyPart(messageBodyPart);

		// Part two is the attachment.
		if (sendAttachment) {
			messageBodyPart = new MimeBodyPart();
			DataSource source = new FileDataSource(file);
			messageBodyPart.setDataHandler(new DataHandler(source));
			try {
				if (IOHelper.isFileCompressed(file)
						&& !filename.endsWith(".zip")) {
					filename = filename + ".zip";
				}
			} catch (IOException e) {
				// Do Nothing.
			}
			messageBodyPart.setFileName(filename);
			multipart.addBodyPart(messageBodyPart);
		}

		// Put parts in message.
		mimeMessage.setContent(multipart);

		// Send message.
		Transport.send(mimeMessage);
	}

	/**
	 * Evaluate the triggering rule.
	 * 
	 * @param context
	 *            The Velocity context.
	 * @return <code>true</code> or <code>false</code>
	 */
	protected boolean evaluateTriggeringRule(VelocityContext context) {
		String triggeringRule = this.triggeringRule;
		try {
			if (StringUtils.isNotBlank(triggeringRule)) {

				// Adjust the rule replacing '\' with '/' within #parse
				// directive if used.
				Pattern pattern = Pattern
						.compile("^(.*#parse\\s*\\()(.*)(\\).*)$");
				Matcher matcher = pattern.matcher(triggeringRule);
				if (matcher.matches()) {
					triggeringRule = matcher.group(1)
							+ matcher.group(2).replaceAll("\\\\", "/")
							+ matcher.group(3);
				}

				if (log.isDebugEnabled()) {
					log.debug("triggeringRule: '" + triggeringRule + "'");
				}

				// Evaluate the rule.
				StringWriter writer = new StringWriter();
				velocity.evaluate(context, writer, "", new StringReader(
						triggeringRule));

				if (log.isDebugEnabled()) {
					log.debug("triggeringRule evaluates to '"
							+ writer.toString().trim() + "'");
				}

				return new Boolean(writer.toString().trim()).booleanValue();

			} else {
				return true;
			}
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Prepare the email message body from the message template.
	 * 
	 * @param context
	 *            The Velocity context.
	 * @return The email message.
	 */
	protected String prepareMessage(VelocityContext context) {
		try {
			if (StringUtils.isNotBlank(messageTemplate)) {
				StringWriter writer = new StringWriter();
				velocity.evaluate(context, writer, "", new StringReader(
						messageTemplate));
				return writer.toString();

			} else if (StringUtils.isNotBlank(messageTemplateFile)) {
				StringWriter writer = new StringWriter();
				File template = new File(messageTemplateFile);
				velocity.mergeTemplate(template.getAbsolutePath(),
						getEncoding(), context, writer);
				return writer.toString();

			} else {
				return "";
			}
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Build the Velocity context to render templates and triggering rule.
	 * 
	 * @param requestFormat
	 *            The request file format.
	 * @param responseFormat
	 *            The response file format.
	 * @param file
	 *            The import or export results file.
	 * @param pipeline
	 *            The workflow pipeline.
	 * @return The Velocity context.
	 */
	protected VelocityContext buildVelocityContext(
			MessageFormat.Enum requestFormat,
			MessageFormat.Enum responseFormat, File file, Pipeline pipeline,
			GlobalConfig config) {

		// Create the context for template rendering.
		VelocityContext context = new VelocityContext();
		context.put("date", new DateTool());
		context.put("workflow", new WorkflowTool(pipeline, config));
		context.put("math", new MathTool());
		context.put("string", new StringTool());
		context.put("logic", new LogicTool());
		if (requestFormat.equals(MessageFormat.SQ_XML)) {
			// We are processing an export.
			if (responseFormat.equals(MessageFormat.CSV)) {
				// The response is in CSV format.
				context.put("file", new CsvExportFileTool(file,
						isCsvHeaderPresent(), getCsvDelimiter(),
						getCsvQuoteCharacter(), getEncoding()));
			} else {
				// The response is in XML format.
				context.put("file", new XmlExportFileTool(file));
			}
		} else {
			// We are processing an import in CSV.
			if (responseFormat.equals(MessageFormat.CSV)) {
				context.put("file", new CsvImportFileTool(file,
						isCsvHeaderPresent(), getCsvDelimiter(),
						getCsvQuoteCharacter(), getEncoding()));
			}
		}
		return context;
	}

}
