package com.taleo.integration.client.customstep.mail;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.exception.VelocityException;
import org.apache.velocity.runtime.RuntimeConstants;
import org.apache.velocity.runtime.log.NullLogChute;
import org.apache.velocity.runtime.resource.loader.FileResourceLoader;
import org.apache.velocity.tools.generic.DateTool;
import org.apache.velocity.tools.generic.MathTool;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.SAXReader;
import org.dom4j.io.XMLWriter;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.integration.client.customstep.velocity.StringTool;
import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Post-processing step to send individual emails from the results of an export.
 * <p>
 * It executes over files containing individual emails. Those files must have
 * been created previously using
 * {@link com.taleo.integration.client.customstep.xml.SplitFilePostStep} and
 * match the structure presented below.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>SMTPHost</b>: The SMTP host.</li>
 * <br>
 * <br>
 * <li><b>port</b>: The SMTP port (default: 25).</li>
 * <br>
 * <br>
 * <li><b>filePattern</b>: The pattern for selecting files containing emails.
 * Can be specified as an absolute or relative path and may use wildcards as
 * defined in {@link org.apache.commons.io.filefilter.WildcardFileFilter} in the
 * file names only. Pattern is case sensitive. If the pattern includes multiple
 * files, they will be processed in ascending name order.</li>
 * <br>
 * <br>
 * <li><b>archiveFolder</b>: Optional - Folder to move file after successfully
 * sending an email. If absent, files are deleted after processing.</li>
 * <br>
 * <br>
 * <li><b>errorFolder</b>: Optional - Folder to move file when an error occurred
 * while sending an email. If absent, files are deleted on error.</li>
 * <br>
 * <br>
 * <li><b>onHold</b>: Optional - Indicate if the email is only prepared,
 * resolving the subject and content, and not actually sent neither moved to
 * archive folder. However, it will still be moved to error folder if an error
 * is detected while preparing the email. (default: false)</li>
 * </ul>
 * <p>
 * 
 * Email files are structured as follows.
 * 
 * <pre>
 * &lt;email&gt;
 *   &lt;from&gt;&lt;/from&gt;
 *   &lt;to&gt;&lt;/to&gt;
 *   &lt;cc&gt;&lt;/cc&gt;
 *   &lt;bcc&gt;&lt;/bcc&gt;
 *   &lt;subject&gt;...&lt;/subject&gt;
 *   &lt;content mimeType=""&gt;...&lt;/content&gt;
 *   &lt;tokens&gt;
 *     &lt;token name="xyz" value=""/&gt;
 *     ...
 *   &lt;/tokens&gt;
 * &lt;/email&gt;
 * </pre>
 * 
 * The XML tags are:<br>
 * <ul>
 * <li><b>from</b>: Mandatory - The sender of the email.</li>
 * <br>
 * <br>
 * <li><b>to</b>: Mandatory - The receiver list separated by comma.</li>
 * <br>
 * <br>
 * <li><b>cc</b>: Optional - The cc receiver list separated by comma.</li>
 * <br>
 * <br>
 * <li><b>bcc</b>: Optional - The bcc receiver list separated by comma.</li>
 * <br>
 * <br>
 * <li><b>subject</b>: Mandatory - The subject of the email.</li>
 * <br>
 * <br>
 * <li><b>content</b>: Mandatory - The content of the email.
 * <ul>
 * <li>The <b>mimeType</b> attribute is either <code>text/plain</code> or
 * <code>text/html</code>. Other mime types have not been tested. (default:
 * <code>text/plain</code>)</li>
 * <li>The <b>xml:space="preserve"</b> attribute can be used to force
 * maintaining spaces and line feeds during processing.</li>
 * </ul>
 * </li>
 * <li><b>tokens</b>: The whole document is interpreted as a Velocity template
 * and may include tokens defined in <b>token</b> elements, using the following
 * syntax: <code>$tokens.get("xyz")</code>.</li>
 * </ul>
 * 
 * After processing, the status element is added with the following structure:
 * 
 * <pre>
 * &lt;email&gt;
 *   ...
 *   &lt;status&gt;
 *     &lt;success date=""/&gt;
 *     <i>or</i>
 *     &lt;error date=""&gt;...&lt;/error&gt;
 *   &lt;/status&gt;
 * &lt;/email&gt;
 * </pre>
 * 
 * Those XML tags are:<br>
 * <ul>
 * <li><b>success</b>: Indicates successful processing.</li>
 * <br>
 * <br>
 * <li><b>error</b>: Indicates an error with corresponding details.</li>
 * </ul>
 * <p>
 * The following variables are also made available to build the email template:
 * <ul>
 * <li><b>$date</b>: A Date tool instance (
 * {@link org.apache.velocity.tools.generic.DateTool}) to access various date
 * functions.</li>
 * <br>
 * <br>
 * <li><b>$math</b>: A Math tool instance (
 * {@link org.apache.velocity.tools.generic.MathTool}) to access various
 * mathematical functions.</li>
 * <br>
 * <br>
 * <li><b>$string</b>: A String tool instance (
 * {@link com.taleo.integration.client.customstep.velocity.StringTool}) to
 * access various string functions.</li>
 * </ul>
 * <p>
 * Example uses of these variables in the templates:
 * <ul>
 * <li>$date.get("yyyy-MM-dd")</li>
 * <li>$date.getYear()</li>
 * <li>$date.getMonth()</li>
 * <li>$date.getDay()</li>
 * <li>$math.add($date.getYear(), 1)</li>
 * <li>$string.rightPad("123", 10)</li>
 * <li>$string.rightPad("123", 10, "_")</li>
 * <li>$string.leftPad("123", 10)</li>
 * <li>$string.leftPad("123", 10, "0")</li>
 * <li>$string.substring("123456", 2, 3)</li>
 * <li>$string.substringAfter("123456", "3")</li>
 * </ul>
 * 
 * If an error prevents processing the complete list of files (e.g. mail server
 * down), those that have not been processed remain in place for processing on
 * the next run. If an error is specific to a given file, this will not
 * interrupt the process, the file will simply be moved to the error folder.
 * Note that this step does not handle bouncing emails. Therefore, it is
 * important to set a meaningful sender address (from) in order to receive that
 * kind of error.
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://velocity.apache.org/index.html">Apache Velocity
 * library</a> (velocity-dep-1.5.jar)</li>
 * <li><a href="http://commons.apache.org/net/">Java mail library</a> (mail.jar
 * - provided with TCC)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/SendEmailsPostStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class SendEmailsPostStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(SendEmailsPostStep.class);

	private final static DateFormat DATE_FORMAT = new SimpleDateFormat(
			"yyyy-MM-dd HH:mm:ss");

	private final static String DEFAULT_MIME_TYPE = "text/plain";

	/**
	 * The Velocity engine.
	 */
	private VelocityEngine velocity = null;

	/**
	 * Constructor without arguments.
	 */
	public SendEmailsPostStep() {
		super();
		// Register valid parameters.
		registerParameterName("SMTPHost");
		registerParameterName("SMTPPort");
		registerParameterName("filePattern");
		registerParameterName("archiveFolder");
		registerParameterName("errorFolder");
		registerParameterName("onHold");

		// Create the velocity engine.
		this.velocity = new VelocityEngine();
		Properties props = new Properties();
		// Disable default logging.
		props.put(RuntimeConstants.RUNTIME_LOG_LOGSYSTEM_CLASS,
				NullLogChute.class.getName());
		props.put("resource.loader", "file");
		props.put("file.resource.loader.class",
				FileResourceLoader.class.getName());
		// Empty loader path allows specifying templates with an
		// absolute path.
		props.put("file.resource.loader.path", "");
		props.put("file.resource.loader.cache", "false");
		try {
			velocity.init(props);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * The default port.
	 */
	private final int DEFAULT_PORT = 25;

	/**
	 * The SMTP host.
	 */
	private String smtpHost;

	/**
	 * The SMTP port.
	 */
	private int smtpPort = DEFAULT_PORT;

	/**
	 * The file pattern.
	 */
	private String filePattern;

	/**
	 * The archive folder.
	 */
	private String archiveFolder;

	/**
	 * The error folder.
	 */
	private String errorFolder;

	/**
	 * The onHold flag.
	 */
	private boolean onHold = false;

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "SendEmailsPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to send individual emails from the results of an export.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "SendEmailsPostStep";
	}

	/**
	 * @return
	 */
	public String getSmtpHost() {
		return smtpHost;
	}

	/**
	 * @param smtpHost
	 */
	public void setSmtpHost(String smtpHost) {
		this.smtpHost = smtpHost;
	}

	/**
	 * @return
	 */
	public int getSmtpPort() {
		return smtpPort;
	}

	/**
	 * @param port
	 */
	public void setSmtpPort(int smtpPort) {
		this.smtpPort = smtpPort;
	}

	/**
	 * Get the file pattern.
	 * 
	 * @return
	 */
	public String getFilePattern() {
		return filePattern;
	}

	/**
	 * Set the file pattern.
	 * 
	 * @param filePattern
	 */
	public void setFilePattern(String filePattern) {
		this.filePattern = filePattern;
	}

	/**
	 * Get the archive folder.
	 * 
	 * @return
	 */
	public String getArchiveFolder() {
		return archiveFolder;
	}

	/**
	 * Set the archive folder.
	 * 
	 * @param archiveFolder
	 */
	public void setArchiveFolder(String archiveFolder) {
		this.archiveFolder = archiveFolder;
	}

	/**
	 * Get the error folder.
	 * 
	 * @return
	 */
	public String getErrorFolder() {
		return archiveFolder;
	}

	/**
	 * Set the error folder.
	 * 
	 * @param errorFolder
	 */
	public void setErrorFolder(String errorFolder) {
		this.errorFolder = errorFolder;
	}

	/**
	 * Get the onHold flag.
	 * 
	 * @return
	 */
	public boolean isOnHold() {
		return onHold;
	}

	/**
	 * Set the onHold flag.
	 * 
	 * @param onHold
	 */
	public void setOnHold(boolean onHold) {
		this.onHold = onHold;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// First, get the basic supported pipeline for a post-processing step
		// (this handles the constraints on the last step type)
		SupportedPipeline supportedPipeline = getPostProcessSupportPipeline();

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);
		
		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		setSmtpHost(parameters.getProperty("SMTPHost"));
		if (StringUtils.isBlank(smtpHost)) {
			throw new IllegalArgumentException(
					"The 'SmtpHost' parameter is mandatory.");
		}

		try {
			setSmtpPort(Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("SMTPPort"),
					Integer.toString(smtpPort))));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The 'SMTPPort' parameter must have a numeric value.");
		}

		filePattern = parameters.getProperty("filePattern");

		archiveFolder = parameters.getProperty("archiveFolder");

		errorFolder = parameters.getProperty("errorFolder");

		onHold = Boolean.valueOf(
				StringUtils.defaultIfEmpty(parameters.getProperty("onHold"),
						"false")).booleanValue();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// If not active, return immediately.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}
		try {
			// Execute.
			execute();
		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Execute the step over files matching the file pattern and containing each
	 * one a single email record.
	 * 
	 * @throws Exception
	 */
	protected void execute() throws Exception {
		// Get email session.
		Properties props = new Properties();
		props.put("mail.smtp.host", smtpHost);
		props.put("mail.smtp.port", Integer.toString(smtpPort));
		Session session = Session.getInstance(props);

		// Get the files to process.
		File[] files = new File[0];
		files = FileUtils.getMatchingFiles(new File(filePattern));
		FileUtils.sortFilesByNameAscending(files, true);
		log.info(files.length + " file(s) match the pattern.");

		// Process each file.
		for (int i = 0; i < files.length; i++) {
			execute(session, files[i]);
		}
	}

	/**
	 * Execute the step over a file containing a single email record.
	 * 
	 * @param session
	 *            The email session.
	 * @param file
	 *            The file to process.
	 * @throws Exception
	 */
	protected void execute(Session session, File file) throws Exception {
		log.info("Processing file " + file.getPath());

		Document document = null;
		String originalXML = null;
		String error = null;
		String targetFolder = null;
		if (!onHold) {
			targetFolder = archiveFolder;
		}

		try {
			// Load the XML document.
			SAXReader reader = new SAXReader();
			document = reader.read(new InputStreamReader(new FileInputStream(
					file), getEncoding()));

			// Build the Velocity context.
			VelocityContext context = buildVelocityContext(document);

			// Take a snapshot of the original XML.
			originalXML = document.asXML();

			// Render the document.
			document = render(document, context);

			if (log.isDebugEnabled()) {
				log.debug("Processing: " + document.asXML());
			}

			// Send the email.
			sendEmail(session, document);

		} catch (IllegalArgumentException e) {
			targetFolder = errorFolder;
			error = e.getMessage();

		} catch (DocumentException e) {
			targetFolder = errorFolder;
			error = e.getMessage();

		} catch (MessagingException e) {
			targetFolder = errorFolder;
			error = e.getMessage();
		}

		// Set the status in the XML element.
		if (document != null) {
			Element status = document.getRootElement().element("status");
			if (status == null) {
				status = document.getRootElement().addElement("status");
			}
			if (error == null) {
				if (!onHold) {
					status.addElement("success").addAttribute("date",
							DATE_FORMAT.format(new Date()));
				}
			} else {
				status.addElement("error")
						.addAttribute("date", DATE_FORMAT.format(new Date()))
						.addText(error);
				log.warn("Error while processing file " + file.getPath() + ": "
						+ error);
			}

			// Write file if XML changed.
			if (!document.asXML().equals(originalXML)) {
				OutputFormat format = OutputFormat.createPrettyPrint();
				format.setNewLineAfterDeclaration(false);
				XMLWriter writer = new XMLWriter(new OutputStreamWriter(
						new FileOutputStream(file), getEncoding()), format);
				writer.write(document);
				writer.close();
			}
		} else {
			log.error("Error while processing file " + file.getPath() + ": "
					+ error);
		}

		// Move file to the target folder.
		if (StringUtils.isNotBlank(targetFolder)) {
			File dest = new File(targetFolder, file.getName());
			if (!dest.equals(file)) {
				log.info("Moving file " + file.getName() + " to "
						+ dest.getPath());
				dest.delete();
				org.apache.commons.io.FileUtils.moveFile(file, dest);
			}
		} else {
			if (!onHold) {
				log.info("Deleting file " + file);
				file.delete();
			}
		}
	}

	/**
	 * Send an email with the content of an XML document.
	 * 
	 * @param session
	 *            The email session.
	 * @param document
	 *            The document to process.
	 * @throws IllegalArgumentException
	 * @throws MessagingException
	 * @throws IOException
	 * @throws DocumentException
	 * @throws VelocityException
	 */
	protected void sendEmail(Session session, Document document)
			throws IllegalArgumentException, MessagingException, IOException,
			VelocityException, DocumentException {

		// Get the root element.
		Element root = document.getRootElement();

		// Get elements.
		Element from = root.element("from");
		Element to = root.element("to");
		Element cc = root.element("cc");
		Element bcc = root.element("bcc");
		Element subject = root.element("subject");
		Element content = root.element("content");

		// Validate elements.
		if (from == null || !from.hasContent()) {
			throw new IllegalArgumentException(
					"The 'from' element is mandatory.");
		}
		if (to == null || !to.hasContent()) {
			throw new IllegalArgumentException("The 'to' element is mandatory.");
		}
		if (subject == null || !subject.hasContent()) {
			throw new IllegalArgumentException(
					"The 'subject' element is mandatory.");
		}
		if (content == null || !content.hasContent()) {
			throw new IllegalArgumentException(
					"The 'content' element is mandatory.");
		}

		// Get attributes.
		String mimeType = content.attributeValue("mimeType");
		if (StringUtils.isEmpty(mimeType)) {
			mimeType = DEFAULT_MIME_TYPE;
		}

		if (onHold) {
			log.info("No message was sent because step is on hold.");

		} else {
			// Set message properties.
			MimeMessage message = new MimeMessage(session);
			try {
				message.setFrom(new InternetAddress(from.getText()));
				message.setRecipients(MimeMessage.RecipientType.TO,
						InternetAddress.parse(to.getText(), false));
				if (cc != null) {
					message.setRecipients(MimeMessage.RecipientType.CC,
							InternetAddress.parse(cc.getText(), false));
				}
				if (bcc != null) {
					message.setRecipients(MimeMessage.RecipientType.BCC,
							InternetAddress.parse(bcc.getText(), false));
				}
				message.setSubject(subject.getText(), getEncoding());
				message.setContent(content.getText(), mimeType);
			} catch (AddressException e) {
				throw new IllegalArgumentException(e);
			}

			// Send message.
			Transport.send(message);
		}
	}

	/**
	 * Build the Velocity context with the tokens element defined within the
	 * given file.
	 * 
	 * @param file
	 *            The file containing the document.
	 * @return
	 * @throws DocumentException
	 * @throws FileNotFoundException
	 */
	protected VelocityContext buildVelocityContext(Document document)
			throws DocumentException, FileNotFoundException {

		// Context for Velocity template rendering.
		VelocityContext context = new VelocityContext();
		context.put("date", new DateTool());
		context.put("math", new MathTool());
		context.put("string", new StringTool());

		// Get tokens.
		Element tokens = document.getRootElement().element("tokens");
		if (tokens != null) {
			Iterator tokensIterator = tokens.elementIterator("token");

			// Build a map of tokens and set it in the context.
			Map map = new HashMap();
			for (Iterator i = tokensIterator; i.hasNext();) {
				Element token = (Element) i.next();
				map.put(token.attributeValue("name"),
						token.attributeValue("value"));
			}
			context.put("tokens", map);
		}
		return context;
	}

	/**
	 * Render the document in the given Velocity context.
	 * 
	 * @param document
	 * @param context
	 * @return
	 * @throws IOException
	 * @throws DocumentException
	 */
	protected Document render(Document document, VelocityContext context)
			throws IOException, DocumentException {

		// Render document in Velocity context.
		StringWriter writer = new StringWriter();
		velocity.evaluate(context, writer, "",
				new StringReader(document.asXML()));

		// Reload the rendered XML document.
		return DocumentHelper.parseText(writer.toString());
	}
}
