package com.taleo.integration.client.customstep.paging;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.ParseException;
import java.util.Date;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.util.DateUtils;

/**
 * Manage the paging state.
 * <p>
 * Reads and writes the paging state (page index and record count) on disk. Also
 * stores paging parameters (paging size, decrease factor, current run date),
 * because those cannot change until the whole paging cycle is complete.
 * <p>
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public class PagingManager {

	private static Logger log = Logger.getLogger(PagingManager.class);

	/** The Paging file extension. */
	final static String PAGING_EXTENSION = "pgn";

	/** Key for the PagingManager in the pipeline. */
	public final static String PIPELINE_KEY = "PagingManager";

	/**
	 * The file to store and retrieve paging state.
	 */
	private File pagingFile;

	/**
	 * The paging size.
	 */
	private int pagingSize = 0;

	/**
	 * The decrease factor.
	 */
	private int decreaseFactor = 0;

	/**
	 * The last run date.
	 */
	private Date lastRunDate;

	/**
	 * The current run date.
	 */
	private Date currentRunDate;

	/**
	 * The paging size adjusted for the decrease factor.
	 */
	private int adjustedPagingSize = 0;

	/**
	 * The current page index (1 based). Equals 0 if paging was not started.
	 */
	private int currentPageIndex = 0;

	/**
	 * The current recordCount. Equals -1 if not set for the current page.
	 */
	private int currentRecordCount = -1;

	/**
	 * Constructor with the file to read/write the paging state, the paging
	 * size, the decrease factor, the last run date and the current run date.
	 * <p>
	 * If paging size, decrease factor, last run date or current run date have
	 * changed since the last cycle, while paging was not complete, throw an
	 * exception because paging will be inconsistent.
	 * 
	 * @param file
	 *            The file to store and retrieve the paging state.
	 * @param pagingSize
	 *            The paging size.
	 * @param decreaseFactor
	 *            The decrease factor.
	 * @param lastRunDate
	 *            The last run date, may be null.
	 * @param currentRunDate
	 *            The current run date, may be null.
	 * 
	 * @throws IOException
	 */
	public PagingManager(File file, int pagingSize, int decreaseFactor,
			Date lastRunDate, Date currentRunDate) throws IOException {
		if (pagingSize < 0) {
			throw new IllegalArgumentException(
					"The 'pagingSize' parameter must be >= 0");
		}
		if (decreaseFactor < 0) {
			throw new IllegalArgumentException(
					"The 'decreaseFactor' parameter must be >= 0");
		}
		if (decreaseFactor >= 100) {
			throw new IllegalArgumentException(
					"The 'decreaseFactor' parameter must be < 100");
		}
		this.pagingFile = file;
		this.pagingSize = pagingSize;
		this.decreaseFactor = decreaseFactor;
		this.lastRunDate = lastRunDate;
		this.currentRunDate = currentRunDate;

		// Load the file if it exists.
		if (this.pagingFile.exists()) {

			BufferedReader reader = null;
			try {
				reader = new BufferedReader(new FileReader(this.pagingFile));
				// Skip header.
				reader.readLine();

				// Read paging state.
				String state;
				state = reader.readLine();
				if (state != null && state.trim().length() > 0) {
					try {
						String[] arr = StringUtils.splitPreserveAllTokens(
								state, '|');
						this.currentPageIndex = Integer.parseInt(arr[0]);
						this.currentRecordCount = Integer.parseInt(arr[1]);

						int _pagingSize = Integer.parseInt(arr[2]);
						int _decreaseFactor = Integer.parseInt(arr[3]);
						Date _lastRunDate = null;
						if (!StringUtils.isEmpty(arr[4])) {
							_lastRunDate = DateUtils
									.parseDateFromISO8601(arr[4]);
						}
						Date _currentRunDate = null;
						if (!StringUtils.isEmpty(arr[5])) {
							_currentRunDate = DateUtils
									.parseDateFromISO8601(arr[5]);
						}

						// Adjust currentPageIndex if pagingSize has changed.
						boolean pagingSizeChanged = false;
						if (this.pagingSize != _pagingSize) {
							pagingSizeChanged = true;
							int totalCount = currentPageIndex * _pagingSize;
							if (this.pagingSize < _pagingSize) {
								// Allow the change only if compatible with
								// total count.
								if (totalCount % this.pagingSize != 0) {
									throw new IllegalStateException(
											"Paging size has changed in an inconsistent way from "
													+ _pagingSize
													+ " to "
													+ this.pagingSize
													+ ". If reduced, it must divide the total count ("
													+ totalCount
													+ ") exactly without a remainder, for example try "
													+ _pagingSize
													/ smallestIntDivisor(_pagingSize)
													+ ".");
								}
							} else {
								if (totalCount % this.pagingSize != 0) {
									log.info("Paging size change to "
											+ this.pagingSize
											+ " has been postponed until the total count is divisible by it.");
									// Set it back to previous value.
									this.pagingSize = _pagingSize;
									pagingSizeChanged = false;
								}
							}
							if (pagingSizeChanged) {
								// Adjust the currentPageIndex.
								this.currentPageIndex = totalCount
										/ this.pagingSize;
								log.info("Current page index has been recalculated to "
										+ this.currentPageIndex
										+ " following paging size change ("
										+ this.pagingSize + ").");
								// Adjust the currentRecordCount. This may force
								// an extra loop, but is safer.
								this.currentRecordCount = this.pagingSize;
							}
						}

						// Check if paging is still consistent.
						boolean consistent = true;
						consistent = consistent
								&& (_decreaseFactor == this.decreaseFactor);
						consistent = consistent
								&& ((_lastRunDate != null && _lastRunDate
										.equals(this.lastRunDate)) || (_lastRunDate == null && this.lastRunDate == null));
						consistent = consistent
								&& ((_currentRunDate != null && _currentRunDate
										.equals(this.currentRunDate)) || (_currentRunDate == null && this.currentRunDate == null));
						if (!consistent) {
							log.info("Paging state from file: " + state);
							log.info("Paging state from run : "
									+ getPagingState());

							throw new IllegalStateException(
									"Paging parameters have changed and are now inconsistent. Please delete the paging file "
											+ this.pagingFile.getPath()
											+ " and restart.");
						}

					} catch (ArrayIndexOutOfBoundsException e) {
						throw new RuntimeException(
								"Invalid paging file content: '" + state + "'.",
								e);
					} catch (ParseException e) {
						throw new RuntimeException(
								"Invalid paging file content: '" + state + "'.",
								e);
					}
				}
			} finally {
				if (reader != null) {
					reader.close();
				}
			}
		}

		// Reset the adjusted paging size.
		resetAdjustedPagingSize();
	}

	/**
	 * Reset the adjusted paging size, considering the decrease factor.
	 */
	private void resetAdjustedPagingSize() {
		adjustedPagingSize = pagingSize;
		if (adjustedPagingSize > 0) {
			for (int i = 1; i < currentPageIndex; i++) {
				adjustedPagingSize = (int) Math.max(
						1,
						Math.floor(adjustedPagingSize
								* (1d - decreaseFactor / 100d)));
			}
		}
	}

	/**
	 * Add a new page.
	 */
	public void addNewPage() {
		this.currentPageIndex += 1;
		this.currentRecordCount = -1;
		resetAdjustedPagingSize();
	}

	/**
	 * Get the paging size.
	 * 
	 * @return
	 */
	public int getPagingSize() {
		return pagingSize;
	}

	/**
	 * Get the paging size adjusted for the decrease factor.
	 * 
	 * @return
	 */
	public int getAdjustedPagingSize() {
		return adjustedPagingSize;
	}

	/**
	 * Get the current page index (1 based).
	 * <p>
	 * Equals 0 if there is no current page.
	 * 
	 * @return
	 */
	public int getCurrentPageIndex() {
		return currentPageIndex;
	}

	/**
	 * Get the current record count (corresponding to the current page index).
	 * <p>
	 * Equals -1 if not set for the current page.
	 * 
	 * @return
	 */
	public int getCurrentRecordCount() {
		return currentRecordCount;
	}

	/**
	 * Set the current record count.
	 * 
	 * @param currentRecordCount
	 */
	public void setCurrentRecordCount(int currentRecordCount) {
		if (currentPageIndex == 0) {
			throw new IllegalStateException(
					"Current record count cannot be set before adding a page.");
		}
		if (currentRecordCount < 0) {
			throw new IllegalArgumentException(
					"The 'currentRecordCount' parameter must be >= 0");
		}
		this.currentRecordCount = currentRecordCount;
	}

	/**
	 * Indicate if paging is new.
	 * 
	 * @return
	 */
	public boolean isNew() {
		return (currentPageIndex == 0 || (currentPageIndex == 1 && currentRecordCount == -1));
	}

	/**
	 * Indicate if paging is complete.
	 * 
	 * @return
	 */
	public boolean isComplete() {
		return (!isNew() && (currentRecordCount == 0 || adjustedPagingSize == 0 || currentRecordCount < adjustedPagingSize));
	}

	/**
	 * Save the current state in the paging file.
	 * 
	 * @throws IOException
	 */
	public void save() throws IOException {
		BufferedWriter writer = null;
		try {

			writer = new BufferedWriter(new FileWriter(this.pagingFile));
			// Write header.
			writer.write("pageIndex|recordCount|pagingSize|decreaseFactor|lastRunDate|currentRunDate");
			writer.newLine();

			// Write data.
			writer.write(getPagingState());
			writer.newLine();
			writer.flush();

			if (log.isDebugEnabled()) {
				log.debug("Loading paging state: " + getPagingState());
			}
		} finally {
			if (writer != null) {
				writer.close();
			}
		}
	}

	/**
	 * Get the paging state in a single line.
	 * 
	 * @return
	 */
	private String getPagingState() {
		return currentPageIndex
				+ "|"
				+ currentRecordCount
				+ "|"
				+ pagingSize
				+ "|"
				+ decreaseFactor
				+ "|"
				+ (lastRunDate == null ? "" : DateUtils
						.formatDateAsISO8601(lastRunDate))
				+ "|"
				+ (currentRunDate == null ? "" : DateUtils
						.formatDateAsISO8601(currentRunDate));
	}

	/**
	 * Delete the paging file.
	 * 
	 * @throws IOException
	 */
	public void delete() throws IOException {
		if (pagingFile.exists()) {
			FileUtils.forceDelete(pagingFile);
		}
	}

	/**
	 * Get the smallest integer value that divides the given number without a
	 * remainder.
	 * 
	 * @param number
	 * @return
	 */
	private int smallestIntDivisor(int number) {
		if (number == 0 || number == 1) {
			return number;
		}
		int divisor = 2;
		while (true) {
			if (number % divisor == 0)
				return divisor;
			divisor += 1;
		}
	}

}
