package com.taleo.integration.client.customstep.paging;

import java.io.IOException;

import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.count.CountManager;
import com.taleo.integration.client.customstep.count.ExportCountPostStep;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.workflow.WorkflowHelper;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.WorkflowError;
import com.taleo.ws.integration.client.WorkflowErrorCode;

/**
 * Post-processing step to manage the end of the paging process.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep}.
 * <p>
 * This post-processing step works in pair with {@link PagingPreStep} and
 * {@link ExportCountPostStep}. Its only purpose is to save the paging state at
 * the end of the workflow. It gets the record count from the
 * {@link CountManager} that must have has been set in the pipeline by
 * {@link ExportCountPostStep}.
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file. <img src="doc-files/PagingPostStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class PagingPostStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(PagingPostStep.class);

	/**
	 * Constructor without arguments.
	 */
	public PagingPostStep() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "PagingPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step for paging";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "PagingPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		// Get the PagingManager that was put in the pipeline by the
		// PagingPreStep.
		PagingManager pagingManager = (PagingManager) pipeline
				.documentProperties().get(PagingManager.PIPELINE_KEY);

		// If the PagingManager is null, quit immediately.
		if (pagingManager == null) {
			log.info("The step is skipped because the PagingPreStep was "
					+ "either inactive or skipped.");
			return;
		}

		// Get the CountManager from the pipeline.
		CountManager countManager = (CountManager) pipeline
				.documentProperties().get(CountManager.PIPELINE_KEY);
		if (countManager == null) {
			throw new StepException(
					"No CountManager was found in the pipeline.");
		}

		try {
			// Set the current record count.
			pagingManager.setCurrentRecordCount(countManager
					.getCurrentRecordCount());

			log.info("Iteration " + pagingManager.getCurrentPageIndex()
					+ " contains " + pagingManager.getCurrentRecordCount()
					+ " records.");

			// Save the paging state.
			pagingManager.save();

		} catch (IOException e) {
			WorkflowError error = WorkflowHelper.createWorkflowError(
					WorkflowErrorCode.IO, e);
			throw new StepException(error);

		} catch (Exception e) {
			throw new StepException(e);
		}
	}
}
