package com.taleo.integration.client.customstep.paging;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.math.BigInteger;
import java.util.Date;

import javax.xml.namespace.QName;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.apache.xmlbeans.XmlOptions;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.lrd.RunDateManager;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.integration.client.util.QueryRuntimeHelper;
import com.taleo.integration.client.workflow.WorkflowHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.WorkflowError;
import com.taleo.ws.integration.client.WorkflowErrorCode;
import com.taleo.ws.integration.query.QueryDocument.Query;

/**
 * Pre-processing step to manage the beginning of the paging process for large
 * extracts.
 * <p>
 * This pre-processing step works in pair with {@link PagingPostStep}. It sets
 * the current page index and the paging size into the query. The PagingPostStep
 * manages the end of the loop.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>pagingSize</b>: Optional - Number of records per page. If absent, the
 * paging size set in the query advanced options will apply. If this one is also
 * absent, paging size is set to 0 which means all records.</li>
 * <br>
 * <br>
 * <li><b>decreaseFactor</b>: Optional - A number representing a safe maximum
 * percentage of deletion that we can expect in this query while the paging
 * occurs (Typical values are 1 or 2, meaning 1 or 2%). See below for a detailed
 * explanation of this strategy. (Default: 0)</li>
 * <br>
 * <br>
 * <li><b>pagingFilename</b>: Optional - The name of the file to read and store
 * the paging state info. The '.pgn' extension will be added if not present in
 * the filename. If absent, the configuration identifier will be used. Can be
 * supplied as a filename only in which case it will be placed in the temporary
 * folder. Otherwise, it can be set as an absolute or relative path.</li>
 * </ul>
 * <p>
 * Paging continues until the record count from the last run is less than the
 * paging size. When that happens, the process exits abruptly with code 0. All
 * subsequent steps are then ignored. The exit code can be set differently using
 * the system property named
 * <code>com.taleo.client.customstep.PagingPreStep.CompleteExitCode</code>.
 * <p>
 * <b>Note:</b> Exiting with a specific code is currently the only option to end
 * TCC on demand. The down side is that the monitoring file remains incomplete.
 * <p>
 * This implementation of paging does not maintain state between pages at the
 * database level. Consequently, this process is sensible to data moving across
 * pages while paging occurs, possibly causing missing records. One way to
 * prevent this is to ensure consistent ordering of the data from one page to
 * the other. If default sorting cannot guarantee to be consistent, a sort order
 * must be forced.
 * <p>
 * If records are added while paging, this may cause duplicate records in the
 * result. In fact, a record inserted in a previously extracted page will push
 * the other records down and the last one that was extracted will appear again
 * in the following page. If required, duplicates can be eliminated by applying
 * a merge that removes them at the end of paging.
 * <p>
 * Also, removed or deleted records may cause shifting of pages and missing
 * records. In fact, records that are removed from already extracted pages
 * during paging will cause other records that should have been extracted in the
 * current page to logically shift to the last extracted page and thus be missed
 * by the current page. The strategy we propose for queries that can have
 * deletions or removals (by filtering on a moving date, e.g. last modified
 * date) is to apply a decrease factor on the paging size on each successive
 * run. This factor represents a safe maximum percentage of deletion/removal we
 * can expect while paging occurs. Doing this, we are likely to extract again
 * the last records of the previous page, but some of them may also be the ones
 * we would have missed otherwise. Duplicates that this strategy will create are
 * a lesser problem than missing records, especially when we apply a merge that
 * removes them at the end. A typical value for this factor is 1%, which for
 * paging size of 100000 represents 1000 removals during the couple minutes that
 * a single page is extracted.
 * <p>
 * In practice, the following rules should apply:
 * <ul>
 * <li>Extracts should be sorted by ascending id or creation date so the new
 * records are extracted last. Sorting should also be on a field or a
 * combination of fields that make up a unique key, to ensure consistent paging.
 * </li>
 * <li>Extracts from tables where deletions can occur or using a filter on
 * moving dates (e.g. last modified date) should use the decreaseFactor
 * parameter as explained above and preferably use a merge post step to
 * eliminate duplicates.</li>
 * <li>With this strategy, the number of records in each page must not reach
 * down to 1, because then there is no more room for decrease. If you must use
 * this strategy with small page counts, make it so that the filter is not based
 * on LRD. This can be done by querying the IDs first and using them to build a
 * static filter. Then paging with a decrease factor of 0 is safe.</li>
 * <li>When doing final extracts in a zone that is closed to transactions, no
 * deletions or modifications can occur, so the decrease factor can simply be
 * set to 0.</li>
 * <li>If paging is used in conjunction with Last Run Date filtering, paging
 * steps must be nested inside LRD Pre and Post steps.</li>
 * <li>If pagingSize is modified while paging is in progress (before reaching
 * the last page), this step will calculate and adjust current page number
 * accordingly, given the new pagingSize leads to an integer page number. If
 * this is not the case, then two scenarios are considered:
 * <ul>
 * <li>If the new pagingSize is less than the previous one, an exception will be
 * raised immediately.</li>
 * <li>If the new pagingSize is greater than the previous one, the change will
 * be postponed until it leads to an integer page number.</li>
 * </ul>
 * </li>
 * </ul>
 * <p>
 * Here are examples of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/PagingPreStep-1.jpg" />
 * <p>
 * <img src="doc-files/PagingPreStep-2.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class PagingPreStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(PagingPreStep.class);

	/** Name of the page index query attribute. */
	private final static String PAGE_INDEX_ATTRIBUTE = "pageindex";

	/**
	 * Constant for the ExitCode system property.
	 */
	private static final String PROP_COMPLETE_EXIT_CODE = "com.taleo.client.customstep.PagingPreStep.CompleteExitCode";

	/**
	 * Default value for
	 * com.taleo.client.customstep.PagingPreStep.CompleteExitCode system
	 * property.
	 */
	private static final int DEFAULT_COMPLETE_EXIT_CODE = 0;

	/**
	 * Constructor without arguments.
	 */
	public PagingPreStep() {
		super();
		// Register valid parameters.
		registerParameterName("pagingSize");
		registerParameterName("decreaseFactor");
		registerParameterName("pagingFilename");
	}

	/**
	 * The global config.
	 */
	private GlobalConfig globalConfig;

	/**
	 * The paging size.
	 */
	protected int pagingSize = 0;

	/**
	 * The decrease factor.
	 */
	protected int decreaseFactor = 0;

	/**
	 * The paging filename.
	 */
	private String pagingFilename;

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "PagingPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Pre-processing step for paging";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "PagingPreStep";
	}

	/**
	 * Get the paging size.
	 * 
	 * @return
	 */
	public int getPagingSize() {
		return pagingSize;
	}

	/**
	 * Set the paging size.
	 * 
	 * @return
	 */
	public void setPagingSize(int pagingSize) {
		if (pagingSize < 0) {
			throw new IllegalArgumentException(
					"The 'pagingSize' parameter must be >= 0");
		}
		this.pagingSize = pagingSize;
	}

	/**
	 * Get the decrease factor.
	 * 
	 * @return
	 */
	public int getDecreaseFactor() {
		return decreaseFactor;
	}

	/**
	 * Set the decrease factor.
	 * 
	 * @return
	 */
	public void setDecreaseFactor(int decreaseFactor) {
		if (decreaseFactor < 0) {
			throw new IllegalArgumentException(
					"The 'decreaseFactor' parameter must be >= 0");
		}
		if (decreaseFactor >= 100) {
			throw new IllegalArgumentException(
					"The 'decreaseFactor' parameter must be < 100");
		}
		this.decreaseFactor = decreaseFactor;
	}

	/**
	 * Get the paging filename.
	 * 
	 * @return
	 */
	public String getPagingFilename() {
		return pagingFilename;
	}

	/**
	 * Set the paging filename.
	 * 
	 * @param pagingFilename
	 */
	public void setPagingFilename(String pagingFilename) {
		this.pagingFilename = pagingFilename;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);
		this.globalConfig = config;

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		try {
			setPagingSize(Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("pagingSize"), "0")));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The pagingSize parameter must be numeric.");
		}

		try {
			setDecreaseFactor(Integer.parseInt(StringUtils.defaultIfEmpty(
					parameters.getProperty("decreaseFactor"), "0")));
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"The decreaseFactor parameter must be numeric.");
		}

		setPagingFilename(parameters.getProperty("pagingFilename"));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the request info and the query instance.
			Message requestInfo = pipeline.getRequestMessage();
			String queryFilename = requestInfo.getCurrentFile();
			if (queryFilename == null) {
				queryFilename = requestInfo.getOriginalFile();
			}
			File queryFile = new File(queryFilename);

			// Decompress if needed.
			queryFile = com.taleo.integration.client.customstep.util.IOHelper
					.getDecompressedFile(queryFile);

			// Load the Query.
			Query query = QueryRuntimeHelper.loadQueryInStep(queryFile,
					productCode);

			// Get the paging size defined in the query
			// and overwrite with the provided pagingSize if > 0.
			BigInteger size = query.getPagingsize();
			if (pagingSize > 0) {
				size = BigInteger.valueOf(pagingSize);
			}
			// If null, use 0 as default.
			if (size == null) {
				size = BigInteger.valueOf(0);
			}

			// Get the paging filename.
			String pgn;
			if (pagingFilename != null) {
				pgn = pagingFilename;
			} else {
				pgn = globalConfig.getGeneral().getConfigurationIdentifier();
			}
			if (!PagingManager.PAGING_EXTENSION.equals(FilenameUtils
					.getExtension(pgn))) {
				pgn += "." + PagingManager.PAGING_EXTENSION;
			}

			File pgnFile = null;
			if (StringUtils.isBlank(FilenameUtils.getPath(pgn))) {
				// Create a paging file in the temp folder if no path specified.
				pgnFile = new File(getTempFolder(), pgn);
			} else {
				pgnFile = new File(pgn);
			}

			log.info("Using the following paging file: '"
					+ pgnFile.getAbsolutePath() + "'.");

			// Get the RunDateManager that was put in the pipeline by the
			// LRDPreStep.
			RunDateManager runDateManager = (RunDateManager) pipeline
					.documentProperties().get(RunDateManager.PIPELINE_KEY);

			// If the RunDateManager is present, get its last run date and
			// current run date.
			Date lastRunDate = null;
			Date currentRunDate = null;
			if (runDateManager != null) {
				lastRunDate = runDateManager.getLastRunDate();
				currentRunDate = runDateManager.getRunDate();
			}

			// Create the paging manager and set it in the pipeline.
			PagingManager pagingManager = new PagingManager(pgnFile,
					size.intValue(), decreaseFactor, lastRunDate,
					currentRunDate);
			pipeline.documentProperties().put(PagingManager.PIPELINE_KEY,
					pagingManager);

			// Exit if paging is complete.
			if (pagingManager.isComplete()) {

				// Delete the paging file.
				pagingManager.delete();

				// Note: TCC currently does not provide any means to abort
				// execution. The only way is to kill the process.
				int exitCode = DEFAULT_COMPLETE_EXIT_CODE;
				String prop = System.getProperty(PROP_COMPLETE_EXIT_CODE);
				if (StringUtils.isNotBlank(prop)) {
					try {
						exitCode = Integer.parseInt(prop);
					} catch (NumberFormatException e) {
						// Invalid property value: keep default value.
						log.warn("Invalid value '" + prop
								+ "' for system property "
								+ PROP_COMPLETE_EXIT_CODE);
					}
				}
				log.info("The run was aborted since the paging is complete. "
						+ "Exiting with code: " + exitCode + "\n");
				System.exit(exitCode);

			} else {
				// Otherwise, add a new page.
				pagingManager.addNewPage();
			}

			log.info("Using pagingSize = "
					+ pagingManager.getAdjustedPagingSize()
					+ " and pageIndex = " + pagingManager.getCurrentPageIndex()
					+ ".");

			// Set the paging size and page index in the query.
			File outFile = createTempFile();
			setPagingSizeAndPageIndex(query, outFile,
					pagingManager.getAdjustedPagingSize(),
					pagingManager.getCurrentPageIndex());

			// Set the transformed query in the request.
			requestInfo.setCurrentFile(outFile.getAbsolutePath());

		} catch (IOException e) {
			WorkflowError error = WorkflowHelper.createWorkflowError(
					WorkflowErrorCode.IO, e);
			throw new StepException(error);

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Set the specified paging size and page index in the query and write the
	 * modified query in the given output file.
	 * 
	 * @param query
	 *            The SQ-XML query.
	 * @param outFile
	 *            The output file.
	 * @param pagingSize
	 *            The paging size.
	 * @param pageIndex
	 *            The page index.
	 * @throws StepException
	 * @throws IOException
	 */
	protected void setPagingSizeAndPageIndex(Query query, File outFile,
			int pagingSize, int pageIndex) throws StepException, IOException {

		// Set the paging size.
		query.setPagingsize(BigInteger.valueOf(pagingSize));

		// Set the page index.
		String[] attributes = new String[0];
		if (query.getAttributes() != null) {
			attributes = query.getAttributes().split(",");
		}
		String attr = PAGE_INDEX_ATTRIBUTE + "=" + pageIndex;
		for (int i = 0; i < attributes.length; i++) {
			if (!attributes[i].split("=")[0].equals(PAGE_INDEX_ATTRIBUTE)) {
				attr += ",";
				attr += attributes[i];
			}
		}
		query.setAttributes(attr);

		// Save the query in the outFile.
		XmlOptions opts = new XmlOptions();
		opts.setSaveSyntheticDocumentElement(new QName(query.getDomNode()
				.getNamespaceURI(), "query", "quer"));
		OutputStream os = IOHelper.getOutputStream(
				new FileOutputStream(outFile), getClass().getName());
		query.save(os, opts);
		os.close();
	}
}
