package com.taleo.integration.client.customstep.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.Comparator;

import org.apache.commons.io.filefilter.WildcardFileFilter;
import org.apache.commons.lang.StringUtils;

import com.taleo.integration.client.util.IOHelper;

/**
 * Utility class that wraps some file functions.
 * 
 * @author Romain Guay, Taleo Corporation
 * 
 */
public final class FileUtils {

	/**
	 * Create a temporary file using the specified folder path and filename
	 * prefix. If the folder specified is null or empty, the file is created in
	 * the default system temporary file folder. If the specified folder does
	 * not exists the folder will be created. File is not deleted on exit. The
	 * suffix is determined by the fact that temporary files are compressed or
	 * not, as defined by System property
	 * <code>com.taleo.integration.compress.tempfile</code>.
	 * 
	 * @param folder
	 *            The folder in which to create the file
	 * @param prefix
	 *            The filename prefix.
	 * @return The newly created temporary file.
	 * @throws IOException
	 */
	public static File createTempFile(String folder, String prefix)
			throws IOException {
		return createTempFile(folder, prefix, false);
	}

	/**
	 * Create a temporary file using the specified folder path and filename
	 * prefix and a flag to indicate delete on exit. The suffix is determined by
	 * the fact that temporary files are compressed or not, as defined by System
	 * property <code>com.taleo.integration.compress.tempfile</code>.
	 * 
	 * @param folder
	 *            The folder in which to create the file
	 * @param prefix
	 *            The filename prefix.
	 * @param deleteOnExit
	 *            Delete file on exit.
	 * @return The newly created temporary file.
	 * @throws IOException
	 */
	public static File createTempFile(String folder, String prefix,
			boolean deleteOnExit) throws IOException {
		File file;

		// Set suffix.
		String suffix = null;
		if (IOHelper.isCompressingTemporaryFile()) {
			suffix = ".zip";
		}

		// Create the file.
		if (folder != null && folder.length() > 0) {
			File _folder = new File(folder);
			if (!_folder.exists()) {
				if (!_folder.mkdirs()) {
					throw new IOException(
							"Cannot create the specified folder: " + folder);
				}
			}
			file = File.createTempFile(prefix, suffix, _folder);
		} else {
			file = File.createTempFile(prefix, suffix);
		}
		if (deleteOnExit) {
			file.deleteOnExit();
		}
		return file;
	}

	/**
	 * Sort an array of files by their name.
	 * 
	 * @param files
	 *            The array of files.
	 * @param asc
	 *            Sort ascending.
	 */
	public static void sortFilesByNameAscending(File[] files, final boolean asc) {
		Arrays.sort(files, new Comparator() {
			public int compare(Object o1, Object o2) {
				int compare = ((File) o1).getName().compareTo(
						((File) o2).getName());
				if (!asc)
					compare = -compare;
				return compare;
			}
		});
	}

	/**
	 * Limits the length of a filename to 180 characters. Also replace illegal
	 * characters (: \ / * ? | < >) with "_" and all control characters (HEX 00
	 * to 1F) with an empty string.
	 * 
	 * @param filename
	 *            The filename only, excluding path.
	 * @return
	 */
	public static String sanitizeFilename(String filename) {
		int index = filename.lastIndexOf(".");
		if (index > 0) {
			filename = StringUtils.left(filename.substring(0, index), 180)
					+ "." + filename.substring(index + 1);
		} else {
			filename = StringUtils.left(filename, 180);
		}
		filename = filename.replaceAll("[:\\\\/*?|<>]", "_");
		filename = filename.replaceAll("[\\u0000-\\u001F]", "");
		return filename;
	}

	/**
	 * Get the line count.
	 * 
	 * @param filename
	 * @return
	 * @throws IOException
	 */
	public static int getLineCount(File file) throws IOException {
		int i = 0;
		BufferedReader reader = null;
		try {
			reader = new BufferedReader(new InputStreamReader(
					IOHelper.getInputStream(file)));
			while (reader.readLine() != null) {
				i++;
			}
		} finally {
			if (reader != null) {
				reader.close();
			}
		}
		return i;
	}

	/**
	 * Return the file content as a String.
	 * 
	 * @param file
	 *            The file.
	 * @param encoding
	 *            The encoding of the file.
	 * @return
	 * @throws IOException
	 */
	public static String readFileToString(File file, String encoding)
			throws IOException {
		BufferedReader reader = new BufferedReader(new InputStreamReader(
				IOHelper.getInputStream(file), encoding));
		String line = null;
		StringBuilder stringBuilder = new StringBuilder();
		String ls = System.getProperty("line.separator");

		while ((line = reader.readLine()) != null) {
			stringBuilder.append(line);
			stringBuilder.append(ls);
		}
		return stringBuilder.toString();
	}

	/**
	 * Get files matching a pattern.
	 * 
	 * @param pattern
	 *            The file pattern. If file only, it is considered in the
	 *            current directory.
	 * @return The list of files matching the pattern. If the parent directory
	 *         does not exist, return an empty list.
	 * @throws IOException
	 * @see WildcardFileFilter
	 */
	public static File[] getMatchingFiles(File pattern) throws IOException {
		// Get the parent directly an default to current.
		File dir = pattern.getParentFile();
		if (dir == null) {
			dir = new File(System.getProperty("user.dir"));
		}

		// Return an empty list if not a directory.
		if (!dir.isDirectory()) {
			return new File[0];
		}

		return dir.listFiles((FileFilter) new WildcardFileFilter(
				new String[] { pattern.getName() }));
	}

}
