package com.taleo.integration.client.customstep.util;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.NTCredentials;
import org.apache.commons.httpclient.UsernamePasswordCredentials;
import org.apache.commons.httpclient.auth.AuthPolicy;
import org.apache.commons.httpclient.auth.AuthScope;
import org.apache.log4j.Logger;

/**
 * Utility class for HTTP clients.
 * 
 * @author Romain Guay, Oracle Corporation
 * 
 */
public final class HttpUtils {

	private static Logger log = Logger.getLogger(HttpUtils.class);

	// Default port for the HTTP proxy.
	private static final int DEFAULT_HTTP_PROXY_PORT = 80;

	// Authentication scheme order.
	private static final List DEFAULT_AUTH_PREFS = new ArrayList(3);
	static {
		DEFAULT_AUTH_PREFS.add(AuthPolicy.NTLM);
		DEFAULT_AUTH_PREFS.add(AuthPolicy.DIGEST);
		DEFAULT_AUTH_PREFS.add(AuthPolicy.BASIC);
	}

	/**
	 * Creates a HTTP client instance for the communication. The protocol and
	 * host are used to determine whether to communication through a HTTP proxy,
	 * based on the standard JVM settings.
	 * 
	 * @return The created httpClient
	 */
	public static HttpClient getHttpClient() {
		// Create an HttpClient.
		HttpClient httpClient = new HttpClient();

		// Set the socket timeout (this is NOT the connection timeout)
		httpClient.getParams().setSoTimeout(0);

		// Determine the standard HTTP proxy information.
		String proxyHost = System.getProperty("http.proxyHost");
		Integer proxyPort = Integer.getInteger("http.proxyPort",
				DEFAULT_HTTP_PROXY_PORT);
		String proxyUser = System.getProperty("http.proxyUser");
		String proxyPassword = System.getProperty("http.proxyPassword");

		// Check if a HTTP proxy was provided.
		if (proxyHost != null && proxyHost.trim().length() > 0) {
			// Set the proxy host.
			log.info("Using HTTP proxy " + proxyHost + ":"
					+ proxyPort.intValue());
			httpClient.getHostConfiguration().setProxy(proxyHost,
					proxyPort.intValue());

			// Check if proxy credentials were provided.
			if (proxyUser != null && proxyUser.trim().length() > 0
					&& proxyPassword != null) {
				// Set the proxy credentials.
				log.info("Using HTTP proxy credentials for " + proxyUser);

				// Set the authentication scheme order.
				httpClient.getParams().setParameter(
						AuthPolicy.AUTH_SCHEME_PRIORITY, DEFAULT_AUTH_PREFS);

				// Set the credentials for the proxy authentication.
				httpClient.getState().setProxyCredentials(AuthScope.ANY,
						createProxyCredentials(proxyUser, proxyPassword));
			}
		}

		// Return the HttpClient instance
		return httpClient;
	}	
	

	/**
	 * Creates the correct instance of credentials based on the proxy parameters
	 * set in the current JVM. This credential is to be used when trying to
	 * authenticate using a proxy.
	 * 
	 * @param pUserName
	 *            The user name to use for the credentials
	 * @param pPassword
	 *            The password for that user
	 * @return The newly created instance of credentials to use for the proxy.
	 */
	public static UsernamePasswordCredentials createProxyCredentials(
			String pUserName, String pPassword) {
		UsernamePasswordCredentials cred;

		// Check if the NTDomain has been specified for NTLM authentication.
		String ntDomain = System.getProperty("http.proxyNTDomain");
		if (ntDomain != null && ntDomain.trim().length() > 0) {
			// Check if the NT host has been specified. Use the current host if
			// not.
			String ntHost = System.getProperty("http.proxyNTHost");
			if (ntHost == null) {
				try {
					ntHost = InetAddress.getLocalHost().getHostName();
				} catch (UnknownHostException e) {
					// Default to empty in that case.
					ntHost = "";
				}
			}

			cred = new NTCredentials(pUserName, pPassword, ntHost, ntDomain);
		} else {
			// Default to normal credentials if no NT information available
			cred = new UsernamePasswordCredentials(pUserName, pPassword);
		}
		return cred;
	}	

}
