package com.taleo.integration.client.customstep.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;

/**
 * IO helper functions, in extension to
 * com.taleo.integration.client.util.IOHelper.
 * 
 * @author Romain Guay, Oracle Corporation
 * 
 */
public class IOHelper {

	/**
	 * Get the original file compressed or the file itself if it is already
	 * compressed.
	 * 
	 * @param pOriginalFile
	 *            the original file
	 * @param pTempFolder
	 *            the temporary files folder
	 * @param pEntry
	 *            the actual entry name of the compressed file. Will be ignored
	 *            if the attribute com.taleo.integration.compress.tempfile as
	 *            been set to false, as the output won't be a
	 *            {@link ZipOutputStream}.
	 * @return the compressed file
	 * @throws IOException
	 */
	public static File getCompressedFile(File pOriginalFile, File pTempFolder,
			String pEntry) throws IOException {
		if (com.taleo.integration.client.util.IOHelper
				.isFileCompressed(pOriginalFile)) {
			return pOriginalFile;
		}

		InputStream is = null;
		OutputStream os = null;
		try {
			is = new BufferedInputStream(new FileInputStream(pOriginalFile));

			File outFile = com.taleo.integration.client.util.IOHelper
					.createTempFile(pTempFolder.getPath());

			os = new BufferedOutputStream(
					com.taleo.integration.client.util.IOHelper.getOutputStream(
							new FileOutputStream(outFile), pEntry));

			IOUtils.copy(is, os);

			return outFile;
		} finally {
			if (is != null) {
				is.close();
			}
			if (os != null) {
				os.close();
			}
		}
	}

	/**
	 * Get the decompressed file or the file itself if it is already
	 * decompressed.
	 * 
	 * @param pCompressedFile
	 *            the compressed file
	 * @return the decompressed file
	 * @throws IOException
	 */
	public static File getDecompressedFile(File pCompressedFile)
			throws IOException {
		if (!com.taleo.integration.client.util.IOHelper
				.isFileCompressed(pCompressedFile)) {
			return pCompressedFile;
		}

		String inFile = pCompressedFile.getAbsolutePath();

		File outFile = new File(inFile.substring(0,
				FilenameUtils.indexOfExtension(inFile)));

		getDecompressedFile(pCompressedFile, outFile);

		return outFile;
	}

	/**
	 * Get the decompressed file or the file itself if it is already
	 * decompressed, into the destination file.
	 * 
	 * @param pCompressedFile
	 *            the compressed file
	 * @param pDestFile
	 *            the destination file
	 * @throws IOException
	 */
	public static void getDecompressedFile(File pCompressedFile, File pDestFile)
			throws IOException {
		if (!com.taleo.integration.client.util.IOHelper
				.isFileCompressed(pCompressedFile)) {
			FileUtils.copyFile(pCompressedFile, pDestFile);
			return;
		}

		ZipInputStream is = null;
		OutputStream os = null;
		try {
			String inFile = pCompressedFile.getAbsolutePath();

			is = new ZipInputStream(new BufferedInputStream(
					new FileInputStream(inFile)));
			is.getNextEntry();

			os = new BufferedOutputStream(new FileOutputStream(pDestFile));

			IOUtils.copy(is, os);
		} finally {
			if (is != null) {
				is.close();
			}
			if (os != null) {
				os.close();
			}
		}
	}

}
