package com.taleo.integration.client.customstep.util;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class contains a static function that will escape special character used
 * within a string to ensure that the string will not parse as a regular
 * expression. This is important when accepting user input and then using that
 * input within a function that accepts regular expressions as an argument. To
 * ensure that the user's input does not interfere with the function all of the
 * regular expression characters are escaped.
 * 
 * @author NickDMax for DreamInCode 2009
 * 
 */
public class RegexUtils {

	/**
	 * By using a precompiled pattern we might speed things up, especially if
	 * there are multiple uses.
	 */
	private static final Pattern GRAB_SP_CHARS = Pattern
			.compile("([\\\\*+\\[\\](){}\\$.?\\^|])");

	/**
	 * This function will escape special characters within a string to ensure
	 * that the string will not be parsed as a regular expression. This is
	 * helpful with accepting using input that needs to be used in functions
	 * that take a regular expression as an argument (such as
	 * String.replaceAll(), or String.split()).
	 * 
	 * @param inStr
	 *            - argument which we wish to escape.
	 * @return - Resulting string with the following characters escaped:
	 *         [](){}+*^?$.\
	 */
	public static String escapeRegex(String inStr) {
		Matcher match = GRAB_SP_CHARS.matcher(inStr);
		return match.replaceAll("\\\\$1");
	}

	/**
	 * This procedure is used to ensure that a string is not interpreted as a
	 * regular expression. This is useful when using the String.replaceAll() or
	 * String.split() functions with user input. This will escape special
	 * characters from the input string so that it will not be interpreted as a
	 * regular expression. This alternate version does not use a precompiled
	 * pattern so it may be copied and pasted into your code as is.
	 * 
	 * @param strIn
	 *            - a string that we want to ensure does not interfere with a
	 *            regular expression.
	 * @return - strIn with the following chars escaped: [](){}+*^?$.\
	 */
	public static String escRegEx(String inStr) {
		// will replace the char with \char
		return inStr.replaceAll("([\\\\*+\\[\\](){}\\$.?\\^|])", "\\\\$1");
	}

}