package com.taleo.integration.client.customstep.util;

/**
 * Helper class for retry processing.
 * 
 * @author Romain Guay, Taleo Corporation
 * 
 */
public final class RetryHelper {
	private int currentAttempt = 1;

	private int nbRetries;

	private int retryInterval;

	private int lastWaited;

	private int retryFactor;

	public static final int NB_RETRIES_DEFAULT = 0;

	public static final int RETRY_INTERVAL_DEFAULT = 0;

	public static final int RETRY_FACTOR_DEFAULT = 1;

	/**
	 * Constructs a <code>RetryHelper</code> instance.
	 * 
	 * @param nbRetries
	 *            max number of retry attempts.
	 * @param retryFactor
	 *            multiplication factor for retrying jobs.
	 * @param retryInterval
	 *            time to wait (seconds).
	 */
	public RetryHelper(int nbRetries, int retryFactor, int retryInterval) {
		this.nbRetries = nbRetries;
		this.retryInterval = retryInterval;
		this.retryFactor = retryFactor;

		// Make sure values are valid.
		if (this.nbRetries < 0) {
			this.nbRetries = NB_RETRIES_DEFAULT;
		}
		if (this.retryInterval < 0) {
			this.retryInterval = RETRY_INTERVAL_DEFAULT;
		}
		if (this.retryFactor <= 0) {
			this.retryFactor = RETRY_FACTOR_DEFAULT;
		}
	}

	/**
	 * Returns the last waited time (seconds).
	 */
	public int getLastWaited() {
		return lastWaited;
	}

	/**
	 * Returns the maximum number of attempts.
	 */
	public int getMaxAttempts() {
		return nbRetries + 1;
	}

	/**
	 * Returns the current attempt.
	 */
	public int getCurrentAttempt() {
		return currentAttempt;
	}

	/**
	 * Returns the number of retries.
	 * 
	 * @return the number of retries.
	 */
	public int getNbRetries() {
		return nbRetries;
	}

	/**
	 * Returns the retryFactor.
	 * 
	 * @return the retryFactor.
	 */
	public int getRetryFactor() {
		return retryFactor;
	}

	/**
	 * Returns the retry interval.
	 * 
	 * @return the retry interval.
	 */
	public int getRetryInterval() {
		return retryInterval;
	}

	/**
	 * If the execution of the method did not complete and we still have retry
	 * attempts, wait for the next retry.
	 */
	public void waitForNextRetry() {
		// Update the last waited time.
		lastWaited = lastWaited == 0 ? retryInterval : lastWaited * retryFactor;

		// Increase the current attempt.
		currentAttempt++;

		try {
			// Wait before retrying.
			Thread.sleep(1000 * lastWaited);
		} catch (InterruptedException e) {
			// Ignore this type of error and retry the process.
		}
	}

}
