/*
 * @(#)CsvReader.java
 *
 * Copyright 1999-2005 by Taleo Corporation,
 * 330, St-Vallier East, Suite 400, Quebec City, Quebec, G1K 9C5, CANADA
 * All rights reserved.
 */

package com.taleo.integration.client.customstep.util.csv;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import com.taleo.integration.client.util.PeekReader;

/**
 * Character delimited (CSV file) reader.
 * 
 * @author bdrolet
 * @since 3.0 SP6
 */
public class CSVReader extends Reader {
	/**
	 * This is the reader of the source csv file.
	 */
	private Reader reader;

	/**
	 * This is the buffer in which the read line will be stored characters
	 */
	private StringBuffer buf;

	/**
	 * This is the list of the column name of the csv file
	 */
	private List header;

	/**
	 * This is the list of value for each column of the csv
	 */
	private List workList;

	/**
	 * Indicate if the source have a header information in the csv file
	 */
	private boolean hasHeader;

	/**
	 * Indicate if the header of the csv file had been read
	 */
	private boolean headerInitialized;

	/**
	 * This is the buffer in which the read characters will be stored
	 */
	private char[] readBuffer;

	/**
	 * Indicate if the end of the underlying reader has been reached
	 */
	private boolean end;

	/**
	 * The current start position in the pre-read buffer
	 */
	private int posStart;

	/**
	 * The current end position in the pre-read buffer
	 */
	private int posEnd;

	/**
	 * The current remaining number of character that are available from the
	 * pre-read buffer.
	 */
	private int nbBuffered;

	/**
	 * The current quote delimiter for the csv.
	 */
	private char quoteDelimiter;

	/**
	 * The current delimiter for the csv column.
	 */
	private char delimiter;

	/**
	 * The current line feed character sequence for the csv line.
	 */
	private String lineFeed;

	/**
	 * Used to keep an Exception if one occurs
	 */
	private Exception error;

	/**
	 * Define a escaping delimiter
	 */
	private String escapingDelimiter;

	/**
	 * Define a default quote character for the csv column
	 */
	private String escapingCarriageReturn;

	/**
	 * Define a default quote character for the csv column
	 */
	private String escapingLineFeed;

	/**
	 * The current line number reached by the reader
	 */
	private int currentLine;

	private Properties properties;

	/**
	 * Define a default quote character for the csv column (ISS).
	 */
	public static final String TALEO_ESCAPING_MODE = "taleoEscaping";

	/**
	 * Define a default quote character for the csv column (excel).
	 */
	public static final String STANDARD_ESCAPING_MODE = "StandartEscaping";

	/**
	 * Error message format when the csv is not well form.
	 */
	public static final String UNEXPECTED_LINE_FORMAT = "Unexpected line format: ";

	/**
	 * Error message format when the csv is not well form.
	 */
	public static final String UNSUPPORTED_TALEO_FORMAT_FOR_STRING_RESULT = "Taleo format is not supported with the read line string result.";

	/**
	 * Error message format when the csv is not well form.
	 */
	public static final String ERROR_INVALID_READER = "The reader cannot be null.";

	/**
	 * Define the size of the bufferRead.
	 */
	private static final int READ_CACHE_SIZE = 2048;

	/**
	 * Error message format when an invalid character is found in place of
	 * another one.
	 */
	private static final String INVALID_CHARACTER_FOUND = "{0}The data in row #{1}, column #{2} is"
			+ " invalid. The character #{3} was a {4}. A delimiter character is expected.";

	/**
	 * Default constructor.
	 * 
	 * @param pReader
	 *            The reader
	 */
	public CSVReader(Reader pReader) {
		this(pReader, true);
	}

	/**
	 * Main constructor.
	 * 
	 * @param pReader
	 *            The reader
	 * @param pHasHeader
	 *            Flag indicating if a header line is present
	 */
	public CSVReader(Reader pReader, boolean pHasHeader) {
		this(pReader, pHasHeader, CSVConstant.DEFAULT_DELIMITER,
				CSVConstant.DEFAULT_QUOTE_CHARACTER);
	}

	/**
	 * Main constructor.
	 * 
	 * @param pReader
	 *            The reader
	 * @param pHasHeader
	 *            Flag indicating if a header line is present
	 * @param pDelimiter
	 *            The value delimiter
	 * @param pQuoteDelimiter
	 *            The quoting character
	 */
	public CSVReader(Reader pReader, boolean pHasHeader, char pDelimiter,
			char pQuoteDelimiter) {
		this(pReader, pHasHeader, pDelimiter, pQuoteDelimiter,
				CSVConstant.DEFAULT_LINE_FEED_STR);
	}

	/**
	 * Main constructor.
	 * 
	 * @param pReader
	 *            The reader
	 * @param pHasHeader
	 *            Flag indicating if a header line is present
	 * @param pDelimiter
	 *            The value delimiter
	 * @param pQuoteDelimiter
	 *            The quoting character
	 * @param pLineFeed
	 *            The line feed character
	 */
	public CSVReader(Reader pReader, boolean pHasHeader, char pDelimiter,
			char pQuoteDelimiter, String pLineFeed) {
		this(pReader, pHasHeader, pDelimiter, pQuoteDelimiter, null, null, null);

		lineFeed = pLineFeed;
	}

	/**
	 * Main constructor.
	 * 
	 * @param pReader
	 *            The reader
	 * @param pHasHeader
	 *            Flag indicating if a header line is present
	 * @param pDelimiter
	 *            The value delimiter
	 * @param pQuoteDelimiter
	 *            The quoting character
	 * @param pEscapingDelimiter
	 *            The escaping sequence for the delimiter
	 * @param pEscapingCarriageReturn
	 *            The escaping sequence for the carriage return
	 * @param pEscapingLineFeed
	 *            The escaping sequence for the line feed
	 */
	public CSVReader(Reader pReader, boolean pHasHeader, char pDelimiter,
			char pQuoteDelimiter, String pEscapingDelimiter,
			String pEscapingCarriageReturn, String pEscapingLineFeed) {
		if (pReader == null) {
			throw new IllegalArgumentException(ERROR_INVALID_READER);
		}

		header = new ArrayList();
		workList = new ArrayList();

		reader = pReader;
		if (!(reader instanceof BufferedReader)) {
			reader = new BufferedReader(reader);
		}
		reader = new PeekReader(reader,
				CsvPropertiesSectionConstants.SIZEOF_PEEKREADER_BUFFER);

		try {
			properties = checkProperties(reader);
		} catch (IllegalStateException e) {
			error = e;
			end = true;
		}

		readBuffer = new char[READ_CACHE_SIZE];

		hasHeader = pHasHeader;
		headerInitialized = false;

		buf = new StringBuffer();

		delimiter = getProperty("delimiter", pDelimiter);
		quoteDelimiter = getProperty("quoteDelimiter",
				pQuoteDelimiter);
		lineFeed = CSVConstant.DEFAULT_LINE_FEED_STR;

		escapingDelimiter = getProperty(
				"escapingDelimiter", pEscapingDelimiter);
		escapingCarriageReturn = getProperty(
				"escapingCarriageReturn",
				pEscapingCarriageReturn);
		escapingLineFeed = getProperty(
				"escapingLineFeed", pEscapingLineFeed);

		preRead();
	}

	/**
	 * Return true if the pre-read buffer is empty.
	 * 
	 * @return true if the pre-read buffer content is empty
	 */
	public boolean bufferIsEmpty() {
		return ((PeekReader) reader).bufferIsEmpty();
	}

	/**
	 * {@inheritDoc}
	 */
	public void close() throws IOException {
		reader.close();
	}

	/**
	 * {@inheritDoc}
	 */
	public int read(char[] pCbuf, int pOff, int pLen) throws IOException {
		int retVal = 0;
		int ch;

		// CHECKSTYLE:OFF InnerAssignment
		// ----------------------------------------------
		// Removing inner assignment checkstyle warning:
		// ----------------------------------------------
		// This code is working and we do not want to refactor it simply to
		// remove the checkstyle
		// warning.
		// 
		// Remove the inner assignment if you need to do any modification to
		// this block of code.
		// ----------------------------------------------
		while (retVal < pLen && (ch = read()) != -1) {
			pCbuf[pOff + (retVal++)] = (char) ch;
		}
		// CHECKSTYLE:ON InnerAssignment

		return retVal != 0 ? retVal : -1;
	}

	/**
	 * Read a line of the csv file and populate the values in the list argument.
	 * 
	 * @param pLine
	 *            The map of lines
	 * @return If the operation was successful.
	 */
	public boolean readLine(Map pLine) throws IOException, InvalidCSVFileFormat {
		boolean result = true;
		pLine.clear();

		String escapingMode = escapingMode();

		result = checkHeader();

		if (result) {
			if (escapingMode.equals(STANDARD_ESCAPING_MODE)) {
				result = readLine(workList);
			} else {
				result = readLineTaleo(workList);
			}

			if (result) {
				if (hasHeader) {
					createMapList(pLine, header, workList);
				} else {
					createMapList(pLine, workList);
				}
			}
		}

		return result;
	}

	private Properties checkProperties(Reader pReader) {
		CsvPropertiesSectionReader propReader = new CsvPropertiesSectionReader(
				pReader);
		try {
			return propReader.readProperties();
		} catch (InvalidCSVFileFormat e) {
			throw new RuntimeException(e);
		}
	}

	private boolean checkHeader() throws IOException, InvalidCSVFileFormat {
		boolean result = true;
		if (hasHeader && !headerInitialized) {
			headerInitialized = true;
			if (escapingMode().equals(STANDARD_ESCAPING_MODE)) {
				result = readLine(header);
			} else {
				result = readLineTaleo(header);
			}
		}
		return result;
	}

	private String escapingMode() {
		String retEscapingMode = null;

		if ((escapingCarriageReturn != null) && (escapingLineFeed != null)
				&& (escapingDelimiter != null)) {
			retEscapingMode = TALEO_ESCAPING_MODE;
		} else {
			retEscapingMode = STANDARD_ESCAPING_MODE;
		}
		return retEscapingMode;

	}

	/**
	 * 
	 * Create a map containing the header column name associate to his value.
	 * 
	 * @param pLine
	 *            The map of lines
	 * @param pHeader
	 *            The list of the header values
	 * @param pWorkList
	 *            The list of the work values associates to the header one.
	 * 
	 * @return The map of the header associate to his work values
	 * @since 1.0
	 */
	private Map createMapList(Map pLine, List pHeader, List pWorkList)
			throws InvalidCSVFileFormat {
		for (int i = 0; i < pWorkList.size(); i++) {
			String value = (String) pWorkList.get(i);

			if (value != null && value.length() > 0) {
				String index = String.valueOf(i);
				if (pHeader != null) {
					try {
						index = (String) pHeader.get(i);
					} catch (IndexOutOfBoundsException e) {
						String errorMsg = MessageFormat.format(
								"{0}{2} values columns were found at row #{1} whereas only {3} header"
										+ " columns were defined",
								new Object[] { UNEXPECTED_LINE_FORMAT,
										new Integer(currentLine),
										new Integer(pWorkList.size()),
										new Integer(pHeader.size()), });
						throw new InvalidCSVFileFormat(errorMsg, e);
					}
				}
				pLine.put(index, value);
			}
		}

		return pLine;
	}

	/**
	 * 
	 * Create a map containing the work values.
	 * 
	 * @param pLine
	 *            The Map of lines
	 * @param pWorkList
	 *            The list of the work values
	 * 
	 * @return The map of work values
	 * @since 1.0
	 */
	private Map createMapList(Map pLine, List pWorkList)
			throws InvalidCSVFileFormat {
		return createMapList(pLine, null, pWorkList);
	}

	/**
	 * Read a single line. This method will read until a line is available, an
	 * I/O error occurs, an InvalidCsvFileFormat error occurs or the end of the
	 * stream is reached.
	 * 
	 * The Default delimiter are the comma.
	 * 
	 * @param pLine
	 *            The list of lines
	 * 
	 * @return return true when a line has been read, or false if the end of the
	 *         stream has been reached
	 */
	public boolean readLine(List pLine) throws IOException,
			InvalidCSVFileFormat {
		boolean retValue = false;

		String escapingMode = escapingMode();

		if (escapingMode.equals(STANDARD_ESCAPING_MODE)) {
			retValue = readLine(pLine, delimiter, quoteDelimiter, lineFeed);
		} else {
			retValue = readLineTaleo(pLine, delimiter, quoteDelimiter);
		}

		return retValue;

	}

	/**
	 * Read a single line. This method will read until a line is available, an
	 * I/O error occurs, an InvalidCsvFileFormat error occurs or the end of the
	 * stream is reached.
	 * 
	 * The Default delimiter are the comma.
	 * 
	 * @param pLine
	 *            The list of lines
	 * 
	 * @return return true when a line has been read, or false if the end of the
	 *         stream has been reached
	 * 
	 * @throws InvalidCSVFileFormat
	 * @throws IOException
	 *             - If an I/O error occurs
	 */
	private boolean readLineTaleo(List pLine) throws IOException,
			InvalidCSVFileFormat {
		return readLineTaleo(pLine, delimiter, quoteDelimiter);
	}

	/**
	 * Read a single line. This method will read until a line is available, an
	 * I/O error occurs, an InvalidCsvFileFormat error occurs or the end of the
	 * stream is reached.
	 * 
	 * @param pLine
	 *            The list of lines
	 * @param pDelimiter
	 *            The delimiter
	 * @param pQuoteCaracter
	 *            The quoting character
	 * @param pLineFeed
	 *            The line feed character
	 * @return return true, or false if the end of the stream has been reached
	 * 
	 * @throws InvalidCSVFileFormat
	 * @throws IOException
	 *             - If an I/O error occurs
	 */
	public boolean readLine(List pLine, char pDelimiter, char pQuoteCaracter,
			String pLineFeed) throws IOException, InvalidCSVFileFormat {
		currentLine++;
		String lineFeedSequence = pLineFeed;
		int lineFeedSeqPos = 0;
		StringBuffer lineFeedBuffer = new StringBuffer();

		checkHeader();

		pLine.clear();
		buf.setLength(0);

		int ch = 0;

		boolean betweenDoubleQuotes = false;
		boolean lastWasDoubleQuote = false;

		// use the know if we are at the end of the row
		boolean readLine = false;

		// use to did not write the character into the list value
		boolean canEraseCharFromList = false;

		// initialize the variable
		StringBuffer columnValue = new StringBuffer();
		columnValue.setLength(0);

		// Indicate if the character need to be append to the line
		boolean canAppendChar = false;

		// Indicate if we have a empty column value
		boolean lastReadCharIsDelimiter = false;

		// Count the number of characters processed in the line
		int columnChar = 0;

		// Count the number of records processed in the line
		int columnRecord = 0;

		// read the source file

		// CHECKSTYLE:OFF InnerAssignment
		// ----------------------------------------------
		// Removing inner assignment checkstyle warning:
		// ----------------------------------------------
		// This code is working and we do not want to refactor it simply to
		// remove the checkstyle
		// warning.
		// 
		// Remove the inner assignment if you need to do any modification to
		// this block of code.
		// ----------------------------------------------

		while ((!readLine) && ((ch = read()) > -1))
		// CHECKSTYLE:ON InnerAssignment
		{
			columnChar++;
			char readChar = (char) ch;
			if (readChar == pQuoteCaracter) {
				// check if the double quote are at the beginning of the column
				if (betweenDoubleQuotes) {
					// check if the double quote are at the end of the column
					if (lastWasDoubleQuote) {
						canAppendChar = true;
						lastWasDoubleQuote = false;
					} else {
						canAppendChar = false;
						lastWasDoubleQuote = true;
					}
				} else {
					canAppendChar = false;
					betweenDoubleQuotes = true;
				}
			}

			// delimiter of each column
			else if (readChar == pDelimiter) {
				if (betweenDoubleQuotes) {
					if (lastWasDoubleQuote) {
						// add the column value to the return line
						pLine.add(columnValue.toString());
						columnRecord++;
						// reset the variable
						columnValue.setLength(0);

						// reset the value of boolean
						canAppendChar = true;
						betweenDoubleQuotes = false;
						canEraseCharFromList = true;
					} else {
						canAppendChar = true;
					}
				} else if (!betweenDoubleQuotes && !lastWasDoubleQuote) {
					// add the column value to the return line
					pLine.add(columnValue.toString());
					columnRecord++;
					// reset the variable
					columnValue.setLength(0);
					canEraseCharFromList = true;
				}
				lastWasDoubleQuote = false;
				lastReadCharIsDelimiter = true;
			}

			// check if we have a carriage return
			else if (readChar == CSVConstant.CARRIAGE_RETURN) {
				canAppendChar = true;

				// when the character is in a quote delimiter, the parser
				// write the character in the list value for the column
				if (!(betweenDoubleQuotes && !lastWasDoubleQuote)) {
					canEraseCharFromList = true;
				}
			}

			// check if we have a line feed
			else if (readChar == CSVConstant.LINE_FEED) {
				// when the character is in a quote delimiter, the parser
				// write the character in the list value for the column
				if (!(betweenDoubleQuotes && !lastWasDoubleQuote)) {
					canEraseCharFromList = true;
				}

				if ((!betweenDoubleQuotes && !lastWasDoubleQuote)
						|| (betweenDoubleQuotes && lastWasDoubleQuote)) {
					// the end of the line has been reached
					readLine = true;
				}
			}
			// check if we have a line feed sequence
			else if (readChar == lineFeedSequence.charAt(lineFeedSeqPos)) {
				// when the character is in a quote delimiter, the parser
				// write the character in the list value for the column
				if (betweenDoubleQuotes && !lastWasDoubleQuote) {
					canEraseCharFromList = false;
				} else {
					canEraseCharFromList = true;
					lineFeedBuffer.append(readChar);
					lineFeedSeqPos++;
				}

				if (lineFeedBuffer.toString().equals(lineFeedSequence)) {
					readLine = true;
					lineFeedBuffer.setLength(0);
				}
			} else
			// all others characters need
			{
				// validate if the csv is correctly formated
				if (lastWasDoubleQuote) {
					String errorMsg = MessageFormat
							.format(INVALID_CHARACTER_FOUND,
									new Object[] { UNEXPECTED_LINE_FORMAT,
											new Integer(currentLine),
											new Integer(++columnRecord),
											new Integer(columnChar),
											"(double) quote", });
					throw new InvalidCSVFileFormat(errorMsg);
				}
				lastWasDoubleQuote = false;
				canAppendChar = true;
				canEraseCharFromList = false;
			}

			if (canAppendChar) {
				buf.append(readChar);
				// we did not want to add the delimiter in the column value
				if (!canEraseCharFromList) {
					columnValue.append(readChar);
					lastReadCharIsDelimiter = false;
				}
			}

			// Since TT91094 Reinitialize the value
			canEraseCharFromList = false;
		}

		if ((columnValue.length() > 0) || lastReadCharIsDelimiter) {
			// add the column value to the return line
			pLine.add(columnValue.toString());
			columnRecord++;
			// reset the variable
			columnValue.setLength(0);
		}

		// validate if the csv is correctly formated
		if (betweenDoubleQuotes && !lastWasDoubleQuote) {
			String errorMsg = MessageFormat
					.format(
							"{0}The row data is missing a (double) quote at"
									+ " the end of the line #{1}. The data in the last column is invalid: {2}",
							new Object[] { UNEXPECTED_LINE_FORMAT,
									new Integer(currentLine), buf.toString(), });
			throw new InvalidCSVFileFormat(errorMsg);
		}

		return readLine || !pLine.isEmpty();
	}

	/**
	 * Read a single line. This method will read until a line is available, an
	 * I/O error occurs, an InvalidCsvFileFormat error occurs or the end of the
	 * stream is reached.
	 * 
	 * @param pLine
	 *            The list of lines
	 * @param pDelimiter
	 *            The delimiter
	 * @param pQuoteCaracter
	 *            The quoting character
	 * 
	 * @return return true, or false if the end of the stream has been reached
	 * 
	 * @throws InvalidCSVFileFormat
	 * @throws IOException
	 *             - If an I/O error occurs
	 */
	public boolean readLineTaleo(List pLine, char pDelimiter,
			char pQuoteCaracter) throws IOException, InvalidCSVFileFormat {
		currentLine++;

		pLine.clear();
		buf.setLength(0);

		int ch = 0;

		boolean betweenDoubleQuotes = false;
		boolean lastWasDoubleQuote = false;

		// use the know if we are at the end of the row
		boolean carriageReturn = false;
		boolean readLine = false;

		// use to did not write the delimiter value
		boolean canEraseCharFromList = false;

		// initialize the variable
		StringBuffer columnValue = new StringBuffer();
		columnValue.setLength(0);

		// Indicate if the character need to be append to the line
		boolean canAppendChar = false;

		// Indicate if we have a empty column value
		boolean lastReadCharIsDelimiter = false;

		// Count the number of characters processed in the line
		int columnChar = 0;

		// Count the number of records processed in the line
		int columnRecord = 0;

		// CHECKSTYLE:OFF InnerAssignment
		// ----------------------------------------------
		// Removing inner assignment checkstyle warning:
		// ----------------------------------------------
		// This code is working and we do not want to refactor it simply to
		// remove the checkstyle
		// warning.
		// 
		// Remove the inner assignment if you need to do any modification to
		// this block of code.
		// ----------------------------------------------

		// read the source file
		while ((!readLine) && ((ch = read()) > -1))
		// CHECKSTYLE:ON InnerAssignment
		{
			columnChar++;
			char readChar = (char) ch;
			if (readChar == pQuoteCaracter) {
				// check if the double quote are at the beginning of the column
				if (betweenDoubleQuotes) {
					// check if the double quote are at the end of the column
					if (lastWasDoubleQuote) {
						canAppendChar = true;
						lastWasDoubleQuote = false;
					} else {
						canAppendChar = false;
						lastWasDoubleQuote = true;
					}
				} else {
					canAppendChar = false;
					betweenDoubleQuotes = true;
				}
			}

			// delimiter of each column
			else if (readChar == pDelimiter) {
				if (betweenDoubleQuotes) {
					if (lastWasDoubleQuote) {
						// add the column value to the return line
						pLine.add(columnValue.toString());
						columnRecord++;
						// reset the variable
						columnValue.setLength(0);

						// reset the value of boolean
						canAppendChar = true;
						carriageReturn = false;
						betweenDoubleQuotes = false;

						// Since TT91094
						// lastWasDoubleQuote = false;
						canEraseCharFromList = true;
					} else {
						canAppendChar = true;
					}
				} else if (!betweenDoubleQuotes && !lastWasDoubleQuote) {
					// add the column value to the return line
					pLine.add(columnValue.toString());
					columnRecord++;
					// reset the variable
					columnValue.setLength(0);
					canEraseCharFromList = true;
				}
				lastWasDoubleQuote = false;
				lastReadCharIsDelimiter = true;
			}

			// check if we have a carriage return
			else if (escapingCarriageReturn.toString().startsWith(
					String.valueOf(readChar))
					|| escapingLineFeed.toString().startsWith(
							String.valueOf(readChar))
					|| escapingDelimiter.toString().startsWith(
							String.valueOf(readChar))) {
				StringBuffer escapingBuf = new StringBuffer();
				// add the character into the buffer
				escapingBuf.append(readChar);

				// read an other char
				ch = read();
				readChar = (char) ch;

				// add the character into the buffer
				escapingBuf.append(readChar);

				/*
				 * loop while the escapingCarriageReturn, or escapingLineFeed,or
				 * escapingDelimiterescaping start with the following escaping
				 * buffer string. The escaping buffer must not be equals to
				 * theses three variables.
				 */
				while ((escapingCarriageReturn.toString().startsWith(
						escapingBuf.toString())
						|| escapingLineFeed.toString().startsWith(
								escapingBuf.toString()) || escapingDelimiter
						.toString().startsWith(escapingBuf.toString()))
						&& (ch != -1)
						&& (!escapingCarriageReturn.toString().equals(
								escapingBuf.toString())
								&& !escapingLineFeed.toString().equals(
										escapingBuf.toString()) && !escapingDelimiter
								.toString().equals(escapingBuf.toString()))) {
					// read the next char
					ch = read();

					if (ch != -1) {
						readChar = (char) ch;
						// add the rad character into the buffered escaping
						// variable
						escapingBuf.append(readChar);
					} else {
						// We read all the file and not escaping variable
						// correspond to the
						// escaping string. Write all buffered character into
						// the output buffer
						buf.append(escapingBuf.toString());
						columnValue.append(escapingBuf.toString());

						// Do not write the last character into the output
						// buffer because it was
						// already write with the escaping buffer.
						canAppendChar = false;
					}
				}

				if (escapingCarriageReturn.equals(escapingBuf.toString())) {
					if (betweenDoubleQuotes && !lastWasDoubleQuote) {
						buf.append(CSVConstant.CARRIAGE_RETURN);
						columnValue.append(CSVConstant.CARRIAGE_RETURN);
						canAppendChar = false;
					} else {
						// The Reader was at the end of the line, but the line
						// have a unexpected
						// Carriage Return
						String errorMsg = MessageFormat.format(
								INVALID_CHARACTER_FOUND, new Object[] {
										UNEXPECTED_LINE_FORMAT,
										new Integer(currentLine),
										new Integer(++columnRecord),
										new Integer(columnChar),
										"carriage return", });
						throw new InvalidCSVFileFormat(errorMsg);
					}
				}

				if (escapingLineFeed.equals(escapingBuf.toString())) {
					if (betweenDoubleQuotes && !lastWasDoubleQuote) {
						buf.append(CSVConstant.LINE_FEED);
						columnValue.append(CSVConstant.LINE_FEED);
						canAppendChar = false;
					} else {
						// the Reader was at the end of the line, but the line
						// have a unexpected
						// line feed
						String errorMsg = MessageFormat
								.format(INVALID_CHARACTER_FOUND, new Object[] {
										UNEXPECTED_LINE_FORMAT,
										new Integer(currentLine),
										new Integer(++columnRecord),
										new Integer(columnChar), "line feed", });
						throw new InvalidCSVFileFormat(errorMsg);
					}
				}

				if (escapingDelimiter.equals(escapingBuf.toString())) {
					if (betweenDoubleQuotes && !lastWasDoubleQuote) {
						buf.append(delimiter);
						columnValue.append(delimiter);
						canAppendChar = false;
					} else {
						// the Reader was at the end of the line, but the line
						// have a unexpected
						// delimiter
						String errorMsg = MessageFormat
								.format(INVALID_CHARACTER_FOUND, new Object[] {
										UNEXPECTED_LINE_FORMAT,
										new Integer(currentLine),
										new Integer(++columnRecord),
										new Integer(columnChar), "delimiter", });
						throw new InvalidCSVFileFormat(errorMsg);
					}
				}
			}

			// check if we have a carriage return
			else if (readChar == CSVConstant.CARRIAGE_RETURN) {
				carriageReturn = true;
				canAppendChar = true;

				// when the character is in a quote delimiter, the parser
				// write the character in the list value for the column
				if (betweenDoubleQuotes && !lastWasDoubleQuote) {
					canEraseCharFromList = false;
				} else {
					canEraseCharFromList = true;
				}
			}

			// check if we have a line feed
			else if (readChar == CSVConstant.LINE_FEED) {
				// when the character is in a quote delimiter, the parser
				// write the character in the list value for the column
				if (betweenDoubleQuotes && !lastWasDoubleQuote) {
					canEraseCharFromList = false;
				} else {
					canEraseCharFromList = true;
				}

				if (((!betweenDoubleQuotes && !lastWasDoubleQuote) || (betweenDoubleQuotes && lastWasDoubleQuote))
						&& carriageReturn) {
					// the end of the line has been reached
					readLine = true;
				}
			} else
			// all others characters need
			{
				// validate if the csv is correctly formated
				if (lastWasDoubleQuote) {
					String errorMsg = MessageFormat
							.format(INVALID_CHARACTER_FOUND,
									new Object[] { UNEXPECTED_LINE_FORMAT,
											new Integer(currentLine),
											new Integer(++columnRecord),
											new Integer(columnChar),
											"(double) quote", });
					throw new InvalidCSVFileFormat(errorMsg);
				}
				lastWasDoubleQuote = false;
				canAppendChar = true;
				canEraseCharFromList = false;
			}

			if (canAppendChar) {
				buf.append(readChar);
				// we did not want to add the some character in the column value
				if (!canEraseCharFromList) {
					columnValue.append(readChar);
					lastReadCharIsDelimiter = false;
				}
				canEraseCharFromList = false;
			}
		}
		if ((columnValue.length() > 0) || lastReadCharIsDelimiter) {
			// add the column value to the return line
			pLine.add(columnValue.toString());
			columnRecord++;
			// reset the variable
			columnValue.setLength(0);
		}

		// validate if the csv is correctly formated
		if (betweenDoubleQuotes && !lastWasDoubleQuote) {
			String errorMsg = MessageFormat
					.format(
							"{0}The row data has finished unexpectedly at"
									+ " line #{1}. The data in the last column is invalid: {2}",
							new Object[] { UNEXPECTED_LINE_FORMAT,
									new Integer(currentLine), buf.toString(), });
			throw new InvalidCSVFileFormat(errorMsg);
		}

		return readLine || !pLine.isEmpty();
	}

	/**
	 * Returns the properties.
	 * 
	 * @return the properties.
	 */
	public Properties getProperties() {
		if (error != null) {
			throw (RuntimeException) error;
		}

		return properties;
	}

	private String getProperty(String pKey, String pDefaultValue) {
		String value = pDefaultValue;

		if (properties != null) {
			value = properties.getProperty(pKey, value);
		}

		return value;
	}

	private char getProperty(String pKey, char pDefaultValue) {
		char value = pDefaultValue;

		if (properties != null) {
			value = properties.getProperty(pKey, String.valueOf(value)).charAt(
					0);
		}

		return value;
	}

	/**
	 * Get the list of headers.
	 * 
	 * @return The list of headers
	 */
	public List getHeadingNames() throws IOException, InvalidCSVFileFormat {
		checkHeader();
		return header;
	}

	/**
	 * {@inheritDoc}
	 */
	public int read() throws IOException {
		int retVal = -1;
		if (nbBuffered > 0) {
			// retrieve the char from the pre-read buffer
			retVal = readBuffer[posStart];
			posStart = posStart < (readBuffer.length - 1) ? posStart + 1 : 0;
			nbBuffered--;

			// fetch some chars from the reader if required, only fetch when the
			// pre-read buffer is
			// empty
			if (nbBuffered == 0) {
				preRead();
			}
		} else {
			// Throw the stored error (if any) when the pre-read buffer is empty
			if (error != null) {
				if (error instanceof IOException) {
					throw (IOException) error;
				}

				throw (RuntimeException) error;
			}
		}
		return retVal;
	}

	/**
	 * Fill the pre-read buffer with the next available chars from the
	 * underlying reader.
	 * 
	 * @throws IOException
	 */
	private void preRead() {
		int length = readBuffer.length;
		int remaining = length - nbBuffered;
		int nbRead = 0;
		int nbToCopy = 0;
		try {
			while (!end && remaining > 0) {
				nbToCopy = length - posEnd;
				if (nbToCopy > remaining) {
					nbToCopy = remaining;
				}
				nbRead = reader.read(readBuffer, posEnd, nbToCopy);

				end = nbRead == -1;

				if (nbRead != -1) {
					remaining -= nbRead;
					nbBuffered += nbRead;
					posEnd += nbRead;
					if (posEnd >= length) {
						posEnd = 0;
					}
				}
			}
		} catch (IOException ioe) {
			// If an error occurred, keep it and throw it when the buffer is
			// emptied.
			error = ioe;
			end = true;
		}
	}
}