/*
 * @(#)CsvPropertiesSectionReader.java
 *
 * Copyright 1999-2005 by Taleo Corporation,
 * 330, St-Vallier East, Suite 400, Quebec City, Quebec, G1K 9C5, CANADA
 * All rights reserved.
 */
package com.taleo.integration.client.customstep.util.csv;

import java.io.IOException;
import java.io.Reader;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.taleo.integration.client.util.PeekReader;

/**
 * <code>CsvPropertiesSectionReader</code> reads the Taleo legacy properties
 * section.
 */
public class CsvPropertiesSectionReader {
	private static final Pattern PROPERTY_PATTERN = Pattern
			.compile("([.A-Za-z0-9_-]+)=(.*)");

	private static final int INITIAL_LINE_BUFFER_SIZE = 64;

	private static final int MAX_PROPERTIES_SIZE = 10 * 1024;

	private Reader reader;

	/**
	 * Constructs a <code>CsvPropertiesSectionReader</code>
	 * 
	 * If the Reader may contain a CSV content WITHOUT
	 * #BEGINPROPERTIES/#ENDPROPERTIES section, you MUST provide a PeekReader,
	 * using a prefetch buffer of at least the size required to parse whether
	 * #BEGINPROPERTIES is present in the file.
	 * 
	 * Using any other Reader or using a PeekReader using an insufficient size
	 * to check the presence or not of the #BEGINPROPERTIES section will assume
	 * that these section IS ALWAYS present in the file and will try parsing it.
	 * 
	 * @param pReader
	 *            the reader to parse for CSV properties
	 */
	public CsvPropertiesSectionReader(Reader pReader) {
		super();
		reader = pReader;
	}

	/**
	 * Read the properties section and put values in a Properties object.
	 * 
	 * @return a Properties object if the section is present, otherwise null
	 * @throws InvalidCSVFileFormat
	 */
	public Properties readProperties() throws InvalidCSVFileFormat {
		boolean parseSection = true;
		if (reader instanceof PeekReader) {
			parseSection = checkForSectionPresence((PeekReader) reader);
		}

		Properties result = null;
		if (parseSection) {
			result = new Properties();
			try {
				boolean done = false;
				boolean inHeader = false;

				StringBuffer line = new StringBuffer(INITIAL_LINE_BUFFER_SIZE);
				int count = readLine(line, MAX_PROPERTIES_SIZE);

				while (count < MAX_PROPERTIES_SIZE && !done) {
					String lineStr = line.toString();
					if (!inHeader) {
						assertBeginProperties(lineStr);
						inHeader = true;
					} else {
						if (equalStrings(
								CsvPropertiesSectionConstants.ENDPROPERTIES,
								lineStr)) {
							done = true;
						} else {
							addProperty(result, line);
						}
					}

					if (!done) {
						count += readLine(line, MAX_PROPERTIES_SIZE - count);
					}
				}

				if (!done && count > MAX_PROPERTIES_SIZE) {
					throw new InvalidCSVFileFormat(
							"Header is too large, a maximum of "
									+ MAX_PROPERTIES_SIZE
									+ " characters is allowed.");
				}
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
		}
		return result;
	}

	private boolean checkForSectionPresence(PeekReader pReader) {
		if (pReader.bufferStartsWith(" ") || pReader.bufferStartsWith("\r")
				|| pReader.bufferStartsWith("\n")
				|| pReader.bufferStartsWith("\t")) {
			throw new IllegalStateException(
					"[CSV_INVALID_FORMAT] The input file cannot start with "
							+ "spaces, new line or tab character.");
		}

		return pReader
				.bufferStartsWith(CsvPropertiesSectionConstants.BEGINPROPERTIES);
	}

	private void addProperty(Properties pProperties, StringBuffer pLine)
			throws InvalidCSVFileFormat {
		Matcher matcher = PROPERTY_PATTERN.matcher(pLine);

		if (!matcher.matches()) {
			throw new InvalidCSVFileFormat("Property is not well formed: "
					+ pLine);
		}

		String key = matcher.group(1).trim();
		String value = matcher.group(2);
		value = value != null ? value.trim() : null;

		pProperties.setProperty(key, value);
	}

	private int readLine(StringBuffer pBuffer, int pMaxRead) throws IOException {
		pBuffer.setLength(0);

		int ch = reader.read();
		int count = 1;
		boolean done = false;

		while (ch != -1 && count < pMaxRead && !done) {
			char c = (char) ch;
			if (!isBlank(c)) {
				if (isNewLine(c)) {
					done = true;
				} else {
					pBuffer.append(c);
				}
			}

			if (!done) {
				ch = reader.read();
				count++;
			}
		}

		if (!done) {
			count++;
		}

		return count;
	}

	private void assertBeginProperties(String pLine)
			throws InvalidCSVFileFormat {
		if (!equalStrings(CsvPropertiesSectionConstants.BEGINPROPERTIES, pLine)) {
			throw new InvalidCSVFileFormat(
					"Expecting a properties header section at the "
							+ "beginning of the file.");
		}
	}

	private boolean isNewLine(char pChar) {
		return pChar == '\n';
	}

	private boolean isBlank(char pChar) {
		return pChar == ' ' || pChar == '\t' || pChar == '\r';
	}

	private boolean equalStrings(String pExpected, String pActual) {
		return pActual != null && pActual.trim().equals(pExpected);
	}
}
