package com.taleo.integration.client.customstep.util.csv;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Reader;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;

/**
 * Utility class for CSV.
 * 
 * @author Romain Guay, Taleo Corporation
 * 
 */
public class CsvUtils {

	/**
	 * Read the first line from the given reader.
	 * 
	 * @param reader
	 *            The reader containing the CSV.
	 * @param csvDelimiter
	 * @param csvQuoteCharacter
	 * @return If the first line is the header, it is returned. If there is no
	 *         header, the first record is returned. If the file is empty, an
	 *         empty array is returned.
	 * @throws InvalidCSVFileFormat
	 * @throws IOException
	 */
	public static String[] readFirstLine(Reader reader, char csvDelimiter,
			char csvQuoteCharacter) throws IOException, InvalidCSVFileFormat {
		return readLine(reader, csvDelimiter, csvQuoteCharacter, 1);
	}

	/**
	 * Read the given line number from the given reader.
	 * 
	 * @param reader
	 *            The reader containing the CSV.
	 * @param csvDelimiter
	 * @param csvQuoteCharacter
	 * @param lineNumber
	 *            The line number (1 based).
	 * @return The line in a String array. If the file is empty, an empty array
	 *         is returned.
	 * @throws InvalidCSVFileFormat
	 * @throws IOException
	 */
	public static String[] readLine(Reader reader, char csvDelimiter,
			char csvQuoteCharacter, int lineNumber) throws IOException,
			InvalidCSVFileFormat {
		CSVReader csvReader = null;
		try {
			csvReader = new CSVReader(reader, false, csvDelimiter,
					csvQuoteCharacter);
			List values = new ArrayList();
			// Read lines before the one desired.
			for (int i = 1; i < lineNumber; i++) {
				csvReader.readLine(values);
			}
			// Read the desired line.
			csvReader.readLine(values);
			return (String[]) values.toArray(new String[] {});

		} finally {
			if (reader != null) {
				reader.close();
			}
		}
	}

	/**
	 * Get the index for the given column name or index. If column is numeric,
	 * it is considered an index already, otherwise the index is found in the
	 * header. Column name comparison ignores case.
	 * 
	 * @param column
	 *            The column name or index.
	 * @param header
	 *            The CSV header. May be null if column is numeric.
	 * @return The column index.
	 * @throws ArrayIndexOutOfBoundsException
	 *             If the column name is not found in the header.
	 */
	public static int getColumnIndex(String column, String[] header)
			throws ArrayIndexOutOfBoundsException {
		if (StringUtils.isNumeric(column)) {
			int index = new Integer(column).intValue();
			if (index < 0) {
				throw new IllegalArgumentException("Argument 'column' ("
						+ column + ") must be >= 0.");
			}
			return index;
		} else {
			if (header == null) {
				throw new IllegalArgumentException("Argument 'column' ("
						+ column
						+ ") must be numeric if the file has no header.");
			}
			for (int i = 0; i < header.length; i++) {
				if (column.equalsIgnoreCase(header[i])) {
					return i;
				}
			}
		}
		throw new ArrayIndexOutOfBoundsException("Column not found: " + column
				+ " in header " + ArrayUtils.toString(header) + ".");
	}

	/**
	 * Get the columns indexes, transforming names into indexes if required.
	 * Column name comparison ignores case.
	 * 
	 * @param columns
	 *            The columns name or index.
	 * @param header
	 *            The CSV header. May be null if all columns are numeric.
	 * @return The column indexes.
	 * @throws ArrayIndexOutOfBoundsException
	 *             If a column name is not found in the header.
	 */
	public static int[] getColumnIndexes(String[] columns, String[] header) {
		int[] indexes = new int[columns.length];
		for (int i = 0; i < columns.length; i++) {
			indexes[i] = getColumnIndex(columns[i], header);
		}
		return indexes;
	}

	/**
	 * Serialize an object into a Base64 string.
	 * 
	 * @param object
	 *            The object.
	 * @return
	 * @throws IOException
	 */

	public static String serializeBase64(Object object) throws IOException {
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		ObjectOutputStream oos = new ObjectOutputStream(bos);
		try {
			oos.writeObject(object);
			oos.flush();
			return new String(Base64.encodeBase64(bos.toByteArray()));
		} finally {
			oos.close();
		}
	}

	/**
	 * Deserialize a Base64 string into an object.
	 * 
	 * @param base64
	 *            The Base64 string.
	 * @return
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public static Object deserializeBase64(String base64) throws IOException,
			ClassNotFoundException {
		ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(
				Base64.decodeBase64(base64.getBytes())));
		try {
			return ois.readObject();
		} finally {
			ois.close();
		}
	}

}
