package com.taleo.integration.client.customstep.util.csv;

import java.io.IOException;
import java.io.Writer;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.lang.StringUtils;

/**
 * Class to write CSV.
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public class CsvWriter {

	/**
	 * The underlying writer.
	 */
	private Writer writer;

	/**
	 * The CSV delimiter.
	 */
	private char delimiter = CSVConstant.DEFAULT_DELIMITER;

	/**
	 * The CSV quote character.
	 */
	private char quoteCharacter = CSVConstant.DEFAULT_QUOTE_CHARACTER;

	/**
	 * Indicator to force quote character.
	 */
	private boolean forceQuoteCharacter = false;

	/**
	 * Private variable for quote character replacement.
	 */
	private String quoteCharacterString = new String(
			new char[] { quoteCharacter });

	/**
	 * Line separator string.
	 */
	private String lineSeparator = System.getProperty("line.separator");

	/**
	 * Constructor with the underlying writer.
	 * 
	 * @param writer
	 *            The underlying writer.
	 */
	public CsvWriter(Writer writer) {
		this.writer = writer;
	}

	/**
	 * Constructor with the underlying writer, the delimiter, quote character
	 * and header presence indicator.
	 * 
	 * @param writer
	 *            The underlying writer.
	 * @param delimiter
	 *            The CSV delimiter.
	 * @param quoteCharacter
	 *            The CSV quote character.
	 */
	public CsvWriter(Writer writer, char delimiter, char quoteCharacter) {
		this(writer);
		setDelimiter(delimiter);
		setQuoteCharacter(quoteCharacter);
	}

	/**
	 * @return
	 */
	public char getDelimiter() {
		return delimiter;
	}

	/**
	 * @param delimiter
	 */
	public void setDelimiter(char delimiter) {
		this.delimiter = delimiter;
	}

	/**
	 * @return
	 */
	public char getQuoteCharacter() {
		return quoteCharacter;
	}

	/**
	 * @param quoteCharacter
	 */
	public void setQuoteCharacter(char quoteCharacter) {
		this.quoteCharacter = quoteCharacter;
		this.quoteCharacterString = new String(new char[] { quoteCharacter });
	}

	/**
	 * @return
	 */
	public String getLineSeparator() {
		return lineSeparator;
	}

	/**
	 * @param lineSeparator
	 */
	public void setLineSeparator(String lineSeparator) {
		this.lineSeparator = lineSeparator;
	}

	/**
	 * Get the indicator that forces quote character on every field.
	 * 
	 * @return
	 */
	public boolean isForceQuoteCharacter() {
		return forceQuoteCharacter;
	}

	/**
	 * Set the indicator that forces quote character on every field.
	 * 
	 * @param forceQuoteCharacter
	 */
	public void setForceQuoteCharacter(boolean forceQuoteCharacter) {
		this.forceQuoteCharacter = forceQuoteCharacter;
	}

	/**
	 * Write a line in the CSV.
	 * 
	 * @param values
	 *            The list of values to write.
	 * @throws IOException
	 */
	public void writeLine(List values) throws IOException {
		// Write the values.
		int i = 0;
		for (Iterator iter = values.iterator(); iter.hasNext();) {
			write(iter.next(), i++);
		}

		// Write a line separator.
		writer.write(lineSeparator);
	}

	/**
	 * Write a line in the CSV.
	 * 
	 * @param values
	 *            The list of values to write.
	 * @throws IOException
	 */
	public void writeLine(String[] values) throws IOException {
		// Write the values.
		for (int i = 0; i < values.length; i++) {
			write(values[i], i);
		}

		// Write a line separator.
		writer.write(lineSeparator);
	}

	/**
	 * Write a value, considering its column index.
	 * 
	 * @param value
	 *            The value to write.
	 * @param index
	 *            The index of the CSV column.
	 * @throws IOException
	 */
	private void write(Object value, int index) throws IOException {
		// Convert value to string.
		String stringValue = "";
		if (value != null) {
			stringValue = value.toString();
		}

		// Write the delimiter except for the first column.
		if (index > 0) {
			writer.write(delimiter);
		}

		// If the value contains the delimiter, quote character, CR or LF, or if
		// the quote character is forced then write a quote character.
		boolean needsQuote = false;
		if (forceQuoteCharacter || stringValue.indexOf(delimiter) > -1
				|| stringValue.indexOf(quoteCharacter) > -1
				|| stringValue.indexOf((char) 10) > -1
				|| stringValue.indexOf((char) 13) > -1) {
			needsQuote = true;
			writer.write(quoteCharacter);
		}

		// If the value contains the quote character, double it.
		if (stringValue.indexOf(quoteCharacter) > -1) {
			stringValue = StringUtils.replace(stringValue,
					quoteCharacterString, quoteCharacterString
							+ quoteCharacterString);
		}

		// Write the value.
		writer.write(stringValue);

		// Write a closing quote character if needed.
		if (needsQuote) {
			writer.write(quoteCharacter);
		}
	}

	/**
	 * Close the writer.
	 * 
	 * @throws IOException
	 */
	public void close() throws IOException {
		writer.close();
	}

	/**
	 * Flush the writer.
	 * 
	 * @throws IOException
	 */
	public void flush() throws IOException {
		writer.flush();
	}

}
