package com.taleo.integration.client.customstep.util.csv;

import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Comparator to sort CSV rows in hierarchical order (parent/child). Rows are of
 * type String[].
 * <p>
 * Must be initialized by setting the parent key for each key, using the
 * <code>setParent(String, Sting)</code> method. Row key does not have to be
 * unique, many rows with the same key will be sorted equal. However, two rows
 * with the same key but different parents, will cause an error.
 * <p>
 * Does not detect circular relationships and will not fail if that happens.
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class HierarchicalRowComparator implements Comparator {

	/**
	 * Column index for the row key.
	 */
	private int keyIndex = -1;

	/**
	 * Sort in top down (parent first) order (default: true).
	 */
	private boolean topDown = true;

	/**
	 * Indicate if case is considered when comparing keys (default: true).
	 */
	private boolean ignoreCase = true;

	/**
	 * Map of child to parent key.
	 */
	private Map parentMap = new HashMap();

	/**
	 * Constructor with the column index of the row key.
	 * 
	 * @param keyIndex
	 *            The column index of the row key.
	 */
	public HierarchicalRowComparator(int keyIndex) {
		this.keyIndex = keyIndex;
	}

	/**
	 * Constructor with the column index of the row key and the sort order.
	 * 
	 * @param keyIndex
	 *            The column index of the row key.
	 * @param topDown
	 *            <code>true</code> for top down (parent first) sorting.
	 */
	public HierarchicalRowComparator(int keyIndex, boolean topDown) {
		this(keyIndex);
		this.topDown = topDown;
	}

	/**
	 * Constructor with the column index of the row key, the sort order and the
	 * case sensitiveness.
	 * 
	 * @param keyIndex
	 *            The column index of the row key.
	 * @param topDown
	 *            <code>true</code> for top down (parent first) sorting.
	 * @param ignoreCase
	 *            Indicate if case is considered when comparing keys.
	 */
	public HierarchicalRowComparator(int keyIndex, boolean topDown,
			boolean ignoreCase) {
		this(keyIndex, topDown);
		this.ignoreCase = ignoreCase;
	}

	/**
	 * Set the parent key for a given child key. Parent key can be null if at
	 * the top of the hierarchy.
	 * 
	 * @param key
	 *            The child key.
	 * @param parentKey
	 *            The parent key.
	 */
	public void setParent(String key, String parentKey) {
		if (key == null) {
			throw new IllegalArgumentException("Argument 'key' cannot be null.");
		}
		if (key.equals(parentKey)) {
			throw new IllegalArgumentException("Key '" + key
					+ "' cannot equal its parentKey.");
		}
		if (isParentSet(key) && !equals(getParent(key), parentKey)) {
			throw new IllegalArgumentException("Key '" + key
					+ "' is already linked to parent '" + getParent(key)
					+ "', it cannot be linked to " + parentKey + ".");
		}
		parentMap.put(key, parentKey);
	}

	/**
	 * Get the parent key for a given child key.
	 * 
	 * @param key
	 *            The child key.
	 * @return The parent key.
	 */
	public String getParent(String key) {
		return (String) parentMap.get(key);
	}

	/**
	 * Indicate if parent key has been set for a given child key.
	 * 
	 * @param key
	 *            The child key.
	 * @return
	 */
	private boolean isParentSet(String key) {
		return parentMap.containsKey(key);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Comparator#compare(T, T)
	 */
	public int compare(Object obj1, Object obj2) {
		if (obj1 instanceof String[]) {
			return compare((String[]) obj1, (String[]) obj2);
		} else if (obj1 instanceof List) {
			return compare((List) obj1, (List) obj2);
		} else {
			throw new UnsupportedOperationException(
					"Cannot compare arguments of the given type.");
		}
	}

	/**
	 * Compare rows as String[].
	 * 
	 * @param row1
	 * @param row2
	 * @return
	 */
	public int compare(String[] row1, String[] row2) {
		return compare(row1[keyIndex], row2[keyIndex]);
	}

	/**
	 * Compare rows as List.
	 * 
	 * @param row1
	 * @param row2
	 * @return
	 */
	public int compare(List row1, List row2) {
		return compare(row1.get(keyIndex), row2.get(keyIndex));
	}

	/**
	 * Compare row keys.
	 * 
	 * @param key1
	 *            Key of the first row.
	 * @param key2
	 *            Key of the second row.
	 * @return
	 */
	private int compare(String key1, String key2) {
		String parentKey1 = getParent(key1);
		String parentKey2 = getParent(key2);

		if (equals(key1, key2)) {
			return 0;
		} else if (equals(parentKey1, parentKey2)) {
			return (topDown ? key1.compareTo(key2) : key2.compareTo(key1));
		} else if (parentKey1 == null || parentKey1.equals("")) {
			return (topDown ? -1 : 1);
		} else if (parentKey2 == null || parentKey2.equals("")) {
			return (topDown ? 1 : -1);
		} else if (equals(parentKey2, key1)) {
			return (topDown ? -1 : 1);
		} else if (equals(parentKey1, key2)) {
			return (topDown ? 1 : -1);
		} else {
			return compare(getParent(key1), getParent(key2));
		}
	}

	/**
	 * Determine equality.
	 * 
	 * @param key1
	 * @param key2
	 * @return
	 */
	private boolean equals(String key1, String key2) {
		if (key1 == null && key2 == null) {
			return true;
		} else if (key1 == null) {
			return false;
		} else {
			if (ignoreCase) {
				return key1.equalsIgnoreCase(key2);
			} else {
				return key1.equals(key2);
			}
		}
	}
}
