package com.taleo.integration.client.customstep.util.csv;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.Constants;
import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.integration.client.util.IOHelper;

/**
 * Implementation of the Merge Sort algorithm to sort large CSV files.
 * <p>
 * Based on: <a href=
 * "http://www.codeodor.com/index.cfm/2007/5/10/Sorting-really-BIG-files/1194"
 * >Sorting-really-BIG-files</a>
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class MergeSort {

	private static Logger log = Logger.getLogger(MergeSort.class);

	/**
	 * The default chunk size (KB).
	 */
	public final static int DEFAULT_CHUNK_SIZE = 10000;

	/**
	 * The chunk size (KB) in which the file will be split for executing the
	 * sort process.
	 */
	private int chunkSize = DEFAULT_CHUNK_SIZE;

	/**
	 * Remove rows having the same sorting order with regards to the provided
	 * sort criteria. If duplicates are present, only the last one is kept.
	 */
	private boolean removeDuplicates = false;

	/**
	 * Remove rows having the same sorting order with regards to the provided
	 * sort criteria. If duplicates are present, all of them are removed.
	 */
	private boolean removeAllDuplicates = false;

	/**
	 * The CSV delimiter.
	 */
	private char csvDelimiter = CSVConstant.DEFAULT_DELIMITER;

	/**
	 * The CSV quote character.
	 */
	private char csvQuoteCharacter = CSVConstant.DEFAULT_QUOTE_CHARACTER;

	/**
	 * The CSV header presence indicator.
	 */
	private boolean csvHeaderPresent = false;

	/**
	 * The encoding used for reading ad writing files (Default: UTF-8).
	 * 
	 * @see java.nio.charset.Charset
	 */
	private String encoding = Constants.UTF8;

	/**
	 * The working folder for the merge process.
	 */
	private String workingFolder = null;

	/**
	 * Get the chunk size.
	 * 
	 * @return
	 */
	public int getChunkSize() {
		return chunkSize;
	}

	/**
	 * Set the chunk size.
	 * 
	 * @param chunkSize
	 */
	public void setChunkSize(int chunkSize) {
		this.chunkSize = chunkSize;
	}

	/**
	 * Get the remove duplicates flag.
	 * 
	 * @return
	 */
	public boolean isRemoveDuplicates() {
		return removeDuplicates;
	}

	/**
	 * Set the remove duplicates flag.
	 * 
	 * @param removeDuplicates
	 */
	public void setRemoveDuplicates(boolean removeDuplicates) {
		this.removeDuplicates = removeDuplicates;
	}

	/**
	 * Get the remove all duplicates flag.
	 * 
	 * @return
	 */
	public boolean isRemoveAllDuplicates() {
		return removeAllDuplicates;
	}

	/**
	 * Set the remove all duplicates flag.
	 * 
	 * @param removeAllDuplicates
	 */
	public void setRemoveAllDuplicates(boolean removeAllDuplicates) {
		this.removeAllDuplicates = removeAllDuplicates;
	}

	/**
	 * Get the CSV delimiter.
	 * 
	 * @return
	 */
	public char getCsvDelimiter() {
		return csvDelimiter;
	}

	/**
	 * Set the CSV delimiter.
	 * 
	 * @param csvDelimiter
	 */
	public void setCsvDelimiter(char csvDelimiter) {
		this.csvDelimiter = csvDelimiter;
	}

	/**
	 * Get the CSV quote character.
	 * 
	 * @return
	 */
	public char getCsvQuoteCharacter() {
		return csvQuoteCharacter;
	}

	/**
	 * Set the CSV quote character.
	 * 
	 * @param csvQuoteCharacter
	 */
	public void setCsvQuoteCharacter(char csvQuoteCharacter) {
		this.csvQuoteCharacter = csvQuoteCharacter;
	}

	/**
	 * Get the CSV header present flag.
	 * 
	 * @return
	 */
	public boolean isCsvHeaderPresent() {
		return csvHeaderPresent;
	}

	/**
	 * Set the CSV header present flag.
	 * 
	 * @param csvHeaderPresent
	 */
	public void setCsvHeaderPresent(boolean csvHeaderPresent) {
		this.csvHeaderPresent = csvHeaderPresent;
	}

	/**
	 * Get the encoding.
	 * 
	 * @return
	 */
	public String getEncoding() {
		return encoding;
	}

	/**
	 * Set the encoding.
	 * 
	 * @param encoding
	 */
	public void setEncoding(String encoding) {
		// Validate encoding by loading it into a Charset.
		Charset.forName(encoding);
		this.encoding = encoding;
	}

	/**
	 * Get the working folder.
	 * 
	 * @return
	 */
	public String getWorkingFolder() {
		return workingFolder;
	}

	/**
	 * Set the working folder.
	 * 
	 * @param workingFolder
	 */
	public void setWorkingFolder(String workingFolder) {
		this.workingFolder = workingFolder;
	}

	/**
	 * Sort the inFile according to the given Comparator and write result in the
	 * outFile. If outFile is the same as inFile, it will be overwritten.
	 * <p>
	 * The Comparator must be able to compare String arrays (String[])
	 * corresponding to a record in the file.
	 * 
	 * @param inFile
	 *            The input file.
	 * @param comparator
	 *            The row comparator.
	 * @param outFile
	 *            The output file.
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	public void sort(File inFile, Comparator comparator, File outFile)
			throws IOException, InvalidCSVFileFormat {
		CSVReader reader = null;
		CsvWriter writer = null;
		try {
			List values = new ArrayList();

			// Prepare a reader on the input file.
			reader = new CSVReader(new InputStreamReader(
					IOHelper.getInputStream(inFile), encoding), false,
					getCsvDelimiter(), getCsvQuoteCharacter());

			// Read the header.
			String[] header = null;
			if (isCsvHeaderPresent()) {
				reader.readLine(values);
				header = (String[]) values.toArray(new String[values.size()]);
			}

			// Read file into multiple chunks.
			String[] row;
			List chunk = new ArrayList();
			List chunkFiles = new ArrayList();
			boolean eof = false;
			while (!eof) {
				int currChunkSize = 0;

				// Read until chunk size is reached.
				while (currChunkSize <= chunkSize * 1024) {
					eof = !reader.readLine(values);
					if (eof)
						break;
					row = (String[]) values.toArray(new String[values.size()]);
					chunk.add(row);
					for (int i = 0; i < row.length; i++) {
						currChunkSize += row[i].getBytes().length;
					}
				}

				// Sort the rows for the current chunk.
				chunk = mergeSort(chunk, comparator);

				// Write current chunk to disk.
				File chunkFile = FileUtils.createTempFile(workingFolder,
						getClass().getName());
				chunkFile.deleteOnExit();

				writer = new CsvWriter(new BufferedWriter(
						new OutputStreamWriter(new FileOutputStream(chunkFile),
								encoding)), getCsvDelimiter(),
						getCsvQuoteCharacter());
				if (isCsvHeaderPresent()) {
					writer.writeLine(header);
				}
				for (int i = 0; i < chunk.size(); i++) {
					writer.writeLine((String[]) chunk.get(i));
				}
				writer.close();
				chunkFiles.add(chunkFile);
				chunk.clear();
			}

			// Merge chunk files.
			mergeFiles(chunkFiles, comparator, outFile);

		} finally {
			if (reader != null) {
				reader.close();
			}
			if (writer != null) {
				writer.close();
			}
		}
	}

	/**
	 * Apply MergeSort algorithm to sort an array of rows based on the given
	 * sort criteria.
	 * 
	 * @param rows
	 *            Array of rows to sort.
	 * @param comparator
	 *            The row comparator.
	 * @return
	 */
	private List mergeSort(List rows, Comparator comparator) {
		List left = new ArrayList();
		List right = new ArrayList();
		if (rows.size() <= 1)
			return rows;
		else {
			int middle = rows.size() / 2;
			for (int i = 0; i < middle; i++)
				left.add(rows.get(i));
			for (int j = middle; j < rows.size(); j++)
				right.add(rows.get(j));
			left = mergeSort(left, comparator);
			right = mergeSort(right, comparator);
			return merge(left, right, comparator);
		}
	}

	/**
	 * Merge the left and right lists from mergeSort back together.
	 * 
	 * @param left
	 *            Left list of rows.
	 * @param right
	 *            Right list of rows.
	 * @param comparator
	 *            The row comparator.
	 * @return
	 */
	private List merge(List left, List right, Comparator comparator) {
		List result = new ArrayList();

		while (left.size() > 0 && right.size() > 0) {
			if (comparator.compare(left.get(0), right.get(0)) <= 0) {
				result.add(left.get(0));
				left.remove(0);
			} else {
				result.add(right.get(0));
				right.remove(0);
			}
		}
		if (left.size() > 0) {
			for (int i = 0; i < left.size(); i++)
				result.add(left.get(i));
		}
		if (right.size() > 0) {
			for (int i = 0; i < right.size(); i++)
				result.add(right.get(i));
		}
		return result;
	}

	/**
	 * Merge the sorted chunk files, applying the same sorting criteria.
	 * 
	 * @param chunkFiles
	 *            Array of chunk files to be merged.
	 * @param comparator
	 *            The row comparator.
	 * @param outFile
	 *            The output file.
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	private void mergeFiles(List chunkFiles, Comparator comparator, File outFile)
			throws IOException, InvalidCSVFileFormat {
		ArrayList readers = new ArrayList();
		CsvWriter writer = null;

		try {
			boolean someFileStillHasRows = false;
			boolean eof = false;
			List values = new ArrayList();

			// Prepare a writer on the output file.
			writer = new CsvWriter(new BufferedWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(new FileOutputStream(outFile),
							outFile.getName()), encoding)), getCsvDelimiter(),
					getCsvQuoteCharacter());

			// Read one row from each chunk file.
			String[] row = null;
			String[] pendingRow = null;
			int duplicateCount = 0;
			ArrayList filerows = new ArrayList();
			for (int i = 0; i < chunkFiles.size(); i++) {
				File chunkFile = (File) chunkFiles.get(i);

				// Prepare a reader on each of the chunk files.
				CSVReader reader = new CSVReader(new InputStreamReader(
						new FileInputStream(chunkFile), encoding), false,
						getCsvDelimiter(), getCsvQuoteCharacter());
				readers.add(reader);

				// Read the header if present and write it for the first chunk.
				if (isCsvHeaderPresent()) {
					reader.readLine(values);
					if (i == 0) {
						String[] header = (String[]) values
								.toArray(new String[values.size()]);
						writer.writeLine(header);
					}
				}

				// Read the first row.
				eof = !reader.readLine(values);
				if (!eof) {
					row = (String[]) values.toArray(new String[values.size()]);
					filerows.add(row);
					someFileStillHasRows = true;
				} else {
					filerows.add(null);
				}
			}

			// Loop while any chunk file has rows.
			while (someFileStillHasRows) {
				String[] min;
				int minIndex = 0;

				// Get row from first file.
				row = (String[]) filerows.get(0);
				if (row != null) {
					min = row;
					minIndex = 0;
				} else {
					min = null;
					minIndex = -1;
				}

				// Check which one is min.
				for (int i = 1; i < filerows.size(); i++) {
					row = (String[]) filerows.get(i);
					if (min != null) {
						if (row != null && comparator.compare(row, min) < 0) {
							min = row;
							minIndex = i;
						}
					} else {
						if (row != null) {
							min = row;
							minIndex = i;
						}
					}
				}

				// Write minIndex row to file.
				if (minIndex >= 0) {
					// Remove duplicates if required.
					if (removeDuplicates || removeAllDuplicates) {
						if (pendingRow != null) {
							if (comparator.compare(pendingRow,
									(String[]) filerows.get(minIndex)) != 0) {
								// If not equal, write pending row to the
								// sorted file. However, drop if record is part
								// of a duplicate set and removeAllDuplicates is
								// true.
								if (!removeAllDuplicates || duplicateCount == 0) {
									writer.writeLine(pendingRow);
								} else {
									logDroppedRow(pendingRow, comparator);
								}
								duplicateCount = 0;
							} else {
								logDroppedRow(pendingRow, comparator);
								duplicateCount++;
							}
						}

						// Replace pending row.
						pendingRow = (String[]) filerows.get(minIndex);

					} else {
						// Write row to the sorted file.
						writer.writeLine((String[]) filerows.get(minIndex));
					}

					// Get another row from the file that had the min.
					eof = !((CSVReader) readers.get(minIndex)).readLine(values);
					if (!eof) {
						filerows.set(minIndex, (String[]) values
								.toArray(new String[values.size()]));
					} else {
						filerows.set(minIndex, null);
					}
				} else {
					someFileStillHasRows = false;
				}

				// Check if one still has rows.
				for (int i = 0; i < filerows.size(); i++) {
					someFileStillHasRows = false;
					if (filerows.get(i) != null) {
						if (minIndex < 0) {
							throw new RuntimeException(
									"mindex < 0 and found row not null"
											+ Arrays.toString((String[]) filerows
													.get(i)));
						}
						someFileStillHasRows = true;
						break;
					}
				}

				// Write the last one not covered above.
				if (!someFileStillHasRows) {
					for (int i = 0; i < filerows.size(); i++) {
						if (filerows.get(i) == null) {
							eof = !((CSVReader) readers.get(i))
									.readLine(values);
							if (!eof) {
								someFileStillHasRows = true;
								filerows.set(i, (String[]) values
										.toArray(new String[values.size()]));
							}
						}
					}
				}

				// Write the last pending row if any.
				if (!someFileStillHasRows && pendingRow != null) {
					if (!removeAllDuplicates || duplicateCount == 0) {
						writer.writeLine(pendingRow);
					} else {
						logDroppedRow(pendingRow, comparator);
					}
				}
			}

		} finally {
			// Close all the files.
			if (writer != null) {
				writer.close();
			}
			for (int i = 0; i < readers.size(); i++) {
				((CSVReader) readers.get(i)).close();
			}
		}
	}

	/**
	 * Log dropped row.
	 * 
	 * @param droppedRow
	 * @param comparator
	 */
	private void logDroppedRow(String[] droppedRow, Comparator comparator) {
		if (droppedRow != null) {
			// Build an array of record keys for
			// logging, if comparing with a RowComparator.
			if (comparator instanceof RowComparator) {
				SortCriterion[] sortCriteria = ((RowComparator) comparator)
						.getSortCriteria();
				String[] keys = new String[sortCriteria.length];
				for (int i = 0; i < sortCriteria.length; i++) {
					SortCriterion sortCriterion = sortCriteria[i];
					keys[i] = droppedRow[sortCriterion.getColumn()];
				}
				log.info("Dropping duplicate record identified by key(s): "
						+ Arrays.toString(keys));
			} else {
				// Log the entire record, by default.
				log.info("Dropping duplicate record: "
						+ Arrays.toString(droppedRow));
			}
		}
	}

}
