package com.taleo.integration.client.customstep.util.csv;

import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.math.NumberUtils;
import org.apache.log4j.Logger;

/**
 * Comparator for CSV rows. Rows are of type String[].
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class RowComparator implements Comparator {

	private static Logger log = Logger.getLogger(RowComparator.class);

	private SortCriterion[] sortCriteria;

	/**
	 * Constructor with an array of SortCriterion.
	 * 
	 * @param sortCriteria
	 *            The array of SortCriterion.
	 */
	public RowComparator(SortCriterion[] sortCriteria) {
		this.sortCriteria = sortCriteria;

		if (log.isDebugEnabled()) {
			log.debug("Sorting on: " + Arrays.toString(sortCriteria));
		}
	}

	/**
	 * Get the array of SortCriterion.
	 * 
	 * @return
	 */
	public SortCriterion[] getSortCriteria() {
		return sortCriteria;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Comparator#compare(T, T)
	 */
	public int compare(Object obj1, Object obj2) {
		if (obj1 instanceof String[]) {
			return compare((String[]) obj1, (String[]) obj2);
		} else if (obj1 instanceof List) {
			return compare((List) obj1, (List) obj2);
		} else {
			throw new UnsupportedOperationException(
					"Cannot compare arguments of the given type.");
		}
	}

	/**
	 * Compare rows as String[].
	 * 
	 * @param row1
	 * @param row2
	 * @return
	 */
	public int compare(String[] row1, String[] row2) {
		int compare = 0;
		String string1 = null;
		String string2 = null;

		// Compare successively on each criterion.
		for (int i = 0; i < sortCriteria.length; i++) {
			SortCriterion c = sortCriteria[i];
			try {
				string1 = row1[c.getColumn()];
			} catch (ArrayIndexOutOfBoundsException e) {
				throw new ArrayIndexOutOfBoundsException("Index "
						+ e.getMessage() + " is invalid for row "
						+ ArrayUtils.toString(row1) + ".");
			}
			try {
				string2 = row2[c.getColumn()];
			} catch (ArrayIndexOutOfBoundsException e) {
				throw new ArrayIndexOutOfBoundsException("Index "
						+ e.getMessage() + " is invalid for row "
						+ ArrayUtils.toString(row2) + ".");
			}
			if (c.isNumeric()) {
				Double double1 = NumberUtils.createDouble(string1);
				Double double2 = NumberUtils.createDouble(string2);
				if (double1.compareTo(double2) > 0) {
					compare = (c.isAscending() ? 1 : -1);
					break;
				} else if (double1.compareTo(double2) < 0) {
					compare = (c.isAscending() ? -1 : 1);
					break;
				}
			} else if (c.isIgnoreCase()) {
				if (string1.compareToIgnoreCase(string2) > 0) {
					compare = (c.isAscending() ? 1 : -1);
					break;
				} else if (string1.compareToIgnoreCase(string2) < 0) {
					compare = (c.isAscending() ? -1 : 1);
					break;
				}
			} else {
				if (string1.compareTo(string2) > 0) {
					compare = (c.isAscending() ? 1 : -1);
					break;
				} else if (string1.compareTo(string2) < 0) {
					compare = (c.isAscending() ? -1 : 1);
					break;
				}
			}
		}

		if (log.isTraceEnabled()) {
			String operator = " = ";
			if (compare < 0) {
				operator = " < ";
			} else if (compare > 0) {
				operator = " > ";
			}
			log.trace(Arrays.toString(row1) + operator + Arrays.toString(row2));
		}

		return compare;
	}

	/**
	 * Compare rows as List.
	 * 
	 * @param row1
	 * @param row2
	 * @return
	 */
	public int compare(List row1, List row2) {
		return compare(row1.toArray(new String[row1.size()]),
				row2.toArray(new String[row2.size()]));
	}
}
