package com.taleo.integration.client.customstep.util.xml;

import java.io.InputStream;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.NamespaceContext;
import javax.xml.xpath.XPathExpression;

import org.apache.log4j.Logger;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.Namespace;
import org.dom4j.io.SAXReader;

/**
 * Implementation of NamespaceContext that introspect an XML fragment to collect
 * the namespaces it contains.
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public class SimpleNamespaceContext implements NamespaceContext {

	private static Logger log = Logger.getLogger(SimpleNamespaceContext.class);

	// Namespaces indexed by prefix.
	private Map namespaces = new Hashtable();

	/**
	 * Constructor with the XML fragment.
	 * <p>
	 * Note that this namespace context cannot be used to resolve XPath
	 * expressions that use elements in the default namespace.
	 * 
	 * @param xml
	 *            The XML fragment as an InputStream.
	 */
	public SimpleNamespaceContext(InputStream xml) {
		this(xml, "");
	}

	/**
	 * Constructor with the XML fragment and the default namespace prefix.
	 * <p>
	 * Allows to set an explicit prefix for the default namespace. This is
	 * useful because {@link XPathExpression} evaluation does not work with the
	 * default namespace. This prefix can then be used in the XPath expression
	 * to evaluate instead of no namespace prefix.
	 * 
	 * @param xml
	 *            The XML fragment as an InputStream.
	 * @param defaultNamespacePrefix
	 *            The substitute prefix to use for the default namespace.
	 */
	public SimpleNamespaceContext(InputStream xml, String defaultNamespacePrefix) {

		// Collect all record elements in a list.
		SAXReader reader = new SAXReader();
		Document document;
		try {
			document = reader.read(xml);
		} catch (DocumentException e) {
			throw new RuntimeException(e);
		}
		List elements = document.selectNodes("//*");

		// Extract all namespaces.
		Iterator iter = elements.iterator();
		while (iter.hasNext()) {
			Element element = (Element) iter.next();
			Namespace ns = element.getNamespace();
			String nsPrefix = ns.getPrefix();
			if (ns != null) {
				// Use the substitute default namespace prefix.
				if (ns.getPrefix() == "") {
					nsPrefix = defaultNamespacePrefix.trim();
				}
				if (!namespaces.containsKey(nsPrefix)) {
					// Add the namespace.
					namespaces.put(nsPrefix, ns);

					if (log.isDebugEnabled()) {
						log.debug("Adding namespace 'xmlns"
								+ (nsPrefix.equals("") ? "" : ":") + nsPrefix
								+ "=" + ns.getURI() + "'");
					}
				}
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * javax.xml.namespace.NamespaceContext#getNamespaceURI(java.lang.String )
	 */
	public String getNamespaceURI(String prefix) {
		Namespace ns = (Namespace) namespaces.get(prefix);
		if (ns == null) {
			throw new RuntimeException("Undefined namespace prefix '" + prefix
					+ "'.");
		}
		return ns.getURI();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.xml.namespace.NamespaceContext#getPrefix(java.lang.String)
	 */
	public String getPrefix(String uri) {
		// This method isn't necessary for XPath processing.
		throw new UnsupportedOperationException(
				"This class was designed for XPath processing only.");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.xml.namespace.NamespaceContext#getPrefixes(java.lang.String)
	 */
	public Iterator getPrefixes(String uri) {
		// This method isn't necessary for XPath processing.
		throw new UnsupportedOperationException(
				"This class was designed for XPath processing only.");
	}
}
