package com.taleo.integration.client.customstep.util.xml;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.namespace.NamespaceContext;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import javax.xml.xpath.XPathVariableResolver;

import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * XPath Helper.
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public class XPathHelper {

	private NamespaceContext nsContext;

	private Map variables = new HashMap();

	/**
	 * Constructor without argument.
	 * <p>
	 * Note that this instance cannot evaluate XPath expressions that use any
	 * namespace.
	 */
	public XPathHelper() {
		this(new NamespaceContext() {
			public String getNamespaceURI(String arg0) {
				return null;
			}

			public String getPrefix(String arg0) {
				return null;
			}

			public Iterator getPrefixes(String arg0) {
				return null;
			}
		});
	}

	/**
	 * Constructor with the namespace context.
	 * <p>
	 * Note that this instance cannot evaluate XPath expressions that use the
	 * default namespace.
	 * 
	 * @param nsContext
	 *            The namespace context to evaluate the XPath expression.
	 */
	public XPathHelper(NamespaceContext nsContext) {
		if (nsContext == null) {
			throw new IllegalArgumentException("Argument '" + nsContext
					+ "' is required.");
		}
		this.nsContext = nsContext;
	}

	/**
	 * Set a variable to be resolved while evaluating XPath.
	 * 
	 * @param name
	 *            The variable name.
	 * @param value
	 *            The variable value.
	 */
	public void setVariable(String name, Object value) {
		variables.put(name, value);
	}

	/**
	 * Get a variable value.
	 * 
	 * @param name
	 *            The variable name.
	 * @return
	 */
	public Object getVariable(Object name) {
		return variables.get(name);
	}

	/**
	 * Evaluate an XPath expression on an XML fragment.
	 * 
	 * @param xml
	 *            The XML fragment as an InputStream.
	 * @param xPath
	 *            The XPathExpression as a String.
	 * @param returnType
	 *            The return type as defined in {@link XPathConstants}.
	 * @return The result of evaluating the XPath expression.
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws IOException
	 * @throws XPathExpressionException
	 */
	public Object evaluate(InputStream xml, String xPath, QName returnType)
			throws ParserConfigurationException, SAXException, IOException,
			XPathExpressionException {

		// Parse the XML stream.
		DocumentBuilderFactory dfactory = DocumentBuilderFactory.newInstance();
		dfactory.setNamespaceAware(true);
		DocumentBuilder docBuilder = dfactory.newDocumentBuilder();
		Node node = docBuilder.parse(new InputSource(xml));

		// Evaluate the XPath expression.
		XPathFactory factory = XPathFactory.newInstance();
		XPath _xPath = factory.newXPath();
		_xPath.setNamespaceContext(nsContext);
		_xPath.setXPathVariableResolver(new XPathVariableResolver() {
			public Object resolveVariable(QName name) {
				return getVariable(name.getLocalPart());
			}
		});
		XPathExpression xPathExpression = _xPath.compile(xPath);
		return xPathExpression.evaluate(node, returnType);
	}

	/**
	 * Strip the namespace prefixes from an XPath expression.
	 * <p>
	 * This is useful because {@link org.dom4j.io.SAXReader} does not work with
	 * namespaces.
	 * <p>
	 * For example, the XPath "/x:A/y:B/C" would result in "/A/B/C".
	 * 
	 * @param xPath
	 *            The XPath expression.
	 * @return
	 */
	public static String stripNamespacePrefixes(String xPath) {
		final Pattern pattern = Pattern.compile("([/\\(\\[])\\s*\\w+:(\\w+)");
		Matcher matcher = pattern.matcher(xPath);
		return matcher.replaceAll("$1$2");
	}

}
