package com.taleo.integration.client.customstep.util.xml;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.io.StringWriter;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * XML Utilities.
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public abstract class XmlUtils {

	/**
	 * Indicates if the specified file is a well formed XML document.
	 * 
	 * @param file
	 *            The file containing the document.
	 * @return
	 * @throws IOException
	 */
	public static boolean isWellFormed(File file) throws IOException {
		boolean isWellFormed = true;

		// Try to parse the XML document.
		try {
			SAXParserFactory factory = SAXParserFactory.newInstance();
			factory.newSAXParser().parse(file, new DefaultHandler());
		} catch (SAXException e) {
			isWellFormed = false;
		} catch (ParserConfigurationException e) {
			throw new RuntimeException(e);
		}
		return isWellFormed;
	}

	/**
	 * Indicates if the specified XML string is well formed.
	 * 
	 * @param xml
	 *            The XML string.
	 * @return
	 * @throws IOException
	 */
	public static boolean isWellFormed(String xml) throws IOException {
		boolean isWellFormed = true;

		// Try to parse the XML string.
		try {
			SAXParserFactory factory = SAXParserFactory.newInstance();
			InputSource source = new InputSource(new StringReader(xml));
			factory.newSAXParser().parse(source, new DefaultHandler());
		} catch (SAXException e) {
			isWellFormed = false;
		} catch (ParserConfigurationException e) {
			throw new RuntimeException(e);
		}
		return isWellFormed;
	}

	/**
	 * Create an empty XML Document.
	 * 
	 * @return
	 */
	public static Document createDomDocument() {
		try {
			DocumentBuilderFactory factory = createDocumentBuilderFactory();
			DocumentBuilder builder = factory.newDocumentBuilder();
			return builder.newDocument();
		} catch (ParserConfigurationException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Create an XML Document from an XML string.
	 * 
	 * @param xml
	 * @return
	 * @throws IOException
	 */
	public static Document createDomDocument(String xml) throws SAXException,
			IOException {
		return createDomDocument(new ByteArrayInputStream(xml.getBytes()));
	}

	/**
	 * Create an XML Document from an InputStream.
	 * 
	 * @param is
	 * @return
	 * @throws IOException
	 */
	public static Document createDomDocument(InputStream is) throws IOException {
		try {
			DocumentBuilderFactory factory = createDocumentBuilderFactory();
			DocumentBuilder builder = factory.newDocumentBuilder();
			return builder.parse(is);
		} catch (SAXException e) {
			throw new RuntimeException(e);
		} catch (ParserConfigurationException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Create a DocumentBuilderFactory.
	 * 
	 * @return
	 */
	private static DocumentBuilderFactory createDocumentBuilderFactory() {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setNamespaceAware(true);
		return factory;
	}

	/**
	 * Write a DOM document to a file.
	 * 
	 * @param document
	 * @param file
	 * @throws IOException
	 */
	public static void writeXml(Document document, File file)
			throws IOException {
		OutputStream os = null;
		try {
			os = new FileOutputStream(file);
			writeXml(document, os);
		} finally {
			if (os != null) {
				os.close();
			}
		}
	}

	/**
	 * Write a DOM document to an OutputStream.
	 * 
	 * @param document
	 * @param os
	 * @throws IOException
	 */
	public static void writeXml(Document document, OutputStream os)
			throws IOException {
		try {
			Source source = new DOMSource(document);
			Result result = new StreamResult(os);
			Transformer transformer = TransformerFactory.newInstance()
					.newTransformer();
			transformer.setOutputProperty(
					javax.xml.transform.OutputKeys.INDENT, "yes");
			transformer.setOutputProperty(
					"{http://xml.apache.org/xslt}indent-amount", "2");
			transformer.transform(source, result);
		} catch (TransformerException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Convert a DOM node to an XML string.
	 * 
	 * @param node
	 * @param indent
	 * @return
	 */
	public static String toXml(Node node, int indent) {
		try {
			TransformerFactory factory = TransformerFactory.newInstance();
			Transformer transformer = factory.newTransformer();
			if (indent > 0) {
				transformer.setOutputProperty(
						javax.xml.transform.OutputKeys.INDENT, "yes");
				transformer.setOutputProperty(
						"{http://xml.apache.org/xslt}indent-amount",
						Integer.toString(indent));
			}
			StringWriter writer = new StringWriter();
			Result result = new StreamResult(writer);
			Source source = new DOMSource(node);
			transformer.transform(source, result);
			writer.close();
			return writer.toString();
		} catch (TransformerException e) {
			throw new RuntimeException(e);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Convert a DOM node to an XML string.
	 * 
	 * @param node
	 * @return
	 */
	public static String toXml(Node node) {
		return toXml(node, 0);
	}

	/**
	 * Strip invalid XML characters from a string.
	 * 
	 * @param text
	 * @return
	 */
	public static String stripInvalidXMLCharacters(String text) {
		return text
				.replaceAll(
						"[^\\u0009\\u000A\\u000D\\u0020-\\uD7FF\\uE000-\\uFFFD\\u10000-\\u10FFFF]",
						"");
	}
}
