package com.taleo.integration.client.customstep.velocity;

import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import com.taleo.integration.client.customstep.Constants;
import com.taleo.integration.client.customstep.util.csv.CSVReader;
import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.util.IOHelper;

/**
 * Class to expose a CSV export results file properties and content into
 * Velocity templates.
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public class CsvExportFileTool extends FileTool {

	/**
	 * The export results file.
	 */
	private File file;

	/**
	 * The CSV header presence indicator.
	 */
	private boolean headerPresent;

	/**
	 * The CSV delimiter.
	 */
	private char csvDelimiter;

	/**
	 * The CSV quote character.
	 */
	private char csvQuoteCharacter;

	/**
	 * The target file.
	 */
	private File targetFile;

	/**
	 * The line count.
	 */
	private int lineCount = 0;

	/**
	 * Indicates if the fields have been initialized.
	 */
	private boolean initialized = false;

	/**
	 * The encoding of the file.
	 */
	private String encoding = Constants.UTF8;

	/**
	 * Constructor with the export file, CSV header indicator, CSV delimiter,
	 * quotation character and encoding.
	 * 
	 * @param file
	 *            The export file.
	 * @param encoding
	 *            The encoding.
	 */
	public CsvExportFileTool(File file, boolean headerPresent,
			char csvDelimiter, char csvQuoteCharacter, String encoding) {
		this.file = file;
		this.headerPresent = headerPresent;
		this.csvDelimiter = csvDelimiter;
		this.csvQuoteCharacter = csvQuoteCharacter;
		this.encoding = encoding;
	}

	/**
	 * Constructor with the export file, CSV header indicator, CSV delimiter,
	 * quotation character, encoding and the target file.
	 * <p>
	 * If specified, the targetFile is used to provide the file name and path in
	 * place of the file itself which is used for the contents. This is useful
	 * when the tool is used in the process of writing the target file and the
	 * name/path must be written within.
	 * 
	 * @param file
	 *            The export file.
	 * @param headerPresent
	 *            Indicates if the CSV header is present in the file.
	 * @param csvDelimiter
	 *            CSV delimiter.
	 * @param csvQuoteCharacter
	 *            CSV quote character.
	 * @param encoding
	 *            The encoding.
	 * @param targetFile
	 *            The target file.
	 */
	public CsvExportFileTool(File file, boolean headerPresent,
			char csvDelimiter, char csvQuoteCharacter, String encoding,
			File targetFile) {
		this(file, headerPresent, csvDelimiter, csvQuoteCharacter, encoding);
		this.targetFile = targetFile;
	}

	/**
	 * Return the file name including the extension.
	 * 
	 * @return The file name.
	 */
	public String name() {
		if (targetFile != null) {
			return targetFile.getName();
		}
		return file.getName();
	}

	/**
	 * Return the file absolute path.
	 * 
	 * @return The file path.
	 */
	public String path() {
		if (targetFile != null) {
			return targetFile.getAbsolutePath();
		}
		return file.getAbsolutePath();
	}

	/**
	 * Return the file last modified date.
	 * 
	 * @return The file last modified date.
	 */
	public Date lastModified() {
		return new Date(file.lastModified());
	}

	/**
	 * Return the file size.
	 * 
	 * @return The file size in bytes.
	 */
	public long size() {
		return file.length();
	}

	/**
	 * Return the file content.
	 * 
	 * @return The file content.
	 * @throws IOException
	 */
	public String content() throws IOException {
		return content(file.getAbsolutePath(), encoding);
	}

	/**
	 * Return the line count including the header row if present.
	 * 
	 * @return The line count.
	 * @throws IOException
	 */
	public int lineCount() throws IOException {
		if (!initialized) {
			initialize();
		}
		return lineCount;
	}

	/**
	 * Return the record count excluding the header row if present.
	 * 
	 * @return The record count.
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	public int recordCount() throws IOException {
		int recordCount = lineCount();
		// If the CSV header is present, remove one line.
		if (headerPresent) {
			recordCount--;
		}
		return recordCount;
	}

	/**
	 * Initialize the line count from the file.
	 * 
	 * @throws IOException
	 */
	private void initialize() throws IOException {
		CSVReader reader = null;

		try {
			reader = new CSVReader(new InputStreamReader(
					IOHelper.getInputStream(file), encoding), false,
					csvDelimiter, csvQuoteCharacter);
			List values = new ArrayList();
			while (reader.readLine(values)) {
				lineCount++;
			}
			initialized = true;

		} catch (InvalidCSVFileFormat e) {
			throw new RuntimeException(e);
		} finally {
			if (reader != null) {
				reader.close();
			}
		}
	}
}
