package com.taleo.integration.client.customstep.velocity;

import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;

import com.taleo.integration.client.customstep.Constants;
import com.taleo.integration.client.customstep.util.csv.CSVReader;
import com.taleo.integration.client.customstep.util.csv.CsvWriter;
import com.taleo.integration.client.customstep.util.csv.InvalidCSVFileFormat;
import com.taleo.integration.client.util.IOHelper;

/**
 * Class to expose a CSV import results file properties and content into
 * Velocity templates.
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public class CsvImportFileTool extends FileTool {

	/**
	 * The index of columns in the results.
	 */
	private final static int STATUS_INDEX = 2;
	private final static int TRANSACTION_TYPE_INDEX = 3;

	/**
	 * The value for success status.
	 */
	private final static String STATUS_SUCCESS = "success";

	/**
	 * The value for failure status.
	 */
	private final static String STATUS_ERROR = "error";

	/**
	 * The value for create transaction type.
	 */
	private final static String TYPE_CREATE = "create";

	/**
	 * The value for update transaction type.
	 */
	private final static String TYPE_UPDATE = "update";

	/**
	 * The value for delete transaction type.
	 */
	private final static String TYPE_DELETE = "delete";

	/**
	 * The CSV import results file.
	 */
	private File file;

	/**
	 * The CSV header presence indicator.
	 */
	private boolean csvHeaderPresent;

	/**
	 * The CSV delimiter.
	 */
	private char csvDelimiter;

	/**
	 * The CSV quote character.
	 */
	private char csvQuoteCharacter;

	/**
	 * The target file.
	 */
	private File targetFile;

	/**
	 * The header row.
	 */
	private String csvHeader;

	/**
	 * The errors results in a list.
	 */
	private List errors = new ArrayList();

	/**
	 * The encoding of the file.
	 */
	private String encoding = Constants.UTF8;

	/**
	 * The counters taken from the file.
	 */
	private int lineCount = 0;

	private int successCount = 0;

	private int createCount = 0;

	private int updateCount = 0;

	private int deleteCount = 0;

	/**
	 * Indicates if the fields have been initialized.
	 */
	private boolean initialized = false;

	/**
	 * Constructor with the import results file, CSV header indicator, CSV
	 * delimiter, quotation character and encoding.
	 * 
	 * @param file
	 *            The export file.
	 * @param headerPresent
	 *            Indicates if the CSV header is present in the file.
	 * @param csvDelimiter
	 *            The CSV delimiter.
	 * @param csvQuoteCharacter
	 *            The CSV quote character.
	 * @param encoding
	 *            The encoding.
	 */
	public CsvImportFileTool(File file, boolean headerPresent,
			char csvDelimiter, char csvQuoteCharacter, String encoding) {
		this.file = file;
		this.csvHeaderPresent = headerPresent;
		this.csvDelimiter = csvDelimiter;
		this.csvQuoteCharacter = csvQuoteCharacter;
		this.encoding = encoding;
	}

	/**
	 * Constructor with the import results file, CSV header indicator, CSV
	 * delimiter, quotation character, encoding and the target file.
	 * <p>
	 * If specified, the targetFile is used to provide the file name and path in
	 * place of the file itself which is used for the contents. This is useful
	 * when the tool is used in the process of writing the target file and the
	 * name/path must be written within.
	 * 
	 * @param file
	 *            The export file.
	 * @param headerPresent
	 *            Indicates if the CSV header is present in the file.
	 * @param csvDelimiter
	 *            The CSV delimiter.
	 * @param csvQuoteCharacter
	 *            The CSV quote character.
	 * @param encoding
	 *            The encoding.
	 * @param targetFile
	 *            The target file.
	 */
	public CsvImportFileTool(File file, boolean headerPresent,
			char csvDelimiter, char csvQuoteCharacter, String encoding,
			File targetFile) {
		this(file, headerPresent, csvDelimiter, csvQuoteCharacter, encoding);
		this.targetFile = targetFile;
	}

	/**
	 * Return the file name including the extension.
	 * 
	 * @return The file name.
	 */
	public String name() {
		if (targetFile != null) {
			return targetFile.getName();
		}
		return file.getName();
	}

	/**
	 * Return the file absolute path.
	 * 
	 * @return The file path.
	 */
	public String path() {
		if (targetFile != null) {
			return targetFile.getAbsolutePath();
		}
		return file.getAbsolutePath();
	}

	/**
	 * Return the file last modified date.
	 * 
	 * @return The file last modified date.
	 */
	public Date lastModified() {
		return new Date(file.lastModified());
	}

	/**
	 * Return the file size.
	 * 
	 * @return The file size in bytes.
	 */
	public long size() {
		return file.length();
	}

	/**
	 * Return the file content.
	 * 
	 * @return The file content.
	 * @throws IOException
	 */
	public String content() throws IOException {
		return content(file.getAbsolutePath(), encoding);
	}

	/**
	 * Return the line count including the header row if present.
	 * 
	 * @return The line count.
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	public int lineCount() throws IOException {
		if (!initialized) {
			initialize();
		}
		return lineCount;
	}

	/**
	 * Return the record count excluding the header row if present.
	 * 
	 * @return The record count.
	 * @throws IOException
	 * @throws InvalidCSVFileFormat
	 */
	public int recordCount() throws IOException {
		int recordCount = lineCount();
		// If the CSV header is present, remove one line.
		if (csvHeaderPresent) {
			recordCount--;
		}
		return recordCount;
	}

	/**
	 * Return the success count.
	 * 
	 * @return The success count.
	 * @throws IOException
	 */
	public int successCount() throws IOException {
		if (!initialized) {
			initialize();
		}
		return successCount;
	}

	/**
	 * Return the create count.
	 * 
	 * @return The create count.
	 * @throws IOException
	 */
	public int createCount() throws IOException {
		if (!initialized) {
			initialize();
		}
		return createCount;
	}

	/**
	 * Return the update count.
	 * 
	 * @return The update count.
	 * @throws IOException
	 */
	public int updateCount() throws IOException {
		if (!initialized) {
			initialize();
		}
		return updateCount;
	}

	/**
	 * Return the delete count.
	 * 
	 * @return The delete count.
	 * @throws IOException
	 */
	public int deleteCount() throws IOException {
		if (!initialized) {
			initialize();
		}
		return deleteCount;
	}

	/**
	 * Return the CSV header row if present.
	 * 
	 * @return
	 */
	public String headerRow() {
		return this.csvHeader;
	}

	/**
	 * Return the error count.
	 * 
	 * @return The error count.
	 * @throws IOException
	 */
	public int errorCount() throws IOException {
		return errorCount(null);
	}

	/**
	 * Return the error count, excluding the ones where the given regex pattern
	 * is found at least once.
	 * <p>
	 * Please make sure to follow regex syntax, taking into consideration that
	 * some characters may need to be escaped. Also, multiple exclusions can be
	 * separated with the | character meaning OR in regex.
	 * 
	 * @param excludePattern
	 *            The regex exclude pattern.
	 * @return The error count.
	 * @throws IOException
	 * @see java.util.regex.Pattern
	 */
	public int errorCount(String excludePattern) throws IOException {
		if (!initialized) {
			initialize();
		}
		Pattern pattern = null;
		if (StringUtils.isNotBlank(excludePattern)) {
			pattern = Pattern.compile(excludePattern, Pattern.DOTALL);
		}
		int errorCount = 0;
		for (Iterator iterator = errors.iterator(); iterator.hasNext();) {
			String error = (String) iterator.next();
			if (StringUtils.isBlank(excludePattern)
					|| !pattern.matcher(error).find()) {
				errorCount++;
			}
		}
		return errorCount;
	}

	/**
	 * Return the error content.
	 * 
	 * @return The error content.
	 * @throws IOException
	 */
	public String errorContent() throws IOException {
		return errorContent(null);
	}

	/**
	 * Return the error content, excluding the ones where the given regex
	 * pattern is found at least once.
	 * <p>
	 * Please make sure to follow regex syntax, taking into consideration that
	 * some characters may need to be escaped. Also, multiple exclusions can be
	 * separated with the | character meaning OR in regex.
	 * 
	 * @param excludePattern
	 *            The regex exclude pattern.
	 * @return
	 * @throws IOException
	 * @see java.util.regex.Pattern
	 */
	public String errorContent(String excludePattern) throws IOException {
		if (!initialized) {
			initialize();
		}
		StringBuilder sb = new StringBuilder();
		if (csvHeaderPresent) {
			sb.append(this.csvHeader);
		}
		Pattern pattern = null;
		if (StringUtils.isNotBlank(excludePattern)) {
			pattern = Pattern.compile(excludePattern, Pattern.DOTALL);
		}
		for (Iterator iterator = errors.iterator(); iterator.hasNext();) {
			String error = (String) iterator.next();
			if (StringUtils.isBlank(excludePattern)
					|| !pattern.matcher(error).find()) {
				sb.append(error);
			}
		}
		return sb.toString();
	}

	/**
	 * Initialize the content fields and counters from the file.
	 * 
	 * @throws IOException
	 */
	private void initialize() throws IOException {
		CSVReader reader = null;
		try {
			reader = new CSVReader(new InputStreamReader(
					IOHelper.getInputStream(file), encoding), false,
					csvDelimiter, csvQuoteCharacter);
			List values = new ArrayList();
			if (csvHeaderPresent) {
				lineCount++;
				reader.readLine(values);
				this.csvHeader = getCsvString(values);
			}
			while (reader.readLine(values)) {
				lineCount++;

				// Get the status and type of transaction.
				String status = (String) values.get(STATUS_INDEX);
				String type = (String) values.get(TRANSACTION_TYPE_INDEX);
				int index = type.lastIndexOf('.');
				type = type.substring(index + 1);

				// Increment the counters.
				if (status.equals(STATUS_SUCCESS)) {
					successCount++;
					if (type.equals(TYPE_CREATE)) {
						createCount++;
					} else if (type.equals(TYPE_UPDATE)) {
						updateCount++;
					} else if (type.equals(TYPE_DELETE)) {
						deleteCount++;
					}
				} else if (status.equals(STATUS_ERROR)) {
					this.errors.add(getCsvString(values));
				}
			}

			initialized = true;

		} catch (InvalidCSVFileFormat e) {
			throw new RuntimeException(e);
		} finally {
			if (reader != null) {
				reader.close();
			}
		}
	}

	/**
	 * Get the CSV string for the provided field values.
	 * 
	 * @param values
	 * @return
	 * @throws IOException
	 */
	private String getCsvString(List values) throws IOException {
		StringWriter stringWriter = null;
		CsvWriter csvWriter = null;
		try {
			stringWriter = new StringWriter();
			csvWriter = new CsvWriter(stringWriter, csvDelimiter,
					csvQuoteCharacter);
			csvWriter.writeLine(values);
			return stringWriter.toString();
		} finally {
			csvWriter.close();
		}
	}
}
