package com.taleo.integration.client.customstep.velocity;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Date;
import java.util.Iterator;

import javax.xml.namespace.NamespaceContext;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.xml.sax.InputSource;

import com.taleo.integration.client.customstep.Constants;

/**
 * Class to expose an XML export results file properties and content into
 * Velocity templates.
 * 
 * @author Romain Guay - Taleo Corporation
 * 
 */
public class XmlExportFileTool extends FileTool {

	/**
	 * The export results file.
	 */
	private File file;

	/**
	 * Constructor with the export file.
	 * 
	 * @param file
	 *            The export file.
	 */
	public XmlExportFileTool(File file) {
		this.file = file;
	}

	/**
	 * Return the file name. This is the last name in the path sequence,
	 * including the extension.
	 * 
	 * @return The file name.
	 */
	public String name() {
		return file.getName();
	}

	/**
	 * Return the file absolute path.
	 * 
	 * @return The file path.
	 */
	public String path() {
		return file.getAbsolutePath();
	}

	/**
	 * Return the file last modified date.
	 * 
	 * @return The file last modified date.
	 */
	public Date lastModified() {
		return new Date(file.lastModified());
	}

	/**
	 * Return the file size.
	 * 
	 * @return The file size in bytes.
	 */
	public long size() {
		return file.length();
	}

	/**
	 * Return the file content.
	 * 
	 * @return The file content.
	 * @throws IOException
	 */
	public String content() throws IOException {
		return content(file.getAbsolutePath());
	}

	/**
	 * Return the line count including the header row if present.
	 * 
	 * @return The line count.
	 * @throws IOException
	 */
	public int lineCount() throws IOException {
		return lineCount(file.getAbsolutePath());
	}

	/**
	 * Return the record count.
	 * 
	 * @return The record count.
	 * @throws IOException
	 */
	public int recordCount() throws IOException {
		// Build the namespace context.
		NamespaceContext nsContext = new NamespaceContext() {
			public String getNamespaceURI(String prefix) {
				if ("itk".equals(prefix)) {
					return Constants.ITK_NAMESPACE;
				} else {
					return "";
				}
			}

			// This method isn't necessary for XPath processing.
			public String getPrefix(String uri) {
				throw new UnsupportedOperationException();
			}

			// This method isn't necessary for XPath processing.
			public Iterator getPrefixes(String uri) {
				throw new UnsupportedOperationException();
			}
		};

		// Get the record count with XPath.
		int recordCount = 0;
		XPathFactory factory = XPathFactory.newInstance();
		XPath xPath = factory.newXPath();
		xPath.setNamespaceContext(nsContext);
		InputSource inputSource = new InputSource(new FileInputStream(file));
		try {
			XPathExpression xPathExpression = xPath
					.compile("count(//itk:ExportTXML/*) + count(//itk:ExportXML/*)");
			recordCount = Integer.parseInt(xPathExpression
					.evaluate(inputSource));
		} catch (XPathExpressionException e) {
			throw new RuntimeException(e);
		}

		return recordCount;
	}
}
