package com.taleo.integration.client.customstep.xml;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.dom4j.DocumentException;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.ElementHandler;
import org.dom4j.ElementPath;
import org.dom4j.io.DocumentSource;
import org.dom4j.io.SAXReader;
import org.xml.sax.SAXException;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.xml.XPathHelper;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.MessageFormat;

/**
 * Post-processing step to apply an XSL transformation to the result.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>xslFile</b>: The filename of the XSL transformation. This value may be
 * supplied as a relative or absolute path.</li>
 * <br>
 * <li><b>outputFile</b>: Optional - The filename for the output of the
 * transformation. If absent the output becomes the current document in the
 * execution pipeline. This value may be supplied as a relative or absolute
 * path.</li>
 * <br>
 * <li><b>outputFormat</b>: Optional - Since XSL can change the format of the
 * output, this parameter allows to set the output format. Must be one of
 * T_SOAP, T_XML, XML, CSV, TEXT, OTHER. Default: Same format as what was
 * received as input.</li>
 * <br>
 * <br>
 * <p>
 * If the XML file is very large, the following parameters can be used to apply
 * XSL over individual records within the file and merging the result using a
 * template.
 * <li><b>recordXPath</b>: Optional - The XPath expression leading to the
 * individual records to apply XSL. This parameter is insensible to namespace
 * prefixes, so they can be provided or not. This XPath also does not support
 * wildcards, but it supports the | to enter multiple alternative XPaths.<br>
 * Example:
 * <ul>
 * <li>
 * /Document/Content/ExportXML/Candidate | /SomeOtherPath</li>
 * <li>
 * /m:submitDocument/Document/Content/ImportEntities/UDSElement-merge |
 * /m:submitDocument/Document/Content/ImportEntities/UDSElement-update</li>
 * </ul>
 * </li>
 * <li><b>templateFile</b>: Optional - File containing the XML template with the
 * [RECORD_OUTPUT] token to identify where to write records after the
 * transformation. This value may be supplied as a relative or absolute path. If
 * absent while the recordXPath is present, the following template will be used
 * by default.
 * 
 * <pre>
 * &lt;XML&gt;
 * [RECORD_OUTPUT]
 * &lt;/XML&gt;
 * </pre>
 * 
 * <li>All other parameters are passed as arguments to the XSL transformation.</li>
 * </ul>
 * <p>
 * This step differs from the TCC built-in XSL transformation step in four ways:
 * <ul>
 * <li>It applies either to the whole XML document or to individual records. The
 * TCC built-in version applies only the second strategy.</li>
 * <br>
 * <li>The record XPath can be set explicitly when looping over records, while
 * in the TCC version, the XPath is static.</li>
 * <br>
 * <li>It allows to set a different output than the execution pipeline. This
 * allows for example, to create a dynamic template with values from the
 * document header, for a subsequent TCC built-in XSL transformation.</li>
 * <br>
 * <li>It allows to define parameters for the XSL transformation.</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * <li><a href="http://dom4j.org/dom4j-1.6.1/">dom4j library</a>
 * (dom4j-1.6.1.jar)</li>
 * </ul>
 * It also depends on the following library if the record looping strategy is
 * used.
 * <ul>
 * <li><a href="http://jaxen.codehaus.org/">jaxen library</a> (jaxen-1.1.jar)</li>
 * </ul>
 * 
 * @author Romain Guay, Taleo Corporation
 */
public abstract class AbstractXslStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(AbstractXslStep.class);

	/**
	 * Constructor without arguments.
	 */
	public AbstractXslStep() {
		super();
		// Register valid parameters.
		registerParameterName("xslFile");
		registerParameterName("outputFile");
		registerParameterName("outputFormat");
		registerParameterName("recordXPath");
		registerParameterName("templateFile");
	}

	/**
	 * Validate the parameters names.
	 * <p>
	 * Override the default method to allow for extra parameters to be passed to
	 * the XSL transformation.
	 */
	protected void validateParameterNames() {
		Enumeration keys = parameters.keys();
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			if (!parameterNames.contains(key)) {
				log.info("Using XSL parameter " + key + "="
						+ parameters.getProperty(key));
			}
		}
	}

	/**
	 * The XSL file name.
	 */
	protected String xslFile;

	/**
	 * The output file name.
	 */
	protected String outputFile;

	/**
	 * The output format.
	 */
	protected MessageFormat.Enum outputFormat;

	/**
	 * The XSL parameters.
	 */
	protected Properties xslParameters = new Properties();

	/**
	 * The record XPath.
	 */
	protected String recordXPath;

	/**
	 * The template filename.
	 */
	protected String templateFile;

	/**
	 * Get the XSL file.
	 * 
	 * @return
	 */
	public String getXslFile() {
		return xslFile;
	}

	/**
	 * Set the XSL file.
	 * 
	 * @param xslFile
	 */
	public void setXslFile(String xslFile) {
		this.xslFile = xslFile;
	}

	/**
	 * Get the output file.
	 * 
	 * @return
	 */
	public String getOutputFile() {
		return outputFile;
	}

	/**
	 * Set the output file.
	 * 
	 * @param outputFile
	 */
	public void setOutputFile(String outputFile) {
		this.outputFile = outputFile;
	}

	/**
	 * Get the output format.
	 * 
	 * @return
	 */
	public MessageFormat.Enum getOutputFormat() {
		return outputFormat;
	}

	/**
	 * Set the output format.
	 * 
	 * @param outputFormat
	 */
	public void setOutputFormat(MessageFormat.Enum outputFormat) {
		this.outputFormat = outputFormat;
	}

	/**
	 * Get the XSL parameter by its name.
	 * 
	 * @param name
	 * 
	 * @return
	 */
	public String getXslParameter(String name) {
		return xslParameters.getProperty(name);
	}

	/**
	 * Set the XSL parameter by its name.
	 * 
	 * @param name
	 * @param value
	 */
	public void setXslParameter(String name, String value) {
		this.xslParameters.put(name, value);
	}

	/**
	 * Get the XPath expression for the records.
	 * 
	 * @return
	 */
	public String getRecordXPath() {
		return recordXPath;
	}

	/**
	 * Set the XPath expression for the records.
	 * 
	 * @param recordXPath
	 */
	public void setRecordXPath(String recordXPath) {
		this.recordXPath = recordXPath;
	}

	/**
	 * Get the template filename.
	 * 
	 * @return
	 */
	public String getTemplateFile() {
		return templateFile;
	}

	/**
	 * Set the template filename.
	 * 
	 * @param templateFile
	 */
	public void setTemplateFile(String templateFile) {
		this.templateFile = templateFile;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		setXslFile(parameters.getProperty("xslFile"));
		if (StringUtils.isBlank(xslFile)) {
			throw new IllegalArgumentException(
					"The 'xslFile' parameter is mandatory.");
		}

		setOutputFile(parameters.getProperty("outputFile"));

		String outputFormat = parameters.getProperty("outputFormat");
		if (StringUtils.isNotBlank(outputFormat)) {
			setOutputFormat(MessageFormat.Enum.forString(outputFormat));
		}

		setRecordXPath(parameters.getProperty("recordXPath"));
		setTemplateFile(parameters.getProperty("templateFile"));

		// Get the remaining parameters as XSL parameters.
		Enumeration keys = parameters.keys();
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			if (!parameterNames.contains(key)) {
				setXslParameter(key, parameters.get(key).toString());
			}
		}
	}

	/**
	 * Execute the XSL transformation.
	 * 
	 * @param inputFile
	 *            The input file.
	 * @param outputFile
	 *            The output file.
	 * @throws TransformerException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws DocumentException
	 */
	protected void execute(File inputFile, File outputFile)
			throws TransformerException, IOException,
			ParserConfigurationException, SAXException, DocumentException {

		File xsl = new File(xslFile);

		// Test for existing xsl file, because transform does not do it.
		if (!xsl.exists()) {
			throw new TransformerException("File not found: "
					+ xsl.getAbsolutePath());
		}

		// Ensure output folder exists.
		org.apache.commons.io.FileUtils.forceMkdir(outputFile.getParentFile());

		// Create an XSL transformer.
		TransformerFactory tFactory = TransformerFactory.newInstance();
		Transformer transformer = tFactory
				.newTransformer(new StreamSource(xsl));

		// Set parameters.
		Iterator it = xslParameters.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry entry = (Map.Entry) it.next();
			transformer.setParameter((String) entry.getKey(), entry.getValue());
		}

		// Get a reader on the input file.
		Reader reader = new InputStreamReader(
				IOHelper.getInputStream(inputFile), getEncoding());

		// Get a writer on the output file.
		Writer writer = new OutputStreamWriter(IOHelper.getOutputStream(
				new FileOutputStream(outputFile), outputFile.getName()),
				getEncoding());

		if (StringUtils.isBlank(recordXPath)) {
			// Prepare source and result.
			Source source = new StreamSource(reader);
			Result result = new StreamResult(writer);

			// Apply the transformation.
			transformer.transform(source, result);

		} else {
			// Get the template.
			Template template = null;
			if (StringUtils.isBlank(templateFile)) {
				template = new Template();
			} else {
				template = new Template(new File(templateFile));
			}

			// Write file header.
			writer.write(template.getHeader());

			// Create a SAXReader with a handler on the records.
			SAXReader saxReader = new SAXReader();
			RecordHandler handler = new RecordHandler(transformer, writer);

			String[] recordXPaths = recordXPath.split("\\|");
			for (int i = 0; i < recordXPaths.length; i++) {
				// Strip the namespace prefixes because SAXReader does not
				// support them.
				recordXPaths[i] = XPathHelper
						.stripNamespacePrefixes(recordXPaths[i]);
				// Add a handler for each XPath.
				saxReader.addHandler(recordXPaths[i].trim(), handler);
			}

			// Read the input file.
			saxReader.read(reader);

			// Write file footer.
			writer.write(template.getFooter());
		}

		// Close.
		reader.close();
		writer.close();
	}

	/**
	 * Handler to process individual records.
	 */
	private class RecordHandler implements ElementHandler {

		Transformer transformer;

		Writer writer;

		private int index = 0;

		/**
		 * Constructor
		 * 
		 * @param transformer
		 *            The XSL transformer.
		 * @param writer
		 *            The writer.
		 */
		public RecordHandler(Transformer transformer, Writer writer) {
			this.transformer = transformer;
			this.writer = writer;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.dom4j.ElementHandler#onStart(org.dom4j.ElementPath)
		 */
		public void onStart(ElementPath path) {
			// Increment the record index.
			index++;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.dom4j.ElementHandler#onEnd(org.dom4j.ElementPath)
		 */
		public void onEnd(ElementPath path) {

			Element record = path.getCurrent();
			String xml = record.asXML();
			if (log.isDebugEnabled()) {
				log.debug("Record Index=" + index + ": \n" + xml);
			}

			try {
				// Add a line termination.
				if (index > 1) {
					writer.write(System.getProperty("line.separator"));
				}

				// Create a document source on the record.
				DocumentSource source = new DocumentSource(
						DocumentHelper.parseText(xml));

				// Apply transformation.
				Result result = new StreamResult(writer);
				transformer.transform(source, result);

				writer.flush();
				record.detach();

			} catch (Exception e) {
				// Log the error, because its detail is not carried by SAX
				// parsing.
				log.error("", e);
				throw new RuntimeException(e);
			}
		}
	}

	/**
	 * Template class to help extract header and footer.
	 */
	private class Template {

		/**
		 * The token that needs to be present in the template file for the
		 * placeholder of the records data.
		 */
		public static final String TEMPLATE_TOKEN = "[RECORD_OUTPUT]";

		/**
		 * The template string.
		 */
		private String templateString;

		/**
		 * The token index.
		 */
		private int tokenIndex;

		/**
		 * Constructor without arguments. Uses the default template.
		 */
		public Template() {
			this.templateString = "<XML>\n" + TEMPLATE_TOKEN + "\n</XML>";
		}

		/**
		 * Constructor with the template file.
		 * 
		 * @param file
		 * @throws IOException
		 */
		public Template(File file) throws IOException {
			this.templateString = org.apache.commons.io.FileUtils
					.readFileToString(file, getEncoding());
			this.tokenIndex = templateString.indexOf(TEMPLATE_TOKEN);
			if (this.tokenIndex == -1) {
				throw new IllegalArgumentException(
						"The template must contain the token for records: "
								+ TEMPLATE_TOKEN);
			}
		}

		/**
		 * Get header.
		 * 
		 */
		public String getHeader() {
			return templateString.substring(0, tokenIndex);
		}

		/**
		 * Get footer.
		 */
		public String getFooter() {
			return templateString.substring(tokenIndex
					+ TEMPLATE_TOKEN.length());
		}
	}

}
