package com.taleo.integration.client.customstep.xml;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.zip.GZIPInputStream;

import javax.xml.namespace.NamespaceContext;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.dom4j.Element;
import org.dom4j.ElementHandler;
import org.dom4j.ElementPath;
import org.dom4j.io.SAXReader;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.tools.document.DocumentInfo;
import com.taleo.integration.client.customstep.tools.document.DocumentInfoBuilder;
import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.integration.client.customstep.util.xml.SimpleNamespaceContext;
import com.taleo.integration.client.customstep.util.xml.XPathHelper;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedMessage;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Post-processing step to extract attached files content in Base64_GZip format
 * and create files on disk.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and offers two
 * different APIs.
 * <p>
 * <ol>
 * <li><b>Primary API</b>
 * <p>
 * The primary API consumes an XML document with the following structure and
 * produces one file for each &lt;file&gt; tag.
 * 
 * <pre>
 * &lt;files&gt;
 *   &lt;file path="..."&gt;
 *     &lt;content&gt;... Base64_GZip content ...&lt;/content&gt;
 *   &lt;/file&gt;
 * &lt;/files&gt;
 * </pre>
 * 
 * Where:
 * <ul>
 * <li><b>path</b>: The complete path of the file to be created. Invalid
 * characters in filename will be replaced with an underscore "_".</li>
 * <li><b>content</b>: The Base64_GZip file content.</li>
 * <li>Any other tag or attribute can be added and will simply be ignored while
 * extracting files.</li>
 * <li>If &lt;content&gt; is the only tag inside &lt;file&gt;, it can be omitted
 * and the Base64_GZip content can be dropped directly inside &lt;file&gt;.</li>
 * </ul>
 * </li>
 * <p>
 * <li><b>Secondary API</b>
 * <p>
 * The second API uses the following parameters:
 * <ul>
 * <li><b>recordXPath</b>: The XPath expression leading to the individual
 * records that make up single files content. This parameter is insensible to
 * namespace prefixes, but you may want to follow the same general note on XPath
 * expressions below. <BR>
 * Example:
 * <ul>
 * <li>
 * <code>/default:Document/default:Content/default:ExportTXML/e:OnboardingProcess/e:ProcessDocuments/e:OnboardingProcessDocument</code>
 * </li>
 * </ul>
 * <br>
 * <li><b>filenameXPath</b>: The XPath expression based at record root and
 * leading to the filename (excluding path). The resulting filename is limited
 * to 180 characters and invalid characters in filename will be replaced with an
 * underscore "_". See general note on XPath expressions below.<BR>
 * Examples:
 * <ul>
 * <li><code>e:OnboardingProcessDocument/e:filename</code></li>
 * <li><code>concat(e:OnboardingProcessDocument/e:No, '.pdf')</code></li>
 * </ul>
 * The following variables can also be used:
 * <ul>
 * <li><b>$index</b>: The index of the record in the document.</li>
 * <li><b>$date</b>: The current date is established at the beginning of the
 * step. Consequently, the same date will be applied to all individual files.</li>
 * <li><b>$document_author</b>: The document author.</li>
 * <li><b>$document_creator</b>: The document creator.</li>
 * <li><b>$document_subject</b>: The document subject.</li>
 * <li><b>$document_title</b>: The document title.</li>
 * </ul>
 * Examples:
 * <ul>
 * <li><code>concat(e:Candidate/e:EmployeeID, $date, '.xml')</code></li>
 * <li><code>concat($date, '_', $index, '.xml')</code></li>
 * <li><code>$document_title</code></li>
 * </ul>
 * </li>
 * <br>
 * <li><b>dateFormat</b>: The format for the $date variable, default:
 * "yyyyMMdd_HHmmss". See {@link SimpleDateFormat} for details.</li>
 * <br>
 * <br>
 * <li><b>contentXPath</b>: The XPath expression based at record root and
 * leading to the Base64 file content. See general note on XPath expressions
 * below.<BR>
 * Example:
 * <ul>
 * <li><code>e:OnboardingProcessDocument/e:content</code></li>
 * </ul>
 * <br>
 * <li><b>outputDirectory</b>: <i>@deprecated - Use the outputDirectoryXPath
 * instead.</i> The directory to put the extracted files. It can be set as
 * absolute or relative to the current directory. <BR>
 * Example:
 * <ul>
 * <li><code>results</code></li>
 * </ul>
 * </li></li><br>
 * <li><b>outputDirectoryXPath</b>: The XPath expression that builds to
 * directory to put the result files. It can be set as absolute or relative to
 * the current directory. See general note on XPath expressions below.
 * <p>
 * It can use the same $date and $index variables as the filenameXPath.<BR>
 * Example:
 * <ul>
 * <li><code>'results'</code></li>
 * <li><code>concat('results/', e:Candidate/e:Number, '_', $date)</code></li>
 * </ul>
 * <p>
 * All <b>XPath expressions</b> must not use wildcards and must use the same
 * namespace prefixes that are declared in the document, except for the default
 * namespace for which the prefix 'default' must be used instead of no prefix.
 * This 'default' prefix may be replaced if needed, by setting the
 * 'com.taleo.integration.client.xml.default.ns.prefix' system property.</li>
 * </ul></li>
 * </ol>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://dom4j.org/dom4j-1.6.1/">dom4j library</a>
 * (dom4j-1.6.1.jar)</li>
 * <li><a href="http://jaxen.codehaus.org/">jaxen library</a> (jaxen-1.1.jar)</li>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * If any of the $document_* variable is used, the following jars must also be
 * present in the classpath:
 * <ul>
 * <li><a
 * href="http://opensource.intarsys.de/home/en/index.php?n=OpenSource.JPod">JPod
 * PDF Library</a> (/lib/jPod.jar)</li>
 * <li><a
 * href="http://opensource.intarsys.de/home/en/index.php?n=OpenSource.JPod"
 * >intarsys runtime library</a> (/lib/isrt.jar)</li>
 * <li><a href="http://freshmeat.net/projects/mimeutil/">MimeUtil library</a>
 * (/lib/mime-util-1.3.2.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file, using the primary API:
 * <p>
 * <img src="doc-files/ExtractAttachedFilePostStep-1.jpg" />
 * <p>
 * Here are examples of how to define this custom step in the TCC configuration
 * file, using the secondary API:
 * <p>
 * <img src="doc-files/ExtractAttachedFilePostStep-2.jpg" />
 * <p>
 * <img src="doc-files/ExtractAttachedFilePostStep-3.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 * @author Sivakumaran Krishnan, Taleo Corporation
 * @deprecated Use
 *             {@link com.taleo.integration.client.customstep.xml.ExtractFilesPostStep}
 *             with operations parameter set to "DecodeBase64,Gunzip" instead.
 */
public class ExtractAttachedFilePostStep extends BaseCustomStep {

	private static Logger log = Logger
			.getLogger(ExtractAttachedFilePostStep.class);

	private static final String DEFAULT_RECORD_XPATH = "/files/file";

	private static final String DEFAULT_FILENAME_XPATH = "tokenize(/file/@path,'\\\\|/')[last()]";

	private static final String DEFAULT_OUTPUT_DIRECTORY_XPATH = "string-join(tokenize(/file/@path,'\\\\|/')[position() < last()],'"
			+ File.separatorChar + "')";

	private static final String DEFAULT_CONTENT_XPATH = "/file/content/text()|/file/text()";

	/**
	 * Constructor without arguments.
	 */
	public ExtractAttachedFilePostStep() {
		super();
		// Register valid parameters.
		registerParameterName("recordXPath");
		registerParameterName("filenameXPath");
		registerParameterName("dateFormat");
		registerParameterName("contentXPath");
		registerParameterName("outputDirectory");
		registerParameterName("outputDirectoryXPath");
	}

	/**
	 * The XPath expression to identify records.
	 */
	private String recordXPath = DEFAULT_RECORD_XPATH;

	/**
	 * The XPath expression to identify the file name.
	 */
	private String filenameXPath = DEFAULT_FILENAME_XPATH;

	/**
	 * The date format.
	 */
	private DateFormat dateFormat = new SimpleDateFormat("yyyyMMdd_HHmmss");

	/**
	 * The XPath expression to identify Base64_GZip file contents.
	 */
	private String contentXPath = DEFAULT_CONTENT_XPATH;

	/**
	 * The output directory.
	 * 
	 * @deprecated
	 */
	private String outputDirectory;

	/**
	 * The output directory XPath.
	 */
	private String outputDirectoryXPath = DEFAULT_OUTPUT_DIRECTORY_XPATH;

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "ExtractAttachedFilePostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to extract attached file contents in Base64_GZip format and create files from it.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "ExtractAttachedFilePostStep";
	}

	/**
	 * Get the XPath expression for the records.
	 * 
	 * @return
	 */
	public String getRecordXPath() {
		return recordXPath;
	}

	/**
	 * Set the XPath expression for the records.
	 * 
	 * @param recordXPath
	 */
	public void setRecordXPath(String recordXPath) {
		this.recordXPath = recordXPath;
	}

	/**
	 * Get the XPath expression for filename.
	 * 
	 * @return
	 */
	public String getFilenameXPath() {
		return filenameXPath;
	}

	/**
	 * Set the XPath expression for filename.
	 * 
	 * @param filenameXPath
	 */
	public void setFilenameXPath(String filenameXPath) {
		this.filenameXPath = filenameXPath;
	}

	/**
	 * Get the date format.
	 * 
	 * @return
	 */
	public DateFormat getDateFormat() {
		return dateFormat;
	}

	/**
	 * Set the date format.
	 * 
	 * @param dateFormat
	 */
	public void setDateFormat(DateFormat dateFormat) {
		this.dateFormat = dateFormat;
	}

	/**
	 * Get the XPath expression for file content.
	 * 
	 * @return
	 */
	public String getContentXPath() {
		return contentXPath;
	}

	/**
	 * Set the XPath expression for file content.
	 * 
	 * @param contentXPath
	 */
	public void setContentXPath(String contentXPath) {
		this.contentXPath = contentXPath;
	}

	/**
	 * Get the output directory.
	 * 
	 * @deprecated Use getOutputDirectoryXPath instead.
	 * @return
	 */
	public String getOutputDirectory() {
		return outputDirectory;
	}

	/**
	 * Set the output directory.
	 * 
	 * @deprecated Use setOutputDirectoryXPath instead.
	 * @param outputDirectory
	 */
	public void setOutputDirectory(String outputDirectory) {
		this.outputDirectory = outputDirectory;
	}

	/**
	 * Get the output directory XPath.
	 * 
	 * @return
	 */
	public String getOutputDirectoryXPath() {
		return outputDirectoryXPath;
	}

	/**
	 * Set the output directory XPath.
	 * 
	 * @param outputDirectoryXPath
	 */
	public void setOutputDirectoryXPath(String outputDirectoryXPath) {
		this.outputDirectoryXPath = outputDirectoryXPath;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// First, get the basic supported pipeline for a post-processing step
		// (this handles the constraints on the last step type)
		SupportedPipeline supportedPipeline = getPostProcessSupportPipeline();

		// Add the supported response message information
		SupportedMessage supportedMessage = supportedPipeline
				.addNewSupportedResponseMessage();

		// Add the supported response message format
		supportedMessage.addNewSupportedFormats();
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.T_XML);
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.XML);
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.T_SOAP);

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);
		
		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		setRecordXPath(parameters.getProperty("recordXPath",
				DEFAULT_RECORD_XPATH));
		if (StringUtils.isBlank(recordXPath)) {
			throw new IllegalArgumentException(
					"The 'recordXPath' parameter is mandatory.");
		}

		setFilenameXPath(parameters.getProperty("filenameXPath",
				DEFAULT_FILENAME_XPATH));
		if (StringUtils.isBlank(filenameXPath)) {
			throw new IllegalArgumentException(
					"The 'filenameXPath' parameter is mandatory.");
		}

		String dateFormat = parameters.getProperty("dateFormat");
		if (StringUtils.isNotBlank(dateFormat)) {
			setDateFormat(new SimpleDateFormat(dateFormat));
		}

		setContentXPath(parameters.getProperty("contentXPath",
				DEFAULT_CONTENT_XPATH));
		if (StringUtils.isBlank(contentXPath)) {
			throw new IllegalArgumentException(
					"The 'contentXPath' parameter is mandatory.");
		}

		setOutputDirectory(parameters.getProperty("outputDirectory"));
		setOutputDirectoryXPath(parameters.getProperty("outputDirectoryXPath",
				DEFAULT_OUTPUT_DIRECTORY_XPATH));
		if (StringUtils.isBlank(outputDirectory)
				&& StringUtils.isBlank(outputDirectoryXPath)) {
			throw new IllegalArgumentException(
					"The 'outputDirectoryXPath' parameter is mandatory.");
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the input file.
			Message responseMessage = pipeline.getResponseMessage();
			File inFile = null;
			if (responseMessage.isSetCurrentFile()) {
				inFile = new File(responseMessage.getCurrentFile());
			} else {
				inFile = new File(responseMessage.getOriginalFile());
			}

			// Execute.
			execute(inFile);

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Process the input file.
	 * 
	 * @param inputFile
	 *            The input file.
	 * @throws Exception
	 */
	public void execute(File inputFile) throws Exception {

		// Strip the namespace prefixes because SAXReader does not support them.
		String recordXPath = XPathHelper
				.stripNamespacePrefixes(this.recordXPath);

		// Create a SAXReader with a handler on the records.
		SAXReader reader = new SAXReader();

		// Create a handler for the file.
		ExtractAttachedFileElementHandler handler = new ExtractAttachedFileElementHandler();

		// Attach recordXPath to the handler
		reader.addHandler(recordXPath, handler);

		// Read the input file.
		reader.read(IOHelper.getInputStream(inputFile), getEncoding());
		log.info("A total of " + handler.getIndex()
				+ " records have been processed.");

		// Throw the exception if present.
		if (handler.getException() != null) {
			throw handler.getException();
		}
	}

	/**
	 * Handler to create a file with Base64 content inside each record.
	 */
	private class ExtractAttachedFileElementHandler implements ElementHandler {

		private final static String INDEX_VARIABLE = "index";

		private final static String DATE_VARIABLE = "date";

		private final static String DOCUMENT_VARIABLE = "document";

		private final static String AUTHOR_VARIABLE = DOCUMENT_VARIABLE
				+ "_author";

		private final static String CREATOR_VARIABLE = DOCUMENT_VARIABLE
				+ "_creator";

		private final static String SUBJECT_VARIABLE = DOCUMENT_VARIABLE
				+ "_subject";

		private final static String TITLE_VARIABLE = DOCUMENT_VARIABLE
				+ "_title";

		private Exception exception;

		private int index = 0;

		private Date date = new Date();

		private final String defaultNsPrefix = System
				.getProperty(
						"com.taleo.integration.client.xml.default.ns.prefix",
						"default");

		/**
		 * Constructor
		 * 
		 * @throws TransformerConfigurationException
		 */
		ExtractAttachedFileElementHandler()
				throws TransformerConfigurationException {
		}

		/**
		 * Get the first exception.
		 * 
		 * @return
		 */
		public Exception getException() {
			return exception;
		}

		/**
		 * Get the index of the last processed record.
		 * 
		 * @return
		 */
		public int getIndex() {
			return index;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.dom4j.ElementHandler#onStart(org.dom4j.ElementPath)
		 */
		public void onStart(ElementPath path) {
			// Increment the record index.
			index++;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.dom4j.ElementHandler#onEnd(org.dom4j.ElementPath)
		 */
		public void onEnd(ElementPath path) {
			FileOutputStream fos = null;
			InputStream input = null;

			try {
				// ####################################################
				// ### 1. Write record content to a temporary file. ###
				// ####################################################

				// Retrieve the current record element.
				Element record = path.getCurrent();
				String xml = record.asXML();
				if (log.isDebugEnabled()) {
					log.debug("Record Index=" + index + ": \n" + xml);
				}

				// Convert the record element to InputStream for parsing.
				byte[] bytes = xml.getBytes(getEncoding());
				input = new ByteArrayInputStream(bytes);

				// Free dom4j memory consumed by the record element.
				record.detach();

				// Create a namespace context and substitute the default ns
				// prefix.
				NamespaceContext nsContext = new SimpleNamespaceContext(input,
						defaultNsPrefix);

				// Prepare a XPathHelper with nsContext.
				XPathHelper helper = new XPathHelper(nsContext);

				// Parsing routine to get the content using XPath.
				String content;
				try {
					// Evaluate XPath to get the content.
					input.reset();
					content = (String) helper.evaluate(input,
							getContentXPath(), XPathConstants.STRING);

				} catch (XPathExpressionException e) {
					throw new RuntimeException(e);
				}

				if (content.length() == 0) {
					log.warn("File skipped because it has no content.");
					return;
				}

				// Create a temporary file.
				File temp = createTempFile();

				// Prepare the OutputStream for landing the content.
				fos = new FileOutputStream(temp);

				// Decode Base64 Content.
				byte[] zippedBytes = Base64.decodeBase64(content
						.getBytes(getEncoding()));

				// Unzip Base64 Content.
				InputStream unzipedStream = new GZIPInputStream(
						new ByteArrayInputStream(zippedBytes));

				// Copy the Decoded/Unzipped data to the temporary file.
				org.apache.commons.io.IOUtils.copy(unzipedStream, fos);
				fos.close();

				// #####################################################
				// ### 2. Resolve the output directory and filename. ###
				// #####################################################

				// Add date and index variables to XPathHelper.
				helper.setVariable(DATE_VARIABLE, dateFormat.format(this.date));
				helper.setVariable(INDEX_VARIABLE, String.valueOf(this.index));

				// Determine if the filenameXPath or the outputDirectoryXPath
				// use one of the $document variables.
				boolean useDocumentVariable = (StringUtils.contains(
						getFilenameXPath(), "$" + DOCUMENT_VARIABLE) || StringUtils
						.contains(getOutputDirectoryXPath(), "$"
								+ DOCUMENT_VARIABLE));

				if (useDocumentVariable) {
					// Create a DocumentInfoBuilder.
					DocumentInfoBuilder builder = new DocumentInfoBuilder();

					// Warn if $document is used and the file type is not
					// supported.
					if (useDocumentVariable && !builder.isSupported(temp)) {
						log.warn("This file type does not support using $document variables "
								+ "to build the filename or output directory. Default values will be used.");
					}

					// Create a document info on the file and add corresponding
					// variables.
					DocumentInfo info = builder.createDocumentInfo(temp);
					helper.setVariable(AUTHOR_VARIABLE, info.getAuthor());
					helper.setVariable(CREATOR_VARIABLE, info.getCreator());
					helper.setVariable(SUBJECT_VARIABLE, info.getSubject());
					helper.setVariable(TITLE_VARIABLE, info.getTitle());
				}

				// Parsing routine to get the filename and output directory
				// using XPath.
				String filename;
				File outputDirectory = null;
				try {
					// Evaluate XPath to get the filename.
					input.reset();
					filename = (String) helper.evaluate(input,
							getFilenameXPath(), XPathConstants.STRING);

					// Sanitize the filename.
					filename = FileUtils.sanitizeFilename(filename);

					// Evaluate XPath to get the outputDirectory.
					input.reset();
					if (StringUtils.isNotBlank(getOutputDirectory())) {
						outputDirectory = new File(getOutputDirectory());
					} else {
						outputDirectory = new File((String) helper.evaluate(
								input, getOutputDirectoryXPath(),
								XPathConstants.STRING));
					}

					if (!outputDirectory.exists()) {
						log.info("Creating output directory: "
								+ outputDirectory.getPath());
						org.apache.commons.io.FileUtils
								.forceMkdir(outputDirectory);
					}
				} catch (XPathExpressionException e) {
					throw new RuntimeException(e);
				}

				// ##################################################
				// ### 3. Move the file to its final destination. ###
				// ##################################################

				if (filename.length() > 0) {
					File file = new File(outputDirectory, filename);
					if (file.exists()) {
						org.apache.commons.io.FileUtils.forceDelete(file);
					}
					org.apache.commons.io.FileUtils.moveFile(temp, file);
					log.info("Extracted file: " + file);
				} else {
					log.info("No valid filename at index " + index + ".");
				}

			} catch (Exception e) {
				// Store the first exception.
				if (this.exception == null) {
					this.exception = e;
				}
				// Log all exceptions.
				log.error("Error processing record " + index + ".", e);

			} finally {
				try {
					if (fos != null) {
						fos.close();
					}
					if (input != null) {
						input.close();
					}
				} catch (IOException e) {
				}
			}
		}
	}
}
