package com.taleo.integration.client.customstep.xml;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.List;
import java.util.zip.GZIPInputStream;

import javax.xml.namespace.NamespaceContext;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;
import org.dom4j.Element;
import org.dom4j.ElementHandler;
import org.dom4j.ElementPath;
import org.dom4j.io.SAXReader;
import org.w3c.dom.Document;
import org.w3c.tidy.Tidy;
import org.xhtmlrenderer.pdf.ITextRenderer;
import org.xhtmlrenderer.resource.FSEntityResolver;
import org.xml.sax.SAXException;

import com.lowagie.text.DocumentException;
import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.integration.client.customstep.util.xml.SimpleNamespaceContext;
import com.taleo.integration.client.customstep.util.xml.XPathHelper;
import com.taleo.integration.client.customstep.util.xml.XmlUtils;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedMessage;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Post-processing step to extract files content and create those files on disk.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and the
 * following additional parameters:
 * <ul>
 * <li><b>operations</b>: Comma separated list of operations to be applied in
 * sequence to the file content. Currently supported operations are the
 * following:
 * <ul>
 * <li><i>DecodeBase64</i>: Decode Base64</li>
 * <li><i>Gunzip</i>: Gunzip</li>
 * <li><i>HtmlToPdf</i>: Convert HTML to PDF</li>
 * <li><i>XhtmlToPdf</i>: Convert XHTML to PDF</li>
 * </ul>
 * For example, if content is Base64 encoded GZip, the operations will be
 * "DecodeBase64,Gunzip". By default, no operation is performed.</li>
 * <br>
 * <br>
 * <li><b>failOnError</b>: Fail if one of the file errors out (Default: false).
 * If false, only an error log is written.</li>
 * </ul>
 * 
 * <p>
 * This step consumes an XML document with the following structure and produces
 * one file for each &lt;file&gt; tag.
 * 
 * <pre>
 * &lt;files&gt;
 *   &lt;file path="..."&gt;
 *     &lt;content&gt;... content ...&lt;/content&gt;
 *   &lt;/file&gt;
 * &lt;/files&gt;
 * </pre>
 * 
 * Where:
 * <ul>
 * <li><b>path</b>: The complete path of the file to be created. Invalid
 * characters in filename will be replaced with an underscore "_".</li>
 * <li><b>content</b>: The file content.</li>
 * <li>Any other tag or attribute can be added and will simply be ignored while
 * extracting files.</li>
 * <br>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://dom4j.org/dom4j-1.6.1/">dom4j library</a>
 * (dom4j-1.6.1.jar)</li>
 * <li><a href="http://jaxen.codehaus.org/">jaxen library</a> (jaxen-1.1.jar)</li>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * 
 * If using the HtmlToPdf or XhtmlToPdf operation, the following libraries will
 * also be required:
 * <ul>
 * <li><a href="http://code.google.com/p/flying-saucer/">Google Core
 * renderer</a> (core-renderer.jar)</li>
 * <li><a href="http://itextpdf.com/">iText</a> (iText-2.0.8.jar)</li>
 * </ul>
 * 
 * If using the HtmlToPdf operation, the following library will also be
 * required:
 * <ul>
 * <li><a href="http://jtidy.sourceforge.net/">JTidy</a> (jtidy-r938.jar)</li>
 * </ul>
 * 
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/ExtractFilesPostStep.png" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class ExtractFilesPostStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(ExtractFilesPostStep.class);

	private static final String RECORD_XPATH = "/files/file";

	/**
	 * Constructor without arguments.
	 */
	public ExtractFilesPostStep() {
		super();
		// Register valid parameters.
		registerParameterName("operations");
		registerParameterName("failOnError");
	}

	/**
	 * The list of operations.
	 */
	private String[] operations = new String[0];

	/**
	 * The failOnError flag.
	 */
	protected boolean failOnError = false;

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "ExtractFilesPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to extract files content and create files on disk.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "ExtractFilesPostStep";
	}

	/**
	 * Get the operations.
	 * 
	 * @return
	 */
	public String[] getOperations() {
		return operations;
	}

	/**
	 * Set the operations.
	 * 
	 * @param operations
	 *            The list of operations.
	 */
	public void setOperations(String[] operations) {
		for (int i = 0; i < operations.length; i++) {
			List list = Arrays.asList(new String[] { "DecodeBase64", "Gunzip",
					"HtmlToPdf", "XhtmlToPdf" });
			if (!list.contains(operations[i])) {
				throw new IllegalArgumentException(operations[i]
						+ " is not a valid operation.");
			}
		}
		this.operations = operations;
	}

	/**
	 * Get the FailOnError flag.
	 * 
	 * @return
	 */
	public boolean isFailOnError() {
		return failOnError;
	}

	/**
	 * Set the FailOnError flag.
	 * 
	 * @param failOnError
	 */
	public void setFailOnError(boolean failOnError) {
		this.failOnError = failOnError;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// First, get the basic supported pipeline for a post-processing step
		// (this handles the constraints on the last step type)
		SupportedPipeline supportedPipeline = getPostProcessSupportPipeline();

		// Add the supported response message information
		SupportedMessage supportedMessage = supportedPipeline
				.addNewSupportedResponseMessage();

		// Add the supported response message format
		supportedMessage.addNewSupportedFormats();
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.T_XML);
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.XML);
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.T_SOAP);

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		setOperations(parameters.getProperty("operations", "").split(","));

		failOnError = Boolean.valueOf(
				parameters.getProperty("failOnError", "false")).booleanValue();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the input file.
			Message responseMessage = pipeline.getResponseMessage();
			File inFile = null;
			if (responseMessage.isSetCurrentFile()) {
				inFile = new File(responseMessage.getCurrentFile());
			} else {
				inFile = new File(responseMessage.getOriginalFile());
			}

			// Execute.
			execute(inFile);

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Process the input file.
	 * 
	 * @param inputFile
	 *            The input file.
	 * @throws Exception
	 */
	public void execute(File inputFile) throws Exception {
		// Create a SAXReader with a handler on the records.
		SAXReader reader = new SAXReader();

		// Create a handler for the file.
		ExtractAttachedFileElementHandler handler = new ExtractAttachedFileElementHandler();

		// Attach recordXPath to the handler
		reader.addHandler(RECORD_XPATH, handler);

		// Read the input file.
		reader.read(IOHelper.getInputStream(inputFile), getEncoding());
		log.info("A total of " + handler.getIndex()
				+ " records have been processed.");

		// Throw the exception if present.
		if (failOnError && handler.getException() != null) {
			throw handler.getException();
		}
	}

	/**
	 * Handler to create a file with Base64 content inside each record.
	 */
	private class ExtractAttachedFileElementHandler implements ElementHandler {

		private final String OUTPUT_DIRECTORY_XPATH = "string-join(tokenize(/file/@path,'\\\\|/')[position() < last()],'"
				+ File.separatorChar + "')";

		private final String FILENAME_XPATH = "tokenize(/file/@path,'\\\\|/')[last()]";

		private final String CONTENT_XPATH = "/file/content/text()";

		private Exception exception;

		private int index = 0;

		/**
		 * Constructor
		 * 
		 * @throws TransformerConfigurationException
		 */
		ExtractAttachedFileElementHandler()
				throws TransformerConfigurationException {
		}

		/**
		 * Get the first exception.
		 * 
		 * @return
		 */
		public Exception getException() {
			return exception;
		}

		/**
		 * Get the index of the last processed record.
		 * 
		 * @return
		 */
		public int getIndex() {
			return index;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.dom4j.ElementHandler#onStart(org.dom4j.ElementPath)
		 */
		public void onStart(ElementPath path) {
			// Increment the record index.
			index++;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.dom4j.ElementHandler#onEnd(org.dom4j.ElementPath)
		 */
		public void onEnd(ElementPath path) {
			FileOutputStream fos = null;
			InputStream input = null;
			File file = null;
			try {
				// #####################################################
				// ### 1. Read the record content. ###
				// #####################################################
				// Retrieve the current record element.
				Element record = path.getCurrent();
				String xml = record.asXML();
				if (log.isDebugEnabled()) {
					log.debug("Record Index=" + index + ": \n" + xml);
				}

				// Convert the record element to InputStream for parsing.
				input = new ByteArrayInputStream(xml.getBytes(getEncoding()));

				// Free dom4j memory consumed by the record element.
				record.detach();

				// Create a namespace context.
				NamespaceContext nsContext = new SimpleNamespaceContext(input);

				// Prepare a XPathHelper with nsContext.
				XPathHelper helper = new XPathHelper(nsContext);

				// Parsing routine to get the content using XPath.
				String content;
				try {
					// Evaluate XPath to get the content.
					input.reset();
					content = (String) helper.evaluate(input, CONTENT_XPATH,
							XPathConstants.STRING);

				} catch (XPathExpressionException e) {
					throw new RuntimeException(e);
				}

				if (content.length() == 0) {
					log.warn("File skipped because it has no content.");
					return;
				}

				// #####################################################
				// ### 2. Resolve the output directory and filename. ###
				// #####################################################
				String filename;
				File outputDirectory = null;
				try {
					// Evaluate XPath to get the filename.
					input.reset();
					filename = (String) helper.evaluate(input, FILENAME_XPATH,
							XPathConstants.STRING);

					// Sanitize the filename.
					filename = FileUtils.sanitizeFilename(filename);

					// Evaluate XPath to get the outputDirectory.
					input.reset();
					outputDirectory = new File((String) helper.evaluate(input,
							OUTPUT_DIRECTORY_XPATH, XPathConstants.STRING));

					if (!outputDirectory.exists()) {
						log.info("Creating output directory: "
								+ outputDirectory.getPath());
						org.apache.commons.io.FileUtils
								.forceMkdir(outputDirectory);
					}
				} catch (XPathExpressionException e) {
					throw new RuntimeException(e);
				}

				// Use a default filename if none available.
				if (filename.length() == 0) {
					filename = "File_" + index;
					log.warn("No valid filename at index " + index
							+ ". Using this default name: " + filename);
				}

				// Get the file.
				file = new File(outputDirectory, filename);

				// Delete it if it exist already.
				if (file.exists()) {
					org.apache.commons.io.FileUtils.forceDelete(file);
				}

				// ####################################################
				// ### 3. Write record content to the file. ###
				// ####################################################
				// Prepare the OutputStream for writing the content.
				fos = new FileOutputStream(file);

				// Get an InputStream on the file content.
				InputStream is = new ByteArrayInputStream(
						content.getBytes(getEncoding()));

				// Apply each operation in sequence.
				Operation operation = null;
				for (int i = 0; i < getOperations().length; i++) {
					operation = (Operation) Class.forName(
							"com.taleo.integration.client.customstep.xml.ExtractFilesPostStep$"
									+ getOperations()[i]).newInstance();
					is = operation.execute(is, getEncoding());
				}

				// Copy the final input stream to the output stream.
				org.apache.commons.io.IOUtils.copy(is, fos);
				fos.close();
				log.info("Extracted file: " + file.getPath());

			} catch (Exception e) {
				// Store the first exception.
				if (this.exception == null) {
					this.exception = e;
				}
				// Log all exceptions.
				String msg = "Error generating file " + file.getPath() + ".";
				if (log.isDebugEnabled()) {
					log.error(msg, e);
				} else {
					log.error(msg + " - " + e.getMessage());
				}

			} finally {
				try {
					if (fos != null) {
						fos.close();
					}
					if (input != null) {
						input.close();
					}
				} catch (IOException e) {
				}
			}
		}
	}

	/**
	 * Interface for operations on file content.
	 */
	interface Operation {
		public InputStream execute(InputStream is, String encoding)
				throws IOException;
	}

	/**
	 * DecodeBase64 operation.
	 */
	static class DecodeBase64 implements Operation {
		public InputStream execute(InputStream is, String encoding)
				throws IOException {
			byte[] bytes = IOUtils.toByteArray(is);
			return new ByteArrayInputStream(Base64.decodeBase64(bytes));
		}
	}

	/**
	 * Gunzip operation.
	 */
	static class Gunzip implements Operation {
		public InputStream execute(InputStream is, String encoding)
				throws IOException {
			return new GZIPInputStream(is);
		}
	}

	/**
	 * HtmlToPdf operation.
	 */
	static class HtmlToPdf implements Operation {
		public InputStream execute(InputStream is, String encoding)
				throws IOException {
			// Strip invalid XML characters.
			Operation op = new StripInvalidXML();
			is = op.execute(is, encoding);

			// Convert HTML to XHTML.
			op = new HtmlToXhtml();
			is = op.execute(is, encoding);

			// Convert XHTML to PDF.
			op = new XhtmlToPdf();
			is = op.execute(is, encoding);

			return is;
		}
	}

	/**
	 * XhtmlToPdf operation.
	 */
	static class XhtmlToPdf implements Operation {
		public InputStream execute(InputStream is, String encoding)
				throws IOException {
			try {
				// Read InputStream into a Document.
				final DocumentBuilderFactory factory = DocumentBuilderFactory
						.newInstance();
				factory.setValidating(false);
				DocumentBuilder builder = factory.newDocumentBuilder();
				builder.setEntityResolver(FSEntityResolver.instance());
				Document document = builder.parse(is);

				// Prepare an ITextRenderer.
				ITextRenderer renderer = new ITextRenderer();
				renderer.setDocument(document, null);
				renderer.layout();

				// Create PDF.
				ByteArrayOutputStream os = new ByteArrayOutputStream();
				renderer.createPDF(os);

				// Return an InputStream on the result.
				return new ByteArrayInputStream(os.toByteArray());

			} catch (SAXException e) {
				throw new RuntimeException(e);

			} catch (ParserConfigurationException e) {
				throw new RuntimeException(e);

			} catch (DocumentException e) {
				throw new RuntimeException(e);
			}
		}
	}

	/**
	 * StripInvalidXML operation. Used internally for now.
	 */
	static class StripInvalidXML implements Operation {
		public InputStream execute(InputStream is, String encoding)
				throws IOException {
			BufferedReader br = null;
			try {
				// Build a String from the InputStream.
				br = new BufferedReader(new InputStreamReader(is));
				StringBuilder sb = new StringBuilder();
				String line;
				while ((line = br.readLine()) != null) {
					sb.append(line);
				}

				// Replace invalid XML characters.
				String xml = XmlUtils.stripInvalidXMLCharacters(sb.toString());

				// Return an InputStream on the result.
				byte[] bytes = xml.getBytes(encoding);
				return new ByteArrayInputStream(bytes);
			} finally {
				br.close();
			}
		}
	}

	/**
	 * HtmlToXhtml operation. Used internally for now.
	 */
	static class HtmlToXhtml implements Operation {
		public InputStream execute(InputStream is, String encoding)
				throws IOException {

			// Prepare Tidy instance.
			Tidy tidy = new Tidy();
			tidy.setXHTML(true);
			tidy.setQuiet(true);
			tidy.setShowWarnings(false);
			tidy.setInputEncoding(encoding);
			tidy.setOutputEncoding(encoding);

			// Convert to XHTML.
			ByteArrayOutputStream os = new ByteArrayOutputStream();
			tidy.parse(is, os);

			// Return an InputStream on the result.
			return new ByteArrayInputStream(os.toByteArray());
		}
	}
}
