package com.taleo.integration.client.customstep.xml;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.Constants;
import com.taleo.integration.client.customstep.util.xml.XmlUtils;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.ConfigurationBoardRuntimeHelper;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedMessage;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Pre-processing step to replace symbols in a SQ-XML request file. Symbols are
 * strings surrounded with square brackets, just like in the configuration
 * files.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep}.<br>
 * <br>
 * The current implementation gets symbols from the configuration board as well
 * as from Java properties, with the <code>com.taleo.client.symbol</code>
 * prefix. This excludes the default runtime symbols like [CFGFOLDER] and [NOW].
 * <br>
 * <br>
 * One level of indirection is allowed where a symbol value itself contains a
 * symbol which will be replaced in sequence.
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/ReplaceSymbolsInQueryPreStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class ReplaceSymbolsInQueryPreStep extends BaseCustomStep {

	private static Logger log = Logger
			.getLogger(ReplaceSymbolsInQueryPreStep.class);

	public static final String PREFIX_SYMBOL_SYSTEM_PROPERTY = "com.taleo.client.symbol.";

	/** The XSL transformation files. */
	private final static String XSL_TRANSFORM_CONFIG_BOARD_SYMBOLS = "xsl/ConfigBoardSymbols.xsl";
	private final static String XSL_TRANSFORM_REPLACE_SYMBOLS = "xsl/ReplaceSymbols.xsl";

	/**
	 * Constructor without arguments.
	 */
	public ReplaceSymbolsInQueryPreStep() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "ReplaceSymbolsInQueryPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Pre-processing step to replace symbols in SQ-XML requests.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "ReplaceSymbolsInQueryPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// First, get the basic supported pipeline for a post-processing step
		// (this handles the constraints on the last step type)
		SupportedPipeline supportedPipeline = getPreProcessSupportPipeline();

		// Add the supported response message information
		SupportedMessage supportedMessage = supportedPipeline
				.addNewSupportedResponseMessage();

		// Add the supported response message format
		supportedMessage.addNewSupportedFormats();
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.SQ_XML);

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the request info.
			Message requestInfo = pipeline.getRequestMessage();
			String queryFilename = requestInfo.getCurrentFile();
			if (queryFilename == null) {
				queryFilename = requestInfo.getOriginalFile();
			}
			File queryFile = new File(queryFilename);

			// Transform the query file.
			File outFile = createTempFile();
			transform(queryFile, outFile);

			// Set the transformed query in the request.
			requestInfo.setCurrentFile(outFile.getAbsolutePath());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Transform the query file.
	 * 
	 * @param queryFile
	 *            The query file.
	 * @param outFile
	 *            The output file.
	 * @throws TransformerException
	 * @throws IOException
	 */
	protected void transform(File queryFile, File outFile)
			throws TransformerException, IOException {

		// Get the configuration board symbols.
		Document doc = getConfigBoardSymbols();

		// Prepare a list of symbols from System properties.
		Properties symbols = new Properties();
		for (Iterator it = System.getProperties().entrySet().iterator(); it
				.hasNext();) {
			Map.Entry entry = (Map.Entry) it.next();
			if ((entry.getKey().toString())
					.startsWith(PREFIX_SYMBOL_SYSTEM_PROPERTY)) {
				symbols.put(entry.getKey(), entry.getValue());
			}
		}

		// Add symbols to the document starting with those that have themselves
		// a symbol as value, so they will be processed first.
		for (Iterator it = symbols.entrySet().iterator(); it.hasNext();) {
			Map.Entry entry = (Map.Entry) it.next();
			if (entry.getValue().toString().startsWith("[")
					&& entry.getValue().toString().endsWith("]")) {
				addSymbol(doc, entry);
			}
		}
		// Add regular symbols after, so they will be processed last.
		for (Iterator it = symbols.entrySet().iterator(); it.hasNext();) {
			Map.Entry entry = (Map.Entry) it.next();
			if (!(entry.getValue().toString().startsWith("[") && entry
					.getValue().toString().endsWith("]"))) {
				addSymbol(doc, entry);
			}
		}

		if (log.isDebugEnabled()) {
			log.debug("Replacing symbols: " + XmlUtils.toXml(doc));
		}

		// Create an XSL transformer.
		InputStream xsl = this.getClass().getClassLoader()
				.getResourceAsStream(XSL_TRANSFORM_REPLACE_SYMBOLS);
		TransformerFactory tFactory = TransformerFactory.newInstance();
		Transformer transformer = tFactory
				.newTransformer(new StreamSource(xsl));

		// Set parameters.
		transformer.setParameter("Symbols", doc);

		// Prepare streams.
		InputStream is = IOHelper.getInputStream(queryFile);
		OutputStream os = IOHelper.getOutputStream(
				new FileOutputStream(outFile), getClass().getName());

		// Prepare source and result.
		Source source = new StreamSource(is);
		Result result = new StreamResult(os);

		// Apply the transformation.
		transformer.transform(source, result);

		// Close.
		is.close();
		os.close();
	}

	/**
	 * Get the configuration board symbols.
	 * 
	 * @return
	 * @throws TransformerException
	 * @throws IOException
	 */
	private Document getConfigBoardSymbols() throws TransformerException,
			IOException {
		// Get the configuration board file.
		File configBoard = ConfigurationBoardRuntimeHelper
				.getDefaultConfigBoardFile();

		// Create an XSL transformer.
		InputStream xsl = this.getClass().getClassLoader()
				.getResourceAsStream(XSL_TRANSFORM_CONFIG_BOARD_SYMBOLS);
		TransformerFactory tFactory = TransformerFactory.newInstance();
		Transformer transformer = tFactory
				.newTransformer(new StreamSource(xsl));

		// Prepare streams.
		InputStream is = IOHelper.getInputStream(configBoard);

		// Prepare source and result.
		Source source = new StreamSource(is);
		Document doc = XmlUtils.createDomDocument();
		Result result = new DOMResult(doc);

		// Apply transformation.
		transformer.transform(source, result);

		// Close.
		is.close();

		return doc;
	}

	/**
	 * Add a symbol element in the document.
	 * 
	 * @param doc
	 * @param entry
	 */
	private void addSymbol(Document doc, Map.Entry entry) {
		Element symbol = doc.createElementNS(Constants.CLI_NAMESPACE, "Symbol");
		symbol.setAttribute(
				"name",
				entry.getKey().toString()
						.substring(PREFIX_SYMBOL_SYSTEM_PROPERTY.length()));
		symbol.setTextContent(entry.getValue().toString());
		doc.getDocumentElement().appendChild(symbol);
	}
}
