package com.taleo.integration.client.customstep.xml;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.Properties;

import javax.xml.namespace.NamespaceContext;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.dom4j.DocumentException;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.ElementHandler;
import org.dom4j.ElementPath;
import org.dom4j.io.SAXReader;
import org.dom4j.io.XMLWriter;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.FileUtils;
import com.taleo.integration.client.customstep.util.xml.SimpleNamespaceContext;
import com.taleo.integration.client.customstep.util.xml.XPathHelper;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.GlobalConfig;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedMessage;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Post-processing step to split the XML result file into one file for each
 * record.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep} and offers two
 * different APIs.
 * <p>
 * <ol>
 * <li><b>Primary API</b>
 * <p>
 * The primary API consumes an XML document with the following structure and
 * produces one file for each &lt;File&gt; tag. File content must be in XML
 * format included in CDATA section. It can be further transformed using XSL if
 * needed (see below).
 * 
 * <pre>
 * &lt;files&gt;
 *   &lt;file path="..."&gt;
 *     &lt;![CDATA[... file content ...]]&gt;
 *   &lt;/file&gt;
 * &lt;/files&gt;
 * </pre>
 * 
 * Where:
 * <ul>
 * <li><b>path</b>: The complete path of the file to be created. Invalid
 * characters in filename will be replaced with an underscore "_".</li>
 * <li>Any other attribute can be added and will simply be ignored while
 * extracting files.</li>
 * </ul>
 * </li>
 * <p>
 * <li><b>Secondary API</b>
 * <p>
 * The second API uses the following parameters:
 * <ul>
 * <li><b>recordXPath</b>: The XPath expression leading to the individual
 * records that make up single files content. This parameter is insensible to
 * namespace prefixes, but you may want to follow the same general note on XPath
 * expressions below. <BR>
 * Example:
 * <ul>
 * <li>
 * /default:Document/default:Content/default:ExportXML/e:Candidate</li>
 * </ul>
 * </li>
 * <br>
 * <li><b>filenameXPath</b>: The XPath expression based at record root and
 * leading to the filename (excluding path). The resulting filename is limited
 * to 180 characters and invalid characters in filename will be replaced with an
 * underscore "_". See general note on XPath expressions below.<BR>
 * Examples:
 * <ul>
 * <li>concat(e:Candidate/e:EmployeeID, '.xml')</li>
 * <li>e:Candidate/e:RuntimeField[@name='filename']</li>
 * </ul>
 * The following variables can also be used:
 * <ul>
 * <li><b>$index</b>: The index of the record in the document.</li>
 * <li><b>$date</b>: The current date is established at the beginning of the
 * step. Consequently, the same date will be applied to all individual files.</li>
 * </ul>
 * Examples:
 * <ul>
 * <li>concat(e:Candidate/e:EmployeeID, $date, '.xml')</li>
 * <li>concat($date, '_', $index, '.xml')</li>
 * </ul>
 * </li>
 * <br>
 * <li><b>dateFormat</b>: The format for the $date variable, default:
 * "yyyyMMdd_HHmmss". See {@link SimpleDateFormat} for details.</li>
 * <br>
 * <br>
 * <li><b>outputDirectory</b>: <i>@deprecated - Use the outputDirectoryXPath
 * instead.</i> The directory to put the result files. It can be set as absolute
 * or relative to the current directory.<BR>
 * Example:
 * <ul>
 * <li>results</li>
 * </ul>
 * </li>
 * <li><b>outputDirectoryXPath</b>: The XPath expression that builds to
 * directory to put the result files. It can be set as absolute or relative to
 * the current directory. See general note on XPath expressions below.
 * <p>
 * It can use the same $date and $index variables as the filenameXPath.<BR>
 * Examples:
 * <ul>
 * <li>'results'</li>
 * <li>concat('results/', e:Candidate/e:Number, '_', $date)</li>
 * </ul>
 * </li>
 * <p>
 * All <b>XPath expressions</b> must not use wildcards and must use the same
 * namespace prefixes that are declared in the document, except for the default
 * namespace for which the prefix 'default' must be used instead of no prefix.
 * This 'default' prefix may be replaced if needed, by setting the
 * 'com.taleo.integration.client.xml.default.ns.prefix' system property.
 * </ul>
 * </li>
 * </ol>
 * In both cases, the following parameters apply:<br>
 * <br>
 * <ul>
 * <li><b>xslFile</b>: Optional - The absolute or relative path of an XSL
 * transformation to apply to each individual record. This can be used, for
 * example to add a header and a footer to each file.<BR>
 * If the XSL produces an empty file, it will be deleted. This can happen if
 * there is a condition in the XSL transformation that prevents creating any
 * content in the result.</li>
 * <br>
 * <br>
 * <li>All <b>other parameters</b> are passed as arguments to the XSL
 * transformation.</li>
 * </ul>
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://dom4j.org/dom4j-1.6.1/">dom4j library</a>
 * (dom4j-1.6.1.jar)</li>
 * <li><a href="http://jaxen.codehaus.org/">jaxen library</a> (jaxen-1.1.jar)</li>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file, using the primary API:
 * <p>
 * <img src="doc-files/SplitFilePostStep-1.jpg" />
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file, using the secondary API:
 * <p>
 * <img src="doc-files/SplitFilePostStep-2.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class SplitFilePostStep extends BaseCustomStep {

	private static Logger log = Logger.getLogger(SplitFilePostStep.class);

	private static final String DEFAULT_RECORD_XPATH = "/files/file";

	private static final String DEFAULT_FILENAME_XPATH = "tokenize(/file/@path,'\\\\|/')[last()]";

	private static final String DEFAULT_OUTPUT_DIRECTORY_XPATH = "string-join(tokenize(/file/@path,'\\\\|/')[position() < last()],'"
			+ File.separatorChar + "')";

	/**
	 * Constructor without arguments.
	 */
	public SplitFilePostStep() {
		super();
		// Register valid parameters.
		registerParameterName("recordXPath");
		registerParameterName("filenameXPath");
		registerParameterName("dateFormat");
		registerParameterName("outputDirectory");
		registerParameterName("outputDirectoryXPath");
		registerParameterName("xslFile");
	}

	/**
	 * Validate the parameters names.
	 * <p>
	 * Override the default method to allow for extra parameters to be passed to
	 * the XSL transformation.
	 */
	protected void validateParameterNames() {
		Enumeration keys = parameters.keys();
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			if (!parameterNames.contains(key)) {
				log.info("Using XSL parameter " + key + "="
						+ parameters.getProperty(key) + ".");
			}
		}
	}

	/**
	 * The XPATH expression to identify records.
	 */
	private String recordXPath = DEFAULT_RECORD_XPATH;

	/**
	 * The filename XPath.
	 */
	private String filenameXPath = DEFAULT_FILENAME_XPATH;

	/**
	 * The date format.
	 */
	private DateFormat dateFormat = new SimpleDateFormat("yyyyMMdd_HHmmss");

	/**
	 * The output directory.
	 * 
	 * @deprecated
	 */
	private String outputDirectory;

	/**
	 * The output directory XPath.
	 */
	private String outputDirectoryXPath = DEFAULT_OUTPUT_DIRECTORY_XPATH;

	/**
	 * The XSL file name.
	 */
	private String xslFile;

	/**
	 * The XSL parameters.
	 */
	private Properties xslParameters = new Properties();

	/**
	 * Indicate if we are using the primary API.
	 * 
	 * @return
	 */
	private boolean isPrimaryAPI() {
		return DEFAULT_RECORD_XPATH.equals(getRecordXPath());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "SplitFilePostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to split the result file into individual record files.";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "SplitFilePostStep";
	}

	/**
	 * Get the XPath expression for the records.
	 * 
	 * @return
	 */
	public String getRecordXPath() {
		return recordXPath;
	}

	/**
	 * Set the XPath expression for the records.
	 * 
	 * @param recordXPath
	 */
	public void setRecordXPath(String recordXPath) {
		this.recordXPath = recordXPath;
	}

	/**
	 * Get the XPath expression for filename.
	 * 
	 * @return
	 */
	public String getFilenameXPath() {
		return filenameXPath;
	}

	/**
	 * Set the XPath expression for filename.
	 * 
	 * @param filenameXPath
	 */
	public void setFilenameXPath(String filenameXPath) {
		this.filenameXPath = filenameXPath;
	}

	/**
	 * Get the date format.
	 * 
	 * @return
	 */
	public DateFormat getDateFormat() {
		return dateFormat;
	}

	/**
	 * Set the date format.
	 * 
	 * @param dateFormat
	 */
	public void setDateFormat(DateFormat dateFormat) {
		this.dateFormat = dateFormat;
	}

	/**
	 * Get the output directory.
	 * 
	 * @return
	 * @deprecated Use getOutputDirectoryXPath instead.
	 */
	public String getOutputDirectory() {
		return outputDirectory;
	}

	/**
	 * Set the output directory.
	 * 
	 * @param outputDirectory
	 * @deprecated Use setOutputDirectoryXPath instead.
	 */
	public void setOutputDirectory(String outputDirectory) {
		this.outputDirectory = outputDirectory;
	}

	/**
	 * Get the output directory XPath.
	 * 
	 * @return
	 */
	public String getOutputDirectoryXPath() {
		return outputDirectoryXPath;
	}

	/**
	 * Set the output directory XPath.
	 * 
	 * @param outputDirectoryXPath
	 */
	public void setOutputDirectoryXPath(String outputDirectoryXPath) {
		this.outputDirectoryXPath = outputDirectoryXPath;
	}

	/**
	 * Get the XSL file.
	 * 
	 * @return
	 */
	public String getXslFile() {
		return xslFile;
	}

	/**
	 * Set the XSL file.
	 * 
	 * @param xslFile
	 */
	public void setXslFile(String xslFile) {
		this.xslFile = xslFile;
	}

	/**
	 * Get the XSL parameters.
	 * 
	 * @return
	 */
	public Properties getXslParameters() {
		return xslParameters;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// First, get the basic supported pipeline for a post-processing step
		// (this handles the constraints on the last step type)
		SupportedPipeline supportedPipeline = getPostProcessSupportPipeline();

		// Add the supported response message information
		SupportedMessage supportedMessage = supportedPipeline
				.addNewSupportedResponseMessage();

		// Add the supported response message format
		supportedMessage.addNewSupportedFormats();
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.T_XML);
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.XML);
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.T_SOAP);

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seecom.taleo.integration.client.step.BaseCustomStep#init(com.taleo.ws.
	 * integration.client.GlobalConfig)
	 */
	public void init(GlobalConfig config) {
		super.init(config);

		// If not active, return immediately.
		if (!isActive()) {
			return;
		}

		// Get the parameters.
		setRecordXPath(parameters.getProperty("recordXPath",
				DEFAULT_RECORD_XPATH));
		if (StringUtils.isBlank(recordXPath)) {
			throw new IllegalArgumentException(
					"The 'recordXPath' parameter is mandatory.");
		}

		setFilenameXPath(parameters.getProperty("filenameXPath",
				DEFAULT_FILENAME_XPATH));
		if (StringUtils.isBlank(filenameXPath)) {
			throw new IllegalArgumentException(
					"The 'filenameXPath' parameter is mandatory.");
		}

		String dateFormat = parameters.getProperty("dateFormat");
		if (StringUtils.isNotBlank(dateFormat)) {
			setDateFormat(new SimpleDateFormat(dateFormat));
		}

		setOutputDirectory(parameters.getProperty("outputDirectory"));
		setOutputDirectoryXPath(parameters.getProperty("outputDirectoryXPath",
				DEFAULT_OUTPUT_DIRECTORY_XPATH));
		if (StringUtils.isBlank(outputDirectory)
				&& StringUtils.isBlank(outputDirectoryXPath)) {
			throw new IllegalArgumentException(
					"The 'outputDirectoryXPath' parameter is mandatory.");
		}

		setXslFile(parameters.getProperty("xslFile"));

		// Get the remaining parameters as XSL parameters.
		Enumeration keys = parameters.keys();
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			if (!parameterNames.contains(key)) {
				xslParameters.setProperty(key, parameters.get(key).toString());
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the input file.
			Message responseMessage = pipeline.getResponseMessage();
			File inFile = null;
			if (responseMessage.isSetCurrentFile()) {
				inFile = new File(responseMessage.getCurrentFile());
			} else {
				inFile = new File(responseMessage.getOriginalFile());
			}

			// Execute.
			execute(inFile);

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Split the input file.
	 * 
	 * @param inputFile
	 *            The input file.
	 * @throws TransformerConfigurationException
	 * @throws DocumentException
	 * @throws ParserConfigurationException
	 * @throws IOException
	 */
	protected void execute(File inputFile)
			throws TransformerConfigurationException, DocumentException,
			ParserConfigurationException, IOException {

		// Strip the namespace prefixes because SAXReader does not support them.
		String recordXPath = XPathHelper
				.stripNamespacePrefixes(this.recordXPath);

		// Create a SAXReader with a handler on the records.
		SAXReader reader = new SAXReader();
		reader.addHandler(recordXPath, new SplitFileElementHandler());

		// Read the input file.
		reader.read(IOHelper.getInputStream(inputFile), getEncoding());
	}

	/**
	 * Class to create a distinct file for each record while reading an XML
	 * document.
	 * 
	 * @author Romain Guay - Taleo Corporation
	 * 
	 */
	private class SplitFileElementHandler implements ElementHandler {

		private final static String INDEX_VARIABLE = "index";

		private final static String DATE_VARIABLE = "date";

		private Transformer transformer;

		private int index = 0;

		private Date date = new Date();

		private final String defaultNsPrefix = System
				.getProperty(
						"com.taleo.integration.client.xml.default.ns.prefix",
						"default");

		/**
		 * Constructor
		 * 
		 * @throws TransformerConfigurationException
		 * @throws ParserConfigurationException
		 */
		SplitFileElementHandler() throws TransformerConfigurationException,
				ParserConfigurationException {
			// Prepare the XSL transformer if required.
			if (getXslFile() != null) {
				TransformerFactory tFactory = TransformerFactory.newInstance();
				transformer = tFactory.newTransformer(new StreamSource(
						getXslFile()));

				// Add XSL parameters.
				Enumeration keys = getXslParameters().keys();
				while (keys.hasMoreElements()) {
					String key = (String) keys.nextElement();
					transformer.setParameter(key, getXslParameters().get(key));
				}
			}
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.dom4j.ElementHandler#onStart(org.dom4j.ElementPath)
		 */
		public void onStart(ElementPath path) {
			// Increment the record index.
			index++;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.dom4j.ElementHandler#onEnd(org.dom4j.ElementPath)
		 */
		public void onEnd(ElementPath path) {
			OutputStream os = null;
			InputStream is = null;

			try {
				// Retrieve the current record element.
				Element record = path.getCurrent();
				String xml = record.asXML();
				if (log.isDebugEnabled()) {
					log.debug("Record index=" + index + ": \n" + xml);
				}

				// Convert the record element to InputStream for parsing.
				byte[] bytes = xml.getBytes(getEncoding());
				is = new ByteArrayInputStream(bytes);

				// Free dom4j memory consumed by the record element.
				record.detach();

				// Create a namespace context and substitute the default ns
				// prefix.
				NamespaceContext nsContext = new SimpleNamespaceContext(is,
						defaultNsPrefix);

				// Prepare a XPathHelper with nsContext and variables.
				XPathHelper helper = new XPathHelper(nsContext);
				helper.setVariable(DATE_VARIABLE, dateFormat.format(this.date));
				helper.setVariable(INDEX_VARIABLE, String.valueOf(this.index));

				// Parsing routine to get the filename and output directory
				// using XPath.
				String filename = null;
				File outputDirectory = null;
				try {
					// Evaluate XPath to get the filename.
					is.reset();
					filename = (String) helper.evaluate(is, getFilenameXPath(),
							XPathConstants.STRING);

					// Sanitize the filename.
					filename = FileUtils.sanitizeFilename(filename);

					// Evaluate XPath to get the outputDirectory.
					is.reset();
					if (StringUtils.isNotBlank(getOutputDirectory())) {
						outputDirectory = new File(getOutputDirectory());
					} else {
						outputDirectory = new File((String) helper.evaluate(is,
								getOutputDirectoryXPath(),
								XPathConstants.STRING));
					}

					if (!outputDirectory.exists()) {
						log.info("Creating output directory: "
								+ outputDirectory.getPath());
						org.apache.commons.io.FileUtils
								.forceMkdir(outputDirectory);
					}
				} catch (XPathExpressionException e) {
					throw new RuntimeException(e);
				}

				// Prepare the writer for the file.
				File file = new File(outputDirectory, filename);
				os = new FileOutputStream(file);

				// Set the record to the file content if using the primary API.
				if (isPrimaryAPI()) {
					record = DocumentHelper.parseText(record.getText().trim())
							.getRootElement();
				}

				if (transformer == null) {
					// If no XSL transformation, write record directly.
					XMLWriter xmlWriter = new XMLWriter(os);
					xmlWriter.write(record);

					log.info("Created file: " + file);

				} else {
					// Otherwise, transform the record.
					is.close();
					is = new ByteArrayInputStream(record.asXML().getBytes(
							getEncoding()));
					Source source = new StreamSource(is);
					Result result = new StreamResult(os);

					// Transform the record.
					transformer.transform(source, result);

					// Delete file if empty.
					os.close(); // Close to allow reading actual file length.
					if (file.length() == 0) {
						file.delete();
						index--;
						log.info("Empty file: " + file + " was deleted.");
					} else {
						log.info("Created file: " + file);
					}
				}

			} catch (Exception e) {
				// Log the error, because its detail is not carried by SAX
				// parsing.
				log.error("", e);
				throw new RuntimeException(e);

			} finally {
				try {
					if (os != null) {
						os.close();
					}
					if (is != null) {
						is.close();
					}
				} catch (IOException e) {
				}
			}
		}
	}

}
