package com.taleo.integration.client.customstep.xml;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;

import org.apache.log4j.Logger;

import com.taleo.integration.client.customstep.BaseCustomStep;
import com.taleo.integration.client.customstep.util.xml.XmlUtils;
import com.taleo.integration.client.step.StepException;
import com.taleo.integration.client.util.IOHelper;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;

/**
 * Processing step (pre or post) to strip invalid XML characters from a file.
 * The file can be any text file (CSV, XML, ...).
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.BaseCustomStep}.
 * <p>
 * This step can be useful on CSV files that are to be transformed in XML or on
 * XML files that already contain those invalid characters preventing any XSL
 * transformation on them.
 * <p>
 * Valid XML characters are in the following range: #x9 | #xA | #xD |
 * [#x20-#xD7FF] | [#xE000-#xFFFD] | [#x10000-#x10FFFF]<br>
 * (Reference: <a
 * href="http://www.w3.org/TR/2000/REC-xml-20001006#charsets">W3C</a>)
 * <p>
 * This class is dependent on:
 * <ul>
 * <li><a href="http://commons.apache.org/lang/">Apache Commons Lang</a>
 * (commons-lang-2.4.jar)</li>
 * </ul>
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/StripInvalidXMLCharsStep-1.png" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class StripInvalidXMLCharsStep extends BaseCustomStep {

	private static Logger log = Logger
			.getLogger(StripInvalidXMLCharsStep.class);

	public String getDescription() {
		return "Processing step that will strip invalid XML characters from a file.";
	}

	public String getIdentifier() {
		return "StripInvalidXMLCharsStep";
	}

	public String getName() {
		return "StripInvalidXMLCharsStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			boolean isPreProcess = true;

			// Get the input file.
			File inFile = null;
			Message responseMessage = pipeline.getResponseMessage();
			Message requestMessage = pipeline.getRequestMessage();
			if (responseMessage.isSetCurrentFile()) {
				inFile = new File(responseMessage.getCurrentFile());
				isPreProcess = false;
			} else {
				if (requestMessage.isSetCurrentFile()) {
					inFile = new File(requestMessage.getCurrentFile());
				} else {
					inFile = new File(requestMessage.getOriginalFile());
				}
			}

			// Create a temporary file for the output.
			File outputFile = createTempFile();

			// Apply the transformation.
			execute(inFile, outputFile);

			// Set the new current file.
			if (isPreProcess) {
				requestMessage.setCurrentFile(outputFile.getAbsolutePath());
			} else {
				responseMessage.setCurrentFile(outputFile.getAbsolutePath());
			}

		} catch (Exception e) {
			throw new StepException(e);
		}
	}

	/**
	 * Process the input file and send result in the output file.
	 * 
	 * @param inputFile
	 *            The input file.
	 * @param outputFile
	 *            The output file.
	 * @throws Exception
	 */
	protected void execute(File inputFile, File outputFile) throws Exception {
		BufferedReader reader = null;
		BufferedWriter writer = null;
		try {
			reader = new BufferedReader(new InputStreamReader(
					IOHelper.getInputStream(inputFile), getEncoding()));
			writer = new BufferedWriter(new OutputStreamWriter(
					IOHelper.getOutputStream(new FileOutputStream(outputFile),
							outputFile.getName()), getEncoding()));

			String line;
			int lineCount = 0;
			while ((line = reader.readLine()) != null) {
				if (log.isDebugEnabled()) {
					// Write one character at a time and log changes.
					lineCount++;
					String[] chars = line.split("");
					for (int i = 1; i < chars.length; i++) {
						String c = XmlUtils.stripInvalidXMLCharacters(chars[i]);
						if (!c.equals(chars[i])) {
							log.debug("Removing character '"
									+ "\\u"
									+ Integer.toHexString(chars[i]
											.toCharArray()[0]) + "' on line "
									+ lineCount + ", character " + i + ".");
						}
						writer.write(c);
					}
				} else {
					// Otherwise, process one line at a time.
					writer.write(XmlUtils.stripInvalidXMLCharacters(line));
				}
				writer.newLine();
				writer.flush();
			}
		} finally {
			if (reader != null) {
				reader.close();
			}
			if (writer != null) {
				writer.close();
			}
		}
	}
}
