package com.taleo.integration.client.customstep.xml;

import java.io.File;

import org.apache.log4j.Logger;

import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.MessageFormat;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedMessage;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Post-processing step to apply an XSL transformation to the result.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.xml.AbstractXslStep}.
 * <p>
 * Here are examples of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/XslPostStep-1.jpg" />
 * <p>
 * <img src="doc-files/XslPostStep-2.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class XslPostStep extends AbstractXslStep {

	private static Logger log = Logger.getLogger(XslPostStep.class);

	/**
	 * Constructor without arguments.
	 */
	public XslPostStep() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "XSLPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Post-processing step to apply an XSL transformation";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "XSLPostStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// First, get the basic supported pipeline for a post-processing step
		// (this handles the constraints on the last step type)
		SupportedPipeline supportedPipeline = getPostProcessSupportPipeline();

		// Add the supported response message information
		SupportedMessage supportedMessage = supportedPipeline
				.addNewSupportedResponseMessage();

		// Add the supported response message format
		supportedMessage.addNewSupportedFormats();
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.T_XML);
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.XML);
		supportedMessage.getSupportedFormats().addFormat(MessageFormat.T_SOAP);

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the input file.
			Message responseMessage = pipeline.getResponseMessage();
			File inFile = null;
			if (responseMessage.isSetCurrentFile()) {
				inFile = new File(responseMessage.getCurrentFile());
			} else {
				inFile = new File(responseMessage.getOriginalFile());
			}

			// Transform the current file with XSL.
			File outFile = createTempFile();
			execute(inFile, outFile);

			if (this.outputFile != null) {
				// If outputFile is set, decompress the result into it.
				com.taleo.integration.client.customstep.util.IOHelper
						.getDecompressedFile(outFile, new File(outputFile));
			} else {
				// Otherwise, set it as the new current file in the pipeline.
				responseMessage.setCurrentFile(outFile.getAbsolutePath());
			}
			
			// Set the response format.
			if (getOutputFormat() != null) {
				responseMessage.setFormat(getOutputFormat());
			}

			log.info("Response message format: " + responseMessage.getFormat());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}
}
