package com.taleo.integration.client.customstep.xml;

import java.io.File;

import org.apache.log4j.Logger;

import com.taleo.integration.client.step.StepException;
import com.taleo.ws.integration.client.Message;
import com.taleo.ws.integration.client.Pipeline;
import com.taleo.ws.integration.client.SupportedPipeline;

/**
 * Pre-processing step to apply an XSL transformation to the request.
 * <p>
 * This step takes all parameters from its base class
 * {@link com.taleo.integration.client.customstep.xml.AbstractXslStep}.
 * <p>
 * Here is an example of how to define this custom step in the TCC configuration
 * file:
 * <p>
 * <img src="doc-files/XslPreStep-1.jpg" />
 * 
 * @author Romain Guay, Taleo Corporation
 */
public class XslPreStep extends AbstractXslStep {

	private static Logger log = Logger.getLogger(XslPreStep.class);

	/**
	 * Constructor without arguments.
	 */
	public XslPreStep() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getIdentifier()
	 */
	public String getIdentifier() {
		return "XSLPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getDescription()
	 */
	public String getDescription() {
		return "Pre-processing step to apply an XSL transformation";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.taleo.integration.client.step.BaseCustomStep#getName()
	 */
	public String getName() {
		return "XSLPreStep";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.BaseCustomStep#getSupportedPipeline()
	 */
	public SupportedPipeline getSupportedPipeline() {
		// First, get the basic supported pipeline for a post-processing step
		// (this handles the constraints on the last step type)
		SupportedPipeline supportedPipeline = getPreProcessSupportPipeline();

		// Return the complete supported pipeline information
		return supportedPipeline;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * com.taleo.integration.client.step.Step#execute(com.taleo.ws.integration
	 * .client.Pipeline)
	 */
	public void execute(Pipeline pipeline) throws StepException {
		// Quit if inactive.
		if (!isActive()) {
			log.info("The step is inactive.");
			return;
		}

		try {
			// Get the input file.
			Message requestMessage = pipeline.getRequestMessage();
			File inFile = null;
			if (requestMessage.isSetCurrentFile()) {
				inFile = new File(requestMessage.getCurrentFile());
			} else {
				inFile = new File(requestMessage.getOriginalFile());
			}

			// Transform the current file with XSL.
			File outFile = createTempFile();
			execute(inFile, outFile);

			if (this.outputFile != null) {
				// If outputFile is set, decompress the result into it.
				com.taleo.integration.client.customstep.util.IOHelper
						.getDecompressedFile(outFile, new File(outputFile));
			} else {
				// Otherwise, set it as the new current file in the pipeline.
				requestMessage.setCurrentFile(outFile.getAbsolutePath());
			}

			// Set the response format.
			if (getOutputFormat() != null) {
				requestMessage.setFormat(getOutputFormat());
			}

			log.info("Request message format: " + requestMessage.getFormat());

		} catch (Exception e) {
			throw new StepException(e);
		}
	}
}
