<?xml version="1.0" encoding="UTF-8"?>
	<!--
		====================================================================
	-->
	<!--
		Transform any query into a count query, simply returning the count of
		rows instead of the actual rows themselves. The result is always CSV
		format and only has one or two fields: EntityName (if provided) and
		Count. Query may specify to exclude specific projections from the
		count, by adding the tag "exclude-from-count". For example:
		<quer:projection tag="exclude-from-count"> <quer:field
		path="AttachedFiles,Content"/> </quer:projection> By default,
		AllCustomFields projections are dropped.
	-->
	<!--
		The trick applied here is to set the query as a subquery with its own
		subqueries and filterings, and simply count the rows.
	-->
	<!--
		The main query is irrelevant, but it needs to return only one row.
		Since there is no Dual in TCC, we use a lookup entity that is
		available in all products (JobType) and set a filter to get 1 row
		only.
	-->
	<!--
		====================================================================
	-->
<xsl:stylesheet version="2.0"
	xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:quer="http://www.taleo.com/ws/integration/query"
	xmlns:fct="http://www.taleo.com/xsl_functions" xmlns:xs="http://www.w3.org/2001/XMLSchema"
	exclude-result-prefixes="fct xs">
	<xsl:output method="xml" encoding="UTF-8" indent="yes" />
	<!--
		====================================================================
	-->
	<!-- Parameters: -->
	<!-- - csvdelimiter: The CSV delimiter. Default: comma (,) -->
	<!--
		- csvheader: Indicates if a header row will be present in the result.
		Default: false
	-->
	<!--
		- entityName: Name of the entity to display in the first column.
		Default: false
	-->
	<!--
		====================================================================
	-->
	<xsl:param name="csvdelimiter" select="','" />
	<xsl:param name="csvheader" select="'false'" />
	<xsl:param name="entityName" />
	<!--
		====================================================================
	-->
	<!-- Default template to render as is. -->
	<!--
		====================================================================
	-->
	<xsl:template match="@*|*|processing-instruction()|comment()">
		<xsl:copy>
			<xsl:apply-templates select="*|@*|text()|processing-instruction()|comment()" />
		</xsl:copy>
	</xsl:template>
	<!--
		====================================================================
	-->
	<!-- Match main query and set is as a subquery. -->
	<!--
		====================================================================
	-->
	<xsl:template match="/quer:query">
		<xsl:comment
			select="'Use JobType as the main projected class because this lookup entity is available in all products and set a filter on Sequence = 1 to get only one record.'" />
		<xsl:comment
			select="'This trick is required because there is no DUAL in TCC.'" />
		<quer:query productCode="{@productCode}" model="{@model}"
			projectedClass="JobType" locale="en" mode="CSV" csvheader="{$csvheader}"
			csvdelimiter="{$csvdelimiter}" largegraph="true" preventDuplicates="false">
			<quer:subQueries>
				<xsl:comment
					select="'This subquery adds the row numbers for the actual query.'" />
				<quer:query projectedClass="{@projectedClass}"
					alias="{if (@alias) then @alias else 'Count'}" preventDuplicates="{@preventDuplicates}">
					<xsl:apply-templates select="quer:subQueries" />
					<quer:projections>
						<quer:projection alias="RowNumber" id="RowNumber">
							<quer:customValue>
								<quer:expression>rowNum</quer:expression>
							</quer:customValue>
						</quer:projection>
						<xsl:apply-templates select="quer:projections/quer:projection" />
					</quer:projections>
					<xsl:apply-templates select="quer:filterings" />
					<xsl:apply-templates select="quer:joinings" />
				</quer:query>
			</quer:subQueries>
			<quer:projections>
				<xsl:if test="$entityName != ''">
					<quer:projection alias="EntityName">
						<quer:string>
							<xsl:value-of select="$entityName" />
						</quer:string>
					</quer:projection>
				</xsl:if>
				<xsl:comment select="'This projection counts the row numbers.'" />
				<quer:projection alias="Count">
					<quer:count>
						<quer:projection ref="RowNumber" />
					</quer:count>
				</quer:projection>
			</quer:projections>
			<quer:projectionFilterings />
			<quer:filterings>
				<quer:filtering>
					<quer:equal>
						<quer:field path="Sequence" />
						<quer:integer>1</quer:integer>
					</quer:equal>
				</quer:filtering>
			</quer:filterings>
			<quer:sortings />
			<quer:sortingFilterings />
			<quer:groupings />
			<quer:joinings />
		</quer:query>
	</xsl:template>
	<!--
		====================================================================
	-->
	<!-- Drop projections that include the tag "exclude-from-count". -->
	<!-- Note that multiple tags may be present, separated with spaces. -->
	<!--
		====================================================================
	-->
	<xsl:template
		match="quer:projection[fct:list-includes(@tag, 'exclude-from-count')]" />
	<!--
		====================================================================
	-->
	<!-- Drop allCustomFields projection. -->
	<!--
		====================================================================
	-->
	<xsl:template match="quer:projection[quer:allCustomFields]" />
	<!--
		====================================================================
	-->
	<!--
		Indicate if a list of space separated elements includes the given
		element.
	-->
	<!-- Params: -->
	<!-- list: The list content. -->
	<!-- element: The element searched for. -->
	<!--
		====================================================================
	-->
	<xsl:function name="fct:list-includes" as="xs:boolean">
		<xsl:param name="list" />
		<xsl:param name="element" />
		<xsl:value-of select="index-of(tokenize($list, '\s+'), $element) &gt; 0" />
	</xsl:function>
</xsl:stylesheet>
